@@ -53,11 +53,11 @@ removing all space and newline characters."
 ;; Functions used to tweak translation pairs
 
 (defun agda-input-compose (f g)
-  "\x -> concatMap F (G x)"
+  "λ x -> concatMap F (G x)"
     (lambda (x) (agda-input-concat-map f (funcall g x))))
 
 (defun agda-input-or (f g)
-  "\x -> F x ++ G x"
+  "λ x -> F x ++ G x"
     (lambda (x) (append (funcall f x) (funcall g x))))
 
 (defun agda-input-nonempty ()
@@ -184,11 +184,12 @@ order for the change to take effect."
 
   ;; Equality and similar symbols.
 
-  ("eq"  . ,(agda-input-to-string-list "=∼∽≈≋∻∾∿≀≃⋍≂≅ ≌≊≡≣≐≑≒≓≔≕≖≗≘≙≚≛≜≝≞≟≍≎≏≬⋕"))
+  ("eq"  . ,(agda-input-to-string-list "=∼∽≈≋∻∾∿≀≃⋍≂≅ ≌≊≡≣≐≑≒≓≔≕≖≗≘≙≚≛≜≝≞≟≍≎≏≬⋕＝"))
   ("eqn" . ,(agda-input-to-string-list "≠≁ ≉     ≄  ≇≆  ≢                 ≭    "))
 
                     ("=n"  . ("≠"))
-  ("~"    . ("∼"))  ("~n"  . ("≁"))
+  ("~"    . ,(agda-input-to-string-list "∼～"))
+                    ("~n"  . ("≁"))
   ("~~"   . ("≈"))  ("~~n" . ("≉"))
   ("~~~"  . ("≋"))
   (":~"   . ("∻"))
@@ -198,6 +199,7 @@ order for the change to take effect."
   ("~~-"  . ("≊"))
   ("=="   . ("≡"))  ("==n" . ("≢"))
   ("==="  . ("≣"))
+  ("="    . ("＝"))
   (".="   . ("≐"))  (".=." . ("≑"))
   (":="   . ("≔"))  ("=:"  . ("≕"))
   ("=o"   . ("≗"))
@@ -211,9 +213,9 @@ order for the change to take effect."
 
   ;; Inequality and similar symbols.
 
-  ("leq"  . ,(agda-input-to-string-list "<≪⋘≤≦≲ ≶≺≼≾⊂⊆ ⋐⊏⊑ ⊰⊲⊴⋖⋚⋜⋞"))
+  ("leq"  . ,(agda-input-to-string-list "<≪⋘≤≦≲ ≶≺≼≾⊂⊆⫇⫉ ⋐⊏⊑ ⊰⊲⊴⋖⋚⋜⋞＜"))
   ("leqn" . ,(agda-input-to-string-list "≮  ≰≨≴⋦≸⊀ ⋨⊄⊈⊊  ⋢⋤ ⋪⋬   ⋠"))
-  ("geq"  . ,(agda-input-to-string-list ">≫⋙≥≧≳ ≷≻≽≿⊃⊇ ⋑⊐⊒ ⊱⊳⊵⋗⋛⋝⋟"))
+  ("geq"  . ,(agda-input-to-string-list ">≫⋙≥≧≳ ≷≻≽≿⊃⊇⫈⫊ ⋑⊐⊒ ⊱⊳⊵⋗⋛⋝⋟＞"))
   ("geqn" . ,(agda-input-to-string-list "≯  ≱≩≵⋧≹⊁ ⋩⊅⊉⊋  ⋣⋥ ⋫⋭   ⋡"))
 
   ("<="   . ("≤"))  (">="   . ("≥"))
@@ -228,6 +230,8 @@ order for the change to take effect."
   ("subn"  . ("⊄"))  ("supn"  . ("⊅"))
   ("sub="  . ("⊆"))  ("sup="  . ("⊇"))
   ("sub=n" . ("⊈"))  ("sup=n" . ("⊉"))
+  ("sub~"  . ("⫇"))  ("sup~"  . ("⫈"))
+  ("sub~~"  . ("⫉"))  ("sup~~"  . ("⫊"))
 
   ("squb"   . ("⊏"))  ("squp"   . ("⊐"))
   ("squb="  . ("⊑"))  ("squp="  . ("⊒"))
@@ -252,7 +256,7 @@ order for the change to take effect."
   ("glb" . ("⊓"))  ("lub" . ("⊔"))
   ("Glb" . ("⨅"))  ("Lub" . ("⨆"))
 
-  ;; Entailment etc.
+  ;; Entailment, tacks, etc.
 
   ("entails" . ,(agda-input-to-string-list "⊢⊣⊤⊥⊦⊧⊨⊩⊪⊫⊬⊭⊮⊯"))
 
@@ -263,6 +267,16 @@ order for the change to take effect."
   ("||="  . ("⊫"))  ("||=n" . ("⊯"))
   ("|||-" . ("⊪"))
 
+
+  ("tack" . ,(agda-input-to-string-list "⟘⟙⟛⟝⟞⫫⫪"))
+  ("Bot"  . ("⟘")) ;; similar to Glb/Lub vs. glb/lub
+  ("Top"  . ("⟙"))
+  ("-||-" . ("⟛"))
+  ("|--"  . ("⟝"))
+  ("--|"  . ("⟞"))
+  ("bbot" . ("⫫")) ;; similar to bN for blackboard N
+  ("btop" . ("⫪"))
+
   ;; Divisibility, parallelity.
 
   ("|"  . ("∣"))  ("|n"  . ("∤"))
@@ -275,6 +289,8 @@ order for the change to take effect."
   ("exn" . ("∄"))
   ("0"   . ("∅"))
   ("C"   . ("∁"))
+  ("uin"    . ("⟒"))
+  ("din"    . ("⫙"))
 
   ;; Corners, ceilings and floors.
 
@@ -293,24 +309,29 @@ order for the change to take effect."
   ("x"         . ("×"))
   ("o"         . ("∘"))
   ("comp"      . ("∘"))
-  ("."         . ("∙"))
+  ("."         . ,(agda-input-to-string-list "∙．"))
   ("*"         . ("⋆"))
   (".+"        . ("∔"))
   (".-"        . ("∸"))
   (":"         . ,(agda-input-to-string-list "∶⦂ː꞉˸፥፦：﹕︓"))
   (","         . ,(agda-input-to-string-list "ʻ،⸲⸴⹁⹉、︐︑﹐﹑，､"))
-  (";"         . ,(agda-input-to-string-list "؛⁏፤꛶；︔﹔⍮⸵;"))
+  (";"         . ,(agda-input-to-string-list "⨾⨟⁏፤꛶；︔﹔⍮⸵;"))
+  ("++"        . ("⧺"))
   ("::"        . ("∷"))
   ("::-"       . ("∺"))
   ("-:"        . ("∹"))
   ("+ "        . ("⊹"))
+  ("+"         . ("＋"))
+  ("sqrt"      . ("√"))
+  ("surd"      . ("√"))
   ("surd3"     . ("∛"))
   ("surd4"     . ("∜"))
   ("increment" . ("∆"))
   ("inf"       . ("∞"))
-  ("&"         . ("⅋"))
-  ("z;"        . ("⨟"))
+  ("&"         . ,(agda-input-to-string-list "⅋﹠＆"))
+  ("z;"        . ,(agda-input-to-string-list "⨟⨾"))
   ("z:"        . ("⦂"))
+  ("at"        . ,(agda-input-to-string-list "@﹫＠"))
 
   ;; Circled operators.
 
@@ -322,7 +343,7 @@ order for the change to take effect."
   ("oo"  . ("⊚"))
   ("o*"  . ("⊛"))
   ("o="  . ("⊜"))
-  ("o-"  . ("⊝"))
+  ("o-"  . ,(agda-input-to-string-list "⊝⟜"))
 
   ("O+"  . ("⨁"))
   ("Ox"  . ("⨂"))
@@ -331,23 +352,61 @@ order for the change to take effect."
 
   ;; Boxed operators.
 
-  ("b+" . ("⊞"))
-  ("b-" . ("⊟"))
-  ("bx" . ("⊠"))
-  ("b." . ("⊡"))
+  ("b+"  . ("⊞"))
+  ("b-"  . ("⊟"))
+  ("bx"  . ("⊠"))
+  ("b."  . ("⊡"))
+
+  ("b/"  . ("⧄"))
+  ("b\\" . ("⧅"))
+  ("b*"  . ("⧆"))
+  ("bo"  . ("⧇"))
+  ("bsq" . ("⧈"))
+
+  ;; APL boxed operators.
+
+  ("box="       . ("⌸"))
+  ("box?"       . ("⍰"))
+  ("box'"       . ("⍞"))
+  ("box:"       . ("⍠"))
+  ("box/"       . ("⍁"))
+  ("box\\"      . ("⍂"))
+  ("box<"       . ("⍃"))
+  ("box>"       . ("⍄"))
+  ("boxo"       . ("⌻"))
+  ("boxO"       . ("⌼"))
+
+  ("boxcomp"    . ("⌻"))
+  ("boxcircle"  . ("⌼"))
+  ("boxeq"      . ("⌸"))
+  ("boxneq"     . ("⍯"))
+  ("boxeqn"     . ("⍯"))
+
+  ("boxl"       . ("⍇"))
+  ("boxr"       . ("⍈"))
+  ("boxu"       . ("⍐"))
+  ("boxd"       . ("⍗"))
+
+  ("boxdi"      . ("⌺"))
+  ("boxdiv"     . ("⌹"))
+  ("boxwedge"   . ("⍓"))
+  ("boxvee"     . ("⍌"))
+  ("boxdelta"   . ("⍍"))
+  ("boxnabla"   . ("⍔"))
 
   ;; Various symbols.
 
   ("integral" . ,(agda-input-to-string-list "∫∬∭∮∯∰∱∲∳"))
   ("angle"    . ,(agda-input-to-string-list "∟∡∢⊾⊿"))
   ("join"     . ,(agda-input-to-string-list "⋈⋉⋊⋋⋌⨝⟕⟖⟗"))
+  ("esh"      . ("ʃ"))
 
   ;; Arrows.
 
-  ("l"  . ,(agda-input-to-string-list "←⇐⇚⇇⇆↤⇦↞↼↽⇠⇺↜⇽⟵⟸↚⇍⇷ ↹     ↢↩↫⇋⇜⇤⟻⟽⤆↶↺⟲                                     "))
-  ("r"  . ,(agda-input-to-string-list "→⇒⇛⇉⇄↦⇨↠⇀⇁⇢⇻↝⇾⟶⟹↛⇏⇸⇶ ↴    ↣↪↬⇌⇝⇥⟼⟾⤇↷↻⟳⇰⇴⟴⟿ ➵➸➙➔➛➜➝➞➟➠➡➢➣➤➧➨➩➪➫➬➭➮➯➱➲➳➺➻➼➽➾⊸"))
-  ("u"  . ,(agda-input-to-string-list "↑⇑⟰⇈⇅↥⇧↟↿↾⇡⇞          ↰↱➦ ⇪⇫⇬⇭⇮⇯                                           "))
-  ("d"  . ,(agda-input-to-string-list "↓⇓⟱⇊⇵↧⇩↡⇃⇂⇣⇟         ↵↲↳➥ ↯                                                "))
+  ("l"  . ,(agda-input-to-string-list "←⇐⇚⭅⇇⇆↤⇦↞↼↽⇠⇺↜⇽⟵⟸↚⇍⇷ ↹     ↢↩↫⇋⇜⇤⟻⟽⤆↶↺⟲                                     "))
+  ("r"  . ,(agda-input-to-string-list "→⇒⇛⭆⇉⇄↦⇨↠⇀⇁⇢⇻↝⇾⟶⟹↛⇏⇸⇶ ↴    ↣↪↬⇌⇝⇥⟼⟾⤇↷↻⟳⇰⇴⟴⟿ ➵➸➙➔➛➜➝➞➟➠➡➢➣➤➧➨➩➪➫➬➭➮➯➱➲➳➺➻➼➽➾⊸"))
+  ("u"  . ,(agda-input-to-string-list "↑⇑⤊⟰⇈⇅↥⇧↟↿↾⇡⇞          ↰↱➦ ⇪⇫⇬⇭⇮⇯                                           "))
+  ("d"  . ,(agda-input-to-string-list "↓⇓⤋⟱⇊⇵↧⇩↡⇃⇂⇣⇟         ↵↲↳➥ ↯                                                "))
   ("ud" . ,(agda-input-to-string-list "↕⇕   ↨⇳                                                                    "))
   ("lr" . ,(agda-input-to-string-list "↔⇔         ⇼↭⇿⟷⟺↮⇎⇹                                                        "))
   ("ul" . ,(agda-input-to-string-list "↖⇖                        ⇱↸                                               "))
@@ -389,6 +448,7 @@ order for the change to take effect."
   ("r->" . ("↣"))
 
   ("r-o" . ("⊸"))  ("-o"  . ("⊸"))
+  ("l-o" . ("⟜"))
 
   ("dz" . ("↯"))
 
@@ -536,6 +596,112 @@ order for the change to take effect."
   ("b8"   . ("𝟠"))
   ("b9"   . ("𝟡"))
 
+  ;; Mathematical bold letters.
+
+  ("BA"   . ("𝐀"))
+  ("BB"   . ("𝐁"))
+  ("BC"   . ("𝐂"))
+  ("BD"   . ("𝐃"))
+  ("BE"   . ("𝐄"))
+  ("BF"   . ("𝐅"))
+  ("BG"   . ("𝐆"))
+  ("BH"   . ("𝐇"))
+  ("BI"   . ("𝐈"))
+  ("BJ"   . ("𝐉"))
+  ("BK"   . ("𝐊"))
+  ("BL"   . ("𝐋"))
+  ("BM"   . ("𝐌"))
+  ("BN"   . ("𝐍"))
+  ("BO"   . ("𝐎"))
+  ("BP"   . ("𝐏"))
+  ("BQ"   . ("𝐐"))
+  ("BR"   . ("𝐑"))
+  ("BS"   . ("𝐒"))
+  ("BT"   . ("𝐓"))
+  ("BU"   . ("𝐔"))
+  ("BV"   . ("𝐕"))
+  ("BW"   . ("𝐖"))
+  ("BX"   . ("𝐗"))
+  ("BY"   . ("𝐘"))
+  ("BZ"   . ("𝐙"))
+  ("Ba"   . ("𝐚"))
+  ("Bb"   . ("𝐛"))
+  ("Bc"   . ("𝐜"))
+  ("Bd"   . ("𝐝"))
+  ("Be"   . ("𝐞"))
+  ("Bf"   . ("𝐟"))
+  ("Bg"   . ("𝐠"))
+  ("Bh"   . ("𝐡"))
+  ("Bi"   . ("𝐢"))
+  ("Bj"   . ("𝐣"))
+  ("Bk"   . ("𝐤"))
+  ("Bl"   . ("𝐥"))
+  ("Bm"   . ("𝐦"))
+  ("Bn"   . ("𝐧"))
+  ("Bo"   . ("𝐨"))
+  ("Bp"   . ("𝐩"))
+  ("Bq"   . ("𝐪"))
+  ("Br"   . ("𝐫"))
+  ("Bs"   . ("𝐬"))
+  ("Bt"   . ("𝐭"))
+  ("Bu"   . ("𝐮"))
+  ("Bv"   . ("𝐯"))
+  ("Bw"   . ("𝐰"))
+  ("Bx"   . ("𝐱"))
+  ("By"   . ("𝐲"))
+  ("Bz"   . ("𝐳"))
+
+  ;; Mathematical bold Greek letters.
+
+  ("BGA"  . ("𝚨"))
+  ("BGB"  . ("𝚩"))
+  ("BGC"  . ("𝚾"))
+  ("BGD"  . ("𝚫"))
+  ("BGE"  . ("𝚬"))
+  ("BGG"  . ("𝚪"))
+  ("BGH"  . ("𝚮"))
+  ("BGI"  . ("𝚰"))
+  ("BGK"  . ("𝚱"))
+  ("BGL"  . ("𝚲"))
+  ("BGM"  . ("𝚳"))
+  ("BGN"  . ("𝚴"))
+  ("BGO"  . ("𝛀"))
+  ("BOmicron" . ("𝚶"))
+  ("BGF"  . ("𝚽"))
+  ("BPi"  . ("𝚷"))
+  ("BGP"  . ("𝚿"))
+  ("BGR"  . ("𝚸"))
+  ("BGS"  . ("𝚺"))
+  ("BGT"  . ("𝚻"))
+  ("BGTH" . ("𝚯"))
+  ("BGU"  . ("𝚼"))
+  ("BGX"  . ("𝚵"))
+  ("BGZ"  . ("𝚭"))
+  ("BGa"  . ("𝛂"))
+  ("BGb"  . ("𝛃"))
+  ("BGc"  . ("𝛘"))
+  ("BGd"  . ("𝛅"))
+  ("BGe"  . ("𝛆"))
+  ("BGg"  . ("𝛄"))
+  ("BGh"  . ("𝛈"))
+  ("BGi"  . ("𝛊"))
+  ("BGk"  . ("𝛋"))
+  ("BGl"  . ("𝛌"))
+  ("BGm"  . ("𝛍"))
+  ("BGn"  . ("𝛎"))
+  ("BGo"  . ("𝛚"))
+  ("Bomicron" . ("𝛐"))
+  ("BGf"  . ("𝛗"))
+  ("Bpi"  . ("𝛑"))
+  ("BGp"  . ("𝛙"))
+  ("BGr"  . ("𝛒"))
+  ("BGs"  . ("𝛔"))
+  ("BGt"  . ("𝛕"))
+  ("BGth" . ("𝛉"))
+  ("BGu"  . ("𝛖"))
+  ("BGx"  . ("𝛏"))
+  ("BGz"  . ("𝛇"))
+
   ;; Mathematical bold digits.
 
   ("B0"   . ("𝟎"))
@@ -549,6 +715,110 @@ order for the change to take effect."
   ("B8"   . ("𝟖"))
   ("B9"   . ("𝟗"))
 
+  ;; Fullwidth letters.
+
+  ("FA"   . ("Ａ"))
+  ("FB"   . ("Ｂ"))
+  ("FC"   . ("Ｃ"))
+  ("FD"   . ("Ｄ"))
+  ("FE"   . ("Ｅ"))
+  ("FF"   . ("Ｆ"))
+  ("FG"   . ("Ｇ"))
+  ("FH"   . ("Ｈ"))
+  ("FI"   . ("Ｉ"))
+  ("FJ"   . ("Ｊ"))
+  ("FK"   . ("Ｋ"))
+  ("FL"   . ("Ｌ"))
+  ("FM"   . ("Ｍ"))
+  ("FN"   . ("Ｎ"))
+  ("FO"   . ("Ｏ"))
+  ("FP"   . ("Ｐ"))
+  ("FQ"   . ("Ｑ"))
+  ("FR"   . ("Ｒ"))
+  ("FS"   . ("Ｓ"))
+  ("FT"   . ("Ｔ"))
+  ("FU"   . ("Ｕ"))
+  ("FV"   . ("Ｖ"))
+  ("FW"   . ("Ｗ"))
+  ("FX"   . ("Ｘ"))
+  ("FY"   . ("Ｙ"))
+  ("FZ"   . ("Ｚ"))
+  ("Fa"   . ("ａ"))
+  ("Fb"   . ("ｂ"))
+  ("Fc"   . ("ｃ"))
+  ("Fd"   . ("ｄ"))
+  ("Fe"   . ("ｅ"))
+  ("Ff"   . ("ｆ"))
+  ("Fg"   . ("ｇ"))
+  ("Fh"   . ("ｈ"))
+  ("Fi"   . ("ｉ"))
+  ("Fj"   . ("ｊ"))
+  ("Fk"   . ("ｋ"))
+  ("Fl"   . ("ｌ"))
+  ("Fm"   . ("ｍ"))
+  ("Fn"   . ("ｎ"))
+  ("Fo"   . ("ｏ"))
+  ("Fp"   . ("ｐ"))
+  ("Fq"   . ("ｑ"))
+  ("Fr"   . ("ｒ"))
+  ("Fs"   . ("ｓ"))
+  ("Ft"   . ("ｔ"))
+  ("Fu"   . ("ｕ"))
+  ("Fv"   . ("ｖ"))
+  ("Fw"   . ("ｗ"))
+  ("Fx"   . ("ｘ"))
+  ("Fy"   . ("ｙ"))
+  ("Fz"   . ("ｚ"))
+
+  ;; Fullwidth digits.
+
+  ("F0"   . ("０"))
+  ("F1"   . ("１"))
+  ("F2"   . ("２"))
+  ("F3"   . ("３"))
+  ("F4"   . ("４"))
+  ("F5"   . ("５"))
+  ("F6"   . ("６"))
+  ("F7"   . ("７"))
+  ("F8"   . ("８"))
+  ("F9"   . ("９"))
+
+  ;; Fullwidth symbols.
+
+  ("F!"   . ("！"))
+  ("F\""  . ("＂"))
+  ("F#"   . ("＃"))
+  ("F$"   . ("＄"))
+  ("F%"   . ("％"))
+  ("F&"   . ("＆"))
+  ("F'"   . ("＇"))
+  ("F("   . ("（"))
+  ("F)"   . ("）"))
+  ("F*"   . ("＊"))
+  ("F+"   . ("＋"))
+  ("F,"   . ("，"))
+  ("F-"   . ("－"))
+  ("F."   . ("．"))
+  ("F/"   . ("／"))
+  ("F:"   . ("："))
+  ("F;"   . ("；"))
+  ("F<"   . ("＜"))
+  ("F="   . ("＝"))
+  ("F>"   . ("＞"))
+  ("F?"   . ("？"))
+  ("F@"   . ("＠"))
+  ("F["   . ("［"))
+  ("F\\"  . ("＼"))
+  ("F]"   . ("］"))
+  ("F_"   . ("＿"))
+  ("F{"   . ("｛"))
+  ("F|"   . ("｜"))
+  ("F}"   . ("｝"))
+  ("F~"   . ("～"))
+  ("F(("  . ("｟"))
+  ("F))"  . ("｠"))
+  ("Fneg" . ("￢"))
+
   ;; Parentheses.
 
   ("(" . ,(agda-input-to-string-list "([{⁅⁽₍〈⎴⟅⟦⟨⟪⦃〈《「『【〔〖〚︵︷︹︻︽︿﹁﹃﹙﹛﹝（［｛｢❪❬❰❲❴⟮⦅⦗⧼⸨❮⦇⦉"))
@@ -556,8 +826,8 @@ order for the change to take effect."
 
   ("[[" . ("⟦"))
   ("]]" . ("⟧"))
-  ("<"  . ("⟨"))
-  (">"  . ("⟩"))
+  ("<"  . ,(agda-input-to-string-list "⟨<≪⋘≺⊂⋐⊏⊰⊲⋖＜"))
+  (">"  . ,(agda-input-to-string-list "⟩>≫⋙≻⊃⋑⊐⊱⊳⋗＞"))
   ("<<" . ("⟪"))
   (">>" . ("⟫"))
   ("{{" . ("⦃"))
@@ -575,13 +845,13 @@ order for the change to take effect."
   ("(|" . ("⦇"))  ;; Idiom brackets
   ("|)" . ("⦈"))
 
-  ("((" . ("⦅"))  ;; Banana brackets
-  ("))" . ("⦆"))
+  ("((" . ,(agda-input-to-string-list "⦅｟"))  ;; Banana brackets
+  ("))" . ,(agda-input-to-string-list "⦆｠"))
 
   ;; Primes.
 
-  ("'" . ,(agda-input-to-string-list "′″‴⁗"))
-  ("`" . ,(agda-input-to-string-list "‵‶‷"))
+  ("'" . ,(agda-input-to-string-list "′″‴⁗＇"))
+  ("`" . ,(agda-input-to-string-list "‵‶‷｀"))
 
   ;; Fractions.
 
@@ -596,21 +866,25 @@ order for the change to take effect."
 
   ;; Musical symbols.
 
-  ("note" . ,(agda-input-to-string-list "♩♪♫♬"))
-  ("b"    . ("♭"))
-  ("#"    . ("♯"))
+  ("note"    . ,(agda-input-to-string-list "♩♪♫♬"))
+  ("b"       . ("♭"))
+  ("#"       . ("♯"))
+  ("bb"      . ("𝄫"))
+  ("##"      . ("𝄪"))
 
   ;; Other punctuation and symbols.
 
   ("\\"         . ("\\"))
   ("en"         . ("–"))
   ("em"         . ("—"))
+  ("!"          . ("！"))
   ("!!"         . ("‼"))
+  ("?"          . ("？"))
   ("??"         . ("⁇"))
   ("?!"         . ("‽" "⁈"))
   ("!?"         . ("⁉"))
   ("die"        . ,(agda-input-to-string-list "⚀⚁⚂⚃⚄⚅"))
-  ("asterisk"   . ,(agda-input-to-string-list "⁎⁑⁂✢✣✤✥✱✲✳✺✻✼✽❃❉❊❋"))
+  ("asterisk"   . ,(agda-input-to-string-list "⁎⁑⁂✢✣✤✥✱✲✳✺✻✼✽❃❉❊❋＊"))
   ("8<"         . ("✂" "✄"))
   ("tie"        . ("⁀"))
   ("undertie"   . ("‿"))
@@ -618,12 +892,21 @@ order for the change to take effect."
                                                ⍉⍊⍋⍌⍍⍎⍏⍐⍑⍒⍓⍔⍕⍖⍗⍘⍙⍚⍛
                                                ⍜⍝⍞⍟⍠⍡⍢⍣⍤⍥⍦⍧⍨⍩⍪⍫⍬⍭⍮
                                                ⍯⍰⍱⍲⍳⍴⍵⍶⍷⍸⍹⍺⎕"))
+  ("#"          . ("＃"))
+  ("%"          . ("％"))
+  ("&"          . ("＆"))
+  ("*"          . ("＊"))
+  ("/"          . ,(agda-input-to-string-list "／＼"))
+  ("@"          . ("＠"))
+  ("__"         . ("＿"))
+  ("\""         . ("＂"))
+  ("crossmark"  . ("✗"))
 
   ;; Some combining characters.
   ;;
   ;; The following combining characters also have (other)
   ;; translations:
-  ;; ̀ ́ ̂ ̃ ̄ ̆ ̇ ̈ ̋ ̌ ̣ ̧ ̱
+  ;;
 
   ("^--" . ,(agda-input-to-string-list"̅̿"))
   ("_--" . ,(agda-input-to-string-list"̲̳"))
@@ -669,7 +952,7 @@ order for the change to take effect."
   ("Gp"  . ("ψ"))  ("GP"  . ("Ψ"))
   ("Go"  . ("ω"))  ("GO"  . ("Ω"))
 
-  ;; Mathematical characters
+  ;; Mathematical characters.
 
   ("MiA" . ("𝐴"))
   ("MiB" . ("𝐵"))
@@ -723,6 +1006,7 @@ order for the change to take effect."
   ("Mix" . ("𝑥"))
   ("Miy" . ("𝑦"))
   ("Miz" . ("𝑧"))
+
   ("MIA" . ("𝑨"))
   ("MIB" . ("𝑩"))
   ("MIC" . ("𝑪"))
@@ -749,6 +1033,7 @@ order for the change to take effect."
   ("MIX" . ("𝑿"))
   ("MIY" . ("𝒀"))
   ("MIZ" . ("𝒁"))
+
   ("MIa" . ("𝒂"))
   ("MIb" . ("𝒃"))
   ("MIc" . ("𝒄"))
@@ -775,6 +1060,7 @@ order for the change to take effect."
   ("MIx" . ("𝒙"))
   ("MIy" . ("𝒚"))
   ("MIz" . ("𝒛"))
+
   ("McA" . ("𝒜"))
   ("McB" . ("ℬ"))
   ("McC" . ("𝒞"))
@@ -827,6 +1113,7 @@ order for the change to take effect."
   ("Mcx" . ("𝓍"))
   ("Mcy" . ("𝓎"))
   ("Mcz" . ("𝓏"))
+
   ("MCA" . ("𝓐"))
   ("MCB" . ("𝓑"))
   ("MCC" . ("𝓒"))
@@ -879,6 +1166,7 @@ order for the change to take effect."
   ("MCx" . ("𝔁"))
   ("MCy" . ("𝔂"))
   ("MCz" . ("𝔃"))
+
   ("MfA" . ("𝔄"))
   ("MfB" . ("𝔅"))
   ("MfC" . ("ℭ"))
@@ -932,7 +1220,7 @@ order for the change to take effect."
   ("Mfy" . ("𝔶"))
   ("Mfz" . ("𝔷"))
 
-  ;; (Sub / Super) scripts
+  ;; (Sub / Super) scripts.
   ;;
   ;; Unicode 12.1 omits several latin characters from sub/superscript.
   ;; https://www.quora.com/Why-is-there-no-character-for-superscript-q-in-Unicode
@@ -1028,14 +1316,18 @@ order for the change to take effect."
   ;; ("^Y" . ("Y"))
   ;; ("^Z" . ("Z"))
 
+  ("^Ga" . ("ᵅ"))
   ("^Gb" . ("ᵝ"))
   ("^Gg" . ("ᵞ"))
   ("^Gd" . ("ᵟ"))
   ("^Ge" . ("ᵋ"))
   ("^Gth" . ("ᶿ"))
+  ("^Gi" . ("ᶥ"))
   ("^Gf" . ("ᵠ"))
   ("^Gc" . ("ᵡ"))
 
+  ("^GF" . ("ᶲ"))
+
   ;; Some ISO8859-1 characters.
 
   (" "         . (" "))
