@@ -1,6 +1,10 @@
-{-# LANGUAGE CPP #-}
+{-# OPTIONS_GHC -Wunused-imports #-}
 
-module Agda.Compiler.Common where
+module Agda.Compiler.Common
+  ( module Agda.Compiler.Common
+  , IsMain(..)
+  )
+  where
 
 import Prelude hiding ((!!))
 
@@ -9,24 +13,19 @@ import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 import qualified Data.HashMap.Strict as HMap
-import qualified Data.HashSet as HSet
 import Data.Char
 import Data.Function (on)
-#if __GLASGOW_HASKELL__ < 804
-import Data.Semigroup
-#endif
 
 import Control.Monad
 import Control.Monad.State
 
 import Agda.Syntax.Common
-import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Internal as I
 import Agda.Syntax.TopLevelModuleName
 
-import Agda.Interaction.FindFile ( srcFilePath )
 import Agda.Interaction.Options
 import Agda.Interaction.Imports  ( CheckResult, crInterface, crSource, Source(..) )
+import Agda.Interaction.Library
 
 import Agda.TypeChecking.Monad as TCM
 
@@ -35,24 +34,10 @@ import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.List1          ( pattern (:|) )
 import Agda.Utils.Maybe
-import Agda.Utils.Monad          ( ifNotM )
-import Agda.Utils.Pretty
+import Agda.Utils.WithDefault    ( lensCollapseDefault )
 
 import Agda.Utils.Impossible
 
-data IsMain = IsMain | NotMain
-  deriving (Eq, Show)
-
--- | Conjunctive semigroup ('NotMain' is absorbing).
-instance Semigroup IsMain where
-  NotMain <> _ = NotMain
-  _       <> NotMain = NotMain
-  IsMain  <> IsMain = IsMain
-
-instance Monoid IsMain where
-  mempty = IsMain
-  mappend = (<>)
-
 doCompile :: Monoid r => (IsMain -> Interface -> TCM r) -> IsMain -> Interface -> TCM r
 doCompile f isMain i = do
   flip evalStateT Set.empty $ compilePrim $ doCompile' f isMain i
@@ -71,6 +56,7 @@ doCompile f isMain i = do
     agdaPrim = RawTopLevelModuleName
       { rawModuleNameRange = mempty
       , rawModuleNameParts = "Agda" :| "Primitive" : []
+      , rawModuleNameInferred = False
       }
       -- N.B. The Range in TopLevelModuleName is ignored for Ord, so we can set it to mempty.
 
@@ -100,7 +86,7 @@ setInterface i = do
   -- that it doesn't suffice to replace setTCLens' with setTCLens,
   -- because the stPreImportedModules field is strict.
   stImportedModules `setTCLens'`
-    HSet.fromList (map fst (iImportedModules i))
+    Set.fromList (map fst (iImportedModules i))
   stCurrentModule `setTCLens'`
     Just (iModuleName i, iTopLevelModuleName i)
 
@@ -154,13 +140,13 @@ inCompilerEnv checkResult cont = do
     -- the current pragma options persistent when we setCommandLineOptions
     -- below.
     opts <- getsTC $ stPersistentOptions . stPersistentState
-    let compileDir = case optCompileDir opts of
-          Just dir -> dir
-          Nothing  ->
-            -- The default output directory is the project root.
-            let tm = iTopLevelModuleName mainI
-                f  = srcFilePath $ srcOrigin checkedSource
-            in filePath $ projectRoot f tm
+    compileDir <- case optCompileDir opts of
+        Just dir -> pure dir
+        Nothing  -> do
+          -- The default output directory is the project root.
+          let tm = iTopLevelModuleName mainI
+          f <- srcFilePath $ srcOrigin checkedSource
+          pure $ filePath $ projectRoot f tm
     setCommandLineOptions $
       opts { optCompileDir = Just compileDir }
 
@@ -168,16 +154,17 @@ inCompilerEnv checkResult cont = do
     -- Unfortunately, a pragma option is stored in the interface file as
     -- just a list of strings, thus, the solution is a bit of hack:
     -- We match on whether @["--no-main"]@ is one of the stored options.
-    when (["--no-main"] `elem` iFilePragmaOptions mainI) $
-      stPragmaOptions `modifyTCLens` \ o -> o { optCompileNoMain = True }
+    let iFilePragmaStrings = map pragmaStrings . iFilePragmaOptions
+    when (["--no-main"] `elem` iFilePragmaStrings mainI) $
+      setTCLens (stPragmaOptions . lensOptCompileMain . lensCollapseDefault) False
 
     -- Perhaps all pragma options from the top-level module should be
     -- made available to the compiler in a suitable way. Here are more
     -- hacks:
-    when (any ("--cubical" `elem`) (iFilePragmaOptions mainI)) $
-      stPragmaOptions `modifyTCLens` \ o -> o { optCubical = Just CFull }
-    when (any ("--erased-cubical" `elem`) (iFilePragmaOptions mainI)) $
-      stPragmaOptions `modifyTCLens` \ o -> o { optCubical = Just CErased }
+    when (any ("--cubical" `elem`) $ iFilePragmaStrings mainI) $
+      setTCLens (stPragmaOptions . lensOptCubical) $ Just CFull
+    when (any ("--erased-cubical" `elem`) $ iFilePragmaStrings mainI) $
+      setTCLens (stPragmaOptions . lensOptCubical) $ Just CErased
 
     setScope (iInsideScope mainI) -- so that compiler errors don't use overly qualified names
     ignoreAbstractMode cont
