@@ -1,16 +1,19 @@
 module Agda.Compiler.JS.Pretty where
 
+import GHC.Generics (Generic)
+
 import Data.Char ( isAsciiLower, isAsciiUpper, isDigit )
 import Data.List ( intercalate )
 import Data.String ( IsString (fromString) )
 import Data.Semigroup ( Semigroup, (<>) )
-import Data.Set ( Set, toList, singleton, insert, member )
+import Data.Set ( Set, toList, insert, member )
 import qualified Data.Set as Set
-import Data.Map ( Map, toAscList, empty, null )
+import Data.Map ( Map, toAscList )
 import qualified Data.Text as T
 
 import Agda.Syntax.Common ( Nat )
 
+import Agda.Utils.Function ( applyWhen )
 import Agda.Utils.Hash
 import Agda.Utils.List ( indexWithDefault )
 import Agda.Utils.List1 ( List1, pattern (:|), (<|) )
@@ -34,6 +37,9 @@ import Agda.Compiler.JS.Syntax hiding (exports)
 --  standard pretty printers.
 --- This library code is only used in this module, and it is specialized to pretty
 --- print JavaScript code for the Agda backend, so I think its best place is in this module.
+data JSModuleStyle = JSES6 | JSCJS | JSAMD
+  deriving Generic
+
 data Doc
     = Doc String
     | Indent Int Doc
@@ -51,7 +57,7 @@ render :: Bool -> Doc -> String
 render minify = intercalate "\n" . joinLines . map (uncurry mkIndent) . go 0
   where
     joinLines :: [String] -> [String]
-    joinLines = if minify then chunks 0 [] else id
+    joinLines = applyWhen minify $ chunks 0 []
       where
         chunks len acc [] = [concat (reverse acc)]
         chunks len acc (s: ss)
@@ -76,12 +82,12 @@ render minify = intercalate "\n" . joinLines . map (uncurry mkIndent) . go 0
 
     punctuation = (`elem` ("(){}[];:, " :: String))
 
-    go i Space = if minify then [] else [(i, " ")]
+    go i Space = [(i, " ") | not minify]
     go i Empty = []
     go i (Doc s) = [(i, s)]
     go i (Beside d d') = joinBy (\(i, s) (_, s') -> [(i, s ++ s')]) (go i d) (go i d')
     go i (Above d d') = joinBy overlay (go i d) (go i d')
-    go i (Indent j d) = go (i+j) d
+    go i (Indent j d) = go (i + j) d
     go i (Enclose open close d) = go i $ Group $ Above open $ Above d close
     go i (Group d)
         | size ss < 40 = compact ss
@@ -137,7 +143,7 @@ group = Group
 
 indentBy :: Int -> Doc -> Doc
 indentBy i Empty = Empty
-indentBy i (Indent j d) = Indent (i+j) d
+indentBy i (Indent j d) = Indent (i + j) d
 indentBy i d = Indent i d
 
 enclose :: Doc -> Doc -> Doc -> Doc
@@ -198,13 +204,19 @@ unescapes s = text $ concatMap unescape s
 --   if b is true then the output is minified
 
 class Pretty a where
-    pretty :: (Nat, Bool) -> a -> Doc
+    pretty :: (Nat, Bool, JSModuleStyle) -> a -> Doc
+    prettyPrec :: Int -> (Nat, Bool, JSModuleStyle) -> a -> Doc
 
-prettyShow :: Pretty a => Bool -> a -> String
-prettyShow minify = render minify . pretty (0, minify)
+    pretty = prettyPrec 0
+    prettyPrec = const pretty
+
+    {-# MINIMAL pretty | prettyPrec #-}
+
+prettyShow :: Pretty a => Bool -> JSModuleStyle -> a -> String
+prettyShow minify ms = render minify . pretty (0, minify, ms)
 
 instance Pretty a => Pretty (Maybe a) where
-  pretty n = maybe mempty (pretty n)
+  prettyPrec p n = maybe mempty (prettyPrec p n)
 
 instance (Pretty a, Pretty b) => Pretty (a,b) where
   pretty n (x,y) = pretty n x <> ":" <+> pretty n y
@@ -212,7 +224,7 @@ instance (Pretty a, Pretty b) => Pretty (a,b) where
 -- Pretty-print collections
 
 class Pretties a where
-    pretties :: (Nat, Bool) -> a -> [Doc]
+    pretties :: (Nat, Bool, JSModuleStyle) -> a -> [Doc]
 
 instance Pretty a => Pretties [a] where
   pretties n = map (pretty n)
@@ -226,7 +238,7 @@ instance (Pretty a, Pretty b) => Pretties (Map a b) where
 -- Pretty print identifiers
 
 instance Pretty LocalId where
-  pretty (n, _) (LocalId x) = text $ indexWithDefault __IMPOSSIBLE__ vars (n - x - 1)
+  pretty (n, _, _) (LocalId x) = text $ indexWithDefault __IMPOSSIBLE__ vars (n - x - 1)
     where
       vars = ("": map show [0..]) >>= \s -> map (:s) ['a'..'z']
 
@@ -239,44 +251,45 @@ instance Pretty MemberId where
 
 instance Pretty Comment where
   pretty _ (Comment "") = mempty
-  pretty (_, True) _ = mempty
+  pretty (_, True, _) _ = mempty
   pretty _ (Comment s) = text $ "/* " ++ s ++ " */"
 
 -- Pretty print expressions
 
 instance Pretty Exp where
-  pretty n (Self)            = "exports"
-  pretty n (Local x)         = pretty n x
-  pretty n (Global m)        = pretty n m
-  pretty n (Undefined)       = "undefined"
-  pretty n (Null)            = "null"
-  pretty n (String s)        = "\"" <> unescapes (T.unpack s) <> "\""
-  pretty n (Char c)          = "\"" <> unescapes [c] <> "\""
-  pretty n (Integer x)       = "agdaRTS.primIntegerFromString(\"" <> text (show x) <> "\")"
-  pretty n (Double x)        = text $ show x
-  pretty (n, min) (Lambda x e) =
-    mparens (x /= 1) (punctuate "," (pretties (n+x, min) (map LocalId [x-1, x-2 .. 0])))
-    <+> "=>" <+> block (n+x, min) e
-  pretty n (Object o)        = braces $ punctuate "," $ pretties n o
-  pretty n (Array es)        = brackets $ punctuate "," [pretty n c <> pretty n e | (c, e) <- es]
-  pretty n (Apply f es)      = pretty n f <> parens (punctuate "," $ pretties n es)
-  pretty n (Lookup e l)      = pretty n e <> brackets (pretty n l)
-  pretty n (If e f g)        = parens $ pretty n e <> "?" <+> pretty n f <> ":" <+> pretty n g
-  pretty n (PreOp op e)      = parens $ text op <> " " <> pretty n e
-  pretty n (BinOp e op f)    = parens $ pretty n e <> " " <> text op <> " " <> pretty n f
-  pretty n (Const c)         = text c
-  pretty n (PlainJS js)      = text js
+  prettyPrec p n (Self)         = "exports"
+  prettyPrec p n (Local x)      = pretty n x
+  prettyPrec p n (Global m)     = pretty n m
+  prettyPrec p n (Undefined)    = "undefined"
+  prettyPrec p n (Null)         = "null"
+  prettyPrec p n (String s)     = "\"" <> unescapes (T.unpack s) <> "\""
+  prettyPrec p n (Char c)       = "\"" <> unescapes [c] <> "\""
+  prettyPrec p n (Integer x)    = "agdaRTS.primIntegerFromString(\"" <> text (show x) <> "\")"
+  prettyPrec p n (Double x)     = text $ show x
+  prettyPrec p (n, min, ms) (Lambda x e) = mparens (p > 2) $
+    mparens (x /= 1) (punctuate "," (pretties (n + x, min, ms) (map LocalId [x-1, x-2 .. 0])))
+    <+> "=>" <+> block (n + x, min, ms) e
+  prettyPrec p n (Object o)     = braces $ punctuate "," $ pretties n o
+  prettyPrec p n (Array es)     = brackets $ punctuate "," [pretty n c <> pretty n e | (c, e) <- es]
+  prettyPrec p n (Apply f es)   = prettyPrec 17 n f <> parens (punctuate "," $ pretties n es)
+  prettyPrec p n (Lookup e l)   = prettyPrec 17 n e <> brackets (pretty n l)
+  prettyPrec p n (If e f g)     = mparens (p > 2) $
+    prettyPrec 3 n e <> "?" <+> prettyPrec 3 n f <> ":" <+> prettyPrec 2 n g
+  prettyPrec p n (PreOp op e)   = parens $ text op <> " " <> prettyPrec 17 n e
+  prettyPrec p n (BinOp e op f) = parens $ prettyPrec 17 n e <> " " <> text op <> " " <> prettyPrec 17 n f
+  prettyPrec p n (Const c)      = text c
+  prettyPrec p n (PlainJS js)   = text js
 
-block :: (Nat, Bool) -> Exp -> Doc
-block n e = mparens (doNest e) $ pretty n e
+block :: (Nat, Bool, JSModuleStyle) -> Exp -> Doc
+block n e = mparens (doNest e) $ prettyPrec 2 n e
   where
     doNest Object{} = True
     doNest _ = False
 
 modname :: GlobalId -> Doc
-modname (GlobalId ms) = text $ "\"" ++ intercalate "." ms ++ "\""
+modname (GlobalId ms) = text $ intercalate "." ms
 
-exports :: (Nat, Bool) -> Set JSQName -> [Export] -> Doc
+exports :: (Nat, Bool, JSModuleStyle) -> Set JSQName -> [Export] -> Doc
 exports n lss [] = Empty
 exports n lss es0@(Export ls e : es)
   -- If the parent of @ls@ is already defined (or no parent exists), @ls@ can be defined
@@ -295,19 +308,50 @@ instance Pretty [(GlobalId, Export)] where
            | (g, Export ls e) <- es ]
 
 instance Pretty Module where
-  pretty n (Module m is es callMain) = vsep
-    [ importRTS
+  pretty opt@(n, min, JSES6) (Module m is es callMain) = vsep
+    [ "import agdaRTS from \"./agda-rts.mjs\";"
     , imports
-    , exports n Set.empty es
-    , pretty n callMain
+    , "const exports = {};"
+    , exports opt Set.empty es
+    , pretty opt callMain
+    , ";export default exports;"
+    ]
+    $+$ ""
+    where
+      imports = vcat [
+        "import " <> indent (pretty opt e) <+> " from \"./" <> modname e <> ".mjs\";"
+        | e <- toList (globals es <> Set.fromList is)
+        ]
+      les = toList (globals es <> Set.fromList is)
+  pretty opt@(n, min, JSCJS) (Module m is es callMain) = vsep
+    [ "var agdaRTS" <+> "=" <+> "require(\"agda-rts\");"
+    , imports
+    , exports opt Set.empty es
+    , pretty opt callMain
+    ]
+    $+$ ""
+    where
+      imports = vcat [
+        "var " <> indent (pretty opt e) <+> "=" <+> "require(\"" <> modname e <> "\");"
+        | e <- toList (globals es <> Set.fromList is)
+        ]
+      les = toList (globals es <> Set.fromList is)
+  pretty opt@(n, min, JSAMD) (Module m is es callMain) = vsep
+    [ "define(['agda-rts'"
+      <+> hcat [ ", " <+> ("\"" <> modname e <> "\"") | e <- les ]
+      <+> "],"
+    , "function(agdaRTS"
+      <+> hcat [ ", " <+> pretty opt e | e <- les ]
+      <+> ") {"
+    , "var exports = {};"
+    , exports opt Set.empty es
+    , pretty opt callMain
+    , "; return exports; });"
     ]
     $+$ "" -- Final newline
     where
-      importRTS = "var agdaRTS" <+> "=" <+> "require(\"agda-rts\");"
-      imports   = vcat
-        [ "var " <> indent (pretty n e) <+> "=" <+> "require(" <> modname e <> ");"
-        | e <- toList (globals es <> Set.fromList is)
-        ]
+      les = toList (globals es <> Set.fromList is)
+
 
 variableName :: String -> String
 variableName s = if isValidJSIdent s then "z_" ++ s else "h_" ++ show (hashString s)
