@@ -29,6 +29,7 @@ import Agda.TypeChecking.Monad
 
 import Agda.Syntax.Common.Pretty
 
+import Agda.Utils.CallStack ( HasCallStack )
 import Agda.Utils.Impossible
 
 --------------------------------------------------
@@ -90,8 +91,6 @@ data GHCEnv = GHCEnv
   , ghcEnvPathP
   , ghcEnvSub
   , ghcEnvSubIn
-  , ghcEnvId
-  , ghcEnvConId
     :: Maybe QName
     -- Various (possibly) builtin names.
   , ghcEnvIsTCBuiltin :: QName -> Bool
@@ -146,6 +145,9 @@ runHsCompileT' t e s = (flip runStateT s) . (flip runReaderT e) $ t
 runHsCompileT :: HsCompileT m a -> GHCModuleEnv -> m (a, HsCompileState)
 runHsCompileT t e = runHsCompileT' t e mempty
 
+ghcBackendError :: (HasCallStack, MonadTCError m) => GHCBackendError -> m a
+ghcBackendError = locatedTypeError GHCBackendError
+
 --------------------------------------------------
 -- utilities for haskell names
 --------------------------------------------------
