@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE PatternSynonyms #-}
 
 module Agda.Compiler.Treeless.Erase
@@ -6,16 +8,14 @@ module Agda.Compiler.Treeless.Erase
        , isErasable
        ) where
 
-import Control.Arrow (first, second)
-import Control.Monad
-import Control.Monad.State
+import Control.Arrow       ( first, second )
+import Control.Monad.State ( StateT, evalStateT )
+
 import Data.Map (Map)
 import qualified Data.Map as Map
-import qualified Data.List as List
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
-import Agda.Syntax.Position
 import Agda.Syntax.Treeless
 import Agda.Syntax.Literal
 
@@ -26,7 +26,6 @@ import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Primitive
 
-import {-# SOURCE #-} Agda.Compiler.Backend
 import Agda.Compiler.Treeless.Subst
 import Agda.Compiler.Treeless.Unused
 
@@ -35,7 +34,7 @@ import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Memo
 import Agda.Utils.Monad
-import Agda.Utils.Pretty (prettyShow)
+import Agda.Syntax.Common.Pretty (prettyShow)
 import Agda.Utils.IntSet.Infinite (IntSet)
 import qualified Agda.Utils.IntSet.Infinite as IntSet
 
@@ -49,10 +48,10 @@ data ESt = ESt
       -- ^ Memoize computed `TypeInfo` for data/record types `QName`.
   }
 
-funMap :: Lens' (Map QName FunInfo) ESt
+funMap :: Lens' ESt (Map QName FunInfo)
 funMap f r = f (_funMap r) <&> \ a -> r { _funMap = a }
 
-typeMap :: Lens' (Map QName TypeInfo) ESt
+typeMap :: Lens' ESt (Map QName TypeInfo)
 typeMap f r = f (_typeMap r) <&> \ a -> r { _typeMap = a }
 
 -- | Eraser monad.
@@ -110,7 +109,7 @@ eraseTerms q eval t = usedArguments q t *> runE (eraseTop q t)
                    _             -> erase $ subst 0 TErased b
             else tLet e <$> erase b
         TCase x t d bs -> do
-          (d, bs) <- pruneUnreachable x (caseType t) d bs
+          (d, bs) <- pruneUnreachable x (caseErased t) (caseType t) d bs
           d       <- erase d
           bs      <- mapM eraseAlt bs
           tCase x t d bs
@@ -137,17 +136,41 @@ eraseTerms q eval t = usedArguments q t *> runE (eraseTop q t)
     tCase x t d bs
       | isErased d && all (isErased . aBody) bs = pure TErased
       | otherwise = case bs of
-        [TACon c a b] -> do
+        [b@(TACon c _ _)] -> do
           h <- snd <$> getFunInfo c
           case h of
-            NotErasable -> noerase
+            NotErasable -> fallback
             Empty       -> pure TErased
-            Erasable    -> (if a == 0 then pure else erase) $ applySubst (replicate a TErased ++# idS) b
-                              -- might enable more erasure
-        _ -> noerase
+            Erasable    -> erasedBody b
+        _ -> fallback
       where
         noerase = pure $ TCase x t d bs
 
+        erasedBody = \case
+          TACon _ arity body ->
+            (if arity == 0 then pure else erase) $
+               -- might enable more erasure
+            applySubst (replicate arity TErased ++# idS) body
+          TALit _ body   -> pure body
+          TAGuard _ body -> pure body
+
+        fallback = case (caseErased t, bs) of
+          (Erased{}, [b]) ->
+            -- The case variable is erased, and there is exactly one
+            -- case: use the case's body.
+            erasedBody b
+          (Erased{}, [])  ->
+            -- The case variable is erased, and there is no case: use
+            -- the default.
+            pure $ if isErased d then TErased else d
+          (Erased{}, _ : _ : _) ->
+            -- The case variable is erased, and there are at least two
+            -- cases: crash.
+            __IMPOSSIBLE__
+          _ ->
+            -- The case variable is not erased: do not erase anything.
+            noerase
+
     isErased t = t == TErased || isUnreachable t
 
     eraseRel r t | erasable r = pure TErased
@@ -161,29 +184,49 @@ eraseTerms q eval t = usedArguments q t *> runE (eraseTop q t)
         TACon c a <$> erase (applySubst sub b)
       TAGuard g b -> TAGuard   <$> erase g <*> erase b
 
+pruneUnreachable ::
+  Int -> Erased -> CaseType -> TTerm -> [TAlt] -> E (TTerm, [TAlt])
+pruneUnreachable x erased t d bs = case erased of
+  NotErased{} -> pruneUnreachable' x erased t d bs
+  Erased{}    ->
+    -- If the match is on an erased argument, then the first branch
+    -- should match.
+    case bs of
+      []    -> pruneUnreachable' x erased t d            []
+      b : _ -> pruneUnreachable' x erased t tUnreachable [b]
+
 -- | Doesn't have any type information (other than the name of the data type),
 --   so we can't do better than checking if all constructors are present.
-pruneUnreachable :: Int -> CaseType -> TTerm -> [TAlt] -> E (TTerm, [TAlt])
-pruneUnreachable _ (CTData quantity q) d bs' = do
-  -- In an erased setting erased constructors are not treated
-  -- specially.
+pruneUnreachable' ::
+  Int -> Erased -> CaseType -> TTerm -> [TAlt] -> E (TTerm, [TAlt])
+pruneUnreachable' _ erased (CTData q) d bs' = do
+  -- Erased constructors are pruned iff the match is made on a
+  -- non-erased argument.
   cs <- lift $
-        if hasQuantity0 quantity
+        if isErased erased
         then getConstructors q
         else getNotErasedConstructors q
-  let bs | hasQuantity0 quantity = bs'
-         | otherwise             =
+  let bs | isErased erased = bs'
+         | otherwise       =
            flip filter bs' $ \case
              a@TACon{} -> (aCon a) `elem` cs
              TAGuard{} -> True
              TALit{}   -> True
-  let complete =length cs == length [ b | b@TACon{} <- bs ]
-  let d' | complete  = tUnreachable
+  let -- In the case of a match on an erased argument the value d is
+      -- equal to tUnreachable, except perhaps if bs is empty. In the
+      -- latter case complete is True exactly when the type has zero
+      -- constructors (erased or not), in which case it makes sense to
+      -- replace d with tUnreachable.
+      complete = length cs == length [ b | b@TACon{} <- bs ]
+      d' | complete  = tUnreachable
          | otherwise = d
   return (d', bs)
-pruneUnreachable x CTNat d bs = return $ pruneIntCase x d bs (IntSet.below 0)
-pruneUnreachable x CTInt d bs = return $ pruneIntCase x d bs IntSet.empty
-pruneUnreachable _ _ d bs = pure (d, bs)
+pruneUnreachable' x _ CTNat d bs =
+  return $ pruneIntCase x d bs (IntSet.below 0)
+pruneUnreachable' x _ CTInt d bs =
+  return $ pruneIntCase x d bs IntSet.empty
+pruneUnreachable' _ _ _ d bs =
+  pure (d, bs)
 
 -- These are the guards we generate for Int/Nat pattern matching
 pattern Below :: Int -> Integer -> TTerm
@@ -261,11 +304,6 @@ getFunInfo q = memo (funMap . key q) $ getInfo q
 
 isErasable :: QName -> TCM Bool
 isErasable qn =
-  -- The active backend should be set
-  caseMaybeM (viewTC eActiveBackendName) __IMPOSSIBLE__ $ \ bname ->
-  -- However it may not be part of the set of available backends
-  -- in which case we default to not erasable to avoid false negatives.
-  caseMaybeM (lookupBackend bname)       (pure False)   $ \ _ ->
   erasable . snd <$> runE (getFunInfo qn)
 
 telListView :: Type -> TCM (ListTel, Type)
@@ -306,7 +344,6 @@ getTypeInfo t0 = do
   typeInfo :: QName -> E TypeInfo
   typeInfo q = ifM (erasureForbidden q) (return NotErasable) $ {-else-} do
     memoRec (typeMap . key q) Erasable $ do  -- assume recursive occurrences are erasable
-      mId    <- lift $ getName' builtinId
       msizes <- lift $ mapM getBuiltinName
                          [builtinSize, builtinSizeLt]
       def    <- lift $ getConstInfo q
@@ -315,7 +352,6 @@ getTypeInfo t0 = do
                   I.Record{ recConHead = c }  -> Just [conName c]
                   _                           -> Nothing
       case mcs of
-        _ | Just q == mId        -> return NotErasable
         _ | Just q `elem` msizes -> return Erasable
         Just [c] -> do
           (ts, _) <- lift $ typeWithoutParams c
