@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.Compiler.Treeless.Pretty () where
@@ -7,14 +9,23 @@ import Prelude hiding ((!!)) -- don't use partial functions!
 import Control.Arrow (first)
 import Control.Monad.Reader
 import Data.Maybe
-import qualified Data.Map as Map
+import qualified Data.IntMap as IntMap
 
 import Agda.Syntax.Treeless
+import Agda.Syntax.Common.Pretty
+
 import Agda.Compiler.Treeless.Subst
-import Agda.Utils.Pretty
-import Agda.Utils.List
 
 import Agda.Utils.Impossible
+import Agda.Utils.Function
+import Agda.Utils.List
+
+instance Pretty Compiled where
+  pretty Compiled {cTreeless, cArgUsage} =
+    "Compiled {" <?> vcat
+      [ "cTreeless   =" <?> pretty cTreeless
+      , "funCompiled =" <?> pshow cArgUsage
+      ] <?> "}"
 
 data PEnv = PEnv { pPrec :: Int
                  , pFresh :: [String]
@@ -36,12 +47,12 @@ withNames' :: HasFree a => Int -> a -> ([String] -> P b) -> P b
 withNames' n tm k = withNames n' $ k . insBlanks
   where
     fv = freeVars tm
-    n'  = length $ filter (< n) $ Map.keys fv
+    n'  = length $ filter (< n) $ IntMap.keys fv
     insBlanks = go n
       where
         go 0 _ = []
         go i xs0@(~(x : xs))
-          | Map.member (i - 1) fv = x   : go (i - 1) xs
+          | IntMap.member (i - 1) fv = x   : go (i - 1) xs
           | otherwise             = "_" : go (i - 1) xs0
 
 bindName :: String -> P a -> P a
@@ -53,7 +64,7 @@ bindNames xs p = foldr bindName p xs
 paren :: Int -> P Doc -> P Doc
 paren p doc = do
   n <- asks pPrec
-  (if p < n then parens else id) <$> doc
+  applyWhen (p < n) parens <$> doc
 
 prec :: Int -> P a -> P a
 prec p = local $ \ e -> e { pPrec = p }
