@@ -10,20 +10,21 @@ module Agda.Interaction.FindFile
   ( SourceFile(..), InterfaceFile(intFilePath)
   , toIFile, mkInterfaceFile
   , FindError(..), findErrorToTypeError
-  , findFile, findFile', findFile''
+  , findFile, findFile', findFile'_, findFile''
   , findInterfaceFile', findInterfaceFile
   , checkModuleName
-  , moduleName
   , rootNameModule
   , replaceModuleExtension
+  , dropAgdaExtension, hasAgdaExtension, stripAgdaExtension
   ) where
 
 import Prelude hiding (null)
 
 import Control.Monad
 import Control.Monad.Except
+import Control.Monad.State
 import Control.Monad.Trans
-import Data.Maybe (catMaybes)
+import Data.Maybe (catMaybes, fromMaybe, isJust)
 import qualified Data.Map as Map
 import qualified Data.Text as T
 import System.FilePath
@@ -36,23 +37,25 @@ import Agda.Syntax.Parser.Literate (literateExtsShortList)
 import Agda.Syntax.Position
 import Agda.Syntax.TopLevelModuleName
 
-import Agda.Interaction.Options ( optLocalInterfaces )
-
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Benchmark (billTo)
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import {-# SOURCE #-} Agda.TypeChecking.Monad.Options
   (getIncludeDirs, libToTCM)
-import Agda.TypeChecking.Monad.State (topLevelModuleName)
-import Agda.TypeChecking.Warnings (runPM, warning)
+import Agda.TypeChecking.Monad.State ( registerFileIdWithBuiltin, topLevelModuleName )
+import Agda.TypeChecking.Monad.Trace (runPM, setCurrentRange)
 
 import Agda.Version ( version )
 
 import Agda.Utils.Applicative ( (?$>) )
+import Agda.Utils.CallStack   ( HasCallStack )
+import Agda.Utils.FileId
 import Agda.Utils.FileName
 import Agda.Utils.List  ( stripSuffix, nubOn )
 import Agda.Utils.List1 ( List1, pattern (:|) )
+import Agda.Utils.List2 ( List2, pattern List2 )
 import qualified Agda.Utils.List1 as List1
+import qualified Agda.Utils.List2 as List2
 import Agda.Utils.Monad ( ifM, unlessM )
 import Agda.Syntax.Common.Pretty ( Pretty(..), prettyShow )
 import qualified Agda.Syntax.Common.Pretty as P
@@ -60,17 +63,15 @@ import Agda.Utils.Singleton
 
 import Agda.Utils.Impossible
 
--- | Type aliases for source files and interface files.
---   We may only produce one of these if we know for sure that the file
---   does exist. We can always output an @AbsolutePath@ if we are not sure.
+-- This instance isn't producing something pretty.
+-- instance Pretty SourceFile where
+--   pretty = pretty . srcFileId
 
--- TODO: do not export @SourceFile@ and force users to check the
--- @AbsolutePath@ does exist.
-newtype SourceFile    = SourceFile    { srcFilePath :: AbsolutePath } deriving (Eq, Ord, Show)
+-- | File must exist.
 newtype InterfaceFile = InterfaceFile { intFilePath :: AbsolutePath }
 
-instance Pretty SourceFile    where pretty = pretty . srcFilePath
-instance Pretty InterfaceFile where pretty = pretty . intFilePath
+instance Pretty InterfaceFile where
+  pretty = pretty . intFilePath
 
 -- | Makes an interface file from an AbsolutePath candidate.
 --   If the file does not exist, then fail by returning @Nothing@.
@@ -85,31 +86,19 @@ mkInterfaceFile fp = do
 -- | Converts an Agda file name to the corresponding interface file
 --   name. Note that we do not guarantee that the file exists.
 
-toIFile :: SourceFile -> TCM AbsolutePath
-toIFile (SourceFile src) = do
+toIFile :: HasCallStack => SourceFile -> TCM AbsolutePath
+toIFile (SourceFile fi) = do
+  src <- fileFromId fi -- partial function, thus HasCallStack
   let fp = filePath src
   let localIFile = replaceModuleExtension ".agdai" src
   mroot <- libToTCM $ findProjectRoot (takeDirectory fp)
   case mroot of
     Nothing   -> pure localIFile
-    Just root ->
+    Just root -> do
       let buildDir = root </> "_build" </> version </> "agda"
-          fileName = makeRelative root (filePath localIFile)
-          separatedIFile = mkAbsolute $ buildDir </> fileName
-          ifilePreference = ifM (optLocalInterfaces <$> commandLineOptions)
-            (pure (localIFile, separatedIFile))
-            (pure (separatedIFile, localIFile))
-      in do
-        separatedIFileExists <- liftIO $ doesFileExistCaseSensitive $ filePath separatedIFile
-        localIFileExists <- liftIO $ doesFileExistCaseSensitive $ filePath localIFile
-        case (separatedIFileExists, localIFileExists) of
-          (False, False) -> fst <$> ifilePreference
-          (False, True) -> pure localIFile
-          (True, False) -> pure separatedIFile
-          (True, True) -> do
-            ifiles <- ifilePreference
-            warning $ uncurry DuplicateInterfaceFiles ifiles
-            pure $ fst ifiles
+      fileName <- liftIO $ makeRelativeCanonical root (filePath localIFile)
+      let separatedIFile = mkAbsolute $ buildDir </> fileName
+      pure separatedIFile
 
 replaceModuleExtension :: String -> AbsolutePath -> AbsolutePath
 replaceModuleExtension ext@('.':_) = mkAbsolute . (++ ext) .  dropAgdaExtension . filePath
@@ -120,23 +109,25 @@ replaceModuleExtension ext = replaceModuleExtension ('.':ext)
 -- Invariant: All paths are absolute.
 
 data FindError
-  = NotFound [SourceFile]
-    -- ^ The file was not found. It should have had one of the given
-    -- file names.
-  | Ambiguous [SourceFile]
-    -- ^ Several matching files were found.
-    --
-    -- Invariant: The list of matching files has at least two
-    -- elements.
+  = NotFound [AbsolutePath]
+      -- ^ The file was not found.
+      --   It should have had one of the given file names.
+  | Ambiguous (List2 AbsolutePath)
+      -- ^ Several matching files were found.
   deriving Show
 
 -- | Given the module name which the error applies to this function
 -- converts a 'FindError' to a 'TypeError'.
 
 findErrorToTypeError :: TopLevelModuleName -> FindError -> TypeError
-findErrorToTypeError m (NotFound  files) = FileNotFound m (map srcFilePath files)
-findErrorToTypeError m (Ambiguous files) =
-  AmbiguousTopLevelModuleName m (map srcFilePath files)
+findErrorToTypeError m = \case
+  NotFound  files -> FileNotFound m files
+  Ambiguous files -> AmbiguousTopLevelModuleName m files
+
+-- findErrorToTypeError :: MonadFileId m => TopLevelModuleName -> FindError -> m TypeError
+-- findErrorToTypeError m = \case
+--   NotFound  files -> FileNotFound m <$> mapM srcFilePath files
+--   Ambiguous files -> AmbiguousTopLevelModuleName m <$> mapM srcFilePath files
 
 -- | Finds the source file corresponding to a given top-level module
 -- name. The returned paths are absolute.
@@ -156,37 +147,56 @@ findFile m = do
 --   SIDE EFFECT:  Updates 'stModuleToSource'.
 findFile' :: TopLevelModuleName -> TCM (Either FindError SourceFile)
 findFile' m = do
-    dirs         <- getIncludeDirs
-    modFile      <- useTC stModuleToSource
-    (r, modFile) <- liftIO $ findFile'' dirs m modFile
-    stModuleToSource `setTCLens` modFile
+    dirs     <- getIncludeDirs
+    modToSrc <- useTC stModuleToSource
+    (r, modToSrc) <- liftIO $ runStateT (findFile'' dirs m) modToSrc
+    stModuleToSource `setTCLens` modToSrc
     return r
 
+-- | A variant of 'findFile'' which manipulates an extra 'ModuleToSourceId'
+
+findFile'_ ::
+     List1 AbsolutePath
+       -- ^ Include paths.
+  -> TopLevelModuleName
+  -> StateT ModuleToSourceId TCM (Either FindError SourceFile)
+findFile'_ incs m = do
+  dict <- useTC stFileDict
+  m2s  <- get
+  (r, ModuleToSource dict' m2s') <- liftIO $
+    runStateT (findFile'' incs m) $ ModuleToSource dict m2s
+  setTCLens stFileDict dict'
+  put m2s'
+  return r
+
 -- | A variant of 'findFile'' which does not require 'TCM'.
 
-findFile''
-  :: [AbsolutePath]
-  -- ^ Include paths.
+findFile'' ::
+     List1 AbsolutePath
+       -- ^ Include paths.
   -> TopLevelModuleName
-  -> ModuleToSource
-  -- ^ Cached invocations of 'findFile'''. An updated copy is returned.
-  -> IO (Either FindError SourceFile, ModuleToSource)
-findFile'' dirs m modFile =
-  case Map.lookup m modFile of
-    Just f  -> return (Right (SourceFile f), modFile)
+  -> StateT ModuleToSource IO (Either FindError SourceFile)
+findFile'' dirs m = do
+  ModuleToSource dict modToSrc <- get
+  case Map.lookup m modToSrc of
+    Just sf -> return $ Right sf
     Nothing -> do
-      files          <- fileList acceptableFileExts
-      filesShortList <- fileList parseFileExtsShortList
-      existingFiles  <-
-        liftIO $ filterM (doesFileExistCaseSensitive . filePath . srcFilePath) files
-      return $ case nubOn id existingFiles of
-        []     -> (Left (NotFound filesShortList), modFile)
-        [file] -> (Right file, Map.insert m (srcFilePath file) modFile)
-        files  -> (Left (Ambiguous existingFiles), modFile)
+      files          <- liftIO $ fileList agdaFileExtensions
+      existingFiles  <- liftIO $ filterM (doesFileExistCaseSensitive . filePath) files
+      case nubOn id existingFiles of
+        [file]   -> do
+          let (i, dict') = registerFileIdWithBuiltin file dict
+          let src = SourceFile i
+          put $ ModuleToSource dict' $ Map.insert m src modToSrc
+          return (Right src)
+        []       -> do
+          filesShortList <- liftIO $ fileList $ List2.toList parseFileExtsShortList
+          return (Left (NotFound filesShortList))
+        f0:f1:fs -> return (Left (Ambiguous $ List2 f0 f1 fs))
   where
-    fileList exts = mapM (fmap SourceFile . absolute)
+    fileList exts = mapM absolute
                     [ filePath dir </> file
-                    | dir  <- dirs
+                    | dir  <- List1.toList dirs
                     , file <- map (moduleNameToFileName m) exts
                     ]
 
@@ -195,11 +205,12 @@ findFile'' dirs m modFile =
 --
 -- Raises 'Nothing' if the interface file cannot be found.
 
-findInterfaceFile'
-  :: SourceFile                 -- ^ Path to the source file
+findInterfaceFile' :: HasCallStack  -- We are calling partial function toIFile, thus want a call stack.
+  => SourceFile                 -- ^ Path to the source file
   -> TCM (Maybe InterfaceFile)  -- ^ Maybe path to the interface file
 findInterfaceFile' fp = liftIO . mkInterfaceFile =<< toIFile fp
 
+
 -- | Finds the interface file corresponding to a given top-level
 -- module file. The returned paths are absolute.
 --
@@ -207,35 +218,37 @@ findInterfaceFile' fp = liftIO . mkInterfaceFile =<< toIFile fp
 -- 'Nothing' if the source file can be found but not the interface
 -- file.
 
-findInterfaceFile :: TopLevelModuleName -> TCM (Maybe InterfaceFile)
+findInterfaceFile :: HasCallStack  -- because of calling a partial function
+  => TopLevelModuleName -> TCM (Maybe InterfaceFile)
 findInterfaceFile m = findInterfaceFile' =<< findFile m
 
 -- | Ensures that the module name matches the file name. The file
 -- corresponding to the module name (according to the include path)
 -- has to be the same as the given file name.
 
-checkModuleName
-  :: TopLevelModuleName
-     -- ^ The name of the module.
+checkModuleName ::
+     TopLevelModuleName
+       -- ^ The name of the module.
   -> SourceFile
-     -- ^ The file from which it was loaded.
+       -- ^ The file from which it was loaded.
   -> Maybe TopLevelModuleName
-     -- ^ The expected name, coming from an import statement.
+       -- ^ The expected name, coming from an import statement.
   -> TCM ()
-checkModuleName name (SourceFile file) mexpected = do
+checkModuleName name src0 mexpected = do
+  file <- srcFilePath src0
   findFile' name >>= \case
 
     Left (NotFound files)  -> typeError $
       case mexpected of
-        Nothing -> ModuleNameDoesntMatchFileName name (map srcFilePath files)
+        Nothing -> ModuleNameDoesntMatchFileName name files
         Just expected -> ModuleNameUnexpected name expected
 
     Left (Ambiguous files) -> typeError $
-      AmbiguousTopLevelModuleName name (map srcFilePath files)
+      AmbiguousTopLevelModuleName name files
 
     Right src -> do
-      let file' = srcFilePath src
-      file <- liftIO $ absolute (filePath file)
+      file' <- srcFilePath src
+      file  <- liftIO $ absolute $ filePath file
       unlessM (liftIO $ sameFile file file') $
         typeError $ ModuleDefinedInOtherFile name file file'
 
@@ -243,7 +256,7 @@ checkModuleName name (SourceFile file) mexpected = do
   -- that we do not end up with a mismatch between expected
   -- and actual module name.
 
-  forM_ mexpected $ \ expected ->
+  forM_ mexpected \ expected ->
     unless (name == expected) $
       typeError $ OverlappingProjects file name expected
       -- OverlappingProjects is the correct error for
@@ -251,51 +264,21 @@ checkModuleName name (SourceFile file) mexpected = do
       -- -- typeError $ ModuleNameUnexpected name expected
 
 
--- | Computes the module name of the top-level module in the given
--- file.
---
--- If no top-level module name is given, then an attempt is made to
--- use the file name as a module name.
+parseFileExtsShortList :: List2 String
+parseFileExtsShortList = List2.cons ".agda" literateExtsShortList
 
--- TODO: Perhaps it makes sense to move this procedure to some other
--- module.
+-- | Remove an Agda file extension from a filepath, if possible.
+stripAgdaExtension :: FilePath -> Maybe FilePath
+stripAgdaExtension = stripAnyOfExtensions agdaFileExtensions
 
-moduleName
-  :: AbsolutePath
-     -- ^ The path to the file.
-  -> Module
-     -- ^ The parsed module.
-  -> TCM TopLevelModuleName
-moduleName file parsedModule = billTo [Bench.ModuleName] $ do
-  let defaultName = rootNameModule file
-      raw         = rawTopLevelModuleNameForModule parsedModule
-  topLevelModuleName =<< if isNoName raw
-    then do
-      m <- runPM (fst <$> parse moduleNameParser defaultName)
-             `catchError` \_ ->
-           typeError $ GenericError $
-             "The file name " ++ prettyShow file ++
-             " is invalid because it does not correspond to a valid module name."
-      case m of
-        Qual {} ->
-          typeError $ GenericError $
-            "The file name " ++ prettyShow file ++ " is invalid because " ++
-            defaultName ++ " is not an unqualified module name."
-        QName {} ->
-          return $ RawTopLevelModuleName
-            { rawModuleNameRange = getRange m
-            , rawModuleNameParts = singleton (T.pack defaultName)
-            }
-    else return raw
+-- | Check if a file path has an Agda extension.
+hasAgdaExtension :: FilePath -> Bool
+hasAgdaExtension = isJust . stripAgdaExtension
 
-parseFileExtsShortList :: [String]
-parseFileExtsShortList = ".agda" : literateExtsShortList
+-- | Remove an existing Agda file extension from a file path.
+dropAgdaExtension :: FilePath -> FilePath
+dropAgdaExtension = fromMaybe __IMPOSSIBLE__ . stripAgdaExtension
 
-dropAgdaExtension :: String -> String
-dropAgdaExtension s = case catMaybes [ stripSuffix ext s
-                                     | ext <- acceptableFileExts ] of
-    [name] -> name
-    _      -> __IMPOSSIBLE__
 
 rootNameModule :: AbsolutePath -> String
 rootNameModule = dropAgdaExtension . snd . splitFileName . filePath
