@@ -1,4 +1,3 @@
-{-# LANGUAGE GADTs #-}
 -- | Generate an import dependency graph for a given module.
 
 module Agda.Interaction.Highlighting.Dot.Base
@@ -9,7 +8,6 @@ module Agda.Interaction.Highlighting.Dot.Base
 
 import Control.Monad.IO.Class
 
-import qualified Data.Map.Strict as M
 import qualified Data.Set as S
 
 import qualified Data.Text.Lazy as L
