@@ -13,7 +13,6 @@ module Agda.Interaction.Highlighting.FromAbstract
 import Prelude hiding (null)
 
 import Control.Applicative
-import Control.Monad         ( (<=<) )
 import Control.Monad.Reader  ( MonadReader(..), asks, Reader, runReader )
 
 import qualified Data.Map      as Map
@@ -28,7 +27,7 @@ import           Agda.Interaction.Highlighting.Range   ( rToR )  -- Range is amb
 import           Agda.Syntax.Abstract                ( IsProjP(..) )
 import qualified Agda.Syntax.Abstract      as A
 import           Agda.Syntax.Common        as Common
-import           Agda.Syntax.Concrete                ( FieldAssignment'(..) )
+import           Agda.Syntax.Concrete                ( FieldAssignment'(..), TacticAttribute' )
 import qualified Agda.Syntax.Concrete.Name as C
 import           Agda.Syntax.Info                    ( ModuleInfo(..) )
 import           Agda.Syntax.Literal
@@ -40,7 +39,6 @@ import           Agda.Syntax.TopLevelModuleName
 import Agda.TypeChecking.Monad
   hiding (ModuleInfo, MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 
-import           Agda.Utils.FileName
 import           Agda.Utils.Function
 import           Agda.Utils.Functor
 import           Agda.Utils.List                     ( initLast1 )
@@ -48,7 +46,7 @@ import           Agda.Utils.List1                    ( List1 )
 import qualified Agda.Utils.List1          as List1
 import           Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict   as Strict
-import           Agda.Utils.Pretty
+import           Agda.Syntax.Common.Pretty
 import           Agda.Utils.Singleton
 import           Agda.Utils.Size
 
@@ -103,6 +101,8 @@ class Hilite a where
 instance Hilite a => Hilite [a]
 instance Hilite a => Hilite (List1 a)
 instance Hilite a => Hilite (Maybe a)
+instance Hilite a => Hilite (Ranged a)
+instance Hilite a => Hilite (TacticAttribute' a)
 instance Hilite a => Hilite (WithHiding a)
 
 instance Hilite Void where
@@ -160,7 +160,6 @@ instance (Hilite a, Hilite b) => Hilite (a, b) where
 -- | A.LetBinding
 -- ---------------------------------------------------------------------
 -- | getLet                | A.LetBind                   | bound
--- | getLet                | A.LetDeclaredVariable       | bound
 -- ---------------------------------------------------------------------
 -- | A.LamBinding
 -- ---------------------------------------------------------------------
@@ -196,6 +195,11 @@ instance (Hilite a, Hilite b) => Hilite (a, b) where
 instance Hilite A.RecordDirectives where
   hilite (RecordDirectives _ _ _ c) = hilite c
 
+instance Hilite A.RecordConName where
+  hilite = \case
+    A.NamedRecCon x -> hilite x
+    A.FreshRecCon{} -> mempty
+
 instance Hilite A.Declaration where
   hilite = \case
       A.Axiom _ax _di ai _occ x e            -> hl ai <> hl x <> hl e
@@ -203,14 +207,15 @@ instance Hilite A.Declaration where
       A.Field _di x e                        -> hlField x <> hl e
       A.Primitive _di x e                    -> hl x <> hl e
       A.Mutual _mi ds                        -> hl ds
-      A.Section _r x tel ds                  -> hl x <> hl tel <> hl ds
-      A.Apply mi x a _ci dir                 -> hl mi <> hl x <> hl a <> hl dir
+      A.Section _r er x tel ds               -> hl er <> hl x <> hl tel <> hl ds
+      A.Apply mi er x a _ci dir              -> hl mi <> hl er <> hl x <>
+                                                hl a <> hl dir
       A.Import mi x dir                      -> hl mi <> hl x <> hl dir
       A.Open mi x dir                        -> hl mi <> hl x <> hl dir
-      A.FunDef _di x _delayed cs             -> hl x <> hl cs
-      A.DataSig _di x tel e                  -> hl x <> hl tel <> hl e
+      A.FunDef _di x cs                      -> hl x <> hl cs
+      A.DataSig _di er x tel e               -> hl er <> hl x <> hl tel <> hl e
       A.DataDef _di x _uc pars cs            -> hl x <> hl pars <> hl cs
-      A.RecSig _di x tel e                   -> hl x <> hl tel <> hl e
+      A.RecSig _di er x tel e                -> hl er <> hl x <> hl tel <> hl e
       A.RecDef _di x _uc dir bs e ds         -> hl x <> hl dir <> hl bs <> hl e <> hl ds
       A.PatternSynDef x xs p                 -> hl x <> hl xs <> hl p
       A.UnquoteDecl _mi _di xs e             -> hl xs <> hl e
@@ -218,23 +223,26 @@ instance Hilite A.Declaration where
       A.UnquoteData _i xs _uc _j cs e        -> hl xs <> hl cs <> hl e
       A.ScopedDecl s ds                      -> hl ds
       A.Pragma _r pragma                     -> hl pragma
+      A.UnfoldingDecl _r names               -> hl names
     where
     hl      a = hilite a
     hlField x = hiliteField (concreteQualifier x) (concreteBase x) (Just $ bindingSite x)
 
 instance Hilite A.Pragma where
   hilite = \case
-    A.OptionsPragma _strings     -> mempty
-    A.BuiltinPragma b x          -> singleAspect Keyword b <> hilite x
-    A.BuiltinNoDefPragma b k x   -> singleAspect Keyword b <> hiliteQName (Just $ kindOfNameToNameKind k) x
-    A.CompilePragma b x _foreign -> singleAspect Keyword b <> hilite x
-    A.RewritePragma r xs         -> singleAspect Keyword r <> hilite xs
-    A.StaticPragma x             -> hilite x
-    A.EtaPragma x                -> hilite x
-    A.InjectivePragma x          -> hilite x
-    A.NotProjectionLikePragma x  -> hilite x
-    A.InlinePragma _inline x     -> hilite x
-    A.DisplayPragma x ps e       -> hilite x <> hilite ps <> hilite e
+    A.OptionsPragma _strings        -> mempty
+    A.BuiltinPragma b x             -> singleAspect Keyword b <> hilite x
+    A.BuiltinNoDefPragma b k x      -> singleAspect Keyword b <> hiliteQName (Just $ kindOfNameToNameKind k) x
+    A.CompilePragma b x _foreign    -> singleAspect Keyword b <> hilite x
+    A.RewritePragma r xs            -> singleAspect Keyword r <> hilite xs
+    A.StaticPragma x                -> hilite x
+    A.EtaPragma x                   -> hilite x
+    A.InjectivePragma x             -> hilite x
+    A.InjectiveForInferencePragma x -> hilite x
+    A.NotProjectionLikePragma x     -> hilite x
+    A.OverlapPragma x _             -> hilite x
+    A.InlinePragma _inline x        -> hilite x
+    A.DisplayPragma x ps e          -> hilite x <> hilite ps <> hilite e
 
 instance Hilite A.Expr where
   hilite = \case
@@ -258,8 +266,8 @@ instance Hilite A.Expr where
       A.Generalized _qs e           -> hl e
       A.Fun _r a b                  -> hl a <> hl b
       A.Let _r bs e                 -> hl bs <> hl e
-      A.Rec _r ass                  -> hl ass
-      A.RecUpdate _r e ass          -> hl e <> hl ass
+      A.Rec _kwr _r ass             -> hl ass
+      A.RecUpdate _kwr _r e ass     -> hl e <> hl ass
       A.ScopedExpr _ e              -> hl e
       A.Quote _r                    -> mempty
       A.QuoteTerm _r                -> mempty
@@ -283,10 +291,9 @@ instance (Hilite a, IsProjP a) => Hilite (A.Pattern' a) where
       A.AbsurdP _r           -> mempty
       A.LitP _r l            -> hl l
       A.PatternSynP _r qs es -> hilitePatternSynonym qs <> hl es
-      A.RecP _r ps           -> hl ps
+      A.RecP _kwr _r ps      -> hl ps
       A.EqualP _r ps         -> hl ps
       A.WithP _ p            -> hl p
-      A.AnnP _r a p          -> hl p
 
     where
     hl a = hilite a
@@ -326,6 +333,7 @@ instance (HasRange n, Hilite p, Hilite e) => Hilite (RewriteEqn' x n p e) where
   hilite = \case
     Rewrite es    -> hilite $ fmap snd es
     Invert _x pes -> hilite pes
+    LeftLet pes   -> hilite pes
 
 instance Hilite a => Hilite (A.Clause' a) where
   hilite (A.Clause lhs strippedPats rhs wh _catchall) =
@@ -351,10 +359,11 @@ instance Hilite A.ModuleApplication where
 instance Hilite A.LetBinding where
   hilite = \case
       A.LetBind    _r ai x t e     -> hl ai <> hl x <> hl t <> hl e
+      A.LetAxiom   _r ai x t       -> hl ai <> hl x <> hl t
       A.LetPatBind _r p e          -> hl p  <> hl e
-      A.LetApply   mi x es _ci dir -> hl mi <> hl x <> hl es <> hl dir
-      A.LetOpen    mi x dir        -> hl mi <> hl x <> hl dir
-      A.LetDeclaredVariable x      -> hl x
+      A.LetApply mi er x es _c dir -> hl mi <> hl er <> hl x <>
+                                      hl es <> hl dir
+      A.LetOpen mi x dir           -> hl mi <> hl x <> hl dir
     where
     hl x = hilite x
 
@@ -372,7 +381,7 @@ instance Hilite A.LamBinding where
     A.DomainFull bind      -> hilite bind
 
 instance Hilite a => Hilite (A.Binder' a) where
-  hilite (A.Binder p x) = hilite p <> hilite x
+  hilite (A.Binder p _ x) = hilite p <> hilite x
 
 instance Hilite A.BindName where
   hilite (A.BindName x) = hiliteBound x
@@ -392,7 +401,7 @@ instance Hilite ArgInfo where
   hilite (ArgInfo _hiding modality _origin _fv _a) = hilite modality
 
 instance Hilite Modality where
-  hilite (Modality _relevance quantity _cohesion) = hilite quantity
+  hilite (Modality _relevance quantity _cohesion _polarity) = hilite quantity
 
 -- | If the 'Quantity' attribute comes with a 'Range', highlight the
 -- corresponding attribute as 'Symbol'.
