@@ -1,18 +1,15 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE ViewPatterns #-}
 
 -- | Function for generating highlighted and aligned LaTeX from literate
 -- Agda source.
 
 module Agda.Interaction.Highlighting.LaTeX.Base
-  ( generateLaTeXIO
+  ( LaTeXOptions(..)
+  , generateLaTeXIO
   , prepareCommonAssets
-  , runLogLaTeXTWith
-  , logMsgToText
+  , MonadLogLaTeX(logLaTeX)
   , LogMessage(..)
-  , MonadLogLaTeX
-  , LogLaTeXT
-  , LaTeXOptions(..)
+  , logMsgToText
   ) where
 
 import Prelude hiding (log)
@@ -22,13 +19,8 @@ import Data.Char
 import Data.Maybe
 import Data.Function (on)
 import Data.Foldable (toList)
-#if !(MIN_VERSION_base(4,11,0))
-import Data.Semigroup (Semigroup(..))
-#endif
-import qualified Data.Text as T
 
 import Control.Exception.Base (IOException, try)
-import Control.Monad (forM_, mapM_, unless, when)
 import Control.Monad.Trans.Reader as R ( ReaderT(runReaderT))
 import Control.Monad.RWS.Strict
   ( RWST(runRWST)
@@ -41,9 +33,8 @@ import Control.Monad.IO.Class
   )
 
 import System.Directory
-import System.Exit
 import System.FilePath
-import System.Process
+import System.Process ( readProcess )
 
 import Data.Text (Text)
 import qualified Data.Text               as T
@@ -59,12 +50,10 @@ import qualified Data.HashSet as Set
 import qualified Data.IntMap  as IntMap
 import qualified Data.List    as List
 
-import Paths_Agda
-
 import Agda.Syntax.Common
 import Agda.Syntax.Parser.Literate (literateTeX, LayerRole, atomizeLayers)
 import qualified Agda.Syntax.Parser.Literate as L
-import Agda.Syntax.Position (RangeFile, startPos)
+import Agda.Syntax.Position (RangeFile, startPos')
 import Agda.Syntax.TopLevelModuleName
   (TopLevelModuleName, moduleNameParts)
 
@@ -72,6 +61,8 @@ import Agda.Interaction.Highlighting.Precise hiding (toList)
 
 import Agda.TypeChecking.Monad (Interface(..)) --, reportSLn)
 
+import Agda.Setup ( getDataFileName )
+
 import Agda.Utils.Function (applyWhen)
 import Agda.Utils.Functor  ((<&>))
 import Agda.Utils.List     (last1, updateHead, updateLast)
@@ -84,9 +75,6 @@ import Agda.Utils.Impossible
 ------------------------------------------------------------------------
 -- * Logging
 
-class Monad m => MonadLogLaTeX m where
-  logLaTeX :: LogMessage -> m ()
-
 -- | Log LaTeX messages using a provided action.
 --
 -- This could be accomplished by putting logs into the RWST output and splitting it
@@ -95,37 +83,32 @@ class Monad m => MonadLogLaTeX m where
 --
 -- We want the logging to be reasonably polymorphic, avoid space leaks that can occur
 -- with WriterT, and also be usable during outer phases such as directory preparation.
---
--- I'm not certain this is the best way to do it, but whatever.
-type LogLaTeXT m = ReaderT (LogMessage -> m ()) m
 
-instance Monad m => MonadLogLaTeX (LogLaTeXT m) where
-  logLaTeX message = do
-    doLog <- ask
-    lift $ doLog message
-
-runLogLaTeXTWith :: Monad m => (LogMessage -> m ()) -> LogLaTeXT m a -> m a
-runLogLaTeXTWith = flip runReaderT
-
--- Not currently used, but for example:
--- runLogLaTeXWithIO :: MonadIO m => LogLaTeXT m a -> m a
--- runLogLaTeXWithIO = runLogLaTeXTWith $ liftIO . T.putStrLn . logMsgToText
-
-------------------------------------------------------------------------
--- * Datatypes.
-
--- | The @LaTeX@ monad is a combination of @ExceptT@, @RWST@ and
--- a logger @m@. The error part is just used to keep track whether we finished or
--- not, the reader part contains static options used, the writer is where the
--- output goes and the state is for keeping track of the tokens and some
--- other useful info, and the MonadLogLaTeX part is used for printing debugging info.
-
-type LaTeX a = forall m. MonadLogLaTeX m => RWST Env [Output] State m a
-
--- | Output items.
+class Monad m => MonadLogLaTeX m where
+  logLaTeX :: LogMessage -> m ()
 
 data LogMessage = LogMessage Debug Text [Text] deriving Show
 
+------------------------------------------------------------------------
+-- * The monad and its associated data types.
+
+-- | The @LaTeX@ monad is a combination of @RWST@ and a logger @m@.
+--
+-- The reader part contains static options used,
+-- the writer is where the output goes,
+-- the state is for keeping track of the tokens and some other useful info, and
+-- the MonadLogLaTeX part is used for printing debugging info.
+
+type LaTeXT = RWST Env [Output] State
+type LaTeX a = forall m. MonadLogLaTeX m => LaTeXT m a
+-- Andreas, 2025-03-23 we sometimes expand @a -> LaTeX b@
+-- to @forall m. MonadLogLaTeX m => a -> LaTeXT m b@
+-- to combat changes in the type checker of GHC 9 over GHC 8.
+-- Originally (by asr, 2021-02-07) we used eta-expansions in these places
+-- to circumvent the GHC type checker regressions (see Issue #4955).
+
+-- | Output items.
+
 data Output
   = Text !Text
     -- ^ A piece of text.
@@ -209,10 +192,8 @@ data Debug = MoveColumn | NonCode | Code | Spaces | Output | FileSystem
 
 -- | Run function for the @LaTeX@ monad.
 runLaTeX :: MonadLogLaTeX m =>
-  LaTeX a -> Env -> State -> m (a, State, [Output])
--- ASR (2021-02-07). The eta-expansion is required by GHC >= 9.0.1
--- (see Issue #4955).
-runLaTeX l = runRWST l
+  LaTeXT m a -> Env -> State -> m (a, State, [Output])
+runLaTeX = runRWST
 
 emptyState :: State
 emptyState = State
@@ -261,12 +242,9 @@ replaceSpaces = T.map (\c -> if isSpaceNotNewline c then ' ' else c)
 --   by the length of the token. Otherwise, `moveColumnForToken` is a no-op.
 moveColumnForToken :: Token -> LaTeX ()
 moveColumnForToken Token{ text = t } = do
-  unless (isSpaces t) $ do
+  unless (isSpaces t) do
     log MoveColumn t
-    -- ASR (2021-02-07). The eta-expansion is required by GHC >= 9.0.1
-    -- (see Issue #4955).
-    n <- size t
-    moveColumn n
+    moveColumn =<< size t
 
 -- | Merges 'columns' into 'columnsPrev', resets 'column' and
 -- 'columns'
@@ -284,8 +262,8 @@ resetColumn = modify $ \s ->
     where
     leastNew = columnColumn (last1 n ns)
 
-moveColumn :: Int -> LaTeX ()
-moveColumn i = modify $ \s -> s { column = i + column s }
+moveColumn :: MonadLogLaTeX m => Int -> LaTeXT m ()
+moveColumn i = modify \ s -> s { column = i + column s }
 
 -- | Registers a column of the given kind. The column is returned.
 
@@ -368,7 +346,7 @@ log Code text = do
   logHelper Code text ["columns=", tshow cols, "col=", tshow col]
 log debug text = logHelper debug text []
 
-output :: Output -> LaTeX ()
+output :: MonadLogLaTeX m => Output -> LaTeXT m ()
 output item = do
   log Output $ tshow item
   tell [item]
@@ -454,31 +432,26 @@ cmdArg x = "{" <> x <> "}"
 ------------------------------------------------------------------------
 -- * Output generation from a stream of labelled tokens.
 
-processLayers :: [(LayerRole, Tokens)] -> LaTeX ()
--- ASR (2021-02-07). The eta-expansion on @lt@ is required by GHC >=
--- 9.0.1 (see Issue #4955).
-processLayers lt = forM_ lt $ \ (layerRole,toks) -> do
+processLayers :: MonadLogLaTeX m => [(LayerRole, Tokens)] -> LaTeXT m ()
+processLayers = mapM_ \ (layerRole, toks) -> do
   case layerRole of
     L.Markup  -> processMarkup  toks
     L.Comment -> processComment toks
     L.Code    -> processCode    toks
 
 -- | Deals with markup, which is output verbatim.
-processMarkup :: Tokens -> LaTeX ()
--- ASR (2021-02-07). The eta-expansion on @ts@ is required by GHC >=
--- 9.0.1 (see Issue #4955).
-processMarkup ts = forM_ ts $ \ t' -> do
-  moveColumnForToken t'
-  output (Text (text t'))
+processMarkup :: MonadLogLaTeX m => Tokens -> LaTeXT m ()
+processMarkup = mapM_ \ t -> do
+  moveColumnForToken t
+  output $ Text $ text t
 
 -- | Deals with literate text, which is output verbatim
-processComment :: Tokens -> LaTeX ()
--- ASR (2021-02-07). The eta-expansion with @ts@ is required by GHC >=
--- 9.0.1 (see Issue #4955).
-processComment ts = forM_ ts $ \ t' -> do
-  unless ("%" == T.take 1 (T.stripStart (text t'))) $ do
-    moveColumnForToken t'
-  output (Text (text t'))
+processComment :: MonadLogLaTeX m => Tokens -> LaTeXT m ()
+processComment = mapM_ \ t -> do
+  let t' = text t
+  unless ("%" == T.take 1 (T.stripStart t')) $ do
+    moveColumnForToken t
+  output $ Text t'
 
 -- | Deals with code blocks. Every token, except spaces, is pretty
 -- printed as a LaTeX command.
@@ -518,9 +491,7 @@ processCode toks' = do
     ptOpenWhenColumnZero col =
         when (col == 0) $ do
           registerColumnZero
-          -- ASR (2021-02-07). The eta-expansion @\o -> output o@ is
-          -- required by GHC >= 9.0.1 (see Issue #4955).
-          (\o -> output o). Text . ptOpen =<< columnZero
+          output . Text . ptOpen =<< columnZero
 
     -- Translation from OtherAspect to command strings. So far it happens
     -- to correspond to @show@ but it does not have to (cf. fromAspect)
@@ -602,10 +573,8 @@ spaces [] = return ()
 -- Newlines.
 spaces (s@(T.uncons -> Just ('\n', _)) : ss) = do
   col <- gets column
-  when (col == 0) $
-    -- ASR (2021-02-07). The eta-expansion @\o -> output o@ is
-    -- required by GHC >= 9.0.1 (see Issue #4955).
-    (\o -> output o) . Text . ptOpen =<< columnZero
+  when (col == 0) do
+    output . Text . ptOpen =<< columnZero
   output $ Text $ ptClose <> ptNL <>
                   T.replicate (T.length s - 1) ptEmptyLine
   resetColumn
@@ -774,15 +743,6 @@ latexOutRelativeFilePath m =
     (map T.unpack $ List1.toList $ moduleNameParts m) <.>
   "tex"
 
-groupByFst :: forall a b. Eq a => [(a,b)] -> [(a,[b])]
-groupByFst =
-    map (\xs -> case xs of                     -- Float the grouping to the top level
-          []           -> __IMPOSSIBLE__
-          (tag, _): _ -> (tag, map snd xs))
-
-  . List.groupBy ((==) `on` fst)  -- Group together characters in the same
-                                  -- role.
-
 -- | Transforms the source code into LaTeX.
 toLaTeX
   :: MonadLogLaTeX m
@@ -828,22 +788,25 @@ toLaTeX env path source hi =
                 ( -- Split tokens at newlines
                   concatMap stringLiteral
                 . concatMap multiLineComment
-                . map ( \(mi, cs) ->
-                              Token {text = T.pack cs, info = fromMaybe mempty mi}
+                . List1.toList
+                . fmap (\ (mi, cs) -> Token
+                        { text = T.pack $ List1.toList cs
+                        , info = fromMaybe mempty mi
+                        }
                       )
-                . groupByFst
+                . List1.groupByFst1
                 )
             )
       )
-    . groupByFst
+    . List1.groupByFst
 
-  -- Look up the meta info at each position in the highlighting info.
-  . zipWith (\pos (role, char) -> (role, (IntMap.lookup pos infoMap, char)))
-            [1..] . -- Add position in file to each character.
-              -- Map each character to its role
-              atomizeLayers . literateTeX (startPos path)
-
-  $ L.unpack source
+    -- Look up the meta info at each position in the highlighting info.
+    . zipWith (\pos (role, char) -> (role, (IntMap.lookup pos infoMap, char)))
+              [1..]
+    -- Map each character to its role
+    . atomizeLayers
+    . literateTeX (startPos' ())
+    $ L.unpack source
   where
   infoMap = toMap hi
 
