@@ -13,33 +13,30 @@ import Control.Concurrent
 import Control.Concurrent.Async
 import Control.Concurrent.STM.TChan
 import Control.Concurrent.STM.TVar
-
 import qualified Control.Exception  as E
 
-import Control.Monad
 import Control.Monad.Except         ( MonadError(..), ExceptT(..), runExceptT )
 import Control.Monad.IO.Class       ( MonadIO(..) )
-import Control.Monad.Fail           ( MonadFail )
 import Control.Monad.State          ( MonadState(..), gets, modify, runStateT )
 import Control.Monad.STM
-import Control.Monad.Trans          ( lift )
+import Control.Monad.State          ( StateT )
 
 import qualified Data.Char as Char
 import Data.Function (on)
 import qualified Data.List as List
 import qualified Data.Map as Map
+import qualified Data.Set as Set
 import Data.Maybe
 
 import System.Directory
 import System.FilePath
 
 import Agda.TypeChecking.Monad as TCM
-  hiding (initState, setCommandLineOptions)
-import qualified Agda.TypeChecking.Monad as TCM
 import qualified Agda.TypeChecking.Pretty as TCP
-import Agda.TypeChecking.Rules.Term (checkExpr, isType_)
 import Agda.TypeChecking.Errors
-import Agda.TypeChecking.Warnings (runPM, warning)
+import Agda.TypeChecking.Opacity (saturateOpaqueBlocks)
+import Agda.TypeChecking.Rules.Term (checkExpr, isType_)
+import Agda.TypeChecking.Warnings (warning)
 
 import Agda.Syntax.Fixity
 import Agda.Syntax.Position
@@ -51,12 +48,13 @@ import Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Pretty
 import Agda.Syntax.Info (mkDefInfo)
 import Agda.Syntax.Translation.ConcreteToAbstract
-import Agda.Syntax.Translation.AbstractToConcrete hiding (withScope)
+import Agda.Syntax.Translation.AbstractToConcrete
+import Agda.Syntax.Translation.InternalToAbstract (reify)
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.TopLevelModuleName
 
 import Agda.Interaction.Base
-import Agda.Interaction.ExitCode
+import Agda.Interaction.ExitCode (pattern TCMError, exitAgdaWith)
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
 import Agda.Interaction.Options.Lenses as Lenses
@@ -68,72 +66,34 @@ import qualified Agda.Interaction.BasicOps as B
 import Agda.Interaction.Highlighting.Precise hiding (Error, Postulate, singleton)
 import Agda.Interaction.Imports  ( Mode, pattern ScopeCheck, pattern TypeCheck )
 import qualified Agda.Interaction.Imports as Imp
+import Agda.Interaction.Command
+  (CommandM, liftLocalState, revLift, revLiftTC, localStateCommandM)
 import Agda.Interaction.Highlighting.Generate
 
 import Agda.Compiler.Backend
 
-import Agda.Auto.Auto as Auto
+import Agda.Mimer.Mimer as Mimer
+import qualified Control.DeepSeq as DeepSeq
 
 import Agda.Utils.Either
 import Agda.Utils.FileName
 import Agda.Utils.Function
 import Agda.Utils.Hash
+import Agda.Utils.IO (showIOException)
 import Agda.Utils.Lens
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Pretty hiding (Mode)
+import Agda.Syntax.Common.Pretty hiding (Mode)
 import qualified Agda.Utils.ProfileOptions as Profile
 import Agda.Utils.Singleton
 import Agda.Utils.String
 import Agda.Utils.Time
 import Agda.Utils.Tuple
+import Agda.Utils.WithDefault (lensCollapseDefault, lensKeepDefault)
 
 import Agda.Utils.Impossible
 
-------------------------------------------------------------------------
--- The CommandM monad
-
--- | Restore both 'TCState' and 'CommandState'.
-
-localStateCommandM :: CommandM a -> CommandM a
-localStateCommandM m = do
-  cSt <- get
-  tcSt <- getTC
-  x <- m
-  putTC tcSt
-  put cSt
-  return x
-
--- | Restore 'TCState', do not touch 'CommandState'.
-
-liftLocalState :: TCM a -> CommandM a
-liftLocalState = lift . localTCState
-
--- | Build an opposite action to 'lift' for state monads.
-
-revLift
-    :: MonadState st m
-    => (forall c . m c -> st -> k (c, st))      -- ^ run
-    -> (forall b . k b -> m b)                  -- ^ lift
-    -> (forall x . (m a -> k x) -> k x) -> m a  -- ^ reverse lift in double negative position
-revLift run lift' f = do
-    st <- get
-    (a, st') <- lift' $ f (`run` st)
-    put st'
-    return a
-
-revLiftTC
-    :: MonadTCState m
-    => (forall c . m c -> TCState -> k (c, TCState))  -- ^ run
-    -> (forall b . k b -> m b)                        -- ^ lift
-    -> (forall x . (m a -> k x) -> k x) -> m a        -- ^ reverse lift in double negative position
-revLiftTC run lift' f = do
-    st <- getTC
-    (a, st') <- lift' $ f (`run` st)
-    putTC st'
-    return a
-
 -- | Opposite of 'liftIO' for 'CommandM'.
 --
 -- This function should only be applied to computations that are
@@ -186,7 +146,7 @@ getOldInteractionScope :: InteractionId -> CommandM ScopeInfo
 getOldInteractionScope ii = do
   ms <- gets $ Map.lookup ii . oldInteractionScopes
   case ms of
-    Nothing    -> fail $ "not an old interaction point: " ++ show ii
+    Nothing    -> __IMPOSSIBLE_VERBOSE__ $ "not an old interaction point: " ++ show ii
     Just scope -> return scope
 
 -- | Do setup and error handling for a command.
@@ -233,16 +193,18 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
     -- AsyncCancelled, which is used to abort Agda.
     handleNastyErrors :: CommandM () -> CommandM ()
     handleNastyErrors m = commandMToIO $ \ toIO -> do
-      let handle e =
-            Right <$>
-              toIO (handleErr (Just Direct) $
-                        Exception noRange $ text $ E.displayException e)
 
-          asyncHandler e@AsyncCancelled = return (Left e)
+      let asyncHandler e@AsyncCancelled = return (Left e)
 
-          generalHandler (e :: E.SomeException) = handle e
+          ioHandler (e :: E.IOException) = Right <$> do
+            toIO $ handleErr (Just Direct) $ IOException Nothing noRange e
 
-      r <- ((Right <$> toIO m) `E.catch` asyncHandler)
+          generalHandler (e :: E.SomeException) = Right <$> do
+            toIO $ handleErr (Just Direct) $ GenericException $ showIOException e
+
+      r <- (Right <$> toIO m)
+             `E.catch` asyncHandler
+             `E.catch` ioHandler
              `E.catch` generalHandler
       case r of
         Right x -> return x
@@ -252,6 +214,11 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
     -- error. Because this function may switch the focus to another file
     -- the status information is also updated.
     handleErr method e = do
+
+      -- TODO: make a better predicate for this
+      noError <- lift $ null <$> renderError e
+      unless noError do
+
         unsolved <- lift $ computeUnsolvedInfo
         err     <- lift $ errorHighlighting e
         modFile <- lift $ useTC stModuleToSource
@@ -261,12 +228,9 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
         let info = convert $ err <> unsolved
                      -- Errors take precedence over unsolved things.
 
-        -- TODO: make a better predicate for this
-        noError <- lift $ null <$> prettyError e
-
         showImpl <- lift $ optShowImplicit <$> useTC stPragmaOptions
         showIrr <- lift $ optShowIrrelevant <$> useTC stPragmaOptions
-        unless noError $ do
+        do
           mapM_ putResponse $
             [ Resp_DisplayInfo $ Info_Error $ Info_GenericError e ] ++
             tellEmacsToJumpToError (getRange e) ++
@@ -363,15 +327,8 @@ maybeAbort m = do
             _ -> return (Command (Just x))
         Right a -> do
           liftIO $ popAbortedCommands q a
-          putTC $ initState
-            { stPersistentState = stPersistentState tcState
-            , stPreScopeState   =
-                (stPreScopeState initState)
-                  { stPrePragmaOptions =
-                      stPrePragmaOptions
-                        (stPreScopeState tcState)
-                  }
-            }
+          putTC $ set lensPragmaOptions (tcState ^. lensPragmaOptions) $
+            initStateFromPersistentState $ stPersistentState tcState
           put $ (initCommandState (commandQueue commandState))
             { optionsOnReload = optionsOnReload commandState
             }
@@ -454,6 +411,7 @@ initialiseCommandQueue next = do
 
 independent :: Interaction -> Bool
 independent (Cmd_load {})                   = True
+independent Cmd_load_no_metas{}             = True
 independent (Cmd_compile {})                = True
 independent (Cmd_load_highlighting_info {}) = True
 independent Cmd_tokenHighlighting {}        = True
@@ -466,9 +424,11 @@ independent _                               = False
 updateInteractionPointsAfter :: Interaction -> Bool
 updateInteractionPointsAfter Cmd_load{}                          = True
 updateInteractionPointsAfter Cmd_compile{}                       = True
+updateInteractionPointsAfter Cmd_backend_top{}                   = True
+updateInteractionPointsAfter Cmd_backend_hole{}                  = True
 updateInteractionPointsAfter Cmd_constraints{}                   = False
 updateInteractionPointsAfter Cmd_metas{}                         = False
-updateInteractionPointsAfter Cmd_no_metas{}                      = False
+updateInteractionPointsAfter Cmd_load_no_metas{}                 = False
 updateInteractionPointsAfter Cmd_show_module_contents_toplevel{} = False
 updateInteractionPointsAfter Cmd_search_about_toplevel{}         = False
 updateInteractionPointsAfter Cmd_solveAll{}                      = True
@@ -507,6 +467,13 @@ updateInteractionPointsAfter Cmd_exit{}                          = False
 
 -- | Interpret an interaction
 
+getBackendName :: CompilerBackend -> BackendName
+getBackendName = \case
+  LaTeX -> "LaTeX"
+  QuickLaTeX -> "LaTeX"
+  OtherBackend "GHCNoMain" -> "GHC"
+  OtherBackend b -> b
+
 interpret :: Interaction -> CommandM ()
 
 interpret (Cmd_load m argv) =
@@ -516,32 +483,38 @@ interpret (Cmd_load m argv) =
 
 interpret (Cmd_compile backend file argv) =
   cmd_load' file argv allowUnsolved mode $ \ checkResult -> do
-    mw <- lift $ applyFlagsToTCWarnings $ crWarnings checkResult
-    case mw of
-      [] -> do
-        lift $ case backend of
-          LaTeX                    -> callBackend "LaTeX" IsMain checkResult
-          QuickLaTeX               -> callBackend "LaTeX" IsMain checkResult
-          OtherBackend "GHCNoMain" -> callBackend "GHC" NotMain checkResult   -- for backwards compatibility
-          OtherBackend b           -> callBackend b IsMain checkResult
+    ws <- lift $ applyFlagsToTCWarnings $ crWarnings checkResult
+    case null ws of
+      True -> do
+        lift $ callBackend (getBackendName backend) isMain checkResult
         display_info . Info_CompilationOk backend =<< lift B.getWarningsAndNonFatalErrors
-      w@(_:_) -> display_info $ Info_Error $ Info_CompilationError w
+      False -> display_info $ Info_Error $ Info_CompilationError ws
   where
   allowUnsolved = backend `elem` [LaTeX, QuickLaTeX]
   mode | QuickLaTeX <- backend = ScopeCheck
        | otherwise             = TypeCheck
+  isMain | OtherBackend "GHCNoMain" <- backend = NotMain
+         | otherwise                           = IsMain
+
+interpret (Cmd_backend_top backend cmd) =
+  callBackendInteractTop (getBackendName backend) cmd
+
+interpret (Cmd_backend_hole ii rng s backend cmd) =
+  callBackendInteractHole (getBackendName backend) cmd ii rng s
 
 interpret Cmd_constraints =
-    display_info . Info_Constraints =<< lift B.getConstraints
+  display_info . Info_Constraints =<< lift B.getConstraints
 
 interpret (Cmd_metas norm) = do
   ms <- lift $ B.getGoals' norm (max Simplified norm)
   display_info . Info_AllGoalsWarnings ms =<< lift B.getWarningsAndNonFatalErrors
 
-interpret Cmd_no_metas = do
-  metas <- getOpenMetas
-  unless (null metas) $
-    typeError $ GenericError "Unsolved meta-variables"
+interpret (Cmd_load_no_metas file) = do
+  -- Fail if there are open metas.
+  let allowMetas = False
+  cmd_load' file [] allowMetas TypeCheck $ \ result -> do
+    Imp.raiseNonFatalErrors result
+    unlessM (null <$> getOpenMetas) __IMPOSSIBLE__
 
 interpret (Cmd_show_module_contents_toplevel norm s) =
   atTopLevel $ showModuleContents norm noRange s
@@ -584,28 +557,22 @@ interpret (Cmd_compute_toplevel cmode s) = do
 interpret (ShowImplicitArgs showImpl) = do
   opts <- lift commandLineOptions
   setCommandLineOpts $
-    opts { optPragmaOptions =
-             (optPragmaOptions opts) { optShowImplicit = showImpl } }
+    set (lensPragmaOptions . lensOptShowImplicit . lensKeepDefault) showImpl opts
 
 interpret ToggleImplicitArgs = do
   opts <- lift commandLineOptions
-  let ps = optPragmaOptions opts
   setCommandLineOpts $
-    opts { optPragmaOptions =
-             ps { optShowImplicit = not $ optShowImplicit ps } }
+    over (lensPragmaOptions . lensOptShowImplicit . lensCollapseDefault) not opts
 
 interpret (ShowIrrelevantArgs showIrr) = do
   opts <- lift commandLineOptions
   setCommandLineOpts $
-    opts { optPragmaOptions =
-             (optPragmaOptions opts) { optShowIrrelevant = showIrr } }
+    set (lensPragmaOptions . lensOptShowIrrelevant . lensKeepDefault) showIrr opts
 
 interpret ToggleIrrelevantArgs = do
   opts <- lift commandLineOptions
-  let ps = optPragmaOptions opts
   setCommandLineOpts $
-    opts { optPragmaOptions =
-             ps { optShowIrrelevant = not $ optShowIrrelevant ps } }
+    over (lensPragmaOptions . lensOptShowIrrelevant . lensCollapseDefault) not opts
 
 interpret (Cmd_load_highlighting_info source) = do
   l <- asksTC envHighlightingLevel
@@ -615,7 +582,7 @@ interpret (Cmd_load_highlighting_info source) = do
     setCommandLineOpts =<< lift commandLineOptions
     resp <- lift $ liftIO . tellToUpdateHighlighting =<< do
       ex        <- liftIO $ doesFileExist source
-      absSource <- liftIO $ SourceFile <$> absolute source
+      absSource <- srcFromPath =<< liftIO (absolute source)
       if ex
         then
            do
@@ -696,55 +663,45 @@ interpret (Cmd_refine_or_intro pmLambda ii r s) = interpret $
   let s' = trim s
   in (if null s' then Cmd_intro pmLambda else Cmd_refine) ii r s'
 
-interpret (Cmd_autoOne ii rng hint) = do
-  -- Andreas, 2014-07-05 Issue 1226:
-  -- Save the state to have access to even those interaction ids
-  -- that Auto solves (since Auto gives the solution right away).
-  st <- getTC
-  (time , res) <- maybeTimed $ Auto.auto ii rng hint
-  case autoProgress res of
-   Solutions sols -> do
-    lift $ reportSLn "auto" 10 $ "Auto produced the following solutions " ++ show sols
-    forM_ sols $ \(ii', sol) -> do
-      -- Andreas, 2014-07-05 Issue 1226:
-      -- For highlighting, Resp_GiveAction needs to access
-      -- the @oldInteractionScope@s of the interaction points solved by Auto.
-      -- We dig them out from the state before Auto was invoked.
-      insertOldInteractionScope ii' =<< liftLocalState (putTC st >> getInteractionScope ii')
-      -- Andreas, 2014-07-07: NOT TRUE:
-      -- -- Andreas, 2014-07-05: The following should be obsolete,
-      -- -- as Auto has removed the interaction points already:
-      -- modifyTheInteractionPoints $ filter (/= ii)
-      putResponse $ Resp_GiveAction ii' $ Give_String sol
-    -- Andreas, 2014-07-07: Remove the interaction points in one go.
-    modifyTheInteractionPoints (List.\\ (map fst sols))
-    case autoMessage res of
-     Nothing  -> interpret $ Cmd_metas AsIs
-     Just msg -> display_info $ Info_Auto msg
-   FunClauses cs -> do
-    case autoMessage res of
-     Nothing  -> return ()
-     Just msg -> display_info $ Info_Auto msg
-    putResponse $ Resp_MakeCase ii R.Function cs
-   Refinement s -> give_gen WithoutForce ii rng s Refine
-  maybe (return ()) (display_info . Info_Time) time
+interpret (Cmd_autoOne norm ii rng str) = do
+  iscope <- getInteractionScope ii
+  (time, result) <- maybeTimed $ Mimer.mimer norm ii rng str
+  case result of
+    MimerNoResult -> display_info $ Info_Auto "No solution found"
+    MimerExpr str -> do
+      insertOldInteractionScope ii iscope
+      _ <- lift $ B.parseExprIn ii rng str >>= B.give WithForce ii (Just rng)
+      putResponse $ Resp_GiveAction ii $ Give_String str
+      modifyTheInteractionPoints (List.delete ii)
+      maybe (return ()) (display_info . Info_Time) time
+    MimerList sols -> do
+      display_info $ Info_Auto $ unlines $
+        [ "Solutions:" ] ++
+        [ "  " ++ show i ++ ". " ++ s | (i, s) <- sols ]
+    MimerClauses{} -> __IMPOSSIBLE__    -- Mimer can't do case splitting yet
 
-interpret Cmd_autoAll = do
+interpret (Cmd_autoAll norm) = do
   iis <- getInteractionPoints
+  getOldScope <- do
+    st <- getTC
+    pure $ \ ii -> liftLocalState $ putTC st >> getInteractionScope ii
   unless (null iis) $ do
     let time = 1000 `div` length iis
     st <- getTC
-    solved <- forM iis $ \ ii -> do
+    solved <- fmap concat $ forM iis $ \ ii -> do
       rng <- getInteractionRange ii
-      res <- Auto.auto ii rng ("-t " ++ show time ++ "ms")
-      case autoProgress res of
-        Solutions sols -> forM sols $ \ (jj, s) -> do
-            oldInteractionScope <- liftLocalState (putTC st >> getInteractionScope jj)
-            insertOldInteractionScope jj oldInteractionScope
-            putResponse $ Resp_GiveAction ii $ Give_String s
-            return jj
-        _ -> return []
-    modifyTheInteractionPoints (List.\\ concat solved)
+      res <- Mimer.mimer norm ii rng ("-t " ++ show time ++ "ms")
+      case res of
+        MimerNoResult -> pure []
+        MimerExpr str -> do
+          iscope <- getOldScope ii
+          insertOldInteractionScope ii iscope
+          _ <- liftTCM $ B.parseExprIn ii rng str >>= B.give WithoutForce ii (Just rng)
+          putResponse $ Resp_GiveAction ii $ Give_String str
+          pure [ii]
+        MimerList{} -> pure []    -- Don't list solutions in autoAll
+        MimerClauses{} -> __IMPOSSIBLE__  -- Mimer can't do case splitting yet
+    modifyTheInteractionPoints (List.\\ solved)
 
 interpret (Cmd_context norm ii _ _) =
   display_info . Info_Context ii =<< liftLocalState (B.getResponseContext norm ii)
@@ -774,21 +731,21 @@ interpret (Cmd_goal_type_context_infer norm ii rng s) = do
   aux <- if all Char.isSpace s
             then return GoalOnly
             else do
-              typ <- liftLocalState
-                    $ withInteractionId ii
-                    $ B.typeInMeta ii norm =<< B.parseExprIn ii rng s
-              return (GoalAndHave typ)
+              liftLocalState $ withInteractionId ii $ do
+                parsed <- B.parseExprIn ii rng s
+                (typ, faces) <- B.typeAndFacesInMeta ii norm parsed
+                return (GoalAndHave typ faces)
   cmd_goal_type_context_and aux norm ii rng s
 
 interpret (Cmd_goal_type_context_check norm ii rng s) = do
-  term <- liftLocalState $ withInteractionId ii $ do
+  expr <- liftLocalState $ withInteractionId ii $ do
     expr <- B.parseExprIn ii rng s
     goal <- B.typeOfMeta AsIs ii
     term <- case goal of
       OfType _ ty -> checkExpr expr =<< isType_ ty
       _           -> __IMPOSSIBLE__
-    B.normalForm norm term
-  cmd_goal_type_context_and (GoalAndElaboration term) norm ii rng s
+    reify =<< B.normalForm norm term
+  cmd_goal_type_context_and (GoalAndElaboration expr) norm ii rng s
 
 interpret (Cmd_show_module_contents norm ii rng s) =
   liftCommandMT (withInteractionId ii) $ showModuleContents norm rng s
@@ -804,7 +761,7 @@ interpret (Cmd_make_case ii rng s) = do
   liftCommandMT (withInteractionId ii) $ do
     tel <- lift $ lookupSection (qnameModule f) -- don't shadow the names in this telescope
     unicode <- getsTC $ optUseUnicode . getPragmaOptions
-    pcs      :: [Doc]      <- lift $ inTopContext $ addContext tel $ mapM prettyA cs
+    pcs      :: [Doc]      <- lift $ inTopContext $ addContext tel $ mapM prettyAUnqualify cs
     let pcs' :: [String]    = List.map (extlam_dropName unicode casectxt . decorate) pcs
     lift $ reportSDoc "interaction.case" 60 $ TCP.vcat
       [ "InteractionTop.Cmd_make_case"
@@ -821,25 +778,7 @@ interpret (Cmd_make_case ii rng s) = do
         ]
       ]
     putResponse $ Resp_MakeCase ii (makeCaseVariant casectxt) pcs'
-  where
-    decorate = renderStyle (style { mode = OneLineMode })
 
-    makeCaseVariant :: CaseContext -> MakeCaseVariant
-    makeCaseVariant Nothing = R.Function
-    makeCaseVariant Just{}  = R.ExtendedLambda
-
-    -- very dirty hack, string manipulation by dropping the function name
-    -- and replacing the last " = " with " -> ". It's important not to replace
-    -- the equal sign in named implicit with an arrow!
-    extlam_dropName :: UnicodeOrAscii -> CaseContext -> String -> String
-    extlam_dropName _ Nothing x = x
-    extlam_dropName glyphMode Just{}  x
-        = unwords $ reverse $ replEquals $ reverse $ drop 1 $ words x
-      where
-        arrow = render $ _arrow $ specialCharactersForGlyphs glyphMode
-        replEquals ("=" : ws) = arrow : ws
-        replEquals (w   : ws) = w : replEquals ws
-        replEquals []         = []
 
 interpret (Cmd_compute cmode ii rng s) = do
   expr <- liftLocalState $ do
@@ -852,6 +791,27 @@ interpret Cmd_show_version = display_info Info_Version
 interpret Cmd_abort = return ()
 interpret Cmd_exit  = return ()
 
+
+decorate :: Doc -> String
+decorate = renderStyle (style { mode = OneLineMode })
+
+makeCaseVariant :: CaseContext -> MakeCaseVariant
+makeCaseVariant Nothing = R.Function
+makeCaseVariant Just{}  = R.ExtendedLambda
+
+-- very dirty hack, string manipulation by dropping the function name
+-- and replacing the last " = " with " -> ". It's important not to replace
+-- the equal sign in named implicit with an arrow!
+extlam_dropName :: UnicodeOrAscii -> CaseContext -> String -> String
+extlam_dropName _ Nothing x = x
+extlam_dropName glyphMode Just{}  x
+    = unwords $ reverse $ replEquals $ reverse $ drop 1 $ words x
+  where
+    arrow = render $ _arrow $ specialCharactersForGlyphs glyphMode
+    replEquals ("=" : ws) = arrow : ws
+    replEquals (w   : ws) = w : replEquals ws
+    replEquals []         = []
+
 -- | Solved goals already instantiated internally
 -- The second argument potentially limits it to one specific goal.
 solveInstantiatedGoals :: Rewrite -> Maybe InteractionId -> CommandM ()
@@ -886,18 +846,12 @@ cmd_load'
                -- ^ Continuation after successful loading.
   -> CommandM a
 cmd_load' file argv unsolvedOK mode cmd = do
-    fp <- liftIO $ absolute file
-    ex <- liftIO $ doesFileExist $ filePath fp
-    unless ex $ typeError $ GenericError $
-      "The file " ++ file ++ " was not found."
 
     -- Forget the previous "current file" and interaction points.
     modify $ \ st -> st { theInteractionPoints = []
                         , theCurrentFile       = Nothing
                         }
 
-    t <- liftIO $ getModificationTime file
-
     -- Update the status. Because the "current file" is not set the
     -- status is not "Checked".
     displayStatus
@@ -917,7 +871,16 @@ cmd_load' file argv unsolvedOK mode cmd = do
     -- Parse the file.
     --
     -- Note that options are set below.
-    src <- lift $ Imp.parseSource (SourceFile fp)
+    fp  <- liftIO $ absolute file
+    sf  <- liftTCM $ srcFromPath fp
+    src <- lift $ Imp.parseSource sf
+    -- Andreas, 2024-08-03, see test/interaction/FileNotFound:
+    -- Run 'getModificationTime' after 'parseSource',
+    -- otherwise the user gets a weird error for non-existing files.
+    -- (We assume that parsing is fast in comparison to type-checking,
+    -- so it should not matter much whether we get the time stamp
+    -- before or after parsing.)
+    t   <- liftIO $ getModificationTime file
 
     -- Store the warnings.
     warnings <- useTC stTCWarnings
@@ -929,15 +892,15 @@ cmd_load' file argv unsolvedOK mode cmd = do
     let (z, warns) = runOptM $ parseBackendOptions backends argv opts0
     mapM_ (lift . warning . OptionWarning) warns
     case z of
-      Left err -> lift $ typeError $ GenericError err
+      Left err -> lift $ typeError $ OptionError err
       Right (_, opts) -> do
         opts <- lift $ addTrustedExecutables opts
-        let update o = o { optAllowUnsolved = unsolvedOK && optAllowUnsolved o}
-            root     = projectRoot fp $ Imp.srcModuleName src
+        let update = over (lensOptAllowUnsolved . lensKeepDefault) (unsolvedOK &&)
+            root   = projectRoot fp $ Imp.srcModuleName src
         lift $ TCM.setCommandLineOptions' root $ mapPragmaOptions update opts
 
     -- Restore the warnings that were saved above.
-    modifyTCLens stTCWarnings (++ warnings)
+    modifyTCLens stTCWarnings $ Set.union warnings
 
     ok <- lift $ Imp.typeCheckMain mode src
 
@@ -962,7 +925,8 @@ cmd_load' file argv unsolvedOK mode cmd = do
 withCurrentFile :: CommandM a -> CommandM a
 withCurrentFile m = do
   mfile <- gets $ fmap currentFilePath . theCurrentFile
-  localTC (\ e -> e { envCurrentPath = mfile }) m
+  i <- traverse idFromFile mfile
+  localTC (\ e -> e { envCurrentPath = i }) m
 
 atTopLevel :: CommandM a -> CommandM a
 atTopLevel cmd = liftCommandMT B.atTopLevel cmd
@@ -1027,6 +991,13 @@ give_gen force ii rng s0 giveRefine = do
       [ "ce = " ++ show ce
       , "scopePrecedence = " ++ show (scope ^. scopePrecedence)
       ]
+
+    -- Issue 7218: if the give/refine command creates an extended
+    -- lambda, it also needs to be added to the relevant unfolding sets.
+    -- The easiest way to make sure this is consistent is to just re-run
+    -- the saturation procedures.
+    saturateOpaqueBlocks
+
     -- if the command was @Give@, use the literal user input;
     -- Andreas, 2014-01-15, see issue 1020:
     -- Refine could solve a goal by introducing the sole constructor
@@ -1075,7 +1046,7 @@ highlightExpr e =
 -- | Sorts interaction points based on their ranges.
 
 sortInteractionPoints
-  :: (MonadInteractionPoints m, MonadError TCErr m, MonadFail m)
+  :: (MonadInteractionPoints m, MonadError TCErr m, MonadDebug m)
   => [InteractionId] -> m [InteractionId]
 sortInteractionPoints is =
   map fst . List.sortBy (compare `on` snd) <$> do
@@ -1186,8 +1157,8 @@ parseAndDoAtToplevel
   -> CommandM (Maybe CPUTime, a)
 parseAndDoAtToplevel cmd s = do
   localStateCommandM $ do
-    (e, coh) <- lift $ runPM $ parse exprParser s
-    lift $ checkCohesionAttributes coh
+    (e, attrs) <- lift $ runPM $ parse exprParser s
+    lift $ checkAttributes attrs
     maybeTimed $ atTopLevel $ lift $
       cmd =<< concreteToAbstract_ e
 
