@@ -8,47 +8,16 @@
 
 module Agda.Interaction.Options.Lenses where
 
-import Control.Monad.IO.Class   ( MonadIO(..) )
-
-import Data.Set (Set)
-import qualified Data.Set as Set
-
-import System.FilePath ((</>))
-
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.State
-import Agda.Interaction.Library (getPrimitiveLibDir)
 import Agda.Interaction.Options
 
-import Agda.Utils.Lens
-import Agda.Utils.FileName
 import Agda.Utils.WithDefault (pattern Value)
 
 ---------------------------------------------------------------------------
 -- * Pragma options
 ---------------------------------------------------------------------------
 
-class LensPragmaOptions a where
-  getPragmaOptions  :: a -> PragmaOptions
-  setPragmaOptions  :: PragmaOptions -> a -> a
-  mapPragmaOptions  :: (PragmaOptions -> PragmaOptions) -> a -> a
-  lensPragmaOptions :: Lens' a PragmaOptions
-  -- lensPragmaOptions :: forall f. Functor f => (PragmaOptions -> f PragmaOptions) -> a -> f a
-
-  -- default implementations
-  setPragmaOptions     = mapPragmaOptions . const
-  mapPragmaOptions f a = setPragmaOptions (f $ getPragmaOptions a) a
-
-instance LensPragmaOptions CommandLineOptions where
-  getPragmaOptions = optPragmaOptions
-  setPragmaOptions opts st = st { optPragmaOptions = opts }
-  lensPragmaOptions f st = f (optPragmaOptions st) <&> \ opts -> st { optPragmaOptions = opts }
-
-instance LensPragmaOptions TCState where
-  getPragmaOptions = (^. stPragmaOptions)
-  setPragmaOptions = set stPragmaOptions
-  lensPragmaOptions = stPragmaOptions
-
 modifyPragmaOptions :: MonadTCState m => (PragmaOptions -> PragmaOptions) -> m ()
 modifyPragmaOptions = modifyTC . mapPragmaOptions
 
@@ -140,81 +109,6 @@ modifySafeMode = modifyTC . mapSafeMode
 putSafeMode :: MonadTCState m => SafeMode -> m ()
 putSafeMode = modifyTC . setSafeMode
 
--- | These builtins may use postulates, and are still considered --safe
-
-builtinModulesWithSafePostulates :: Set FilePath
-builtinModulesWithSafePostulates =
-  primitiveModules `Set.union` (Set.fromList
-  [ "Agda" </> "Builtin" </> "Bool.agda"
-  , "Agda" </> "Builtin" </> "Char.agda"
-  , "Agda" </> "Builtin" </> "Char" </> "Properties.agda"
-  , "Agda" </> "Builtin" </> "Coinduction.agda"
-  , "Agda" </> "Builtin" </> "Cubical" </> "Equiv.agda"
-  , "Agda" </> "Builtin" </> "Cubical" </> "Glue.agda"
-  , "Agda" </> "Builtin" </> "Cubical" </> "HCompU.agda"
-  , "Agda" </> "Builtin" </> "Cubical" </> "Id.agda"
-  , "Agda" </> "Builtin" </> "Cubical" </> "Path.agda"
-  , "Agda" </> "Builtin" </> "Cubical" </> "Sub.agda"
-  , "Agda" </> "Builtin" </> "Equality" </> "Erase.agda"
-  , "Agda" </> "Builtin" </> "Equality.agda"
-  , "Agda" </> "Builtin" </> "Float.agda"
-  , "Agda" </> "Builtin" </> "Float" </> "Properties.agda"
-  , "Agda" </> "Builtin" </> "FromNat.agda"
-  , "Agda" </> "Builtin" </> "FromNeg.agda"
-  , "Agda" </> "Builtin" </> "FromString.agda"
-  , "Agda" </> "Builtin" </> "Int.agda"
-  , "Agda" </> "Builtin" </> "IO.agda"
-  , "Agda" </> "Builtin" </> "List.agda"
-  , "Agda" </> "Builtin" </> "Maybe.agda"
-  , "Agda" </> "Builtin" </> "Nat.agda"
-  , "Agda" </> "Builtin" </> "Reflection.agda"
-  , "Agda" </> "Builtin" </> "Reflection" </> "Properties.agda"
-  , "Agda" </> "Builtin" </> "Reflection" </> "External.agda"
-  , "Agda" </> "Builtin" </> "Sigma.agda"
-  , "Agda" </> "Builtin" </> "Size.agda"
-  , "Agda" </> "Builtin" </> "Strict.agda"
-  , "Agda" </> "Builtin" </> "String.agda"
-  , "Agda" </> "Builtin" </> "String" </> "Properties.agda"
-  , "Agda" </> "Builtin" </> "Unit.agda"
-  , "Agda" </> "Builtin" </> "Word.agda"
-  , "Agda" </> "Builtin" </> "Word" </> "Properties.agda"
-  ])
-
--- | These builtins may not use postulates under --safe. They are not
---   automatically unsafe, but will be if they use an unsafe feature.
-
-builtinModulesWithUnsafePostulates :: Set FilePath
-builtinModulesWithUnsafePostulates = Set.fromList
-  [ "Agda" </> "Builtin" </> "TrustMe.agda"
-  , "Agda" </> "Builtin" </> "Equality" </> "Rewrite.agda"
-  ]
-
-primitiveModules :: Set FilePath
-primitiveModules = Set.fromList
-  [ "Agda" </> "Primitive.agda"
-  , "Agda" </> "Primitive" </> "Cubical.agda"
-  ]
-
-builtinModules :: Set FilePath
-builtinModules = builtinModulesWithSafePostulates `Set.union`
-                 builtinModulesWithUnsafePostulates
-
-isPrimitiveModule :: MonadIO m => FilePath -> m Bool
-isPrimitiveModule file = do
-  libdirPrim <- liftIO getPrimitiveLibDir
-  return (file `Set.member` Set.map (libdirPrim </>) primitiveModules)
-
-isBuiltinModule :: MonadIO m => FilePath -> m Bool
-isBuiltinModule file = do
-  libdirPrim <- liftIO getPrimitiveLibDir
-  return (file `Set.member` Set.map (libdirPrim </>) builtinModules)
-
-isBuiltinModuleWithSafePostulates :: MonadIO m => FilePath -> m Bool
-isBuiltinModuleWithSafePostulates file = do
-  libdirPrim <- liftIO getPrimitiveLibDir
-  let safeBuiltins   = Set.map (libdirPrim </>) builtinModulesWithSafePostulates
-  return (file `Set.member` safeBuiltins)
-
 ---------------------------------------------------------------------------
 -- ** Include directories
 ---------------------------------------------------------------------------
