@@ -7,7 +7,7 @@ module Agda.Syntax.Abstract.Name
   , FreshNameMode(..)
   ) where
 
-import Prelude hiding (length)
+import Prelude hiding (length, null)
 
 import Control.DeepSeq
 
@@ -20,6 +20,7 @@ import Data.Void
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
+import Agda.Syntax.Common.Pretty
 import Agda.Syntax.Concrete.Name (IsNoName(..), NumHoles(..), NameInScope(..), LensInScope(..), FreshNameMode(..))
 import qualified Agda.Syntax.Concrete.Name as C
 
@@ -28,7 +29,7 @@ import Agda.Utils.Lens
 import qualified Agda.Utils.List as L
 import Agda.Utils.List1 (List1, pattern (:|), (<|))
 import qualified Agda.Utils.List1 as List1
-import Agda.Syntax.Common.Pretty
+import Agda.Utils.Null
 import Agda.Utils.Size
 
 import Agda.Utils.Impossible
@@ -73,7 +74,7 @@ data QNamed a = QNamed
 -- The 'SetRange' instance for module names sets all individual ranges
 -- to the given one.
 newtype ModuleName = MName { mnameToList :: [Name] }
-  deriving (Eq, Ord)
+  deriving (Eq, Ord, NFData, Null)
 
 -- | Ambiguous qualified names. Used for overloaded constructors.
 --
@@ -105,6 +106,9 @@ data Suffix
   | Suffix !Integer
   deriving (Show, Eq, Ord)
 
+instance Null Suffix where
+  empty = NoSuffix
+
 instance NFData Suffix where
   rnf NoSuffix   = ()
   rnf (Suffix _) = ()
@@ -302,6 +306,9 @@ instance Hashable QName where
 instance IsNoName Name where
   isNoName = isNoName . nameConcrete
 
+instance IsNoName ModuleName where
+  isNoName (MName xs) = all isNoName xs
+
 instance NumHoles Name where
   numHoles = numHoles . nameConcrete
 
@@ -484,6 +491,3 @@ instance NFData Name where
 
 instance NFData QName where
   rnf (QName a b) = rnf a `seq` rnf b
-
-instance NFData ModuleName where
-  rnf (MName a) = rnf a
