@@ -10,11 +10,12 @@ import Prelude hiding (null)
 import Control.Arrow           ( (***), second )
 import Control.Monad           ( (>=>) )
 import Control.Monad.Identity  ( Identity(..), runIdentity )
+import Control.Monad.Reader    ( Reader, runReader, asks, local )
 import Control.Applicative     ( liftA2 )
 
-
 import Data.Maybe
 import Data.Monoid
+import Data.Void (Void)
 
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Common
@@ -26,7 +27,10 @@ import Agda.Syntax.Position
 
 import Agda.Utils.Functor
 import Agda.Utils.List
+import Agda.Utils.List1 ( List1, pattern (:|) )
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Null
+import Agda.Utils.Singleton
 
 import Agda.Utils.Impossible
 
@@ -59,12 +63,11 @@ instance MapNamedArgPattern NAP where
       PatternSynP i x ps -> f $ setNamedArg p $ PatternSynP i x $ mapNamedArgPattern f ps
       -- Pattern subpattern(s):
       -- RecP: we copy the NamedArg info to the subpatterns but discard it after recursion
-      RecP i fs          -> f $ setNamedArg p $ RecP i $ map (fmap namedArg) $ mapNamedArgPattern f $ map (fmap (setNamedArg p)) fs
+      RecP kwr i fs      -> f $ setNamedArg p $ RecP kwr i $ map (fmap namedArg) $ mapNamedArgPattern f $ map (fmap (setNamedArg p)) fs
       -- AsP: we hand the NamedArg info to the subpattern
       AsP i x p0         -> f $ updateNamedArg (AsP i x) $ mapNamedArgPattern f $ setNamedArg p p0
       -- WithP: like AsP
       WithP i p0         -> f $ updateNamedArg (WithP i) $ mapNamedArgPattern f $ setNamedArg p p0
-      AnnP i a p0        -> f $ updateNamedArg (AnnP i a) $ mapNamedArgPattern f $ setNamedArg p p0
 
 instance MapNamedArgPattern a => MapNamedArgPattern [a]                  where
 instance MapNamedArgPattern a => MapNamedArgPattern (FieldAssignment' a) where
@@ -140,7 +143,7 @@ instance APatternLike (Pattern' a) where
       AsP _ _ p          -> foldrAPattern f p
       ConP _ _ ps        -> foldrAPattern f ps
       DefP _ _ ps        -> foldrAPattern f ps
-      RecP _ ps          -> foldrAPattern f ps
+      RecP _ _ ps        -> foldrAPattern f ps
       PatternSynP _ _ ps -> foldrAPattern f ps
       WithP _ p          -> foldrAPattern f p
       VarP _             -> mempty
@@ -150,7 +153,6 @@ instance APatternLike (Pattern' a) where
       AbsurdP _          -> mempty
       LitP _ _           -> mempty
       EqualP _ _         -> mempty
-      AnnP _ _ p         -> foldrAPattern f p
 
   traverseAPatternM pre post = pre >=> recurse >=> post
     where
@@ -167,10 +169,9 @@ instance APatternLike (Pattern' a) where
       A.ConP        i ds ps -> A.ConP        i ds <$> traverseAPatternM pre post ps
       A.DefP        i q  ps -> A.DefP        i q  <$> traverseAPatternM pre post ps
       A.AsP         i x  p  -> A.AsP         i x  <$> traverseAPatternM pre post p
-      A.RecP        i    ps -> A.RecP        i    <$> traverseAPatternM pre post ps
+      A.RecP    kwr i    ps -> A.RecP    kwr i    <$> traverseAPatternM pre post ps
       A.PatternSynP i x  ps -> A.PatternSynP i x  <$> traverseAPatternM pre post ps
       A.WithP       i p     -> A.WithP       i    <$> traverseAPatternM pre post p
-      A.AnnP        i a  p  -> A.AnnP        i a  <$> traverseAPatternM pre post p
 
 instance APatternLike a => APatternLike (Arg a) where
   type ADotT (Arg a) = ADotT a
@@ -223,7 +224,6 @@ patternVars p = foldAPattern f p `appEndo` []
     A.EqualP      {} -> mempty
     A.PatternSynP {} -> mempty
     A.WithP _ _      -> mempty
-    A.AnnP        {} -> mempty
 
 -- | Check if a pattern contains a specific (sub)pattern.
 
@@ -251,9 +251,9 @@ containsAsPattern = containsAPattern $ \case
 -- | Check if any user-written pattern variables occur more than once,
 --   and throw the given error if they do.
 checkPatternLinearity :: (Monad m, APatternLike p)
-                      => p -> ([C.Name] -> m ()) -> m ()
-checkPatternLinearity ps err =
-  unlessNull (duplicates $ map nameConcrete $ patternVars ps) $ \ys -> err ys
+  => p -> (List1 C.Name -> m ()) -> m ()
+checkPatternLinearity =
+  List1.unlessNull . duplicates . map nameConcrete . patternVars
 
 
 -- * Specific traversals
@@ -277,11 +277,10 @@ substPattern'
 substPattern' subE s = mapAPattern $ \ p -> case p of
   VarP x            -> fromMaybe p $ lookup (A.unBind x) s
   DotP i e          -> DotP i $ subE e
-  EqualP i es       -> EqualP i $ map (subE *** subE) es
-  AnnP i a p        -> AnnP i (subE a) p
+  EqualP i es       -> EqualP i $ fmap (subE *** subE) es
   -- No action on the other patterns (besides the recursion):
   ConP _ _ _        -> p
-  RecP _ _          -> p
+  RecP _ _ _        -> p
   ProjP _ _ _       -> p
   WildP _           -> p
   AbsurdP _         -> p
@@ -291,6 +290,78 @@ substPattern' subE s = mapAPattern $ \ p -> case p of
   PatternSynP _ _ _ -> p
   WithP _ _         -> p
 
+-- | Convert a pattern to an expression.
+--
+-- Does not support all cases of patterns.
+-- Result has no 'Range' info, except in identifiers.
+--
+-- This function is only used in expanding pattern synonyms
+-- and in "Agda.Syntax.Translation.InternalToAbstract",
+-- so we can cut some corners.
+patternToExpr :: Pattern -> Expr
+patternToExpr p = patToExpr p `runReader` empty
+
+-- | Converting a pattern to an expression.
+--
+--   The 'Hiding' context is remembered to create instance metas
+--   when translating absurd patterns in instance position.
+--
+class PatternToExpr p e where
+  patToExpr :: p -> Reader Hiding e
+
+  default patToExpr :: (Traversable t, PatternToExpr p' e', p ~ t p', e ~ t e')
+    => p -> Reader Hiding e
+  patToExpr = traverse patToExpr
+
+instance PatternToExpr p e => PatternToExpr [p] [e]
+instance PatternToExpr p e => PatternToExpr (Named n p) (Named n e)
+instance PatternToExpr p e => PatternToExpr (FieldAssignment' p) (FieldAssignment' e)
+
+instance PatternToExpr p e => PatternToExpr (Arg p) (Arg e) where
+  patToExpr (Arg ai p) = local (const $ getHiding ai) $ Arg ai <$> patToExpr p
+
+instance PatternToExpr Pattern Expr where
+  patToExpr = \case
+    VarP x             -> return $ Var (unBind x)
+    ConP _ c ps        -> app (Con c) <$> patToExpr ps
+    ProjP _ o ds       -> return $ Proj o ds
+    DefP _ fs ps       -> app (Def $ headAmbQ fs) <$> patToExpr ps
+    WildP _            -> return $ Underscore emptyMetaInfo
+    AsP _ _ p          -> patToExpr p
+    DotP _ e           -> return e
+    -- Issue #7176: An absurd pattern in an instance position should turn into an instance meta:
+    AbsurdP _          -> asks hidingToMetaKind <&> \ k -> Underscore emptyMetaInfo{ metaKind = k }
+    LitP _ l           -> return $ Lit empty l
+    PatternSynP _ c ps -> app (PatternSyn c) <$> patToExpr ps
+    RecP kwr _ as      -> Rec kwr exprNoRange . map Left <$> patToExpr as
+    EqualP{}           -> __IMPOSSIBLE__  -- Andrea TODO: where is this used?
+    WithP r p          -> __IMPOSSIBLE__
+
+-- | Make sure that there are no dot or equality patterns (called on pattern synonyms).
+--   Also disallows annotated patterns.
+--
+noDotOrEqPattern :: forall m e. Monad m
+  => m (A.Pattern' Void)   -- ^ Exception or replacement for dot (etc.) patterns.
+  -> A.Pattern' e          -- ^ In pattern.
+  -> m (A.Pattern' Void)   -- ^ Out pattern.
+noDotOrEqPattern err = dot
+  where
+    dot :: A.Pattern' e -> m (A.Pattern' Void)
+    dot = \case
+      A.VarP x               -> pure $ A.VarP x
+      A.ConP i c args        -> A.ConP i c <$> (traverse $ traverse $ traverse dot) args
+      A.ProjP i o d          -> pure $ A.ProjP i o d
+      A.WildP i              -> pure $ A.WildP i
+      A.AsP i x p            -> A.AsP i x <$> dot p
+      A.DotP{}               -> err
+      A.EqualP{}             -> err   -- Andrea: so we also disallow = patterns, reasonable?
+      A.AbsurdP i            -> pure $ A.AbsurdP i
+      A.LitP i l             -> pure $ A.LitP i l
+      A.DefP i f args        -> A.DefP i f <$> (traverse $ traverse $ traverse dot) args
+      A.PatternSynP i c args -> A.PatternSynP i c <$> (traverse $ traverse $ traverse dot) args
+      A.RecP kwr i fs        -> A.RecP kwr i <$> (traverse $ traverse dot) fs
+      A.WithP i p            -> A.WithP i <$> dot p
+
 
 -- * Other pattern utilities
 ------------------------------------------------------------------------
@@ -313,11 +384,11 @@ trailingWithPatterns = snd . splitOffTrailingWithPatterns
 --
 -- (This view discards 'PatInfo'.)
 data LHSPatternView e
-  = LHSAppP  (NAPs e)
+  = LHSAppP  (NAPs1 e)
       -- ^ Application patterns (non-empty list).
   | LHSProjP ProjOrigin AmbiguousQName (NamedArg (Pattern' e))
       -- ^ A projection pattern.  Is also stored unmodified here.
-  | LHSWithP [Pattern' e]
+  | LHSWithP (List1 (Pattern' e))
       -- ^ With patterns (non-empty list).
       --   These patterns are not prefixed with 'WithP'.
   deriving (Show)
@@ -332,11 +403,11 @@ lhsPatternView (p0 : ps) =
   case namedArg p0 of
     ProjP _i o d -> Just (LHSProjP o d p0, ps)
     -- If the next pattern is a with-pattern, collect more with-patterns
-    WithP _i p   -> Just (LHSWithP (p : map namedArg ps1), ps2)
+    WithP _i p   -> Just (LHSWithP (p :| map namedArg ps1), ps2)
       where
       (ps1, ps2) = spanJust isWithP ps
     -- If the next pattern is an application pattern, collect more of these
-    _ -> Just (LHSAppP (p0 : ps1), ps2)
+    _ -> Just (LHSAppP (p0 :| ps1), ps2)
       where
       (ps1, ps2) = span (\ p -> isNothing (isProjP p) && isNothing (isWithP p)) ps
 
@@ -369,7 +440,7 @@ lhsCoreToSpine = \case
   LHSHead f ps     -> QNamed f ps
   LHSProj d h ps   -> lhsCoreToSpine (namedArg h) <&> (++ (p : ps))
     where p = updateNamedArg (const $ ProjP empty ProjPrefix d) h
-  LHSWith h wps ps -> lhsCoreToSpine h <&> (++ map fromWithPat wps ++ ps)
+  LHSWith h wps ps -> lhsCoreToSpine h <&> (++ map fromWithPat (List1.toList wps) ++ ps)
     where
       fromWithPat :: Arg (Pattern' e) -> NamedArg (Pattern' e)
       fromWithPat = fmap (unnamed . mkWithP)
@@ -383,16 +454,16 @@ lhsCoreApp :: LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
 lhsCoreApp core ps = core { lhsPats = lhsPats core ++ ps }
 
 -- | Add with-patterns to the right.
-lhsCoreWith :: LHSCore' e -> [Arg (Pattern' e)] -> LHSCore' e
-lhsCoreWith (LHSWith core wps []) wps' = LHSWith core (wps ++ wps') []
+lhsCoreWith :: LHSCore' e -> List1 (Arg (Pattern' e)) -> LHSCore' e
+lhsCoreWith (LHSWith core wps []) wps' = LHSWith core (wps <> wps') []
 lhsCoreWith core                  wps' = LHSWith core wps' []
 
 lhsCoreAddChunk :: IsProjP e => LHSCore' e -> LHSPatternView e -> LHSCore' e
 lhsCoreAddChunk core = \case
-  LHSAppP ps               -> lhsCoreApp core ps
+  LHSAppP ps               -> lhsCoreApp core $ List1.toList ps
   LHSWithP wps             -> lhsCoreWith core (defaultArg <$> wps)
   LHSProjP ProjPrefix d np -> LHSProj d (setNamedArg np core) []  -- Prefix projection pattern.
-  LHSProjP _          _ np -> lhsCoreApp core [np]       -- Postfix projection pattern.
+  LHSProjP _          _ np -> lhsCoreApp core (singleton np)      -- Postfix projection pattern.
 
 -- | Add projection, with, and applicative patterns to the right.
 lhsCoreAddSpine :: IsProjP e => LHSCore' e -> [NamedArg (Pattern' e)] -> LHSCore' e
@@ -426,7 +497,7 @@ lhsCoreToPattern lc =
     LHSProj d lhscore aps -> DefP noInfo d $
       fmap (fmap lhsCoreToPattern) lhscore : aps
     LHSWith h wps aps     -> case lhsCoreToPattern h of
-      DefP r q ps         -> DefP r q $ ps ++ map fromWithPat wps ++ aps
+      DefP r q ps         -> DefP r q $ ps ++ map fromWithPat (List1.toList wps) ++ aps
         where
           fromWithPat :: Arg Pattern -> NamedArg Pattern
           fromWithPat = fmap (unnamed . mkWithP)
