@@ -4,7 +4,6 @@ module Agda.Syntax.Abstract.Views where
 import Prelude hiding (null)
 
 import Control.Applicative ( Const(Const), getConst )
-import Control.Arrow (first)
 import Control.Monad.Identity
 
 import Data.Foldable (foldMap)
@@ -14,7 +13,7 @@ import Data.Void
 
 import Agda.Syntax.Common
 import Agda.Syntax.Abstract as A
-import Agda.Syntax.Concrete (FieldAssignment', exprFieldA)
+import Agda.Syntax.Concrete (FieldAssignment', exprFieldA, TacticAttribute')
 import Agda.Syntax.Info
 import Agda.Syntax.Scope.Base (KindOfName(..), conKindOfName, WithKind(..))
 
@@ -42,7 +41,7 @@ appView' e = f (DL.toList es)
   where
   (f, es) = appView'' e
 
-  appView'' e = case e of
+  appView'' = \case
     App i e1 e2
       | Dot _ e2' <- unScope $ namedArg e2
       , Just f <- maybeProjTurnPostfix e2'
@@ -51,7 +50,7 @@ appView' e = f (DL.toList es)
     App i e1 arg | (f, es) <- appView'' e1 ->
       (f, es `DL.snoc` (fmap . fmap) (i,) arg)
     ScopedExpr _ e -> appView'' e
-    _              -> (Application e, mempty)
+    e              -> (Application e, mempty)
 
 maybeProjTurnPostfix :: Expr -> Maybe Expr
 maybeProjTurnPostfix e =
@@ -85,11 +84,10 @@ piView = \case
 unPiView :: PiView -> Expr
 unPiView (PiView tels t) = foldr (uncurry Pi) t tels
 
--- | Gather top-level 'AsP'atterns and 'AnnP'atterns to expose underlying pattern.
-asView :: A.Pattern -> ([Name], [A.Expr], A.Pattern)
-asView (A.AsP _ x p)  = (\(asb, ann, p) -> (unBind x : asb, ann, p)) $ asView p
-asView (A.AnnP _ a p) = (\(asb, ann, p) -> (asb, a : ann, p))        $ asView p
-asView p              = ([], [], p)
+-- | Gather top-level 'AsP'atterns to expose underlying pattern.
+asView :: A.Pattern -> ([Name], A.Pattern)
+asView (A.AsP _ x p)  = (\(asb, p) -> (unBind x : asb, p)) $ asView p
+asView p              = ([], p)
 
 -- | Remove top 'ScopedExpr' wrappers.
 unScope :: Expr -> Expr
@@ -109,12 +107,15 @@ deepUnscopeDecls :: [A.Declaration] -> [A.Declaration]
 deepUnscopeDecls = concatMap deepUnscopeDecl
 
 deepUnscopeDecl :: A.Declaration -> [A.Declaration]
-deepUnscopeDecl (A.ScopedDecl _ ds)              = deepUnscopeDecls ds
-deepUnscopeDecl (A.Mutual i ds)                  = [A.Mutual i (deepUnscopeDecls ds)]
-deepUnscopeDecl (A.Section i m tel ds)           = [A.Section i m (deepUnscope tel) (deepUnscopeDecls ds)]
-deepUnscopeDecl (A.RecDef i x uc dir bs e ds) =
-  [ A.RecDef i x uc dir (deepUnscope bs) (deepUnscope e) (deepUnscopeDecls ds) ]
-deepUnscopeDecl d                                = [deepUnscope d]
+deepUnscopeDecl = \case
+  A.ScopedDecl _ ds           -> deepUnscopeDecls ds
+  A.Mutual i ds               -> [A.Mutual i (deepUnscopeDecls ds)]
+  A.Section i e m tel ds      -> [A.Section i e m (deepUnscope tel)
+                                   (deepUnscopeDecls ds)]
+  A.RecDef i x uc dir bs e ds -> [ A.RecDef i x uc dir (deepUnscope bs)
+                                     (deepUnscope e)
+                                     (deepUnscopeDecls ds) ]
+  d                           -> [deepUnscope d]
 
 -- * Traversal
 ---------------------------------------------------------------------------
@@ -172,8 +173,8 @@ instance ExprLike Expr where
       Generalized  s e           -> Generalized s <$> recurse e
       Fun ei arg e               -> Fun ei <$> recurse arg <*> recurse e
       Let ei bs e                -> Let ei <$> recurse bs <*> recurse e
-      Rec ei bs                  -> Rec ei <$> recurse bs
-      RecUpdate ei e bs          -> RecUpdate ei <$> recurse e <*> recurse bs
+      Rec kwr ei bs              -> Rec kwr ei <$> recurse bs
+      RecUpdate kwr ei e bs      -> RecUpdate kwr ei <$> recurse e <*> recurse bs
       ScopedExpr sc e            -> ScopedExpr sc <$> recurse e
       Quote{}                    -> pure e0
       QuoteTerm{}                -> pure e0
@@ -204,8 +205,8 @@ instance ExprLike Expr where
       Generalized _ e        -> m `mappend` fold e
       Fun _ e e'             -> m `mappend` fold e `mappend` fold e'
       Let _ bs e             -> m `mappend` fold bs `mappend` fold e
-      Rec _ as               -> m `mappend` fold as
-      RecUpdate _ e as       -> m `mappend` fold e `mappend` fold as
+      Rec _ _ as             -> m `mappend` fold as
+      RecUpdate _ _ e as     -> m `mappend` fold e `mappend` fold as
       ScopedExpr _ e         -> m `mappend` fold e
       Quote{}                -> m
       QuoteTerm{}            -> m
@@ -239,8 +240,8 @@ instance ExprLike Expr where
       Generalized s e            -> f =<< Generalized s <$> trav e
       Fun ei arg e               -> f =<< Fun ei <$> trav arg <*> trav e
       Let ei bs e                -> f =<< Let ei <$> trav bs <*> trav e
-      Rec ei bs                  -> f =<< Rec ei <$> trav bs
-      RecUpdate ei e bs          -> f =<< RecUpdate ei <$> trav e <*> trav bs
+      Rec kwr ei bs              -> f =<< Rec kwr ei <$> trav bs
+      RecUpdate kwr ei e bs      -> f =<< RecUpdate kwr ei <$> trav e <*> trav bs
       ScopedExpr sc e            -> f =<< ScopedExpr sc <$> trav e
       Quote{}                    -> f e
       QuoteTerm{}                -> f e
@@ -252,8 +253,10 @@ instance ExprLike Expr where
 instance ExprLike a => ExprLike (Arg a)
 instance ExprLike a => ExprLike (Maybe a)
 instance ExprLike a => ExprLike (Named x a)
+instance ExprLike a => ExprLike (Ranged a)
 instance ExprLike a => ExprLike [a]
 instance ExprLike a => ExprLike (List1 a)
+instance ExprLike a => ExprLike (TacticAttribute' a)
 
 instance (ExprLike a, ExprLike b) => ExprLike (a, b) where
   recurseExpr f (x, y) = (,) <$> recurseExpr f x <*> recurseExpr f y
@@ -328,19 +331,19 @@ instance ExprLike LetBinding where
       recurse e = recurseExpr f e
     case e of
       LetBind li ai x e e'  -> LetBind li ai x <$> recurse e <*> recurse e'
+      LetAxiom li ai x e    -> LetAxiom li ai x <$> recurse e
       LetPatBind li p e     -> LetPatBind li <$> recurse p <*> recurse e
       LetApply{}            -> pure e
       LetOpen{}             -> pure e
-      LetDeclaredVariable _ -> pure e
 
   foldExpr :: forall m. FoldExprFn m LetBinding
   foldExpr f e =
     case e of
       LetBind _ _ _ e e'    -> fold e `mappend` fold e'
+      LetAxiom _ _ _ e      -> fold e
       LetPatBind _ p e      -> fold p `mappend` fold e
       LetApply{}            -> mempty
       LetOpen{}             -> mempty
-      LetDeclaredVariable _ -> mempty
     where
       fold :: FoldExprRecFn m
       fold e = foldExpr f e
@@ -352,10 +355,10 @@ instance ExprLike LetBinding where
       trav e = traverseExpr f e
     case e of
       LetBind li ai x e e'  -> LetBind li ai x <$> trav e <*> trav e'
+      LetAxiom li ai x e    -> LetAxiom li ai x <$> trav e
       LetPatBind li p e     -> LetPatBind li <$> trav p <*> trav e
       LetApply{}            -> pure e
       LetOpen{}             -> pure e
-      LetDeclaredVariable _ -> pure e
 
 instance ExprLike a => ExprLike (Pattern' a) where
 
@@ -382,6 +385,7 @@ instance (ExprLike qn, ExprLike nm, ExprLike p, ExprLike e) => ExprLike (Rewrite
   recurseExpr f = \case
     Rewrite es    -> Rewrite <$> recurseExpr f es
     Invert qn pes -> Invert <$> recurseExpr f qn <*> recurseExpr f pes
+    LeftLet pes   -> LeftLet <$> recurseExpr f pes
 
 instance ExprLike WhereDeclarations where
   recurseExpr f (WhereDecls a b c) = WhereDecls a b <$> recurseExpr f c
@@ -407,9 +411,11 @@ instance ExprLike Pragma where
       CompilePragma{}             -> pure p
       StaticPragma{}              -> pure p
       InjectivePragma{}           -> pure p
+      InjectiveForInferencePragma{} -> pure p
       InlinePragma{}              -> pure p
       EtaPragma{}                 -> pure p
       NotProjectionLikePragma{}   -> pure p
+      OverlapPragma{}             -> pure p
       DisplayPragma f xs e        -> DisplayPragma f <$> rec xs <*> rec e
     where
       rec :: RecurseExprRecFn m
@@ -433,21 +439,22 @@ instance ExprLike Declaration where
       Field i x e               -> Field i x <$> rec e
       Primitive i x e           -> Primitive i x <$> rec e
       Mutual i ds               -> Mutual i <$> rec ds
-      Section i m tel ds        -> Section i m <$> rec tel <*> rec ds
-      Apply i m a ci d          -> (\ a -> Apply i m a ci d) <$> rec a
+      Section i e m tel ds      -> Section i e m <$> rec tel <*> rec ds
+      Apply i e m a ci d        -> (\a -> Apply i e m a ci d) <$> rec a
       Import{}                  -> pure d
       Pragma i p                -> Pragma i <$> rec p
       Open{}                    -> pure d
-      FunDef i f d cs           -> FunDef i f d <$> rec cs
-      DataSig i d tel e         -> DataSig i d <$> rec tel <*> rec e
+      FunDef i f cs             -> FunDef i f <$> rec cs
+      DataSig i er d tel e      -> DataSig i er d <$> rec tel <*> rec e
       DataDef i d uc bs cs      -> DataDef i d uc <$> rec bs <*> rec cs
-      RecSig i r tel e          -> RecSig i r <$> rec tel <*> rec e
+      RecSig i er r tel e       -> RecSig i er r <$> rec tel <*> rec e
       RecDef i r uc dir bs e ds -> RecDef i r uc dir <$> rec bs <*> rec e <*> rec ds
       PatternSynDef f xs p      -> PatternSynDef f xs <$> rec p
       UnquoteDecl i is xs e     -> UnquoteDecl i is xs <$> rec e
       UnquoteDef i xs e         -> UnquoteDef i xs <$> rec e
       UnquoteData i xs uc j cs e -> UnquoteData i xs uc j cs <$> rec e
       ScopedDecl s ds           -> ScopedDecl s <$> rec ds
+      UnfoldingDecl r ds        -> UnfoldingDecl r <$> rec ds
     where
       rec :: RecurseExprRecFn m
       rec e = recurseExpr f e
@@ -489,7 +496,9 @@ instance DeclaredNames KName where
 
 instance DeclaredNames RecordDirectives where
   declaredNames (RecordDirectives i _ _ c) = kc where
-    kc = maybe mempty (singleton . WithKind k) c
+    kc = case c of
+      NamedRecCon c -> singleton $ WithKind k c
+      FreshRecCon{} -> mempty
     k  = maybe ConName (conKindOfName . rangedThing) i
 
 instance DeclaredNames Declaration where
@@ -502,21 +511,22 @@ instance DeclaredNames Declaration where
       Field _ q _                  -> singleton (WithKind FldName q)
       Primitive _ q _              -> singleton (WithKind PrimName q)
       Mutual _ decls               -> declaredNames decls
-      DataSig _ q _ _              -> singleton (WithKind DataName q)
+      DataSig _ _ q _ _            -> singleton (WithKind DataName q)
       DataDef _ q _ _ decls        -> singleton (WithKind DataName q) <> foldMap con decls
-      RecSig _ q _ _               -> singleton (WithKind RecName q)
+      RecSig _ _ q _ _             -> singleton (WithKind RecName q)
       RecDef _ q _ dir _ _ decls   -> singleton (WithKind RecName q) <> declaredNames dir <> declaredNames decls
       PatternSynDef q _ _          -> singleton (WithKind PatternSynName q)
       UnquoteDecl _ _ qs _         -> fromList $ map (WithKind OtherDefName) qs  -- could be Fun or Axiom
       UnquoteDef _ qs _            -> fromList $ map (WithKind FunName) qs       -- cannot be Axiom
-      UnquoteData _ d _ _ cs _     -> singleton (WithKind DataName d) <> (fromList $ map (WithKind ConName) cs) -- singleton _ <> map (WithKind ConName) cs
-      FunDef _ q _ cls             -> singleton (WithKind FunName q) <> declaredNames cls
+      UnquoteData _ d _ _ cs _     -> singleton (WithKind DataName d) <> fromList (map (WithKind ConName) cs) -- singleton _ <> map (WithKind ConName) cs
+      FunDef _ q cls               -> singleton (WithKind FunName q) <> declaredNames cls
       ScopedDecl _ decls           -> declaredNames decls
-      Section _ _ _ decls          -> declaredNames decls
+      Section _ _ _ _ decls        -> declaredNames decls
       Pragma _ pragma              -> declaredNames pragma
       Apply{}                      -> mempty
       Import{}                     -> mempty
       Open{}                       -> mempty
+      UnfoldingDecl{}              -> mempty
     where
     con = \case
       Axiom _ _ _ _ q _ -> singleton $ WithKind ConName q
@@ -531,10 +541,12 @@ instance DeclaredNames Pragma where
     StaticPragma{}            -> mempty
     EtaPragma{}               -> mempty
     InjectivePragma{}         -> mempty
+    InjectiveForInferencePragma{} -> mempty
     InlinePragma{}            -> mempty
     NotProjectionLikePragma{} -> mempty
     DisplayPragma{}           -> mempty
     OptionsPragma{}           -> mempty
+    OverlapPragma{}           -> mempty
 
 instance DeclaredNames Clause where
   declaredNames (Clause _ _ rhs decls _) = declaredNames rhs <> declaredNames decls
@@ -619,7 +631,6 @@ instance DeclaredNames RHS where
 --   declaredNames (LetPatBind _ _ e)      = declaredNames e
 --   declaredNames (LetApply _ _ app _ _)  = declaredNames app
 --   declaredNames LetOpen{}               = mempty
---   declaredNames (LetDeclaredVariable _) = mempty
 
 -- instance DeclaredNames ModuleApplication where
 --   declaredNames (SectionApp bindss _ es) = declaredNames bindss <> declaredNames es
