@@ -11,10 +11,10 @@ import Agda.Syntax.Position
 import Agda.Syntax.Common hiding (TerminationCheck())
 import qualified Agda.Syntax.Common as Common
 import Agda.Syntax.Concrete
-import Agda.Syntax.Concrete.Name
-import Agda.Syntax.Concrete.Pretty
+import Agda.Syntax.Concrete.Name   ()
+import Agda.Syntax.Concrete.Pretty ()
 
-import Agda.Utils.Pretty
+import Agda.Syntax.Common.Pretty
 import Agda.Utils.Impossible
 import Agda.Utils.List1 (List1)
 import qualified Agda.Utils.List1 as List1
@@ -52,14 +52,18 @@ data NiceDeclaration
       --   ('Hiding' should be 'NotHidden'.)
   | NiceField Range Access IsAbstract IsInstance TacticAttribute Name (Arg Expr)
   | PrimitiveFunction Range Access IsAbstract Name (Arg Expr)
-  | NiceMutual Range TerminationCheck CoverageCheck PositivityCheck [NiceDeclaration]
-  | NiceModule Range Access IsAbstract QName Telescope [Declaration]
-  | NiceModuleMacro Range Access Name ModuleApplication OpenShortHand ImportDirective
+  | NiceMutual KwRange TerminationCheck CoverageCheck PositivityCheck [NiceDeclaration]
+  | NiceModule Range Access IsAbstract Erased QName Telescope
+      [Declaration]
+  | NiceModuleMacro Range Access Erased Name ModuleApplication
+      OpenShortHand ImportDirective
   | NiceOpen Range QName ImportDirective
   | NiceImport Range QName (Maybe AsName) OpenShortHand ImportDirective
   | NicePragma Range Pragma
-  | NiceRecSig Range Access IsAbstract PositivityCheck UniverseCheck Name [LamBinding] Expr
-  | NiceDataSig Range Access IsAbstract PositivityCheck UniverseCheck Name [LamBinding] Expr
+  | NiceRecSig Range Erased Access IsAbstract PositivityCheck
+      UniverseCheck Name [LamBinding] Expr
+  | NiceDataSig Range Erased Access IsAbstract PositivityCheck
+      UniverseCheck Name [LamBinding] Expr
   | NiceFunClause Range Access IsAbstract TerminationCheck CoverageCheck Catchall Declaration
     -- ^ An uncategorized function clause, could be a function clause
     --   without type signature or a pattern lhs (e.g. for irrefutable let).
@@ -72,14 +76,15 @@ data NiceDeclaration
       --   Andreas, 2017-01-01: Because of issue #2372, we add 'IsInstance' here.
       --   An alias should know that it is an instance.
   | NiceDataDef Range Origin IsAbstract PositivityCheck UniverseCheck Name [LamBinding] [NiceConstructor]
-  | NiceLoneConstructor Range [NiceConstructor]
-  | NiceRecDef Range Origin IsAbstract PositivityCheck UniverseCheck Name RecordDirectives [LamBinding] [Declaration]
+  | NiceLoneConstructor KwRange [NiceConstructor]
+  | NiceRecDef Range Origin IsAbstract PositivityCheck UniverseCheck Name [RecordDirective] [LamBinding] [Declaration]
       -- ^ @(Maybe Range)@ gives range of the 'pattern' declaration.
-  | NicePatternSyn Range Access Name [Arg Name] Pattern
+  | NicePatternSyn Range Access Name [WithHiding Name] Pattern
   | NiceGeneralize Range Access ArgInfo TacticAttribute Name Expr
   | NiceUnquoteDecl Range Access IsAbstract IsInstance TerminationCheck CoverageCheck [Name] Expr
   | NiceUnquoteDef Range Access IsAbstract TerminationCheck CoverageCheck [Name] Expr
   | NiceUnquoteData Range Access IsAbstract PositivityCheck UniverseCheck Name [Name] Expr
+  | NiceOpaque KwRange [QName] [NiceDeclaration]
   deriving (Show, Generic)
 
 instance NFData NiceDeclaration
@@ -89,8 +94,6 @@ type TerminationCheck = Common.TerminationCheck Measure
 -- | Termination measure is, for now, a variable name.
 type Measure = Name
 
-type Catchall = Bool
-
 -- | Only 'Axiom's.
 type NiceConstructor = NiceTypeSignature
 
@@ -173,7 +176,7 @@ isInterleavedData _ = Nothing
 
 interleavedDecl :: Name -> InterleavedDecl -> [(DeclNum, NiceDeclaration)]
 interleavedDecl k = \case
-  InterleavedData i d@(NiceDataSig _ acc abs pc uc _ pars _) ds ->
+  InterleavedData i d@(NiceDataSig _ _ acc abs pc uc _ pars _) ds ->
     let fpars   = concatMap dropTypeAndModality pars
         r       = getRange (k, fpars)
         ddef cs = NiceDataDef (getRange (r, cs)) UserWritten
@@ -193,15 +196,16 @@ data KindOfBlock
   | FieldBlock      -- ^ @field@.  Ensured by parser.
   | DataBlock       -- ^ @data ... where@.  Here we got a bad error message for Agda-2.5 (Issue 1698).
   | ConstructorBlock  -- ^ @constructor@, in @interleaved mutual@.
-  deriving (Eq, Ord, Show)
+  deriving (Eq, Ord, Show, Generic)
 
+instance NFData KindOfBlock
 
 instance HasRange NiceDeclaration where
   getRange (Axiom r _ _ _ _ _ _)           = r
   getRange (NiceField r _ _ _ _ _ _)       = r
-  getRange (NiceMutual r _ _ _ _)          = r
-  getRange (NiceModule r _ _ _ _ _ )       = r
-  getRange (NiceModuleMacro r _ _ _ _ _)   = r
+  getRange (NiceMutual kwr _ _ _ ds)       = fuseRange kwr ds
+  getRange (NiceModule r _ _ _ _ _ _ )     = r
+  getRange (NiceModuleMacro r _ _ _ _ _ _) = r
   getRange (NiceOpen r _ _)                = r
   getRange (NiceImport r _ _ _ _)          = r
   getRange (NicePragma r _)                = r
@@ -209,16 +213,17 @@ instance HasRange NiceDeclaration where
   getRange (FunSig r _ _ _ _ _ _ _ _ _)    = r
   getRange (FunDef r _ _ _ _ _ _ _)        = r
   getRange (NiceDataDef r _ _ _ _ _ _ _)   = r
-  getRange (NiceLoneConstructor r _)       = r
+  getRange (NiceLoneConstructor kwr ds)    = fuseRange kwr ds
   getRange (NiceRecDef r _ _ _ _ _ _ _ _)  = r
-  getRange (NiceRecSig r _ _ _ _ _ _ _)    = r
-  getRange (NiceDataSig r _ _ _ _ _ _ _)   = r
+  getRange (NiceRecSig r _ _ _ _ _ _ _ _)  = r
+  getRange (NiceDataSig r _ _ _ _ _ _ _ _) = r
   getRange (NicePatternSyn r _ _ _ _)      = r
   getRange (NiceGeneralize r _ _ _ _ _)    = r
   getRange (NiceFunClause r _ _ _ _ _ _)   = r
   getRange (NiceUnquoteDecl r _ _ _ _ _ _ _) = r
   getRange (NiceUnquoteDef r _ _ _ _ _ _)  = r
   getRange (NiceUnquoteData r _ _ _ _ _ _ _) = r
+  getRange (NiceOpaque kwr xs ds)          = getRange (kwr, xs, ds)
 
 instance Pretty NiceDeclaration where
   pretty = \case
@@ -226,18 +231,19 @@ instance Pretty NiceDeclaration where
     NiceField _ _ _ _ _ x _        -> text "field" <+> pretty x
     PrimitiveFunction _ _ _ x _    -> text "primitive" <+> pretty x
     NiceMutual{}                   -> text "mutual"
-    NiceModule _ _ _ x _ _         -> text "module" <+> pretty x <+> text "where"
-    NiceModuleMacro _ _ x _ _ _    -> text "module" <+> pretty x <+> text "= ..."
+    NiceOpaque _ _ ds              -> text "opaque" <+> nest 2 (vcat (map pretty ds))
+    NiceModule _ _ _ _ x _ _       -> text "module" <+> pretty x <+> text "where"
+    NiceModuleMacro _ _ _ x _ _ _  -> text "module" <+> pretty x <+> text "= ..."
     NiceOpen _ x _                 -> text "open" <+> pretty x
     NiceImport _ x _ _ _           -> text "import" <+> pretty x
     NicePragma{}                   -> text "{-# ... #-}"
-    NiceRecSig _ _ _ _ _ x _ _     -> text "record" <+> pretty x
-    NiceDataSig _ _ _ _ _ x _ _    -> text "data" <+> pretty x
+    NiceRecSig _ _ _ _ _ _ x _ _   -> text "record" <+> pretty x
+    NiceDataSig _ _ _ _ _ _ x _ _  -> text "data" <+> pretty x
     NiceFunClause{}                -> text "<function clause>"
     FunSig _ _ _ _ _ _ _ _ x _     -> pretty x <+> colon <+> text "_"
     FunDef _ _ _ _ _ _ x _         -> pretty x <+> text "= _"
     NiceDataDef _ _ _ _ _ x _ _    -> text "data" <+> pretty x <+> text "where"
-    NiceLoneConstructor _ ds       -> text "constructor"
+    NiceLoneConstructor _ _        -> text "data _ where"
     NiceRecDef _ _ _ _ _ x  _ _ _  -> text "record" <+> pretty x <+> text "where"
     NicePatternSyn _ _ x _ _       -> text "pattern" <+> pretty x
     NiceGeneralize _ _ _ _ x _     -> text "variable" <+> pretty x
@@ -268,6 +274,7 @@ declName FunDef{}            = "Function definitions"
 declName NiceRecDef{}        = "Records"
 declName NiceDataDef{}       = "Data types"
 declName NiceLoneConstructor{} = "Constructors"
+declName NiceOpaque{}          = "Opaque blocks"
 
 
 data InMutual
@@ -290,7 +297,9 @@ data DataRecOrFun
     -- ^ Name of a record type
   | FunName TerminationCheck CoverageCheck
     -- ^ Name of a function.
-  deriving Show
+  deriving (Show, Generic)
+
+instance NFData DataRecOrFun
 
 -- Ignore pragmas when checking equality
 instance Eq DataRecOrFun where
