@@ -23,7 +23,6 @@ import GHC.Generics (Generic)
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Literal
-import Agda.Syntax.Concrete.Pretty (prettyHiding)
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Internal.Blockers
 import Agda.Syntax.Internal.Elim
@@ -116,6 +115,7 @@ instance LensAnnotation    (Dom' t e) where
 instance LensRelevance (Dom' t e) where
 instance LensQuantity  (Dom' t e) where
 instance LensCohesion  (Dom' t e) where
+instance LensModalPolarity (Dom' t e) where
 
 argFromDom :: Dom' t a -> Arg a
 argFromDom Dom{domInfo = i, unDom = a} = Arg i a
@@ -154,6 +154,10 @@ data DataOrRecord' p
   deriving (Show, Eq, Generic)
 
 type DataOrRecord = DataOrRecord' PatternOrCopattern
+type DataOrRecord_ = DataOrRecord' ()
+
+pattern IsRecord_ :: DataOrRecord_
+pattern IsRecord_ = IsRecord ()
 
 instance PatternMatchingAllowed DataOrRecord where
   patternMatchingAllowed = \case
@@ -343,14 +347,6 @@ pattern SSet l = Univ USSet l
 
 type Sort = Sort' Term
 
--- | Is this a strict universe inhabitable by data types?
-isStrictDataSort :: Sort' t -> Bool
-isStrictDataSort = \case
-  Univ u _ -> univFibrancy u == IsStrict
-  Inf  u _ -> univFibrancy u == IsStrict
-  _ -> False
-
-
 -- | A level is a maximum expression of a closed level and 0..n
 --   'PlusLevel' expressions each of which is an atom plus a number.
 data Level' t = Max !Integer [PlusLevel' t]
@@ -417,14 +413,8 @@ data Clause = Clause
       --   Used, e.g., by @TermCheck@.
       --   Can be 'Irrelevant' if we encountered an irrelevant projection
       --   pattern on the lhs.
-    , clauseCatchall    :: Bool
+    , clauseCatchall    :: Catchall
       -- ^ Clause has been labelled as CATCHALL.
-    , clauseExact       :: Maybe Bool
-      -- ^ Pattern matching of this clause is exact, no catch-all case.
-      --   Computed by the coverage checker.
-      --   @Nothing@ means coverage checker has not run yet (clause may be inexact).
-      --   @Just False@ means clause is not exact.
-      --   @Just True@ means clause is exact.
     , clauseRecursive   :: Maybe Bool
       -- ^ @clauseBody@ contains recursive calls; computed by termination checker.
       --   @Nothing@ means that termination checker has not run yet,
@@ -471,6 +461,7 @@ defaultPatternInfo = PatternInfo PatOSystem []
 data PatOrigin
   = PatOSystem         -- ^ Pattern inserted by the system
   | PatOSplit          -- ^ Pattern generated by case split
+  | PatOSplitArg ArgName -- ^ Argument to pattern generated by case split
   | PatOVar Name       -- ^ User wrote a variable pattern
   | PatODot            -- ^ User wrote a dot pattern
   | PatOWild           -- ^ User wrote a wildcard pattern
@@ -585,6 +576,7 @@ fromConPatternInfo i = patToConO $ patOrigin $ conPInfo i
     patToConO = \case
       PatOSystem -> ConOSystem
       PatOSplit  -> ConOSplit
+      PatOSplitArg{} -> ConOSystem
       PatOVar{}  -> ConOSystem
       PatODot    -> ConOSystem
       PatOWild   -> ConOSystem
@@ -636,27 +628,6 @@ patternInfo (DefP i _ _)      = Just i
 patternOrigin :: Pattern' x -> Maybe PatOrigin
 patternOrigin = fmap patOrigin . patternInfo
 
--- | Does the pattern perform a match that could fail?
-properlyMatching :: Pattern' a -> Bool
-properlyMatching = properlyMatching' True True
-
-properlyMatching'
-  :: Bool       -- ^ Should absurd patterns count as proper match?
-  -> Bool       -- ^ Should projection patterns count as proper match?
-  -> Pattern' a -- ^ The pattern.
-  -> Bool
-properlyMatching' absP projP = \case
-  p | absP && patternOrigin p == Just PatOAbsurd -> True
-  ConP _ ci ps    -- record constructors do not count as proper matches themselves
-    | conPRecord ci -> List.any (properlyMatching . namedArg) ps
-    | otherwise     -> True
-  LitP{}    -> True
-  DefP{}    -> True
-  ProjP{}   -> projP
-  VarP{}    -> False
-  DotP{}    -> False
-  IApplyP{} -> False
-
 instance IsProjP (Pattern' a) where
   isProjP = \case
     ProjP o d -> Just (o, unambiguous d)
@@ -743,11 +714,16 @@ instance Null (Substitution' a) where
 
 data EqualityView
   = EqualityViewType EqualityTypeData
-  | OtherType Type -- ^ reduced
-  | IdiomType Type -- ^ reduced
+      -- ^ A type of the form @u ≡ v@ decomposed into its parts.
+      --   Used as type for the @rewrite@ expression.
+  | OtherType Type
+      -- ^ A reduced type used as type for a @with@ expression.
+  | IdiomType Type
+      -- ^ A reduced type used as type for the @with@ inspect idiom.
 
 data EqualityTypeData = EqualityTypeData
-    { _eqtSort   :: Sort        -- ^ Sort of this type.
+    { _eqtRange  :: Range       -- ^ Range of the @rewrite@ expression, if any.
+    , _eqtSort   :: Sort        -- ^ Sort of this type.
     , _eqtName   :: QName       -- ^ Builtin EQUALITY.
     , _eqtParams :: Args        -- ^ Hidden.  Empty or @Level@.
     , _eqtType   :: Arg Term    -- ^ Hidden.
@@ -755,16 +731,17 @@ data EqualityTypeData = EqualityTypeData
     , _eqtRhs    :: Arg Term    -- ^ NotHidden.
     }
 
-pattern EqualityType
-  :: Sort
+pattern EqualityType ::
+     Range
+  -> Sort
   -> QName
   -> Args
   -> Arg Term
   -> Arg Term
   -> Arg Term
   -> EqualityView
-pattern EqualityType{ eqtSort, eqtName, eqtParams, eqtType, eqtLhs, eqtRhs } =
-  EqualityViewType (EqualityTypeData eqtSort eqtName eqtParams eqtType eqtLhs eqtRhs)
+pattern EqualityType{ eqtRange, eqtSort, eqtName, eqtParams, eqtType, eqtLhs, eqtRhs } =
+  EqualityViewType (EqualityTypeData eqtRange eqtSort eqtName eqtParams eqtType eqtLhs eqtRhs)
 
 {-# COMPLETE EqualityType, OtherType, IdiomType #-}
 
@@ -848,10 +825,8 @@ dummyLocName cs = maybe __IMPOSSIBLE__ prettyCallSite (headCallSite cs)
 dummyTermWith :: DummyTermKind -> CallStack -> Term
 dummyTermWith kind cs = flip Dummy [] $ concat [kind, ": ", dummyLocName cs]
 
--- | A dummy level to constitute a level/sort created at location.
---   Note: use macro __DUMMY_LEVEL__ !
-dummyLevel :: CallStack -> Level
-dummyLevel = atomicLevel . dummyTermWith "dummyLevel"
+__DUMMY_TERM_WITH__ :: HasCallStack => DummyTermKind -> Term
+__DUMMY_TERM_WITH__ = withCallerCallStack . dummyTermWith
 
 -- | A dummy term created at location.
 --   Note: use macro __DUMMY_TERM__ !
@@ -861,6 +836,11 @@ dummyTerm = dummyTermWith "dummyTerm"
 __DUMMY_TERM__ :: HasCallStack => Term
 __DUMMY_TERM__ = withCallerCallStack dummyTerm
 
+-- | A dummy level to constitute a level/sort created at location.
+--   Note: use macro __DUMMY_LEVEL__ !
+dummyLevel :: CallStack -> Level
+dummyLevel = atomicLevel . dummyTermWith "dummyLevel"
+
 __DUMMY_LEVEL__ :: HasCallStack => Level
 __DUMMY_LEVEL__ = withCallerCallStack dummyLevel
 
@@ -921,13 +901,39 @@ mkProp n = Prop $ ClosedLevel n
 mkSSet :: Integer -> Sort
 mkSSet n = SSet $ ClosedLevel n
 
+impossibleTerm :: CallStack -> Term
+impossibleTerm = flip Dummy [] . show . Impossible
+
+---------------------------------------------------------------------------
+-- * Sorts.
+---------------------------------------------------------------------------
+
 isSort :: Term -> Maybe Sort
 isSort = \case
   Sort s -> Just s
   _      -> Nothing
 
-impossibleTerm :: CallStack -> Term
-impossibleTerm = flip Dummy [] . show . Impossible
+-- | Get the flavor of the universe. 'Nothing' could also mean "don't know".
+sortUniv :: Sort' t -> Maybe Univ
+sortUniv = \case
+  Univ u _ -> Just u
+  Inf  u _ -> Just u
+  _        -> Nothing
+
+-- | Is this a Prop universe?  Answers are yes ('True') or maybe ('False').
+isProp :: Sort' t -> Bool
+isProp = (Just UProp ==) . sortUniv
+
+-- | Is this a strict universe inhabitable by data types?
+isStrictDataSort :: Sort' t -> Bool
+isStrictDataSort = maybe False ((IsStrict ==) . univFibrancy) . sortUniv
+
+-- | Turn a known 'UProp' sort into a 'UType' sort, leave others unchanged.
+propToType :: Sort' t -> Sort' t
+propToType = \case
+  Univ UProp l -> Univ UType l
+  Inf  UProp l -> Inf  UType l
+  s -> s
 
 ---------------------------------------------------------------------------
 -- * Telescopes.
@@ -1049,12 +1055,12 @@ suggests (Suggestion x : xs) = fromMaybe (suggests xs) $ suggestName x
 
 -- | Convert top-level postfix projections into prefix projections.
 unSpine :: Term -> Term
-unSpine = unSpine' $ const True
+unSpine = unSpine' $ \_ _ -> True
 
 -- | Convert 'Proj' projection eliminations
 --   according to their 'ProjOrigin' into
 --   'Def' projection applications.
-unSpine' :: (ProjOrigin -> Bool) -> Term -> Term
+unSpine' :: (ProjOrigin -> QName -> Bool) -> Term -> Term
 unSpine' p v =
   case hasElims v of
     Just (h, es) -> loop h [] es
@@ -1063,9 +1069,9 @@ unSpine' p v =
     loop :: (Elims -> Term) -> Elims -> Elims -> Term
     loop h res es =
       case es of
-        []                   -> v
-        Proj o f : es' | p o -> loop (Def f) [Apply (defaultArg v)] es'
-        e        : es'       -> loop h (e : res) es'
+        []                     -> v
+        Proj o f : es' | p o f -> loop (Def f) [Apply (defaultArg v)] es'
+        e        : es'         -> loop h (e : res) es'
       where v = h $ reverse res
 
 -- | A view distinguishing the neutrals @Var@, @Def@, and @MetaV@ which
@@ -1117,8 +1123,8 @@ instance Null (Tele a) where
 -- | A 'null' clause is one with no patterns and no rhs.
 --   Should not exist in practice.
 instance Null Clause where
-  empty = Clause empty empty empty empty empty empty False Nothing Nothing Nothing empty empty
-  null (Clause _ _ tel pats body _ _ _ _ _ _ wm)
+  empty = Clause empty empty empty empty empty empty empty empty empty empty empty
+  null (Clause _ _ tel pats body _ _ _ _ _ wm)
     =  null tel
     && null pats
     && null body
@@ -1298,8 +1304,8 @@ instance KillRange a => KillRange (Pattern' a) where
       DefP o q ps      -> killRangeN (DefP o) q ps
 
 instance KillRange Clause where
-  killRange (Clause rl rf tel ps body t catchall exact recursive unreachable ell wm) =
-    killRangeN Clause rl rf tel ps body t catchall exact recursive unreachable ell wm
+  killRange (Clause rl rf tel ps body t catchall recursive unreachable ell wm) =
+    killRangeN Clause rl rf tel ps body t catchall recursive unreachable ell wm
 
 instance KillRange a => KillRange (Tele a) where
   killRange = fmap killRange
@@ -1341,10 +1347,10 @@ instance Pretty Term where
       Def q els            -> pretty q `pApp` els
       Con c ci vs          -> pretty (conName c) `pApp` vs
       Pi a (NoAbs _ b)     -> mparens (p > 0) $
-        sep [ prettyPrec 1 (unDom a) <+> "->"
+        sep [ pretty (getModality a) <+> prettyPrec 1 (unDom a) <+> "->"
             , nest 2 $ pretty b ]
       Pi a b               -> mparens (p > 0) $
-        sep [ pDom (domInfo a) (text (absName b) <+> ":" <+> pretty (unDom a)) <+> "->"
+        sep [ pDom (domInfo a) (pretty (getModality a) <+> text (absName b) <+> ":" <+> pretty (unDom a)) <+> "->"
             , nest 2 $ pretty (unAbs b) ]
       Sort s      -> prettyPrec p s
       Level l     -> prettyPrec p l
@@ -1360,7 +1366,7 @@ instance Pretty t => Pretty (Abs t) where
   pretty (NoAbs x t) = "NoAbs" <+> (text x <> ".") <+> pretty t
 
 instance (Pretty t, Pretty e) => Pretty (Dom' t e) where
-  pretty dom = pLock <+> pTac <+> pDom dom (pretty $ unDom dom)
+  pretty dom = pLock <+> pTac <+> pDom dom (pretty (getModality dom) <+> pretty (unDom dom))
     where
       pTac | Just t <- domTactic dom = "@" <> parens ("tactic" <+> pretty t)
            | otherwise               = empty
@@ -1510,7 +1516,7 @@ instance NFData PlusLevel where
 instance NFData e => NFData (Dom e) where
   rnf (Dom a c d e f) = rnf a `seq` rnf c `seq` rnf d `seq` rnf e `seq` rnf f
 
-instance NFData DataOrRecord
+instance NFData a => NFData (DataOrRecord' a)
 instance NFData ConHead
 instance NFData a => NFData (Abs a)
 instance NFData a => NFData (Tele a)
