@@ -1,14 +1,14 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.Syntax.Internal.Elim where
 
 import Control.DeepSeq
 
 import Agda.Syntax.Common
-import Agda.Syntax.Concrete.Pretty () -- Pretty Arg instance
 import Agda.Syntax.Position
 import Agda.Syntax.Abstract.Name
 
-import Agda.Utils.Pretty
+import Agda.Syntax.Common.Pretty
 import Agda.Utils.Empty
 import Agda.Utils.Maybe
 import Agda.Utils.Tuple
