@@ -1,15 +1,12 @@
-{-# LANGUAGE CPP          #-}
 
 -- | Tree traversal for internal syntax.
 
 module Agda.Syntax.Internal.Generic where
 
-#if __GLASGOW_HASKELL__ < 804
-import Data.Monoid ((<>))
-#endif
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Utils.Functor
+import Agda.Utils.List1 (List1)
 
 -- | Generic term traversal.
 --
@@ -61,6 +58,7 @@ instance TermLike a => TermLike (Elim' a)      where
 instance TermLike a => TermLike (Arg a)        where
 instance TermLike a => TermLike (Dom a)        where
 instance TermLike a => TermLike [a]            where
+instance TermLike a => TermLike (List1 a)      where
 instance TermLike a => TermLike (Maybe a)      where
 instance TermLike a => TermLike (Blocked a)    where
 instance TermLike a => TermLike (Abs a)        where
@@ -125,12 +123,11 @@ instance TermLike Type where
 
 instance TermLike Sort where
   traverseTermM f = \case
-    Type l     -> Type <$> traverseTermM f l
-    Prop l     -> Prop <$> traverseTermM f l
+    Univ u l   -> Univ u <$> traverseTermM f l
     s@(Inf _ _)-> pure s
-    SSet l     -> SSet <$> traverseTermM f l
     s@SizeUniv -> pure s
     s@LockUniv -> pure s
+    s@LevelUniv -> pure s
     s@IntervalUniv -> pure s
     PiSort a b c -> PiSort   <$> traverseTermM f a <*> traverseTermM f b <*> traverseTermM f c
     FunSort a b -> FunSort   <$> traverseTermM f a <*> traverseTermM f b
@@ -140,12 +137,11 @@ instance TermLike Sort where
     s@(DummyS _) -> pure s
 
   foldTerm f = \case
-    Type l     -> foldTerm f l
-    Prop l     -> foldTerm f l
+    Univ _ l   -> foldTerm f l
     Inf _ _    -> mempty
-    SSet l     -> foldTerm f l
     SizeUniv   -> mempty
     LockUniv   -> mempty
+    LevelUniv  -> mempty
     IntervalUniv -> mempty
     PiSort a b c -> foldTerm f a <> foldTerm f b <> foldTerm f c
     FunSort a b -> foldTerm f a <> foldTerm f b
@@ -161,7 +157,7 @@ instance TermLike EqualityView where
       <$> traverseTermM f t
     IdiomType t -> IdiomType
       <$> traverseTermM f t
-    EqualityType s eq l t a b -> EqualityType s eq
+    EqualityType r s eq l t a b -> EqualityType r s eq
       <$> traverse (traverseTermM f) l
       <*> traverseTermM f t
       <*> traverseTermM f a
@@ -170,7 +166,7 @@ instance TermLike EqualityView where
   foldTerm f = \case
     OtherType t -> foldTerm f t
     IdiomType t -> foldTerm f t
-    EqualityType s eq l t a b -> foldTerm f (l ++ [t, a, b])
+    EqualityType _r _s _eq l t a b -> foldTerm f (l ++ [t, a, b])
 
 -- | Put it in a monad to make it possible to do strictly.
 copyTerm :: (TermLike a, Monad m) => a -> m a
