@@ -1,9 +1,10 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 -- | Extract all names and meta-variables from things.
 
 module Agda.Syntax.Internal.Names where
 
 import Data.HashMap.Strict (HashMap)
-import qualified Data.HashMap.Strict as HMap
 import Data.Map (Map)
 import Data.Set (Set)
 
@@ -97,26 +98,34 @@ instance NamesIn a => NamesIn (Tele a)
 
 instance (NamesIn a, NamesIn b) => NamesIn (a, b) where
   namesAndMetasIn' sg (x, y) =
-    mappend (namesAndMetasIn' sg x) (namesAndMetasIn' sg y)
+    namesAndMetasIn' sg x <> namesAndMetasIn' sg y
+  {-# INLINE namesAndMetasIn' #-}
 
 instance (NamesIn a, NamesIn b, NamesIn c) => NamesIn (a, b, c) where
-  namesAndMetasIn' sg (x, y, z) = namesAndMetasIn' sg (x, (y, z))
+  namesAndMetasIn' sg (x, y, z) =
+    namesAndMetasIn' sg x <> namesAndMetasIn' sg y <> namesAndMetasIn' sg z
+  {-# INLINE namesAndMetasIn' #-}
 
 instance (NamesIn a, NamesIn b, NamesIn c, NamesIn d) => NamesIn (a, b, c, d) where
   namesAndMetasIn' sg (x, y, z, u) =
-    namesAndMetasIn' sg ((x, y), (z, u))
+    namesAndMetasIn' sg x <> namesAndMetasIn' sg y <> namesAndMetasIn' sg z <> namesAndMetasIn' sg u
+  {-# INLINE namesAndMetasIn' #-}
 
 instance
   (NamesIn a, NamesIn b, NamesIn c, NamesIn d, NamesIn e) =>
   NamesIn (a, b, c, d, e) where
   namesAndMetasIn' sg (x, y, z, u, v) =
-    namesAndMetasIn' sg ((x, y), (z, (u, v)))
+    namesAndMetasIn' sg x <> namesAndMetasIn' sg y <> namesAndMetasIn' sg z <> namesAndMetasIn' sg u
+    <> namesAndMetasIn' sg v
+  {-# INLINE namesAndMetasIn' #-}
 
 instance
   (NamesIn a, NamesIn b, NamesIn c, NamesIn d, NamesIn e, NamesIn f) =>
   NamesIn (a, b, c, d, e, f) where
   namesAndMetasIn' sg (x, y, z, u, v, w) =
-    namesAndMetasIn' sg ((x, (y, z)), (u, (v, w)))
+    namesAndMetasIn' sg x <> namesAndMetasIn' sg y <> namesAndMetasIn' sg z <> namesAndMetasIn' sg u
+    <> namesAndMetasIn' sg v <> namesAndMetasIn' sg w
+  {-# INLINE namesAndMetasIn' #-}
 
 instance NamesIn CompKit where
   namesAndMetasIn' sg (CompKit a b) = namesAndMetasIn' sg (a,b)
@@ -141,30 +150,30 @@ instance NamesIn Bool where
 
 instance NamesIn Definition where
   namesAndMetasIn' sg
-    (Defn _ _ t _ _ _ _ disp _ _ _ _ _ _ _ _ _ _ def) =
+    (Defn _ _ t _ _ _ disp _ _ _ _ _ _ _ _ _ _ def) =
     namesAndMetasIn' sg (t, def, disp)
 
 instance NamesIn Defn where
   namesAndMetasIn' sg = \case
     Axiom _            -> mempty
     DataOrRecSig _     -> mempty
-    GeneralizableVar   -> mempty
+    GeneralizableVar _ -> mempty
     PrimitiveSort _ s  -> namesAndMetasIn' sg s
     AbstractDefn{}     -> __IMPOSSIBLE__
     -- Andreas 2017-07-27, Q: which names can be in @cc@ which are not already in @cl@?
-    Function cl cc _ _ _ _ _ _ _ _ _ _ el _ _
+    Function cl cc _ _ _ _ _ _ _ _ el _ _ _
       -> namesAndMetasIn' sg (cl, cc, el)
     Datatype _ _ cl cs s _ _ _ trX trD
       -> namesAndMetasIn' sg (cl, cs, s, trX, trD)
     Record _ cl c _ fs recTel _ _ _ _ _ _ comp
       -> namesAndMetasIn' sg (cl, c, fs, recTel, comp)
-    Constructor _ _ c d _ _ kit fs _ _
+    Constructor _ _ c d _ kit fs _ _ _ _
       -> namesAndMetasIn' sg (c, d, kit, fs)
-    Primitive _ _ cl _ cc
+    Primitive _ _ cl _ cc _
       -> namesAndMetasIn' sg (cl, cc)
 
 instance NamesIn Clause where
-  namesAndMetasIn' sg (Clause _ _ tel ps b t _ _ _ _ _ _) =
+  namesAndMetasIn' sg (Clause _ _ tel ps b t _ _ _ _ _) =
     namesAndMetasIn' sg (tel, ps, b, t)
 
 instance NamesIn CompiledClauses where
@@ -183,7 +192,7 @@ instance NamesIn (Pattern' a) where
     VarP _ _        -> mempty
     LitP _ l        -> namesAndMetasIn' sg l
     DotP _ v        -> namesAndMetasIn' sg v
-    ConP c _ args   -> namesAndMetasIn' sg (c, args)
+    ConP c cpi args -> namesAndMetasIn' sg (c, cpi, args)
     DefP o q args   -> namesAndMetasIn' sg (q, args)
     ProjP _ f       -> namesAndMetasIn' sg f
     IApplyP _ t u _ -> namesAndMetasIn' sg (t, u)
@@ -193,12 +202,11 @@ instance NamesIn a => NamesIn (Type' a) where
 
 instance NamesIn Sort where
   namesAndMetasIn' sg = \case
-    Type l      -> namesAndMetasIn' sg l
-    Prop l      -> namesAndMetasIn' sg l
+    Univ _ l    -> namesAndMetasIn' sg l
     Inf _ _     -> mempty
-    SSet l      -> namesAndMetasIn' sg l
     SizeUniv    -> mempty
     LockUniv    -> mempty
+    LevelUniv   -> mempty
     IntervalUniv -> mempty
     PiSort a b c  -> namesAndMetasIn' sg (a, b, c)
     FunSort a b -> namesAndMetasIn' sg (a, b)
@@ -260,8 +268,8 @@ instance NamesIn DisplayTerm where
     DWithApp v us es -> namesAndMetasIn' sg (v, us, es)
     DCon c _ vs      -> namesAndMetasIn' sg (c, vs)
     DDef f es        -> namesAndMetasIn' sg (f, es)
-    DDot v           -> namesAndMetasIn' sg v
-    DTerm v          -> namesAndMetasIn' sg v
+    DDot' v es       -> namesAndMetasIn' sg (v, es)
+    DTerm' v es      -> namesAndMetasIn' sg (v, es)
 
 instance NamesIn a => NamesIn (Builtin a) where
   namesAndMetasIn' sg = \case
@@ -272,7 +280,7 @@ instance NamesIn a => NamesIn (Builtin a) where
 -- | Note that the 'primFunImplementation' is skipped.
 instance NamesIn PrimFun where
   namesAndMetasIn' sg = \case
-    PrimFun x _ _ -> namesAndMetasIn' sg x
+    PrimFun x _ _ _ -> namesAndMetasIn' sg x
 
 instance NamesIn Section where
   namesAndMetasIn' sg = \case
@@ -294,21 +302,20 @@ instance NamesIn NLPType where
 
 instance NamesIn NLPSort where
   namesAndMetasIn' sg = \case
-    PType a       -> namesAndMetasIn' sg a
-    PProp a       -> namesAndMetasIn' sg a
-    PSSet a       -> namesAndMetasIn' sg a
+    PUniv _ a     -> namesAndMetasIn' sg a
     PInf _ _      -> mempty
     PSizeUniv     -> mempty
     PLockUniv     -> mempty
+    PLevelUniv    -> mempty
     PIntervalUniv -> mempty
 
 instance NamesIn RewriteRule where
   namesAndMetasIn' sg = \case
-    RewriteRule a b c d e f _ ->
+    RewriteRule a b c d e f _ _ ->
       namesAndMetasIn' sg (a, b, c, d, e, f)
 
 instance (NamesIn a, NamesIn b) => NamesIn (HashMap a b) where
-  namesAndMetasIn' sg = namesAndMetasIn' sg . HMap.toList
+  namesAndMetasIn' sg map = foldMap (namesAndMetasIn' sg) map
 
 instance NamesIn System where
   namesAndMetasIn' sg (System tel cs) = namesAndMetasIn' sg (tel, cs)
@@ -346,7 +353,7 @@ instance NamesIn TAlt where
 
 instance NamesIn CaseType where
   namesAndMetasIn' sg = \case
-    CTData _ x -> namesAndMetasIn' sg x
+    CTData x   -> namesAndMetasIn' sg x
     CTNat      -> mempty
     CTInt      -> mempty
     CTChar     -> mempty
@@ -355,7 +362,7 @@ instance NamesIn CaseType where
     CTQName    -> mempty
 
 instance NamesIn CaseInfo where
-  namesAndMetasIn' sg (CaseInfo _ t) = namesAndMetasIn' sg t
+  namesAndMetasIn' sg (CaseInfo _ _ t) = namesAndMetasIn' sg t
 
 instance NamesIn Compiled where
   namesAndMetasIn' sg (Compiled t _) = namesAndMetasIn' sg t
@@ -366,6 +373,9 @@ newtype PSyn = PSyn A.PatternSynDefn
 instance NamesIn PSyn where
   namesAndMetasIn' sg (PSyn (_args, p)) = namesAndMetasIn' sg p
 
+instance NamesIn ConPatternInfo where
+  namesAndMetasIn' sg (ConPatternInfo _ _ _ ty _) = namesAndMetasIn' sg ty
+
 instance NamesIn (A.Pattern' a) where
   namesAndMetasIn' sg = \case
     A.VarP _               -> mempty
@@ -377,11 +387,10 @@ instance NamesIn (A.Pattern' a) where
     A.AbsurdP _            -> mempty
     A.LitP _ l             -> namesAndMetasIn' sg l
     A.PatternSynP _ c args -> namesAndMetasIn' sg (c, args)
-    A.RecP _ fs            -> namesAndMetasIn' sg fs
+    A.RecP _ _ fs          -> namesAndMetasIn' sg fs
     A.DotP{}               -> __IMPOSSIBLE__    -- Dot patterns are not allowed in pattern synonyms
     A.EqualP{}             -> __IMPOSSIBLE__    -- Andrea: should we allow these in pattern synonyms?
     A.WithP _ p            -> namesAndMetasIn' sg p
-    A.AnnP _ a p           -> __IMPOSSIBLE__    -- Type annotations are not (yet) allowed in pattern synonyms
 
 instance NamesIn AmbiguousQName where
   namesAndMetasIn' sg (AmbQ cs) = namesAndMetasIn' sg cs
