@@ -9,7 +9,7 @@ module Agda.Syntax.Parser
       -- * Parsers
     , moduleParser
     , moduleNameParser
-    , acceptableFileExts
+    , agdaFileExtensions
     , exprParser
     , exprWhereParser
     , holeContentParser
@@ -72,13 +72,13 @@ warning w = PM (modify (w:))
 
 -- | Embed a 'ParseResult' as 'PM' computation.
 
-wrap :: ParseResult a -> PM (a, CohesionAttributes)
+wrap :: ParseResult a -> PM (a, Attributes)
 wrap (ParseFailed err)  = throwError err
 wrap (ParseOk s x)      = do
   modify' (parseWarnings s ++)
-  return (x, parseCohesion s)
+  return (x, parseAttributes s)
 
-wrapM :: IO (ParseResult a) -> PM (a, CohesionAttributes)
+wrapM :: IO (ParseResult a) -> PM (a, Attributes)
 wrapM m = liftIO m >>= wrap
 
 -- | Returns the contents of the given file.
@@ -90,7 +90,7 @@ readFilePM file =
 
 wrapIOM :: (MonadError e m, MonadIO m) => (IOError -> e) -> IO a -> m a
 wrapIOM f m = do
-  a <- liftIO$ (Right <$> m) `catch` (\err -> return$ Left (err :: IOError))
+  a <- liftIO $ (Right <$> m) `catch` (\ err -> return $ Left (err :: IOError))
   case a of
     Right x  -> return x
     Left err -> throwError (f err)
@@ -107,7 +107,7 @@ data Parser a = Parser
   }
 
 type LiterateParser a =
-  Parser a -> [Layer] -> PM (a, CohesionAttributes)
+  Parser a -> SrcFile -> [Layer] -> PM (a, Attributes)
 
 -- | Initial state for lexing.
 
@@ -121,7 +121,7 @@ layoutLexState = [layout, normal]
 
 -- | Parse without top-level layout.
 
-parse :: Parser a -> String -> PM (a, CohesionAttributes)
+parse :: Parser a -> String -> PM (a, Attributes)
 parse p = wrapM . return . M.parse (parseFlags p) normalLexState (parser p)
 
 -- | Parse with top-level layout.
@@ -130,30 +130,30 @@ parseFileFromString
   :: SrcFile   -- ^ Name of source file.
   -> Parser a  -- ^ Parser to use.
   -> String    -- ^ Contents of source file.
-  -> PM (a, CohesionAttributes)
+  -> PM (a, Attributes)
 parseFileFromString src p = wrapM . return . M.parseFromSrc (parseFlags p) layoutLexState (parser p) src
 
 -- | Parse with top-level layout.
 
 parseLiterateWithoutComments :: LiterateParser a
-parseLiterateWithoutComments p layers = parseFileFromString (literateSrcFile layers) p $ illiterate layers
+parseLiterateWithoutComments p f layers = parseFileFromString f p $ illiterate layers
 
 -- | Parse with top-level layout.
 
 parseLiterateWithComments :: LiterateParser [Token]
-parseLiterateWithComments p layers = do
-  (code, coh) <- parseLiterateWithoutComments p layers
+parseLiterateWithComments p f layers = do
+  (code, coh) <- parseLiterateWithoutComments p f layers
   let literate = filter (not . isCodeLayer) layers
   let (terms, overlaps) = interleaveRanges (map Left code) (map Right literate)
 
   forM_ (map fst overlaps) $ \c ->
-    warning $ OverlappingTokensWarning { warnRange = getRange c }
+    warning $ OverlappingTokensWarning { warnRange = f <$ getRangeWithoutFile c }
 
-  (, coh) <$> (return $ forMaybe terms $ \case
-    Left t                           -> Just t
-    Right (Layer Comment interval s) -> Just $ TokTeX    (interval, s)
-    Right (Layer Markup  interval s) -> Just $ TokMarkup (interval, s)
-    Right (Layer Code _ _)           -> Nothing)
+  return . (,coh) . forMaybe terms $ \case
+      Left t                           -> Just t
+      Right (Layer Comment interval s) -> Just $ TokTeX    (f <$ interval, s)
+      Right (Layer Markup  interval s) -> Just $ TokMarkup (f <$ interval, s)
+      Right (Layer Code _ _)           -> Nothing
 
 
 parseLiterateFile
@@ -164,17 +164,16 @@ parseLiterateFile
   -> String
      -- ^ The file contents. Note that the file is /not/ read from
      -- disk.
-  -> PM (a, CohesionAttributes)
-parseLiterateFile po p path = parseLiterate p p . po (startPos (Just path))
+  -> PM (a, Attributes)
+parseLiterateFile po p path = parseLiterate p p (pure path) . po (startPos' ())
 
-parsePosString ::
-  Parser a -> Position -> String -> PM (a, CohesionAttributes)
+parsePosString :: Parser a -> Position -> String -> PM (a, Attributes)
 parsePosString p pos = wrapM . return . M.parsePosString pos (parseFlags p) normalLexState (parser p)
 
 -- | Extensions supported by `parseFile`.
 
-acceptableFileExts :: [String]
-acceptableFileExts = ".agda" : (fst <$> literateProcessors)
+agdaFileExtensions :: [String]
+agdaFileExtensions = ".agda" : (fst <$> literateProcessors)
 
 parseFile
   :: Show a
@@ -184,7 +183,7 @@ parseFile
   -> String
      -- ^ The file contents. Note that the file is /not/ read from
      -- disk.
-  -> PM ((a, CohesionAttributes), FileType)
+  -> PM ((a, Attributes), FileType)
 parseFile p file input =
   if ".agda" `List.isSuffixOf` path then
     (, AgdaFileType) <$> parseFileFromString (Strict.Just file) p input
@@ -195,7 +194,7 @@ parseFile p file input =
 
     go [] = throwError InvalidExtensionError
                    { errPath = file
-                   , errValidExts = acceptableFileExts
+                   , errValidExts = agdaFileExtensions
                    }
     go ((ext, (po, ft)) : pos)
       | ext `List.isSuffixOf` path =
