@@ -1,11 +1,10 @@
-{-# LANGUAGE ViewPatterns #-}
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 -- | Preprocessors for literate code formats.
 
 module Agda.Syntax.Parser.Literate
   ( literateProcessors
   , literateExtsShortList
-  , literateSrcFile
   , literateTeX
   , literateRsT
   , literateMd
@@ -33,7 +32,9 @@ import Agda.Syntax.Common
 import Agda.Syntax.Position
 
 import Agda.Utils.List
+import Agda.Utils.List1 (List1)
 import qualified Agda.Utils.List1 as List1
+import Agda.Utils.Singleton
 
 import Agda.Utils.Impossible
 
@@ -46,7 +47,7 @@ data LayerRole = Markup | Comment | Code
 
 data Layer = Layer
   { layerRole    :: LayerRole
-  , interval     :: Interval
+  , interval     :: IntervalWithoutFile
   , layerContent :: String
   } deriving Show
 
@@ -54,17 +55,17 @@ data Layer = Layer
 
 type Layers = [Layer]
 
-instance HasRange Layer where
-  getRange = getRange . interval
+instance HasRangeWithoutFile Layer where
+  getRangeWithoutFile = getRangeWithoutFile . interval
 
 -- | Annotates a tokenized string with position information.
 
-mkLayers :: Position -> [(LayerRole, String)] -> Layers
+mkLayers :: PositionWithoutFile -> [(LayerRole, String)] -> Layers
 mkLayers pos []            = emptyLiterate pos
 mkLayers pos ((_,"") : xs) = mkLayers pos xs
                              -- Empty layers are ignored.
 mkLayers pos ((ty,s) : xs) =
-  Layer ty (Interval pos next) s : mkLayers next xs
+  Layer ty (Interval () pos next) s : mkLayers next xs
   where
   next = movePosByString pos s
 
@@ -79,15 +80,11 @@ atomizeLayers = (fmap <$> ((,) . fst) <*> snd) <=< unMkLayers
 --
 --   > f : Processor
 --
---   prop> f pos s /= []
+--   proposition> f pos s /= []
 --
---   prop> f pos s >>= layerContent == s
+--   proposition> f pos s >>= layerContent == s
 
-type Processor = Position -> String -> [Layer]
-
-literateSrcFile :: [Layer] -> SrcFile
-literateSrcFile []                    = __IMPOSSIBLE__
-literateSrcFile (Layer{interval} : _) = getIntervalFile interval
+type Processor = PositionWithoutFile -> String -> [Layer]
 
 -- | List of valid extensions for literate Agda files, and their
 --   corresponding preprocessors.
@@ -103,6 +100,10 @@ literateProcessors =
     , (".tex", (literateTeX, TexFileType))
     , (".md",  (literateMd,  MdFileType ))
     , (".org", (literateOrg, OrgFileType))
+    , (".tree", (literateTree, TreeFileType))
+    -- For now, treat typst as markdown because they use the same
+    -- syntax for code blocks.
+    , (".typ", (literateMd,  TypstFileType))
     ]
 
 -- | Returns @True@ if the role corresponds to Agda code.
@@ -141,8 +142,8 @@ isBlank = (&&) <$> isSpace <*> (/= '\n')
 -- | Short list of extensions for literate Agda files.
 --   For display purposes.
 
-literateExtsShortList :: [String]
-literateExtsShortList = [".lagda"]
+literateExtsShortList :: List1 String
+literateExtsShortList = singleton ".lagda"
 
 -- | Returns a tuple consisting of the first line of the input, and the rest
 --   of the input.
@@ -156,8 +157,8 @@ caseLine a k = \case
 
 -- | Canonical decomposition of an empty literate file.
 
-emptyLiterate :: Position -> [Layer]
-emptyLiterate pos = [Layer Markup (Interval pos pos) ""]
+emptyLiterate :: PositionWithoutFile -> [Layer]
+emptyLiterate pos = [Layer Markup (Interval () pos pos) ""]
 
 -- | Create a regular expression that:
 --   - Must match the whole string
@@ -170,7 +171,7 @@ rex s =
 
 -- | Preprocessor for literate TeX.
 
-literateTeX :: Position -> String -> [Layer]
+literateTeX :: Processor
 literateTeX pos s = mkLayers pos (tex s)
   where
   tex :: String -> [(LayerRole, String)]
@@ -196,7 +197,7 @@ literateTeX pos s = mkLayers pos (tex s)
 
 -- | Preprocessor for Markdown.
 
-literateMd :: Position -> String -> [Layer]
+literateMd :: Processor
 literateMd pos s = mkLayers pos $ md s
   where
   md :: String -> [(LayerRole, String)]
@@ -234,7 +235,7 @@ literateMd pos s = mkLayers pos $ md s
 
 -- | Preprocessor for reStructuredText.
 
-literateRsT :: Position -> String -> [Layer]
+literateRsT :: Processor
 literateRsT pos s = mkLayers pos $ rst s
   where
   rst :: String -> [(LayerRole, String)]
@@ -281,7 +282,7 @@ literateRsT pos s = mkLayers pos $ rst s
 
 -- | Preprocessor for Org mode documents.
 
-literateOrg :: Position -> String -> [Layer]
+literateOrg :: Processor
 literateOrg pos s = mkLayers pos $ org s
   where
   org :: String -> [(LayerRole, String)]
@@ -309,3 +310,29 @@ literateOrg pos s = mkLayers pos $ org s
   rex' :: String -> Regex
   -- Source blocks start with `#+begin_src` but the casing does not matter.
   rex' = makeRegexOpts blankCompOpt{newSyntax = True, caseSensitive = False} blankExecOpt
+
+-- | Preprocessor for Forester documents
+
+literateTree :: Processor
+literateTree pos s = mkLayers pos (tree s)
+  where
+  tree :: String -> [(LayerRole, String)]
+  tree = caseLine [] $ \ line rest ->
+    case tree_begin `matchM` line of
+      Just (getAllTextSubmatches -> [_, pre, _, markup, whitespace]) ->
+        (Comment, pre) : (Markup, markup) :
+        (Code, whitespace) : code rest
+      Just _  -> __IMPOSSIBLE__
+      Nothing -> (Comment, line) : tree rest
+
+  tree_begin = rex "(([^\\%]|\\\\.)*)(\\\\agda\\{[^\n]*)(\n)?"
+
+  code :: String -> [(LayerRole, String)]
+  code = caseLine [] $ \ line rest ->
+    case tree_end `matchM` line of
+      Just (getAllTextSubmatches -> [_, code, markup, post]) ->
+        (Code, code) : (Markup, markup) : (Comment, post) : tree rest
+      Just _  -> __IMPOSSIBLE__
+      Nothing -> (Code, line) : code rest
+
+  tree_end = rex "([[:blank:]]*)(\\})(.*)"
