@@ -10,6 +10,7 @@ module Agda.Syntax.TopLevelModuleName
   ) where
 
 import Agda.Syntax.TopLevelModuleName.Boot
+import Agda.Syntax.Abstract.Name (isNoName)
 
 import Control.DeepSeq
 
@@ -42,6 +43,9 @@ import Agda.Utils.Size
 data RawTopLevelModuleName = RawTopLevelModuleName
   { rawModuleNameRange :: Range
   , rawModuleNameParts :: TopLevelModuleNameParts
+  , rawModuleNameInferred :: !Bool
+      -- ^ Was this module name constructed from a file name
+      --   rather than declared in the file?
   }
   deriving (Show, Generic)
 
@@ -62,11 +66,11 @@ instance HasRange RawTopLevelModuleName where
   getRange = rawModuleNameRange
 
 instance SetRange RawTopLevelModuleName where
-  setRange r (RawTopLevelModuleName _ x) = RawTopLevelModuleName r x
+  setRange r (RawTopLevelModuleName _ x z) = RawTopLevelModuleName r x z
 
 instance KillRange RawTopLevelModuleName where
-  killRange (RawTopLevelModuleName _ x) =
-    RawTopLevelModuleName noRange x
+  killRange (RawTopLevelModuleName _ x z) =
+    RawTopLevelModuleName noRange x z
 
 instance C.IsNoName RawTopLevelModuleName where
   isNoName m = rawModuleNameParts m == singleton "_"
@@ -74,7 +78,7 @@ instance C.IsNoName RawTopLevelModuleName where
 -- | The 'Range' is not forced.
 
 instance NFData RawTopLevelModuleName where
-  rnf (RawTopLevelModuleName _ x) = rnf x
+  rnf (RawTopLevelModuleName _ x _) = rnf x
 
 -- | Turns a raw top-level module name into a string.
 
@@ -97,6 +101,7 @@ rawTopLevelModuleNameForQName q = RawTopLevelModuleName
   { rawModuleNameRange = getRange q
   , rawModuleNameParts =
       fmap (T.pack . C.nameToRawName) $ C.qnameParts q
+  , rawModuleNameInferred = C.isNoName q
   }
 
 -- | Computes the 'RawTopLevelModuleName' corresponding to the given
@@ -106,13 +111,13 @@ rawTopLevelModuleNameForQName q = RawTopLevelModuleName
 
 rawTopLevelModuleNameForModuleName ::
   A.ModuleName -> RawTopLevelModuleName
-rawTopLevelModuleNameForModuleName (A.MName []) = __IMPOSSIBLE__
-rawTopLevelModuleNameForModuleName (A.MName ms) =
+rawTopLevelModuleNameForModuleName x@(A.MName ms) =
   List1.ifNull ms __IMPOSSIBLE__ $ \ms ->
   RawTopLevelModuleName
     { rawModuleNameRange = getRange ms
     , rawModuleNameParts =
         fmap (T.pack . C.nameToRawName . A.nameConcrete) ms
+    , rawModuleNameInferred = isNoName x
     }
 
 -- | Computes the top-level module name.
@@ -156,6 +161,7 @@ rawTopLevelModuleName :: TopLevelModuleName -> RawTopLevelModuleName
 rawTopLevelModuleName m = RawTopLevelModuleName
   { rawModuleNameRange = moduleNameRange m
   , rawModuleNameParts = moduleNameParts m
+  , rawModuleNameInferred = moduleNameInferred m
   }
 
 -- | Converts a raw top-level module name and a hash to a top-level
@@ -171,6 +177,7 @@ unsafeTopLevelModuleName m h = TopLevelModuleName
   { moduleNameRange = rawModuleNameRange m
   , moduleNameParts = rawModuleNameParts m
   , moduleNameId    = h
+  , moduleNameInferred = rawModuleNameInferred m
   }
 
 -- | A corresponding 'C.QName'. The range of each 'Name' part is the
