@@ -1,19 +1,24 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 -- | Functions for abstracting terms over other terms.
+
 module Agda.TypeChecking.Abstract where
 
+import Prelude hiding ( null )
+
 import Control.Monad
 import Control.Monad.Except
 
-import Data.Function (on)
+import Data.Function ( on )
 import qualified Data.HashMap.Strict as HMap
 
 import Agda.Syntax.Common
+import Agda.Syntax.Position ( Range )
 import Agda.Syntax.Internal
 
+import Agda.TypeChecking.Free ( freeIn )
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin ( equalityUnview, primEqualityName )
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.CheckInternal
 import Agda.TypeChecking.Conversion
@@ -21,20 +26,35 @@ import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Sort
 import Agda.TypeChecking.Telescope
+import Agda.TypeChecking.Warnings ( warning )
 
 import Agda.Utils.Functor
 import Agda.Utils.List ( splitExactlyAt, dropEnd )
+import Agda.Utils.Null
+
 import Agda.Utils.Impossible
 
--- | @abstractType a v b[v] = b@ where @a : v@.
-abstractType :: Type -> Term -> Type -> TCM Type
-abstractType a v (El s b) = El (absTerm v s) <$> abstractTerm a v (sort s) b
+-- | @abstractType r a v b[v] = b@ where @a : v@.
+abstractType ::
+     Range     -- ^ Range of the @rewrite@ expression, if any, otherwise empty.
+  -> Type      -- ^ Type of the term to abstract.
+  -> Term      -- ^ Term to abstract.
+  -> Type      -- ^ Type to abstract in.
+  -> TCM Type  -- ^ Type with hole (de Bruijn index 0) for the abstracted term.
+abstractType r a v (El s b) = do
+
+  c <- El (absTerm 0 v s) <$> abstractTerm a v (sort s) b
+  unless (null r || 0 `freeIn` c) do
+    -- Andreas, 2025-07-03, issue #7973
+    -- If with abstraction did not abstract anything, warn the user.
+    setCurrentRange r $ warning RewritesNothing
+  return c
 
 -- | @piAbstractTerm NotHidden v a b[v] = (w : a) -> b[w]@
 --   @piAbstractTerm Hidden    v a b[v] = {w : a} -> b[w]@
 piAbstractTerm :: ArgInfo -> Term -> Type -> Type -> TCM Type
 piAbstractTerm info v a b = do
-  fun <- mkPi (setArgInfo info $ defaultDom ("w", a)) <$> abstractType a v b
+  fun <- mkPi (setArgInfo info $ defaultDom ("w", a)) <$> abstractType empty a v b
   reportSDoc "tc.abstract" 50 $
     sep [ "piAbstract" <+> sep [ prettyTCM v <+> ":", nest 2 $ prettyTCM a ]
         , nest 2 $ "from" <+> prettyTCM b
@@ -56,14 +76,14 @@ piAbstractTerm info v a b = do
 piAbstract :: Arg (Term, EqualityView) -> Type -> TCM Type
 piAbstract (Arg info (v, OtherType a)) b = piAbstractTerm info v a b
 piAbstract (Arg info (v, IdiomType a)) b = do
-  b  <- raise 1 <$> abstractType a v b
+  b  <- raise 1 <$> abstractType empty a v b
   eq <- addContext ("w" :: String, defaultDom a) $ do
     -- manufacture the type @w ≡ v@
     eqName <- primEqualityName
     eqTy <- defType <$> getConstInfo eqName
     -- E.g. @eqTy = eqTel → Set a@ where @eqTel = {a : Level} {A : Set a} (x y : A)@.
     TelV eqTel _ <- telView eqTy
-    tel  <- newTelMeta (telFromList $ dropEnd 2 $ telToList eqTel)
+    tel  <- newTelMeta (telFromList $ dropEnd 3 $ telToList eqTel)
     let eq = Def eqName $ map Apply
                  $ map (setHiding Hidden) tel
                  -- we write `v ≡ w` because this equality is typically used to
@@ -71,24 +91,32 @@ piAbstract (Arg info (v, IdiomType a)) b = do
                  -- in a with-clause.
                  -- If we were to write `w ≡ v`, we would often need to take the
                  -- symmetric of the proof we get to make use of `rewrite`.
-                 ++ [ defaultArg (raise 1 v)
+                 ++ [ setHiding Hidden $ defaultArg $ raise 1 $ unEl a
+                    , defaultArg (raise 1 v)
                     , defaultArg (var 0)
                     ]
+    -- Since the result of this function will be type-checked in
+    -- `withFunctionType`, we can be a little lazy here and put
+    -- a meta for the sort.
     sort <- newSortMeta
-    let ty = El sort eq
-    ty <$ checkType ty
+    return $ El sort eq
 
   pure $ mkPi (setHiding (getHiding info) $ defaultDom ("w", a))
        $ mkPi (setHiding NotHidden        $ defaultDom ("eq", eq))
        $ b
-piAbstract (Arg info (prf, EqualityViewType eqt@(EqualityTypeData _ _ _ (Arg _ a) v _))) b = do
+piAbstract (Arg info (prf, EqualityViewType eqt@(EqualityTypeData r _ _ _ (Arg _ a) v _))) b = do
   s <- sortOf a
   let prfTy :: Type
       prfTy = equalityUnview eqt
       vTy   = El s a
-  b <- abstractType prfTy prf b
+  -- Andreas, 2025-07-03, issue #7973
+  -- We alert the user when the lhs of the equality proof could not be abstracted
+  -- but not when the equality proof itself could not be abstracted.
+  -- Only the former means that the rewrite did not fire.
+  b <- abstractType empty prfTy prf b  -- @empty@ means do not warn
   b <- addContext ("w" :: String, defaultDom prfTy) $
-         abstractType (raise 1 vTy) (unArg $ raise 1 v) b
+         -- Passing range @r@ here means warn if abstraction failed to abstract anything.
+         abstractType r (raise 1 vTy) (unArg $ raise 1 v) b
   return . funType "lhs" vTy . funType "equality" eqTy' . swap01 $ b
   where
     funType str a = mkPi $ setArgInfo info $ defaultDom (str, a)
@@ -139,7 +167,7 @@ abstractTerm a u@Con{} b v = do
         , nest 2 $ sep [ (text . show) v <+> ":", nest 2 $ (text . show) b ] ]
 
   hole <- qualify <$> currentModule <*> freshName_ ("hole" :: String)
-  noMutualBlock $ addConstant' hole defaultArgInfo hole a defaultAxiom
+  noMutualBlock $ addConstant' hole defaultArgInfo a defaultAxiom
 
   args <- map Apply <$> getContextArgs
   let n = length args
@@ -171,21 +199,26 @@ abstractTerm a u@Con{} b v = do
         return v
   reportSDoc "tc.abstract" 50 $ "Resulting abstraction" <?> prettyTCM res
   modifySignature $ updateDefinitions $ HMap.delete hole
-  return $ absTerm (Def hole args) res
+  return $ absTerm 0 (Def hole args) res
 
-abstractTerm _ u _ v = return $ absTerm u v -- Non-constructors can use untyped abstraction
+abstractTerm _ u _ v = return $ absTerm 0 u v -- Non-constructors can use untyped abstraction
 
 class AbsTerm a where
-  -- | @subst u . absTerm u == id@
-  absTerm :: Term -> a -> a
+  -- | @subst j u . absTerm j u == id@
+  absTerm ::
+       Nat   -- ^ De Bruijn index that should be the placeholder for the abstracted term.
+    -> Term  -- ^ Term to abstract.
+    -> a     -- ^ Where to abstract.
+    -> a     -- ^ If the given de Bruijn index is free in the result, abstraction actually happened.
 
 instance AbsTerm Term where
-  absTerm u v | Just es <- u `isPrefixOf` v = Var 0 $ absT es
-              | otherwise                   =
+  absTerm j u v
+    | Just es <- u `isPrefixOf` v = Var j $ absT es
+    | otherwise                   =
     case v of
 -- Andreas, 2013-10-20: the original impl. works only at base types
---    v | u == v  -> Var 0 []  -- incomplete see succeed/WithOfFunctionType
-      Var i vs    -> Var (i + 1) $ absT vs
+--    v | u == v  -> Var j []  -- incomplete see succeed/WithOfFunctionType
+      Var i vs    -> Var (if i < j then i else i + 1) $ absT vs
       Lam h b     -> Lam h $ absT b
       Def c vs    -> Def c $ absT vs
       Con c ci vs -> Con c ci $ absT vs
@@ -198,20 +231,19 @@ instance AbsTerm Term where
       Dummy s es   -> Dummy s $ absT es
       where
         absT :: AbsTerm b => b -> b
-        absT x = absTerm u x
+        absT x = absTerm j u x
 
 instance AbsTerm Type where
-  absTerm u (El s v) = El (absTerm u s) (absTerm u v)
+  absTerm j u (El s v) = El (absTerm j u s) (absTerm j u v)
 
 instance AbsTerm Sort where
-  absTerm u = \case
-    Type n     -> Type $ absS n
-    Prop n     -> Prop $ absS n
-    s@(Inf f n)-> s
-    SSet n     -> SSet $ absS n
+  absTerm j u = \case
+    Univ u n   -> Univ u $ absS n
+    s@Inf{}    -> s
     SizeUniv   -> SizeUniv
     LockUniv   -> LockUniv
-    IntervalUniv -> LockUniv
+    LevelUniv  -> LevelUniv
+    IntervalUniv -> IntervalUniv
     PiSort a s1 s2 -> PiSort (absS a) (absS s1) (absS s2)
     FunSort s1 s2 -> FunSort (absS s1) (absS s2)
     UnivSort s -> UnivSort $ absS s
@@ -220,35 +252,35 @@ instance AbsTerm Sort where
     s@DummyS{} -> s
     where
       absS :: AbsTerm b => b -> b
-      absS x = absTerm u x
+      absS x = absTerm j u x
 
 instance AbsTerm Level where
-  absTerm u (Max n as) = Max n $ absTerm u as
+  absTerm j u (Max n as) = Max n $ absTerm j u as
 
 instance AbsTerm PlusLevel where
-  absTerm u (Plus n l) = Plus n $ absTerm u l
+  absTerm j u (Plus n l) = Plus n $ absTerm j u l
 
 instance AbsTerm a => AbsTerm (Elim' a) where
-  absTerm = fmap . absTerm
+  absTerm j = fmap . absTerm j
 
 instance AbsTerm a => AbsTerm (Arg a) where
-  absTerm = fmap . absTerm
+  absTerm j = fmap . absTerm j
 
 instance AbsTerm a => AbsTerm (Dom a) where
-  absTerm = fmap . absTerm
+  absTerm j = fmap . absTerm j
 
 instance AbsTerm a => AbsTerm [a] where
-  absTerm = fmap . absTerm
+  absTerm j = fmap . absTerm j
 
 instance AbsTerm a => AbsTerm (Maybe a) where
-  absTerm = fmap . absTerm
+  absTerm j = fmap . absTerm j
 
 instance (TermSubst a, AbsTerm a) => AbsTerm (Abs a) where
-  absTerm u (NoAbs x v) = NoAbs x $ absTerm u v
-  absTerm u (Abs   x v) = Abs x $ swap01 $ absTerm (raise 1 u) v
+  absTerm j u (NoAbs x v) = NoAbs x $ absTerm j u v
+  absTerm j u (Abs   x v) = Abs x $ absTerm (j + 1) (raise 1 u) v
 
 instance (AbsTerm a, AbsTerm b) => AbsTerm (a, b) where
-  absTerm u (x, y) = (absTerm u x, absTerm u y)
+  absTerm j u (x, y) = (absTerm j u x, absTerm j u y)
 
 -- | This swaps @var 0@ and @var 1@.
 swap01 :: TermSubst a => a -> a
@@ -292,11 +324,10 @@ instance EqualSy PlusLevel where
 
 instance EqualSy Sort where
   equalSy = curry $ \case
-    (Type l    , Type l'     ) -> equalSy l l'
-    (Prop l    , Prop l'     ) -> equalSy l l'
-    (Inf f m   , Inf f' n    ) -> f == f' && m == n
-    (SSet l    , SSet l'     ) -> equalSy l l'
+    (Univ u l  , Univ u' l'  ) -> u == u' && equalSy l l'
+    (Inf u m   , Inf u' n    ) -> u == u' && m == n
     (SizeUniv  , SizeUniv    ) -> True
+    (LevelUniv , LevelUniv   ) -> True
     (PiSort a b c, PiSort a' b' c') -> equalSy a a' && equalSy b b' && equalSy c c'
     (FunSort a b, FunSort a' b') -> equalSy a a' && equalSy b b'
     (UnivSort a, UnivSort a' ) -> equalSy a a'
