@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 -- | Case trees.
 --
 --   After coverage checking, pattern matching is translated
@@ -23,7 +25,7 @@ import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
 import Agda.Utils.Null
-import Agda.Utils.Pretty
+import Agda.Syntax.Common.Pretty
 
 import Agda.Utils.Impossible
 
@@ -44,10 +46,10 @@ data Case c = Branches
     --   present, there should not be any conBranches or litBranches.
   , litBranches    :: Map Literal c
     -- ^ Map from literal to case subtree.
-  , catchAllBranch :: Maybe c
+  , catchallBranch :: Maybe c
     -- ^ (Possibly additional) catch-all clause.
   , fallThrough :: Maybe Bool
-    -- ^ (if True) In case of non-canonical argument use catchAllBranch.
+    -- ^ (if True) In case of non-canonical argument use catchallBranch.
   , lazyMatch :: Bool
     -- ^ Lazy pattern match. Requires single (non-copattern) branch with no lit
     --   branches and no catch-all.
@@ -86,8 +88,8 @@ etaCase c x = Branches False Map.empty (Just (c, x)) Map.empty Nothing (Just Fal
 projCase :: QName -> c -> Case c
 projCase c x = Branches True (Map.singleton c $ WithArity 0 x) Nothing Map.empty Nothing (Just False) False
 
-catchAll :: c -> Case c
-catchAll x = Branches False Map.empty Nothing Map.empty (Just x) (Just True) False
+catchall :: c -> Case c
+catchall x = Branches False Map.empty Nothing Map.empty (Just x) (Just True) False
 
 -- | Check that the requirements on lazy matching (single inductive case) are
 --   met, and set lazy to False otherwise.
@@ -95,19 +97,19 @@ checkLazyMatch :: Case c -> Case c
 checkLazyMatch b = b { lazyMatch = lazyMatch b && requirements }
   where
     requirements = and
-      [ null (catchAllBranch b)
+      [ null (catchallBranch b)
       , Map.size (conBranches b) <= 1
       , null (litBranches b)
       , not $ projPatterns b ]
 
 -- | Check whether a case tree has a catch-all clause.
-hasCatchAll :: CompiledClauses -> Bool
-hasCatchAll = getAny . loop
+hasCatchall :: CompiledClauses -> Bool
+hasCatchall = getAny . loop
   where
   loop cc = case cc of
     Fail{}    -> mempty
     Done{}    -> mempty
-    Case _ br -> maybe (foldMap loop br) (const $ Any True) $ catchAllBranch br
+    Case _ br -> maybe (foldMap loop br) (const $ Any True) $ catchallBranch br
 
 -- | Check whether a case tree has any projection patterns
 hasProjectionPatterns :: CompiledClauses -> Bool
@@ -196,9 +198,9 @@ instance KillRange c => KillRange (Case c) where
     b lazy
 
 instance KillRange CompiledClauses where
-  killRange (Case i br) = killRange2 Case i br
-  killRange (Done xs v) = killRange2 Done xs v
-  killRange (Fail xs)   = killRange1 Fail xs
+  killRange (Case i br) = killRangeN Case i br
+  killRange (Done xs v) = killRangeN Done xs v
+  killRange (Fail xs)   = killRangeN Fail xs
 
 -- * TermLike instances
 
