@@ -1,17 +1,19 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Constraints where
 
 import Prelude hiding (null)
 
-import Control.Monad
-import Control.Monad.Except
+import Control.Monad.Except ( MonadError )
 
 import qualified Data.List as List
 import qualified Data.Set as Set
 import Data.Either
 
 import Agda.Syntax.Common
+import Agda.Syntax.Common.Pretty ( prettyShow )
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Monad
@@ -21,7 +23,6 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.LevelConstraints
 import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Sort
-import Agda.TypeChecking.MetaVars.Mention
 import Agda.TypeChecking.Warnings
 
 import Agda.TypeChecking.Irrelevance
@@ -36,12 +37,12 @@ import {-# SOURCE #-} Agda.TypeChecking.Lock
 import {-# SOURCE #-} Agda.TypeChecking.CheckInternal ( checkType )
 
 import Agda.Utils.CallStack ( withCurrentCallStack )
-import Agda.Utils.Functor
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
-import Agda.Utils.Null
-import Agda.Utils.Pretty (prettyShow)
+import Agda.Utils.Null ()
 import qualified Agda.Utils.ProfileOptions as Profile
+import Agda.Utils.Singleton
 
 import Agda.Utils.Impossible
 
@@ -64,10 +65,11 @@ addConstraintTCM unblock c = do
         , "unblocker: " , prettyTCM unblock
         ]
       -- Jesper, 2022-10-22: We should never block on a meta that is
-      -- already solved. However, currently this is not always
-      -- strictly adhered to. TODO: pull off the band-aid and
-      -- uncomment the following line.
-      -- whenM (anyM (Set.toList $ allBlockingMetas unblock) isInstantiatedMeta) __IMPOSSIBLE__
+      -- already solved.
+      forM_ (allBlockingMetas unblock) $ \ m ->
+        whenM (isInstantiatedMeta m) $ do
+          reportSDoc "tc.constr.add" 5 $ "Attempted to block on solved meta" <+> prettyTCM m
+          __IMPOSSIBLE__
       -- Need to reduce to reveal possibly blocking metas
       c <- reduce =<< instantiateFull c
       caseMaybeM (simpl c) {-no-} (addConstraint' unblock c) $ {-yes-} \ cs -> do
@@ -95,13 +97,13 @@ addConstraintTCM unblock c = do
           -- Get all level constraints.
           lvlcs <- instantiateFull =<< do
             List.filter (isLvl . clValue) . map theConstraint <$> getAllConstraints
-          unless (null lvlcs) $ do
+          List1.ifNull lvlcs (return Nothing) $ {-else-} \ lvlcs -> do
             reportSDoc "tc.constr.lvl" 40 $ vcat
               [ "simplifying level constraint" <+> prettyTCM c
               , nest 2 $ hang "using" 2 $ prettyTCM lvlcs
               ]
-          -- Try to simplify @c@ using the other constraints.
-          return $ simplifyLevelConstraint c $ map clValue lvlcs
+            -- Try to simplify @c@ using the other constraints.
+            return $ simplifyLevelConstraint c $ fmap clValue lvlcs
         | otherwise = return Nothing
 
 wakeConstraintsTCM :: (ProblemConstraint-> WakeUp) -> TCM ()
@@ -133,6 +135,7 @@ stealConstraintsTCM pid = do
   modifySleepingConstraints $ List.map rename
 
 
+{-# SPECIALIZE noConstraints :: TCM a -> TCM a #-}
 -- | Don't allow the argument to produce any blocking constraints.
 --
 -- WARNING: this does not mean that the given computation cannot
@@ -141,13 +144,26 @@ stealConstraintsTCM pid = do
 noConstraints
   :: (MonadConstraint m, MonadWarning m, MonadError TCErr m, MonadFresh ProblemId m)
   => m a -> m a
-noConstraints problem = do
+noConstraints = noConstraints' False
+
+-- | As noConstraints but also fail for non-blocking constraints.
+reallyNoConstraints
+  :: (MonadConstraint m, MonadWarning m, MonadError TCErr m, MonadFresh ProblemId m)
+  => m a -> m a
+reallyNoConstraints = noConstraints' True
+
+noConstraints'
+  :: (MonadConstraint m, MonadWarning m, MonadError TCErr m, MonadFresh ProblemId m)
+  => Bool -> m a -> m a
+noConstraints' includingNonBlocking problem = do
   (pid, x) <- newProblem problem
-  cs <- getConstraintsForProblem pid
-  unless (null cs) $ do
-    withCurrentCallStack $ \loc -> do
-      w <- warning'_ loc (UnsolvedConstraints cs)
-      typeError' loc $ NonFatalErrors [ w ]
+  let counts | includingNonBlocking = const True
+             | otherwise            = isBlockingConstraint . clValue . theConstraint
+  cs <- List.filter counts <$> getConstraintsForProblem pid
+  List1.ifNull cs (pure ()) \ cs -> do
+    withCurrentCallStack \ loc -> do
+      w <- warning'_ loc $ UnsolvedConstraints cs
+      typeError' loc $ NonFatalErrors $ singleton w
   return x
 
 -- | Run a computation that should succeeds without constraining
@@ -163,6 +179,7 @@ nonConstraining ::
   ) => m a -> m a
 nonConstraining = dontAssignMetas . noConstraints
 
+{-# SPECIALIZE newProblem :: TCM a -> TCM (ProblemId, a) #-}
 -- | Create a fresh problem for the given action.
 newProblem
   :: (MonadFresh ProblemId m, MonadConstraint m)
@@ -175,6 +192,7 @@ newProblem action = do
   solveAwakeConstraints
   return (pid, x)
 
+{-# SPECIALIZE newProblem_ :: TCM a -> TCM ProblemId #-}
 newProblem_
   :: (MonadFresh ProblemId m, MonadConstraint m)
   => m a -> m ProblemId
@@ -201,6 +219,7 @@ whenConstraints action handler =
     stealConstraints pid
     handler
 
+{-# SPECIALIZE wakeupConstraints :: MetaId -> TCM () #-}
 -- | Wake up the constraints depending on the given meta.
 wakeupConstraints :: MonadMetaSolver m => MetaId -> m ()
 wakeupConstraints x = do
@@ -280,14 +299,13 @@ solveConstraint_ (UnBlock m)                =   -- alwaysUnblock since these hav
       -- Ulf, 2018-04-30: The size solver shouldn't touch blocked terms! They have
       -- a twin meta that it's safe to solve.
       InstV{} -> __IMPOSSIBLE__
-      -- Open (whatever that means)
-      Open -> __IMPOSSIBLE__
-      OpenInstance -> __IMPOSSIBLE__
-solveConstraint_ (FindInstance m cands) = findInstance m cands
-solveConstraint_ (CheckFunDef d i q cs _err) = withoutCache $
+      OpenMeta{} -> __IMPOSSIBLE__
+solveConstraint_ (FindInstance _ m cands) = findInstance m cands
+solveConstraint_ (ResolveInstanceHead q) = resolveInstanceHead q
+solveConstraint_ (CheckFunDef i q cs _err) = withoutCache $
   -- re #3498: checking a fundef would normally be cached, but here it's
   -- happening out of order so it would only corrupt the caching log.
-  checkFunDef d i q cs
+  checkFunDef i q cs
 solveConstraint_ (CheckLockedVars a b c d)   = checkLockedVars a b c d
 solveConstraint_ (HasBiggerSort a)      = hasBiggerSort a
 solveConstraint_ (HasPTSRule a b)       = hasPTSRule a b
@@ -299,11 +317,12 @@ solveConstraint_ (UsableAtModality cc ms mod t) = usableAtModality' ms cc mod t
 checkTypeCheckingProblem :: TypeCheckingProblem -> TCM Term
 checkTypeCheckingProblem = \case
   CheckExpr cmp e t              -> checkExpr' cmp e t
-  CheckArgs cmp eh r args t0 t1 k -> checkArguments cmp eh r args t0 t1 k
-  CheckProjAppToKnownPrincipalArg cmp e o ds args t k v0 pt patm ->
-    checkProjAppToKnownPrincipalArg cmp e o ds args t k v0 pt patm
+  CheckArgs cmp eh hd args t0 t1 k -> checkArguments cmp eh hd args t0 t1 k
+  CheckProjAppToKnownPrincipalArg cmp e o ds hd args t k v0 pt patm ->
+    checkProjAppToKnownPrincipalArg cmp e o ds hd args t k v0 pt patm
   CheckLambda cmp args body target -> checkPostponedLambda cmp args body target
   DoQuoteTerm cmp et t           -> doQuoteTerm cmp et t
+  DisambiguateConstructor bcd k  -> disambiguateConstructor' bcd k
 
 debugConstraints :: TCM ()
 debugConstraints = verboseS "tc.constr" 50 $ do
@@ -313,3 +332,29 @@ debugConstraints = verboseS "tc.constr" 50 $ do
     [ "Current constraints"
     , nest 2 $ vcat [ "awake " <+> vcat (map prettyTCM awake)
                     , "asleep" <+> vcat (map prettyTCM sleeping) ] ]
+
+-- Update the blocker after some instantiation or pruning might have happened.
+updateBlocker :: (PureTCM m) => Blocker -> m Blocker
+updateBlocker = instantiate
+
+addAndUnblocker :: (PureTCM m, MonadBlock m) => Blocker -> m a -> m a
+addAndUnblocker u
+  | u == alwaysUnblock = id
+  | otherwise          = catchPatternErr $ \ u' -> do
+      u <- updateBlocker u
+      patternViolation (unblockOnBoth u u')
+
+addOrUnblocker :: (PureTCM m, MonadBlock m) => Blocker -> m a -> m a
+addOrUnblocker u
+  | u == neverUnblock = id
+  | otherwise         = catchPatternErr $ \ u' -> do
+      u <- updateBlocker u
+      patternViolation (unblockOnEither u u')
+
+-- Reduce a term and call the continuation. If the continuation is
+-- blocked, the whole call is blocked either on what blocked the reduction
+-- or on what blocked the continuation (using `blockedOnEither`).
+withReduced
+  :: (Reduce a, IsMeta a, PureTCM m, MonadBlock m)
+  => a -> (a -> m b) -> m b
+withReduced a cont = ifBlocked a (\b a' -> addOrUnblocker b $ cont a') (\_ a' -> cont a')
