@@ -9,6 +9,8 @@ import Control.Monad.Trans (lift)
 import Control.Monad.Trans.Maybe
 
 import Data.Monoid (All(..))
+import Data.IntMap (IntMap)
+import qualified Data.IntMap as IntMap
 import Data.Map (Map)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
@@ -25,6 +27,7 @@ import Agda.TypeChecking.Reduce (instantiate)
 
 import Agda.Utils.Functor
 import Agda.Utils.List
+import Agda.Utils.List1 (List1)
 import Agda.Utils.Maybe
 import Agda.Syntax.Common.Pretty
 
@@ -114,24 +117,39 @@ matchDisplayForm d@(Display n ps v) es
   | length ps > length es = mzero
   | otherwise             = do
       let (es0, es1) = splitAt (length ps) es
-      mm <- match (Window 0 n) ps es0
+
+      -- The 'Display' constructor acts as though it binds the pattern
+      -- pattern variables up to 'n', so a match like
+      --
+      --   Display 1 [@1 @0] x =? [@0 _]
+      --
+      -- should work (it didn't; see LiftDisplayIntermediate). In
+      -- effect, this is because the LHS patterns are in some context
+      -- "Γ . @0", but the RHS term is only in context Γ.
+      --
+      -- Therefore, we should raise the RHS term by the number of
+      -- pattern variables, to bring it into the context of the
+      -- patterns.
+
+      mm <- match (Window 0 n) ps (raise n es0)
       us <- forM [0 .. n - 1] $ \ i -> do
-              -- #5294: Fail if we don't have bindings for all variables. This can
-              --        happen outside parameterised modules when some of the parameters
-              --        are not used in the lhs.
-              Just u <- return $ Map.lookup i mm
-              return u
+        -- #5294: Fail if we don't have bindings for all variables. This can
+        --        happen outside parameterised modules when some of the parameters
+        --        are not used in the lhs.
+        Just u <- return $ IntMap.lookup i mm
+        -- Note that the RHS terms are independent of the pattern variables.
+        return (applySubst (strengthenS __IMPOSSIBLE__ n) <$> u)
       return (d, substWithOrigin (parallelS $ map woThing us) us v `applyE` es1)
 
-type MatchResult = Map Int (WithOrigin Term)
+type MatchResult = IntMap (WithOrigin Term)
 
 unionMatch :: Monad m => MatchResult -> MatchResult -> MaybeT m MatchResult
 unionMatch m1 m2
-  | null (Map.intersection m1 m2) = return $ Map.union m1 m2
+  | null (IntMap.intersection m1 m2) = return $ IntMap.union m1 m2
   | otherwise = mzero  -- Non-linear pattern, fail for now.
 
 unionsMatch :: Monad m => [MatchResult] -> MaybeT m MatchResult
-unionsMatch = foldM unionMatch Map.empty
+unionsMatch = foldM unionMatch mempty
 
 data Window = Window {dbLo, dbHi :: Nat}
 
@@ -152,15 +170,17 @@ class Match a where
   match :: MonadDisplayForm m => Window -> a -> a -> MaybeT m MatchResult
 
 instance Match a => Match [a] where
-  match n xs ys = unionsMatch =<< zipWithM (match n) xs ys
+  match n xs ys
+    | length xs == length ys = unionsMatch =<< zipWithM (match n) xs ys
+    | otherwise = mzero
 
 instance Match a => Match (Arg a) where
-  match n p v = Map.map (setOrigin (getOrigin v)) <$> match n (unArg p) (unArg v)
+  match n p v = IntMap.map (setOrigin (getOrigin v)) <$> match n (unArg p) (unArg v)
 
 instance Match a => Match (Elim' a) where
   match n p v =
     case (p, v) of
-      (Proj _ f, Proj _ f') | f == f' -> return Map.empty
+      (Proj _ f, Proj _ f') | f == f' -> return mempty
       _ | Just a  <- isApplyElim p
         , Just a' <- isApplyElim v    -> match n a a'
       -- we do not care to differentiate between Apply and IApply for
@@ -169,14 +189,14 @@ instance Match a => Match (Elim' a) where
 
 instance Match Term where
   match w p v = lift (instantiate v) >>= \ v -> case (unSpine p, unSpine v) of
-    (Var i [], v)    | Just j <- inWindow w i -> return $ Map.singleton j (WithOrigin Inserted v)
+    (Var i [], v)    | Just j <- inWindow w i -> return $ IntMap.singleton j (WithOrigin Inserted v)
     (Var i (_:_), v) | Just{} <- inWindow w i -> mzero  -- Higher-order pattern, fail for now.
     (Var i ps, Var j vs) | i == j  -> match w ps vs
     (Def c ps, Def d vs) | c == d  -> match w ps vs
     (Con c _ ps, Con d _ vs) | c == d -> match w ps vs
-    (Lit l, Lit l')      | l == l' -> return Map.empty
+    (Lit l, Lit l')      | l == l' -> return mempty
     (Lam h p, Lam h' v)  | h == h' -> match (shiftWindow w) (unAbs p) (unAbs v)
-    (p, v)               | p == v  -> return Map.empty  -- TODO: this is wrong (this is why we lifted the rhs before)
+    (p, v)               | p == v  -> return mempty  -- TODO: this is wrong (this is why we lifted the rhs before)
     (p, Level l)                   -> match w p =<< reallyUnLevelView l
     (Sort ps, Sort pv)             -> match w ps pv
     (p, Sort (Type v))             -> match w p =<< reallyUnLevelView v
@@ -185,7 +205,7 @@ instance Match Term where
 instance Match Sort where
   match w p v = case (p, v) of
     (Type pl, Type vl) -> match w pl vl
-    _ | p == v -> return Map.empty
+    _ | p == v -> return mempty
     _          -> mzero
 
 instance Match Level where
@@ -208,6 +228,9 @@ class SubstWithOrigin a where
 instance SubstWithOrigin a => SubstWithOrigin [a] where
   substWithOrigin rho ots = map (substWithOrigin rho ots)
 
+instance SubstWithOrigin a => SubstWithOrigin (List1 a) where
+  substWithOrigin rho ots = fmap (substWithOrigin rho ots)
+
 instance (SubstWithOrigin a, SubstWithOrigin (Arg a)) => SubstWithOrigin (Elim' a) where
   substWithOrigin rho ots (Apply arg) = Apply $ substWithOrigin rho ots arg
   substWithOrigin rho ots e@Proj{}    = e
