@@ -60,16 +60,14 @@
 module Agda.TypeChecking.Free.Lazy where
 
 import Control.Applicative hiding (empty)
-import Control.Monad        ( guard )
 import Control.Monad.Reader ( MonadReader(..), asks, ReaderT, Reader, runReader )
 
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 import Data.HashSet (HashSet)
 import qualified Data.HashSet as HashSet
-import Data.Semigroup ( Semigroup, (<>) )
-import qualified Data.Set as Set
 import Data.Set (Set)
+import qualified Data.Set as Set
 
 
 import Agda.Syntax.Common
@@ -77,6 +75,7 @@ import Agda.Syntax.Internal
 
 import Agda.Utils.Functor
 import Agda.Utils.Lens
+import Agda.Utils.List1 (List1)
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Semigroup
@@ -539,7 +538,7 @@ instance Free Term where
     Sort s       -> freeVars' s
     Level l      -> freeVars' l
     MetaV m ts   -> underFlexRig (Flexible $ singleton m) $ freeVars' ts
-    DontCare mt  -> underModality (Modality Irrelevant unitQuantity unitCohesion) $ freeVars' mt
+    DontCare mt  -> underModality (Modality irrelevant unitQuantity unitCohesion unitPolarity) $ freeVars' mt
     Dummy{}      -> mempty
 
 instance Free t => Free (Type' t) where
@@ -572,10 +571,12 @@ instance Free Level where
 instance Free t => Free (PlusLevel' t) where
   freeVars' (Plus _ l)    = freeVars' l
 
-instance Free t => Free [t]            where
-instance Free t => Free (Maybe t)      where
-instance Free t => Free (WithHiding t) where
+instance Free t => Free [t]
+instance Free t => Free (List1 t)
+instance Free t => Free (Maybe t)
+instance Free t => Free (WithHiding t)
 instance Free t => Free (Named nm t)
+instance Free t => Free (Ranged t)
 
 instance (Free t, Free u) => Free (t, u) where
   freeVars' (t, u) = freeVars' t `mappend` freeVars' u
@@ -609,4 +610,4 @@ instance Free EqualityView where
   freeVars' = \case
     OtherType t -> freeVars' t
     IdiomType t -> freeVars' t
-    EqualityType s _eq l t a b -> freeVars' (s, l, [t, a, b])
+    EqualityType _r s _eq l t a b -> freeVars' (s, l, [t, a, b])
