@@ -1,4 +1,4 @@
-{-# LANGUAGE BangPatterns #-}
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.TypeChecking.Monad.Caching
   ( -- * Log reading/writing operations
@@ -17,8 +17,7 @@ module Agda.TypeChecking.Monad.Caching
   , restorePostScopeState
   ) where
 
-
-import qualified Data.Map as Map
+import qualified Data.Set as Set
 
 import Agda.Syntax.Common
 
@@ -27,6 +26,7 @@ import Agda.Interaction.Options
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Debug
 
+import qualified Agda.Utils.BiMap as BiMap
 import Agda.Utils.Lens
 import Agda.Utils.Monad
 import Agda.Utils.Null (empty)
@@ -35,11 +35,22 @@ import Agda.Utils.Impossible
 
 -- | To be called before any write or restore calls.
 {-# SPECIALIZE cachingStarts :: TCM () #-}
-cachingStarts :: (MonadTCState m, ReadTCState m) => m ()
+cachingStarts :: (MonadDebug m, MonadTCState m, ReadTCState m) => m ()
 cachingStarts = do
     NameId _ m <- useTC stFreshNameId
     stFreshNameId `setTCLens` NameId 1 m
+    stFreshOpaqueId `setTCLens` OpaqueId 1 m
     stAreWeCaching `setTCLens` True
+    validateCache m -- fixes issue #4835
+    where
+      validateCache m = (localCache readFromCachedLog) >>= \case
+        Just (_ , s) -> do
+          let
+            NameId _ m' = stPostFreshNameId s
+            OpaqueId _ m'' = stPostFreshOpaqueId s
+            stale = or [ m' /= m, m'' /= m ]
+          when stale cleanCachedLog
+        _ -> return ()
 
 areWeCaching :: (ReadTCState m) => m Bool
 areWeCaching = useR stAreWeCaching
@@ -58,19 +69,21 @@ restorePostScopeState :: (MonadDebug m, MonadTCState m) => PostScopeState -> m (
 restorePostScopeState pss = do
   reportSLn "cache" 10 $ "restorePostScopeState"
   modifyTC $ \s ->
-    let ipoints = s^.stInteractionPoints
-        ws = s^.stTCWarnings
+    let ipoints = s ^. stInteractionPoints
+        ws = s ^. stTCWarnings
         pss' = pss{stPostInteractionPoints = stPostInteractionPoints pss `mergeIPMap` ipoints
                   ,stPostTCWarnings = stPostTCWarnings pss `mergeWarnings` ws
+                  ,stPostOpaqueBlocks = s ^. stOpaqueBlocks
+                  ,stPostOpaqueIds = s ^. stOpaqueIds
                   }
     in  s{stPostScopeState = pss'}
   where
-    mergeIPMap lm sm = Map.mapWithKey (\k v -> maybe v (`mergeIP` v) (Map.lookup k lm)) sm
+    mergeIPMap lm sm = BiMap.mapWithKey (\k v -> maybe v (`mergeIP` v) (BiMap.lookup k lm)) sm
     -- see #1338 on why we need to use the new ranges.
     mergeIP li si = li { ipRange = ipRange si }
 
-    mergeWarnings loading current = [ w | w <- current, not $ tcWarningCached w ]
-                                 ++ [ w | w <- loading,       tcWarningCached w ]
+    mergeWarnings loading current = Set.filter (not . tcWarningCached) current
+                        `Set.union` Set.filter (      tcWarningCached) loading
 
 {-# SPECIALIZE modifyCache :: (Maybe LoadedFileCache -> Maybe LoadedFileCache) -> TCM () #-}
 modifyCache
