@@ -35,9 +35,20 @@ solvingProblems pids m = verboseBracket "tc.constr.solve" 50 ("working on proble
   return x
 
 isProblemSolved :: (MonadTCEnv m, ReadTCState m) => ProblemId -> m Bool
-isProblemSolved pid =
+isProblemSolved = isProblemSolved' False
+
+isProblemCompletelySolved :: (MonadTCEnv m, ReadTCState m) => ProblemId -> m Bool
+isProblemCompletelySolved = isProblemSolved' True
+
+isProblemSolved' :: (MonadTCEnv m, ReadTCState m) => Bool -> ProblemId -> m Bool
+isProblemSolved' completely pid =
   and2M (not . Set.member pid <$> asksTC envActiveProblems)
-        (not . any (Set.member pid . constraintProblems) <$> getAllConstraints)
+        (not . any belongsToUs <$> getAllConstraints)
+  where
+    belongsToUs c
+      | Set.notMember pid (constraintProblems c)         = False
+      | isBlockingConstraint (clValue $ theConstraint c) = True
+      | otherwise                                        = completely -- Ignore non-blocking unless `completely`
 
 {-# SPECIALIZE getConstraintsForProblem :: ProblemId -> TCM Constraints #-}
 getConstraintsForProblem :: ReadTCState m => ProblemId -> m Constraints
@@ -181,31 +192,37 @@ addAwakeConstraint' = addConstraintTo stAwakeConstraints
 
 addConstraintTo :: Lens' TCState Constraints -> Blocker -> Constraint -> TCM ()
 addConstraintTo bucket unblock c = do
-    pc <- build
+    pc <- buildConstraint unblock c
     stDirty `setTCLens` True
     bucket `modifyTCLens` (pc :)
-  where
-    build | isBlocking c = buildConstraint unblock c
-          | otherwise    = buildProblemConstraint_ unblock c
-    isBlocking = \case
-      SortCmp{}        -> False
-      LevelCmp{}       -> False
-      FindInstance{}   -> False
-      HasBiggerSort{}  -> False
-      HasPTSRule{}     -> False
-      CheckDataSort{}  -> False
-      ValueCmp{}       -> True
-      ValueCmpOnFace{} -> True
-      ElimCmp{}        -> True
-      UnBlock{}        -> True
-      IsEmpty{}        -> True
-      CheckSizeLtSat{} -> True
-      CheckFunDef{}    -> True
-      UnquoteTactic{}  -> True
-      CheckMetaInst{}  -> True
-      CheckType{}      -> True
-      CheckLockedVars{} -> True
-      UsableAtModality{} -> True
+
+-- | A problem is considered solved if there are no unsolved blocking constraints belonging to it.
+--   There's no really good principle for what constraints are blocking and which are not, but the
+--   general idea is that nothing bad should happen if you assume a non-blocking constraint is
+--   solvable, but it turns out it isn't. For instance, assuming an equality constraint between two
+--   types that turns out to be false can lead to ill typed terms in places where we don't expect
+--   them.
+isBlockingConstraint :: Constraint -> Bool
+isBlockingConstraint = \case
+  SortCmp{}             -> False
+  LevelCmp{}            -> False
+  FindInstance{}        -> False
+  ResolveInstanceHead{} -> False
+  HasBiggerSort{}       -> False
+  HasPTSRule{}          -> False
+  CheckDataSort{}       -> False
+  ValueCmp{}            -> True
+  ValueCmpOnFace{}      -> True
+  ElimCmp{}             -> True
+  UnBlock{}             -> True
+  IsEmpty{}             -> True
+  CheckSizeLtSat{}      -> True
+  CheckFunDef{}         -> True
+  UnquoteTactic{}       -> True
+  CheckMetaInst{}       -> True
+  CheckType{}           -> True
+  CheckLockedVars{}     -> True
+  UsableAtModality{}    -> True
 
 -- | Start solving constraints
 nowSolvingConstraints :: MonadTCEnv m => m a -> m a
@@ -222,11 +239,13 @@ isInstanceConstraint :: Constraint -> Bool
 isInstanceConstraint FindInstance{} = True
 isInstanceConstraint _              = False
 
-shouldPostponeInstanceSearch :: (ReadTCState m, HasOptions m) => m Bool
-shouldPostponeInstanceSearch =
+canDropRecursiveInstance :: (ReadTCState m, HasOptions m) => m Bool
+canDropRecursiveInstance =
   and2M ((^. stConsideringInstance) <$> getTCState)
-        (not . optOverlappingInstances <$> pragmaOptions)
-  `or2M` ((^. stPostponeInstanceSearch) <$> getTCState)
+        (not . optBacktrackingInstances <$> pragmaOptions)
+
+shouldPostponeInstanceSearch :: (ReadTCState m, HasOptions m) => m Bool
+shouldPostponeInstanceSearch = canDropRecursiveInstance `or2M` ((^. stPostponeInstanceSearch) <$> getTCState)
 
 -- | Wake constraints matching the given predicate (and aren't instance
 --   constraints if 'shouldPostponeInstanceSearch').
