@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Monad.Debug
   ( module Agda.TypeChecking.Monad.Debug
@@ -19,14 +20,13 @@ import Control.Monad.Writer
 
 import Data.Maybe
 import Data.Time                    ( getCurrentTime, getCurrentTimeZone, utcToLocalTime )
-import Data.Time.Format.ISO8601.Compat ( iso8601Show )
-  -- This is also exported from Data.Time.Format.ISO8601, but only from time >= 1.9
+import Data.Time.Format.ISO8601     ( iso8601Show )
 
 import {-# SOURCE #-} Agda.TypeChecking.Errors
 import Agda.TypeChecking.Monad.Base
 
 import Agda.Interaction.Options
-import {-# SOURCE #-} Agda.Interaction.Response (Response(..))
+import Agda.Interaction.Response.Base (Response_boot(..))
 
 import Agda.Utils.CallStack ( HasCallStack, withCallerCallStack )
 import Agda.Utils.Lens
@@ -35,7 +35,7 @@ import Agda.Utils.ListT
 import Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
-import Agda.Utils.Pretty
+import Agda.Syntax.Common.Pretty
 import Agda.Utils.ProfileOptions
 import Agda.Utils.Update
 import qualified Agda.Utils.Trie as Trie
@@ -71,10 +71,18 @@ class (Functor m, Applicative m, Monad m) => MonadDebug m where
     => VerboseKey -> VerboseLevel -> String -> m a -> m a
   traceDebugMessage k n s = liftThrough $ traceDebugMessage k n s
 
+#ifdef DEBUG
   default verboseBracket
     :: (MonadTransControl t, MonadDebug n, m ~ t n)
     => VerboseKey -> VerboseLevel -> String -> m a -> m a
   verboseBracket k n s = liftThrough $ verboseBracket k n s
+#else
+  default verboseBracket
+    :: (MonadTransControl t, MonadDebug n, m ~ t n)
+    => VerboseKey -> VerboseLevel -> String -> m a -> m a
+  verboseBracket k n s ma = ma
+  {-# INLINE verboseBracket #-}
+#endif
 
   default getVerbosity
     :: (MonadTrans t, MonadDebug n, m ~ t n)
@@ -157,7 +165,7 @@ instance MonadDebug TCM where
 
   formatDebugMessage k n d = catchAndPrintImpossible k n $ do
     render <$> d `catchError` \ err -> do
-      prettyError err <&> \ s -> vcat
+      renderError err <&> \ s -> vcat
         [ sep $ map text
           [ "Printing debug message"
           , k  ++ ":" ++ show n
@@ -166,11 +174,16 @@ instance MonadDebug TCM where
         , nest 2 $ text s
         ]
 
+#ifdef DEBUG
   verboseBracket k n s = applyWhenVerboseS k n $ \ m -> do
     openVerboseBracket k n s
     (m <* closeVerboseBracket k n) `catchError` \ e -> do
       closeVerboseBracketException k n
       throwError e
+#else
+  verboseBracket k n s ma = ma
+  {-# INLINE verboseBracket #-}
+#endif
 
   getVerbosity      = defaultGetVerbosity
   getProfileOptions = defaultGetProfileOptions
@@ -216,17 +229,44 @@ instance ReportS [String]  where reportS k n = reportSLn  k n . unlines
 instance ReportS [Doc]     where reportS k n = reportSLn  k n . render . vcat
 instance ReportS Doc       where reportS k n = reportSLn  k n . render
 
+#ifdef DEBUG
+
 -- | Conditionally println debug string.
 {-# SPECIALIZE reportSLn :: VerboseKey -> VerboseLevel -> String -> TCM () #-}
 reportSLn :: MonadDebug m => VerboseKey -> VerboseLevel -> String -> m ()
 reportSLn k n s = verboseS k n $ displayDebugMessage k n $ s ++ "\n"
 
+#else
+
+{-# INLINE reportSLn #-}
+reportSLn :: MonadDebug m => VerboseKey -> VerboseLevel -> String -> m ()
+reportSLn _ _ _ = pure ()
+
+#endif
+
+-- | Conditionally println debug string. Works regardless of the debug flag.
+{-# SPECIALIZE reportSLn :: VerboseKey -> VerboseLevel -> String -> TCM () #-}
+alwaysReportSLn :: MonadDebug m => VerboseKey -> VerboseLevel -> String -> m ()
+alwaysReportSLn k n s = verboseS k n $ displayDebugMessage k n $ s ++ "\n"
+
+
 __IMPOSSIBLE_VERBOSE__ :: (HasCallStack, MonadDebug m) => String -> m a
-__IMPOSSIBLE_VERBOSE__ s = do { reportSLn "impossible" 10 s ; throwImpossible err }
+__IMPOSSIBLE_VERBOSE__ s = do
+  -- Andreas, 2023-07-19, issue #6728 is fixed by manually inlining reportSLn here.
+  -- reportSLn "impossible" 10 s
+  -- It seems like GHC 9.6 optimization does otherwise something that throws
+  -- away the debug message.
+  let k = "impossible"
+  let n = 10
+  verboseS k n $ displayDebugMessage k n $ s ++ "\n"
+
+  throwImpossible err
   where
     -- Create the "Impossible" error using *our* caller as the call site.
     err = withCallerCallStack Impossible
 
+#ifdef DEBUG
+
 -- | Conditionally render debug 'Doc' and print it.
 {-# SPECIALIZE reportSDoc :: VerboseKey -> VerboseLevel -> TCM Doc -> TCM () #-}
 reportSDoc :: MonadDebug m => VerboseKey -> VerboseLevel -> TCM Doc -> m ()
@@ -239,6 +279,26 @@ reportResult k n debug action = do
   x <- action
   x <$ reportSDoc k n (debug x)
 
+#else
+
+-- | Conditionally render debug 'Doc' and print it.
+{-# INLINE reportSDoc #-}
+reportSDoc :: MonadDebug m => VerboseKey -> VerboseLevel -> TCM Doc -> m ()
+reportSDoc _ _ _ = pure ()
+
+-- | Debug print the result of a computation.
+{-# INLINE reportResult #-}
+reportResult :: MonadDebug m => VerboseKey -> VerboseLevel -> (a -> TCM Doc) -> m a -> m a
+reportResult _ _ _ action = action
+
+#endif
+
+-- | Conditionally render debug 'Doc' and print it. Works regardless of the debug flag.
+{-# SPECIALIZE reportSDoc :: VerboseKey -> VerboseLevel -> TCM Doc -> TCM () #-}
+alwaysReportSDoc :: MonadDebug m => VerboseKey -> VerboseLevel -> TCM Doc -> m ()
+alwaysReportSDoc k n d = verboseS k n $ do
+  displayDebugMessage k n . (++ "\n") =<< formatDebugMessage k n (locallyTC eIsDebugPrinting (const True) d)
+
 unlessDebugPrinting :: MonadDebug m => m () -> m ()
 unlessDebugPrinting = unlessM isDebugPrinting
 
@@ -263,6 +323,9 @@ instance TraceS [String]  where traceS k n = traceSLn  k n . unlines
 instance TraceS [Doc]     where traceS k n = traceSLn  k n . render . vcat
 instance TraceS Doc       where traceS k n = traceSLn  k n . render
 
+
+#ifdef DEBUG
+
 traceSLn :: MonadDebug m => VerboseKey -> VerboseLevel -> String -> m a -> m a
 traceSLn k n s = applyWhenVerboseS k n $ traceDebugMessage k n $ s ++ "\n"
 
@@ -272,6 +335,20 @@ traceSDoc k n d = applyWhenVerboseS k n $ \cont -> do
   s <- formatDebugMessage k n $ locallyTC eIsDebugPrinting (const True) d
   traceDebugMessage k n (s ++ "\n") cont
 
+#else
+
+{-# INLINE traceSLn #-}
+traceSLn :: MonadDebug m => VerboseKey -> VerboseLevel -> String -> m a -> m a
+traceSLn _ _ _ action = action
+
+-- | Conditionally render debug 'Doc', print it, and then continue.
+{-# INLINE traceSDoc #-}
+traceSDoc :: MonadDebug m => VerboseKey -> VerboseLevel -> TCM Doc -> m a -> m a
+traceSDoc _ _ _ action = action
+
+#endif
+
+
 openVerboseBracket :: MonadDebug m => VerboseKey -> VerboseLevel -> String -> m ()
 openVerboseBracket k n s = displayDebugMessage k n $ "{ " ++ s ++ "\n"
 
