@@ -1,18 +1,16 @@
-module Agda.TypeChecking.Monad.Env where
+{-# OPTIONS_GHC -Wunused-imports #-}
 
+module Agda.TypeChecking.Monad.Env where
 
 import qualified Data.List as List
 
 import Data.Maybe (fromMaybe)
 
-
 import Agda.Syntax.Common
 import Agda.Syntax.Abstract.Name
-import Agda.Syntax.TopLevelModuleName
 
 import Agda.TypeChecking.Monad.Base
 
-import Agda.Utils.FileName
 import qualified Agda.Utils.SmallSet as SmallSet
 
 import Agda.Utils.Impossible
@@ -29,8 +27,10 @@ withCurrentModule m =
     localTC $ \ e -> e { envCurrentModule = m }
 
 -- | Get the path of the currently checked file
-getCurrentPath :: MonadTCEnv m => m AbsolutePath
-getCurrentPath = fromMaybe __IMPOSSIBLE__ <$> asksTC envCurrentPath
+getCurrentPath :: (MonadTCEnv m, MonadFileId m) => m AbsolutePath
+getCurrentPath = do
+  i <- fromMaybe __IMPOSSIBLE__ <$> asksTC envCurrentPath
+  fileFromId i
 
 -- | Get the number of variables bound by anonymous modules.
 {-# SPECIALIZE getAnonymousVariables :: ModuleName -> TCM Nat #-}
@@ -100,7 +100,8 @@ putAllowedReductions = modifyAllowedReductions . const
 
 -- | Reduce @Def f vs@ only if @f@ is a projection.
 onlyReduceProjections :: MonadTCEnv m => m a -> m a
-onlyReduceProjections = putAllowedReductions $ SmallSet.singleton ProjectionReductions
+onlyReduceProjections = modifyAllowedReductions $ SmallSet.intersection $
+  SmallSet.singleton ProjectionReductions
 
 -- | Allow all reductions except for non-terminating functions (default).
 allowAllReductions :: MonadTCEnv m => m a -> m a
@@ -113,7 +114,8 @@ allowNonTerminatingReductions = putAllowedReductions reallyAllReductions
 -- | Allow all reductions when reducing types. Otherwise only allow
 --   inlined functions to be unfolded.
 onlyReduceTypes :: MonadTCEnv m => m a -> m a
-onlyReduceTypes = putAllowedReductions $ SmallSet.fromList [TypeLevelReductions, InlineReductions]
+onlyReduceTypes = modifyAllowedReductions $ SmallSet.intersection $
+  SmallSet.fromList [TypeLevelReductions, InlineReductions]
 
 -- | Update allowed reductions when working on types
 typeLevelReductions :: MonadTCEnv m => m a -> m a
@@ -135,3 +137,9 @@ isInsideDotPattern = asksTC envInsideDotPattern
 -- | Don't use call-by-need evaluation for the given computation.
 callByName :: TCM a -> TCM a
 callByName = localTC $ \ e -> e { envCallByNeed = False }
+
+-- | Don't fold let bindings when printing. This is a bit crude since it disables any folding of let
+--   bindings at all. In many cases it's better to use `removeLetBinding` before printing to drop
+--   the let bindings that should not be folded.
+dontFoldLetBindings :: MonadTCEnv m => m a -> m a
+dontFoldLetBindings = localTC $ \ e -> e { envFoldLetBindings = False }
