@@ -1,3 +1,4 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.TypeChecking.Monad.Imports
   ( addImport
@@ -20,21 +21,45 @@ module Agda.TypeChecking.Monad.Imports
 import Control.Arrow   ( (***) )
 import Control.Monad   ( when )
 
-import qualified Data.HashSet as HSet
+import Data.Maybe (catMaybes)
 import qualified Data.Map as Map
+import Data.Set (Set)
+import qualified Data.Set as Set
 
-import Agda.Syntax.Abstract.Name
+import Agda.Syntax.Common.Pretty
 import Agda.Syntax.TopLevelModuleName
 import Agda.TypeChecking.Monad.Base
 
 import Agda.Utils.List ( caseListM )
-import Agda.Utils.Pretty
-
+import qualified Agda.Utils.List1 as List1
+import qualified Agda.Utils.List2 as List2
+import Agda.Utils.Singleton (singleton)
 
 import Agda.Utils.Impossible
 
+-- | Register the given module as imported in the current state.
+--   Also recursively add its imports to the cumulative imports.
 addImport :: TopLevelModuleName -> TCM ()
-addImport top = modifyTCLens' stImportedModules $ HSet.insert top
+addImport top = do
+  modifyTCLens' stImportedModules $ Set.insert top
+  modifyTCLensM stImportedModulesTransitive $ completeTransitiveImports $ singleton top
+
+-- | @completeTransitiveImports ms ms'@.
+--   Precondition: @ms@ disjoint from @ms'@.
+completeTransitiveImports :: ReadTCState m => Set TopLevelModuleName -> ImportedModules -> m ImportedModules
+completeTransitiveImports ms old = if null ms then return old else do
+
+  -- Add the given imports to the current set.
+  let next = old `Set.union` ms
+
+  -- The interfaces for the modules we added to the transitive imports.
+  is <- catMaybes <$> mapM getVisitedModule (Set.toList ms)
+
+  -- The imports of these modules.
+  let imps = Set.unions $ map (Set.fromList . map fst . iImportedModules . miInterface) is
+
+  -- Recurse on the new imports.
+  completeTransitiveImports (imps `Set.difference` next) next
 
 addImportCycleCheck :: TopLevelModuleName -> TCM a -> TCM a
 addImportCycleCheck m =
@@ -104,5 +129,6 @@ withImportPath path = localTC $ \e -> e { envImportPath = path }
 checkForImportCycle :: TCM ()
 checkForImportCycle = do
   caseListM getImportPath __IMPOSSIBLE__ $ \ m ms -> do
-    when (m `elem` ms) $ typeError $ CyclicModuleDependency
-                                   $ dropWhile (/= m) $ reverse (m:ms)
+    when (m `elem` ms) $ typeError $ CyclicModuleDependency $
+      List2.snoc (List1.fromListSafe __IMPOSSIBLE__ $ dropWhile (/= m) $ reverse ms) m
+        -- NB: we know that ms contains m, so even after dropWhile the list is not empty.
