@@ -3,7 +3,6 @@ module Agda.TypeChecking.Monad.Options where
 
 import Prelude hiding (null)
 
-import Control.Arrow          ( (&&&) )
 import Control.Monad          ( unless, when )
 import Control.Monad.IO.Class ( MonadIO(..) )
 import Control.Monad.Except
@@ -13,16 +12,14 @@ import Control.Monad.Writer
 
 import qualified Data.Graph as Graph
 import Data.List (sort)
-import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
+import qualified Data.Set as Set
 
 import System.Directory
 import System.FilePath
 
 import Agda.Syntax.Common
-import qualified Agda.Syntax.Concrete as C
-import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.TopLevelModuleName
 
 import Agda.TypeChecking.Monad.Debug (reportSDoc)
@@ -31,20 +28,23 @@ import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Imports
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Benchmark
+import Agda.TypeChecking.Monad.Trace
 
 import Agda.Interaction.FindFile
-import Agda.Interaction.Options
+import Agda.Interaction.Options hiding (setPragmaOptions)
 import qualified Agda.Interaction.Options.Lenses as Lens
 import Agda.Interaction.Library
+import Agda.Interaction.Library.Base (LibCache(LibCache), libAbove, libFile, runLibM)
 
 import Agda.Utils.FileName
-import Agda.Utils.Functor
 import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as G
+import Agda.Utils.Lens
 import Agda.Utils.List
-import Agda.Utils.Maybe
+import Agda.Utils.List1 (List1)
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Null
-import Agda.Utils.Pretty
-import Agda.Utils.Tuple
+import Agda.Syntax.Common.Pretty
+import Agda.Utils.Size
 import Agda.Utils.WithDefault
 
 import Agda.Utils.Impossible
@@ -56,11 +56,29 @@ setPragmaOptions opts = do
   -- Check for unsafe pragma options if @--safe@ is on.
   when (Lens.getSafeMode opts) $
     unlessNull (unsafePragmaOptions opts) $ \ unsafe ->
-      warning $ SafeFlagPragma unsafe
+      warning $ SafeFlagPragma $ Set.fromList unsafe
 
   stPragmaOptions `setTCLens` opts
   updateBenchmarkingStatus
 
+-- | Check that that you don't turn on inconsistent options. For instance, if --a implies --b and
+--   you have both --a and --no-b. Only warn for things that have changed compared to the old
+--   options.
+checkPragmaOptionConsistency :: PragmaOptions -> PragmaOptions -> TCM ()
+checkPragmaOptionConsistency oldOpts newOpts = do
+  mapM_ check impliedPragmaOptions
+  where
+    -- Only warn if both flags are set explicitly (Value rather than Default)
+    check (ImpliesPragmaOption nameA valA flagA nameB valB flagB)
+      | flagA newOpts == flagA oldOpts
+      , flagB newOpts == flagB oldOpts  = pure () -- Nothing changed, don't check again.
+      | Value vA <- flagA newOpts, vA == valA
+      , Value vB <- flagB newOpts, vB /= valB = warning $ ConflictingPragmaOptions (nameA .= valA) (nameB .= valB)
+      | otherwise                 = pure ()
+      where
+        name .= True  = name
+        name .= False = "no-" ++ name
+
 -- | Sets the command line options (both persistent and pragma options
 -- are updated).
 --
@@ -87,7 +105,7 @@ setCommandLineOptions' root opts = do
           opts' <- setLibraryPaths root opts
           let incs = optIncludePaths opts'
           setIncludeDirs incs root
-          getIncludeDirs
+          List1.toList <$> getIncludeDirs
         incs -> return incs
       modifyTC $ Lens.setCommandLineOptions opts{ optAbsoluteIncludePaths = incs }
       setPragmaOptions (optPragmaOptions opts)
@@ -95,32 +113,62 @@ setCommandLineOptions' root opts = do
 
 libToTCM :: LibM a -> TCM a
 libToTCM m = do
-  cachedConfs <- useTC stProjectConfigs
-  cachedLibs  <- useTC stAgdaLibFiles
+  libCache <- useTC stLibCache
+  ((z, warns), libCache') <- liftIO $ runLibM m libCache
 
-  ((z, warns), (cachedConfs', cachedLibs')) <- liftIO $
-    (`runStateT` (cachedConfs, cachedLibs)) $ runWriterT $ runExceptT m
+  setTCLens stLibCache libCache'
 
-  modifyTCLens stProjectConfigs $ const cachedConfs'
-  modifyTCLens stAgdaLibFiles   $ const cachedLibs'
-
-  unless (null warns) $ warnings $ map LibraryWarning warns
+  List1.unlessNull warns \ warns -> warnings $ fmap LibraryWarning warns
   case z of
-    Left s  -> typeError $ GenericDocError s
+    Left s  -> typeError $ LibraryError s
     Right x -> return x
 
 -- | Returns the library files for a given file.
+--
+-- Nothing is returned if 'optUseLibs' is 'False'.
+--
+-- An error is raised if 'optUseLibs' is 'True' and a library file is
+-- located too far down the directory hierarchy (see
+-- 'checkLibraryFileNotTooFarDown').
 
 getAgdaLibFiles
   :: AbsolutePath        -- ^ The file name.
   -> TopLevelModuleName  -- ^ The top-level module name.
   -> TCM [AgdaLibFile]
 getAgdaLibFiles f m = do
+  ls <- getAgdaLibFilesWithoutTopLevelModuleName f
+  mapM_ (checkLibraryFileNotTooFarDown m) ls
+  return ls
+
+-- | Returns potential library files for a file without a known
+-- top-level module name.
+--
+-- Once the top-level module name is known one can use
+-- 'checkLibraryFileNotTooFarDown' to check that the potential library
+-- files were not located too far down the directory hierarchy.
+--
+-- Nothing is returned if 'optUseLibs' is 'False'.
+
+getAgdaLibFilesWithoutTopLevelModuleName
+  :: AbsolutePath  -- ^ The file.
+  -> TCM [AgdaLibFile]
+getAgdaLibFilesWithoutTopLevelModuleName f = do
   useLibs <- optUseLibs <$> commandLineOptions
-  if | useLibs   -> libToTCM $ mkLibM [] $ getAgdaLibFiles' root
+  if | useLibs   -> libToTCM $ getAgdaLibFile root
      | otherwise -> return []
   where
-  root = filePath (projectRoot f m)
+  root = takeDirectory $ filePath f
+
+-- | Checks that a library file for the module @A.B.C@ (say) in the
+-- directory @dir/A/B@ is located at least two directories above the
+-- file (not in @dir/A@ or @dir/A/B@).
+
+checkLibraryFileNotTooFarDown ::
+  TopLevelModuleName ->
+  AgdaLibFile ->
+  TCM ()
+checkLibraryFileNotTooFarDown m lib =
+  when (lib ^. libAbove < size m - 1) $ typeError $ LibTooFarDown m lib
 
 -- | Returns the library options for a given file.
 
@@ -171,14 +219,29 @@ addTrustedExecutables o = do
   -- or is a security risk.
   return o{ optTrustedExecutables = trustedExes }
 
+-- | Set pragma options without checking for consistency.
 setOptionsFromPragma :: OptionsPragma -> TCM ()
-setOptionsFromPragma ps = do
+setOptionsFromPragma = setOptionsFromPragma' False
+
+-- | Set pragma options and check them for consistency.
+checkAndSetOptionsFromPragma :: OptionsPragma -> TCM ()
+checkAndSetOptionsFromPragma = setOptionsFromPragma' True
+
+setOptionsFromPragma' :: Bool -> OptionsPragma -> TCM ()
+setOptionsFromPragma' checkConsistency ps = setCurrentRange (pragmaRange ps) $ do
     opts <- commandLineOptions
     let (z, warns) = runOptM (parsePragmaOptions ps opts)
     mapM_ (warning . OptionWarning) warns
     case z of
-      Left err    -> typeError $ GenericError err
-      Right opts' -> setPragmaOptions opts'
+      Left err    -> typeError $ OptionError err
+      Right opts' -> do
+
+        -- Check consistency of implied options
+        when checkConsistency $ do
+          oldOpts <- pragmaOptions
+          checkPragmaOptionConsistency oldOpts opts'
+
+        setPragmaOptions opts'
 
 -- | Disable display forms.
 enableDisplayForms :: MonadTCEnv m => m a -> m a
@@ -199,12 +262,9 @@ displayFormsEnabled = asksTC envDisplayFormsEnabled
 -- Precondition: 'optAbsoluteIncludePaths' must be nonempty (i.e.
 -- 'setCommandLineOptions' must have run).
 
-getIncludeDirs :: HasOptions m => m [AbsolutePath]
+getIncludeDirs :: HasOptions m => m (List1 AbsolutePath)
 getIncludeDirs = do
-  incs <- optAbsoluteIncludePaths <$> commandLineOptions
-  case incs of
-    [] -> __IMPOSSIBLE__
-    _  -> return incs
+  List1.fromListSafe __IMPOSSIBLE__ . optAbsoluteIncludePaths <$> commandLineOptions
 
 -- | Makes the given directories absolute and stores them as include
 -- directories.
@@ -223,24 +283,22 @@ setIncludeDirs incs root = do
   oldIncs <- getsTC Lens.getAbsoluteIncludePaths
 
   -- Add the current dir if no include path is given
-  incs <- return $ if null incs then ["."] else incs
+  incs <- return $ List1.fromListSafe (List1.singleton ".") incs
   -- Make paths absolute
-  incs <- return $  map (mkAbsolute . (filePath root </>)) incs
+  incs <- return $ fmap (mkAbsolute . (filePath root </>)) incs
 
   -- Andreas, 2013-10-30  Add default include dir
-      -- NB: This is an absolute file name, but
-      -- Agda.Utils.FilePath wants to check absoluteness anyway.
-  primdir <- liftIO $ mkAbsolute <$> getPrimitiveLibDir
+  primdir <- useTC stPrimitiveLibDir
       -- We add the default dir at the end, since it is then
       -- printed last in error messages.
       -- Might also be useful to overwrite default imports...
-  incs <- return $ nubOn id $ incs ++ [primdir]
+  incs <- return $ List1.fromListSafe __IMPOSSIBLE__ $ nubOn id $ List1.toList $ incs <> List1.singleton primdir
 
   reportSDoc "setIncludeDirs" 10 $ return $ vcat
     [ "Old include directories:"
     , nest 2 $ vcat $ map pretty oldIncs
     , "New include directories:"
-    , nest 2 $ vcat $ map pretty incs
+    , nest 2 $ vcat $ fmap pretty incs
     ]
 
   -- Check whether the include dirs have changed.  If yes, reset state.
@@ -256,22 +314,20 @@ setIncludeDirs incs root = do
   -- up in a situation in which we use the contents of the file
   -- "old-path/M.agda", when the user actually meant
   -- "new-path/M.agda".
-  when (sort oldIncs /= sort incs) $ do
+  when (sort oldIncs /= sort (List1.toList incs)) $ do
     ho <- getInteractionOutputCallback
     tcWarnings <- useTC stTCWarnings -- restore already generated warnings
-    projectConfs <- useTC stProjectConfigs  -- restore cached project configs & .agda-lib
-    agdaLibFiles <- useTC stAgdaLibFiles    -- files, since they use absolute paths
+    libCache   <- useTC stLibCache   -- restore cached project configs & .agda-lib files, since they use absolute paths
     decodedModules <- getDecodedModules
     (keptDecodedModules, modFile) <- modulesToKeep incs decodedModules
     resetAllState
     setTCLens stTCWarnings tcWarnings
-    setTCLens stProjectConfigs projectConfs
-    setTCLens stAgdaLibFiles agdaLibFiles
+    setTCLens stLibCache libCache
     setInteractionOutputCallback ho
     setDecodedModules keptDecodedModules
-    setTCLens stModuleToSource modFile
+    setTCLens stModuleToSourceId modFile
 
-  Lens.putAbsoluteIncludePaths incs
+  Lens.putAbsoluteIncludePaths $ List1.toList incs
   where
   -- A decoded module is kept if its top-level module name is resolved
   -- to the same absolute path using the old and the new include
@@ -285,9 +341,9 @@ setIncludeDirs incs root = do
   -- ModuleToSource structure, constructed using the new include
   -- directories, is returned.
   modulesToKeep
-    :: [AbsolutePath]  -- New include directories.
+    :: List1 AbsolutePath -- New include directories.
     -> DecodedModules  -- Old decoded modules.
-    -> TCM (DecodedModules, ModuleToSource)
+    -> TCM (DecodedModules, ModuleToSourceId)
   modulesToKeep incs old = process Map.empty Map.empty modules
     where
     -- A graph with one node per module in old, and an edge from m to
@@ -325,8 +381,8 @@ setIncludeDirs incs root = do
       G.sccs' dependencyGraph
 
     process ::
-      Map TopLevelModuleName ModuleInfo -> ModuleToSource ->
-      [ModuleInfo] -> TCM (DecodedModules, ModuleToSource)
+      Map TopLevelModuleName ModuleInfo -> ModuleToSourceId ->
+      [ModuleInfo] -> TCM (DecodedModules, ModuleToSourceId)
     process !keep !modFile [] = return
       ( Map.fromList $
         Map.toList keep
@@ -339,7 +395,7 @@ setIncludeDirs incs root = do
         if not depsKept then return (keep, modFile) else do
         let t = iTopLevelModuleName $ miInterface m
         oldF            <- findFile' t
-        (newF, modFile) <- liftIO $ findFile'' incs t modFile
+        (newF, modFile) <- runStateT (findFile'_ incs t) modFile
         return $ case (oldF, newF) of
           (Right f1, Right f2) | f1 == f2 ->
             (Map.insert t m keep, modFile)
@@ -349,8 +405,11 @@ setIncludeDirs incs root = do
 isPropEnabled :: HasOptions m => m Bool
 isPropEnabled = optProp <$> pragmaOptions
 
+isLevelUniverseEnabled :: HasOptions m => m Bool
+isLevelUniverseEnabled = optLevelUniverse <$> pragmaOptions
+
 isTwoLevelEnabled :: HasOptions m => m Bool
-isTwoLevelEnabled = collapseDefault . optTwoLevel <$> pragmaOptions
+isTwoLevelEnabled = optTwoLevel <$> pragmaOptions
 
 {-# SPECIALIZE hasUniversePolymorphism :: TCM Bool #-}
 hasUniversePolymorphism :: HasOptions m => m Bool
@@ -359,6 +418,9 @@ hasUniversePolymorphism = optUniversePolymorphism <$> pragmaOptions
 showImplicitArguments :: HasOptions m => m Bool
 showImplicitArguments = optShowImplicit <$> pragmaOptions
 
+showGeneralizedArguments :: HasOptions m => m Bool
+showGeneralizedArguments = (\opt -> optShowGeneralized opt) <$> pragmaOptions
+
 showIrrelevantArguments :: HasOptions m => m Bool
 showIrrelevantArguments = optShowIrrelevant <$> pragmaOptions
 
@@ -377,14 +439,18 @@ withShowAllArguments = withShowAllArguments' True
 
 withShowAllArguments' :: ReadTCState m => Bool -> m a -> m a
 withShowAllArguments' yes = withPragmaOptions $ \ opts ->
-  opts { optShowImplicit = yes, optShowIrrelevant = yes }
+  opts { _optShowImplicit = Value yes, _optShowIrrelevant = Value yes }
+
+withoutPrintingGeneralization :: ReadTCState m => m a -> m a
+withoutPrintingGeneralization = withPragmaOptions $ \ opts ->
+  opts { _optShowGeneralized = Value False }
 
 -- | Change 'PragmaOptions' for a computation and restore afterwards.
 withPragmaOptions :: ReadTCState m => (PragmaOptions -> PragmaOptions) -> m a -> m a
 withPragmaOptions = locallyTCState stPragmaOptions
 
 positivityCheckEnabled :: HasOptions m => m Bool
-positivityCheckEnabled = not . optDisablePositivity <$> pragmaOptions
+positivityCheckEnabled = optPositivityCheck <$> pragmaOptions
 
 {-# SPECIALIZE typeInType :: TCM Bool #-}
 typeInType :: HasOptions m => m Bool
@@ -405,7 +471,7 @@ getLanguage :: HasOptions m => m Language
 getLanguage = do
   opts <- pragmaOptions
   return $
-    if not (collapseDefault (optWithoutK opts)) then WithK else
+    if not (optWithoutK opts) then WithK else
     case optCubical opts of
       Just variant -> Cubical variant
       Nothing      -> WithoutK
