@@ -4,14 +4,17 @@ module Agda.TypeChecking.Monad.Trace where
 
 import Prelude hiding (null)
 
-import Control.Monad.Except
-import Control.Monad.Reader
-import Control.Monad.State
-import Control.Monad.Trans.Identity
-import Control.Monad.Writer
+import Control.Monad.Except         ( ExceptT  (ExceptT  ), runExceptT   , throwError )
+import Control.Monad.Reader         ( ReaderT  (ReaderT  ), runReaderT   )
+import Control.Monad.State          ( StateT   (StateT   ), runStateT    )
+import Control.Monad.Trans.Identity ( IdentityT(IdentityT), runIdentityT )
+import Control.Monad.Trans.Maybe    ( MaybeT   (MaybeT   ), runMaybeT    )
+import Control.Monad.Writer         ( WriterT  (WriterT  ), runWriterT   )
 
 import qualified Data.Set as Set
 
+import Agda.Syntax.Common.Pretty
+import Agda.Syntax.Parser (PM, runPMIO)
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Position as P
 
@@ -23,12 +26,11 @@ import Agda.TypeChecking.Monad.Base
   hiding (ModuleInfo, MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.State
+import Agda.TypeChecking.Warnings (warning)
 
 import Agda.Utils.Function
-import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Syntax.Common.Pretty (prettyShow)
 
 ---------------------------------------------------------------------------
 -- * Trace
@@ -115,9 +117,15 @@ class (MonadTCEnv m, ReadTCState m) => MonadTrace m where
     => RemoveTokenBasedHighlighting -> HighlightingInfo -> m ()
   printHighlightingInfo r i = lift $ printHighlightingInfo r i
 
+traceCallCPS' :: MonadTrace m => Call -> (m b -> m b) -> m b -> m b
+traceCallCPS' c k ret = traceCallCPS c (\ret -> k (ret ())) (\() -> ret)
+
 instance MonadTrace m => MonadTrace (IdentityT m) where
   traceClosureCall c f = IdentityT $ traceClosureCall c $ runIdentityT f
 
+instance MonadTrace m => MonadTrace (MaybeT m) where
+  traceClosureCall c f = MaybeT $ traceClosureCall c $ runMaybeT f
+
 instance MonadTrace m => MonadTrace (ReaderT r m) where
   traceClosureCall c f = ReaderT $ \r -> traceClosureCall c $ runReaderT f r
 
@@ -136,17 +144,6 @@ instance MonadTrace TCM where
     -- Since the fix of #2092 we may report an error outside the current file.
     -- (For instance, if we import a module which then happens to have the
     -- wrong name.)
-    -- Thus, we no longer crash, but just report the alien range.
-    -- -- Andreas, 2015-02-09 Make sure we do not set a range
-    -- -- outside the current file
-    verboseS "check.ranges" 90 $
-      Strict.whenJust (rangeFile callRange) $ \f -> do
-        currentFile <- asksTC envCurrentPath
-        when (currentFile /= Just (rangeFilePath f)) $ do
-          reportSLn "check.ranges" 90 $
-            prettyShow call ++
-            " is setting the current range to " ++ show callRange ++
-            " which is outside of the current file " ++ show currentFile
 
     -- Compute update to 'Range' and 'Call' components of 'TCEnv'.
     let withCall = localTC $ foldr (.) id $ concat $
@@ -217,10 +214,11 @@ instance MonadTrace TCM where
   printHighlightingInfo remove info = do
     modToSrc <- useTC stModuleToSource
     method   <- viewTC eHighlightingMethod
-    reportS "highlighting" 50
+    reportSDoc "highlighting" 50 $ pure $ vcat
       [ "Printing highlighting info:"
-      , show info
-      , "  modToSrc = " ++ show modToSrc
+      , nest 2 $ (text . show) info
+      , "File modules:"
+      , nest 2 $ pretty modToSrc
       ]
     unless (null info) $ do
       appInteractionOutputCallback $
@@ -271,3 +269,26 @@ highlightAsTypeChecked rPre r m
     return v
     where
     p rs x = printHighlightingInfo KeepHighlighting (singleton rs x)
+
+---------------------------------------------------------------------------
+-- * Warnings in the parser
+---------------------------------------------------------------------------
+
+-- | Running the Parse monad, raising parser warnings.
+
+runPM :: PM a -> TCM a
+runPM m = do
+  (res, ws) <- runPMIO m
+  forM_ ws \ w -> setCurrentRange w $ warning $ ParseWarning w
+  case res of
+    Left  e -> throwError $ ParserError e
+    Right a -> return a
+
+-- | Running the Parse monad, dropping parser warnings.
+
+runPMDropWarnings :: PM a -> TCM a
+runPMDropWarnings m = do
+  (res, _ws) <- runPMIO m
+  case res of
+    Left  e -> throwError $ ParserError e
+    Right a -> return a
