@@ -19,13 +19,14 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Monad hiding (constructorForm)
-import Agda.TypeChecking.Monad.Builtin (getName',builtinHComp, builtinConId)
+import Agda.TypeChecking.Monad.Builtin (getName', builtinHComp)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records
 
 import Agda.Utils.Empty
 import Agda.Utils.Functor (for, ($>), (<&>))
 import Agda.Utils.Maybe
+import Agda.Utils.Monad ( anyM, or2M)
 import Agda.Utils.Null
 import Agda.Utils.Singleton
 import Agda.Utils.Size
@@ -33,11 +34,11 @@ import Agda.Utils.Tuple
 
 import Agda.Utils.Impossible
 
--- | If matching is inconclusive (@DontKnow@) we want to know whether
---   it is due to a particular meta variable.
+-- | If matching is inconclusive (@DontKnow@) we want to know whether it is on a
+--   lazy pattern and whether it is due to a particular meta variable.
 data Match a = Yes Simplification (IntMap (Arg a))
              | No
-             | DontKnow (Blocked ())
+             | DontKnow OnlyLazy (Blocked ())
   deriving Functor
 
 instance Null (Match a) where
@@ -64,10 +65,10 @@ buildSubstitution err n vs = foldr cons idS $ matchedArgs' n vs
 instance Semigroup (Match a) where
     -- @NotBlocked (StuckOn e)@ means blocked by a variable.
     -- In this case, no instantiation of meta-variables will make progress.
-    DontKnow b <> DontKnow b'      = DontKnow $ b <> b'
-    DontKnow m <> _                = DontKnow m
-    _          <> DontKnow m       = DontKnow m
-    -- One could imagine DontKnow _ <> No = No, but would break the
+    DontKnow l b <> DontKnow l' b' = DontKnow (l <> l') (b <> b')
+    DontKnow l m <> _              = DontKnow l m
+    _            <> DontKnow l m   = DontKnow l m
+    -- One could imagine DontKnow _ _ <> No = No, but would break the
     -- equivalence to case-trees (Issue 2964).
     No         <> _                = No
     _          <> No               = No
@@ -77,6 +78,18 @@ instance Monoid (Match a) where
     mempty = empty
     mappend = (<>)
 
+-- | Whether the inconclusive matches are only on lazy patterns.
+data OnlyLazy = OnlyLazy | NonLazy
+
+instance Semigroup OnlyLazy where
+  NonLazy  <> _        = NonLazy
+  _        <> NonLazy  = NonLazy
+  OnlyLazy <> OnlyLazy = OnlyLazy
+
+instance Monoid OnlyLazy where
+  mempty = OnlyLazy
+  mappend = (<>)
+
 -- | Instead of 'zipWithM', we need to use this lazy version
 --   of combining pattern matching computations.
 
@@ -118,7 +131,10 @@ foldMatch match = loop where
             -- contain ill-typed terms due to eta-expansion at wrong
             -- type.
             return (r <> r', v' : vs)
-          DontKnow m -> return (DontKnow m, v' : vs)
+          DontKnow OnlyLazy _ -> do
+            (r', _vs') <- loop ps vs
+            return (r <> r', v' : vs)
+          DontKnow NonLazy m -> return (DontKnow NonLazy m, v' : vs)
           Yes{} -> do
             (r', vs') <- loop ps vs
             return (r <> r', v' : vs')
@@ -181,6 +197,7 @@ matchCopattern _       elim@Proj{}    = return (No , elim)
 matchCopattern p       (Apply v) = mapSnd Apply <$> matchPattern p v
 matchCopattern p       e@(IApply x y r) = mapSnd (mergeElim e) <$> matchPattern p (defaultArg r)
 
+{-# SPECIALIZE matchPatterns :: [NamedArg DeBruijnPattern] -> [Arg Term] -> TCM (Match Term, [Arg Term]) #-}
 matchPatterns :: MonadMatch m
               => [NamedArg DeBruijnPattern]
               -> [Arg Term]
@@ -221,56 +238,59 @@ matchPattern p u = case (p, u) of
       NotBlocked _ (Lit l')
           | l == l'            -> return (Yes YesSimplification empty , arg')
           | otherwise          -> return (No                          , arg')
-      Blocked b _              -> return (DontKnow $ Blocked b ()     , arg')
-      NotBlocked r t           -> return (DontKnow $ NotBlocked r' () , arg')
+      Blocked b _              -> return (DontKnow NonLazy $ Blocked b ()     , arg')
+      NotBlocked r t           -> return (DontKnow NonLazy $ NotBlocked r' () , arg')
         where r' = stuckOn (Apply arg') r
 
   -- Case constructor pattern.
   (ConP c cpi ps, Arg info v) -> do
-    if not (conPRecord cpi) then fallback c ps (Arg info v) else do
-    isEtaRecordCon (conName c) >>= \case
-      Nothing -> fallback c ps (Arg info v)
-      Just fs -> do
-        -- Case: Eta record constructor.
-        -- This case is necessary if we want to use the clauses before
-        -- record pattern translation (e.g., in type-checking definitions by copatterns).
-        unless (size fs == size ps) __IMPOSSIBLE__
-        mapSnd (Arg info . Con c (fromConPatternInfo cpi) . map Apply) <$> do
-          matchPatterns ps $ for fs $ \ (Arg ai f) -> Arg ai $ v `applyE` [Proj ProjSystem f]
-    where
-    isEtaRecordCon :: HasConstInfo m => QName -> m (Maybe [Arg QName])
-    isEtaRecordCon c = do
-      (theDef <$> getConstInfo c) >>= \case
-        Constructor{ conData = d } -> do
-          (theDef <$> getConstInfo d) >>= \case
-            r@Record{ recFields = fs } | YesEta <- recEtaEquality r -> return $ Just $ map argFromDom fs
-            _ -> return Nothing
-        _ -> __IMPOSSIBLE__
+    let lazy = if conPLazy cpi then OnlyLazy else NonLazy
+    if not (conPRecord cpi) then fallback c lazy ps (Arg info v) else do
+    isEtaRecordConstructor (conName c) >>= \case
+      Nothing -> fallback c lazy ps (Arg info v)
+      -- Case: Eta record constructor.
+      -- This case is necessary if we want to use the clauses before
+      -- record pattern translation (e.g., in type-checking definitions by copatterns).
+      Just (_r, def) -> do
+        reportSDoc "tc.match" 50 $ vcat
+          [ "matchPattern: eta record"
+          , nest 2 $ "c  = " <+> prettyTCM c
+          , nest 2 $ "ps = " <+> prettyTCMPatternList ps
+          , nest 2 $ "v  = " <+> prettyTCM v
+          ]
+        -- Issue #7266: in case we are brazenly matching potentially ill-typed arguments,
+        -- `v` might be an application of a *different* constructor.
+        -- In that case we certainly have no match.
+        case v of
+          Con c' _ _ | c /= c' -> return (No, u)
+          _ -> do
+            let fs = map argFromDom $ _recFields def
+            unless (size fs == size ps) __IMPOSSIBLE__
+            mapSnd (Arg info . Con c (fromConPatternInfo cpi) . map Apply) <$> do
+              matchPatterns ps $ for fs $ \ (Arg ai f) -> Arg ai $ v `applyE` [Proj ProjSystem f]
   (DefP o q ps, v) -> do
     let f (Def q' vs) | q == q' = Just (Def q, vs)
         f _                     = Nothing
-    fallback' f ps v
+    fallback' f NonLazy ps v
  where
     -- Default: not an eta record constructor.
   fallback :: MonadMatch m
-           => ConHead -> [NamedArg DeBruijnPattern] -> Arg Term -> m (Match Term, Arg Term)
-  fallback c ps v = do
+           => ConHead -> OnlyLazy -> [NamedArg DeBruijnPattern] -> Arg Term -> m (Match Term, Arg Term)
+  fallback c lazy ps v = do
     let f (Con c' ci' vs) | c == c' = Just (Con c' ci',vs)
         f _                         = Nothing
-    fallback' f ps v
+    fallback' f lazy ps v
 
   -- Regardless of blocking, constructors and a properly applied @hcomp@
   -- can be matched on.
   isMatchable' :: HasBuiltins m => m (Blocked Term -> Maybe Term)
   isMatchable' = do
-    [mhcomp,mconid] <- mapM getName' [builtinHComp, builtinConId]
+    mhcomp <- getName' builtinHComp
     return $ \ r ->
       case ignoreBlocking r of
         t@Con{} -> Just t
         t@(Def q [l,a,phi,u,u0]) | Just q == mhcomp
                 -> Just t
-        t@(Def q [l,a,x,y,phi,p]) | Just q == mconid
-                -> Just t
         -- TODO this covers the transpIx functions, but it's a hack.
         t@(Def q _) | NotBlocked{blockingStatus = MissingClauses _} <- r -> Just t
         _       -> Nothing
@@ -278,19 +298,17 @@ matchPattern p u = case (p, u) of
   -- DefP hcomp and ConP matching.
   fallback' :: MonadMatch m
             => (Term -> Maybe (Elims -> Term , Elims))
+            -> OnlyLazy
             -> [NamedArg DeBruijnPattern]
             -> Arg Term
             -> m (Match Term, Arg Term)
-  fallback' mtc ps (Arg info v) = do
+  fallback' mtc lazy ps (Arg info v) = do
         isMatchable <- isMatchable'
 
         w <- reduceB v
         -- Unfold delayed (corecursive) definitions one step. This is
         -- only necessary if c is a coinductive constructor, but
-        -- 1) it does not hurt to do it all the time, and
-        -- 2) whatInduction c sometimes crashes because c may point to
-        --    an axiom at this stage (if we are checking the
-        --    projection functions for a record type).
+        -- it does not hurt to do it all the time.
 {-
         w <- case w of
                NotBlocked r (Def f es) ->   -- Andreas, 2014-06-12 TODO: r == ReallyNotBlocked sufficient?
@@ -312,20 +330,18 @@ matchPattern p u = case (p, u) of
           b | Just t <- isMatchable b ->
             case mtc t of
               Just (bld, vs) -> do
-                (m, vs1) <- yesSimplification <$> matchPatterns ps (fromMaybe __IMPOSSIBLE__ $ allApplyElims vs)
-                return (m, Arg info $ bld (mergeElims vs vs1))
+                (m, vs1) <- matchPatterns ps (fromMaybe __IMPOSSIBLE__ $ allApplyElims vs)
+                return (yesSimplification m, Arg info $ bld (mergeElims vs vs1))
               Nothing
                                     -> return (No                          , arg)
-          Blocked b _               -> return (DontKnow $ Blocked b ()     , arg)
-          NotBlocked r _            -> return (DontKnow $ NotBlocked r' () , arg)
+          Blocked b _               -> return (DontKnow lazy $ Blocked b ()     , arg)
+          NotBlocked r _            -> return (DontKnow lazy $ NotBlocked r' () , arg)
             where r' = stuckOn (Apply arg) r
 
--- ASR (08 November 2014). The type of the function could be
---
--- @(Match Term, [Arg Term]) -> (Match Term, [Arg Term])@.
-yesSimplification :: (Match a, b) -> (Match a, b)
-yesSimplification (Yes _ vs, us) = (Yes YesSimplification vs, us)
-yesSimplification r              = r
+yesSimplification :: Match a -> Match a
+yesSimplification = \case
+  Yes _ vs -> Yes YesSimplification vs
+  m -> m
 
 -- Matching patterns against patterns -------------------------------------
 
@@ -365,3 +381,25 @@ matchPatternsP :: MonadMatch m
                -> m (Match DeBruijnPattern)
 matchPatternsP ps qs = do
   mconcat <$> zipWithM matchPatternP (map namedArg ps) qs
+
+
+-- | Does the pattern perform a match that could fail?
+properlyMatching :: HasConstInfo m => Pattern' a -> m Bool
+properlyMatching = properlyMatching' True True
+
+properlyMatching' :: HasConstInfo m
+  => Bool       -- ^ Should absurd patterns count as proper match?
+  -> Bool       -- ^ Should projection patterns count as proper match?
+  -> Pattern' a -- ^ The pattern.
+  -> m Bool
+properlyMatching' absP projP = \case
+  p | absP && patternOrigin p == Just PatOAbsurd -> return True
+  ConP con ci ps    -- eta record constructors do not count as proper matches themselves
+    | conPRecord ci -> (isNothing <$> isEtaRecordConstructor (conName con)) `or2M` anyM (properlyMatching . namedArg) ps
+    | otherwise     -> return True
+  LitP{}    -> return True
+  DefP{}    -> return True
+  ProjP{}   -> return projP
+  VarP{}    -> return False
+  DotP{}    -> return False
+  IApplyP{} -> return False
