@@ -3,13 +3,17 @@
 -- | Occurrences.
 
 module Agda.TypeChecking.Positivity.Occurrence
-  ( Occurrence(..)
+  ( PragmaPolarities
+  , Occurrence(..)
   , OccursWhere(..)
   , Where(..)
   , boundToEverySome
   , productOfEdgesInBoundedWalk
+  , modalPolarityToOccurrence
   ) where
 
+import Prelude hiding (null)
+
 import Control.DeepSeq
 import Control.Monad
 
@@ -22,13 +26,14 @@ import Data.Sequence (Seq)
 import GHC.Generics (Generic)
 
 import Agda.Syntax.Common
+import Agda.Syntax.Common.Pretty
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Position
 
 import Agda.Utils.Graph.AdjacencyMap.Unidirectional (Graph)
 import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as Graph
+import Agda.Utils.List1 (List1)
 import Agda.Utils.Null
-import Agda.Syntax.Common.Pretty
 import Agda.Utils.SemiRing
 import Agda.Utils.Size
 
@@ -38,6 +43,9 @@ import Agda.Utils.Impossible
 
 -- Operations and instances in Agda.TypeChecking.Positivity.
 
+-- | List of polarities stemming from POLARITY pragma.
+type PragmaPolarities = List1 (Ranged Occurrence)
+
 -- | Description of an occurrence.
 data OccursWhere
   = OccursWhere Range (Seq Where) (Seq Where)
@@ -50,21 +58,27 @@ data OccursWhere
 
 instance NFData OccursWhere
 
+instance Null OccursWhere where
+  empty = OccursWhere empty empty empty
+  null (OccursWhere r wh1 wh2) = and [ null r, null wh1, null wh2 ]
+
 -- | One part of the description of an occurrence.
 data Where
   = LeftOfArrow
-  | DefArg QName Nat -- ^ in the nth argument of a define constant
-  | UnderInf         -- ^ in the principal argument of built-in ∞
-  | VarArg           -- ^ as an argument to a bound variable
-  | MetaArg          -- ^ as an argument of a metavariable
-  | ConArgType QName -- ^ in the type of a constructor
-  | IndArgType QName -- ^ in a datatype index of a constructor
+  | DefArg QName Nat       -- ^ in the nth argument of a define constant
+  | UnderInf               -- ^ in the principal argument of built-in ∞
+  | VarArg Occurrence Nat  -- ^ as an argument to a bound variable.
+                           --   The polarity, if given, is the polarity of
+                           --   the argument the occurence is in
+  | MetaArg                -- ^ as an argument of a metavariable
+  | ConArgType QName       -- ^ in the type of a constructor
+  | IndArgType QName       -- ^ in a datatype index of a constructor
   | ConEndpoint QName
-                     -- ^ in an endpoint of a higher constructor
-  | InClause Nat     -- ^ in the nth clause of a defined function
-  | Matched          -- ^ matched against in a clause of a defined function
-  | IsIndex          -- ^ is an index of an inductive family
-  | InDefOf QName    -- ^ in the definition of a constant
+                           -- ^ in an endpoint of a higher constructor
+  | InClause Nat           -- ^ in the nth clause of a defined function
+  | Matched                -- ^ matched against in a clause of a defined function
+  | IsIndex                -- ^ is an index of an inductive family
+  | InDefOf QName          -- ^ in the definition of a constant
   deriving (Show, Eq, Ord, Generic)
 
 instance NFData Where
@@ -98,7 +112,7 @@ instance Pretty Where where
     LeftOfArrow  -> "LeftOfArrow"
     DefArg q i   -> "DefArg"     <+> pretty q <+> pretty i
     UnderInf     -> "UnderInf"
-    VarArg       -> "VarArg"
+    VarArg k i   -> "VarArg" <+> pretty k <+> pretty i
     MetaArg      -> "MetaArg"
     ConArgType q -> "ConArgType" <+> pretty q
     IndArgType q -> "IndArgType" <+> pretty q
@@ -229,3 +243,11 @@ productOfEdgesInBoundedWalk occ g u v bound =
         Just es@(_ : _) -> Just (foldr1 otimes (map Graph.label es))
         Just []         -> __IMPOSSIBLE__
         Nothing         -> Nothing
+
+modalPolarityToOccurrence :: ModalPolarity -> Occurrence
+modalPolarityToOccurrence = \case
+  UnusedPolarity -> Unused
+  StrictlyPositive -> StrictPos
+  Positive -> JustPos
+  Negative -> JustNeg
+  MixedPolarity -> Mixed
