@@ -1,9 +1,7 @@
 {-# LANGUAGE NondecreasingIndentation #-}
-{-# LANGUAGE TypeFamilies #-}
 
 module Agda.TypeChecking.Primitive.Cubical
   ( module Agda.TypeChecking.Primitive.Cubical
-  , module Agda.TypeChecking.Primitive.Cubical.Id
   , module Agda.TypeChecking.Primitive.Cubical.Base
   , module Agda.TypeChecking.Primitive.Cubical.Glue
   , module Agda.TypeChecking.Primitive.Cubical.HCompU
@@ -55,11 +53,10 @@ import qualified Agda.Utils.BoolSet as BoolSet
 import Agda.TypeChecking.Primitive.Cubical.HCompU
 import Agda.TypeChecking.Primitive.Cubical.Glue
 import Agda.TypeChecking.Primitive.Cubical.Base
-import Agda.TypeChecking.Primitive.Cubical.Id
 
 primPOr :: TCM PrimitiveImpl
 primPOr = do
-  requireCubical CErased ""
+  requireCubical CErased
   t    <- runNamesT [] $
           hPi' "a" (els (pure LevelUniv) (cl primLevel))    $ \ a  ->
           nPi' "i" primIntervalType $ \ i  ->
@@ -89,7 +86,7 @@ primPOr = do
 
 primPartial' :: TCM PrimitiveImpl
 primPartial' = do
-  requireCubical CErased ""
+  requireCubical CErased
   t <- runNamesT [] $
        hPi' "a" (els (pure LevelUniv) (cl primLevel)) (\ a ->
         nPi' "φ" primIntervalType $ \ _ ->
@@ -105,7 +102,7 @@ primPartial' = do
 
 primPartialP' :: TCM PrimitiveImpl
 primPartialP' = do
-  requireCubical CErased ""
+  requireCubical CErased
   t <- runNamesT [] $
        hPi' "a" (els (pure LevelUniv) (cl primLevel)) (\ a ->
         nPi' "φ" primIntervalType $ \ phi ->
@@ -120,7 +117,7 @@ primPartialP' = do
 
 primSubOut' :: TCM PrimitiveImpl
 primSubOut' = do
-  requireCubical CErased ""
+  requireCubical CErased
   t    <- runNamesT [] $
           hPi' "a" (els (pure LevelUniv) (cl primLevel)) $ \ a ->
           hPi' "A" (el' (cl primLevelSuc <@> a) (Sort . tmSort <$> a)) $ \ bA ->
@@ -144,7 +141,7 @@ primSubOut' = do
 
 primTrans' :: TCM PrimitiveImpl
 primTrans' = do
-  requireCubical CErased ""
+  requireCubical CErased
   t    <- runNamesT [] $
           hPi' "a" (primIntervalType --> els (pure LevelUniv) (cl primLevel)) $ \ a ->
           nPi' "A" (nPi' "i" primIntervalType $ \ i -> (sort . tmSort <$> (a <@> i))) $ \ bA ->
@@ -155,7 +152,7 @@ primTrans' = do
 
 primHComp' :: TCM PrimitiveImpl
 primHComp' = do
-  requireCubical CErased ""
+  requireCubical CErased
   t    <- runNamesT [] $
           hPi' "a" (els (pure LevelUniv) (cl primLevel)) $ \ a ->
           hPi' "A" (sort . tmSort <$> a) $ \ bA ->
@@ -183,8 +180,8 @@ mkComp s = do
 
   let
     forward la bA r u = pure tTrans
-      <#> (lam "i" $ \i -> la <@> (i `imax` r))
-      <@> (lam "i" $ \i -> bA <@> (i `imax` r))
+      <#> lam "i" (\i -> la <@> (i `imax` r))
+      <@> lam "i" (\i -> bA <@> (i `imax` r))
       <@> r
       <@> u
 
@@ -251,7 +248,8 @@ doPiKanOp cmd t ab = do
         -- We're looking at a fibrant Pi with fibrant domain: Transport
         -- backwards along the domain.
         Type lx -> do
-          [la, bA] <- mapM (open . f) [Level lx, unEl . unDom $ x]
+          la <- open . f $ Level lx
+          bA <- open . f . unEl . unDom $ x
           pure $ Just $ \iOrNot phi a0 ->
             pure tTrans <#> lam "j" (\j -> la <@> iOrNot j)
               <@> lam "j" (\ j -> bA <@> iOrNot j)
@@ -271,10 +269,21 @@ doPiKanOp cmd t ab = do
             Def q [] | Just q == mInterval -> return $ Just $ \_ _ a0 -> a0
             _ -> return Nothing
 
-        _ -> return Nothing
+        Prop _     -> pure Nothing
+        SSet _     -> pure Nothing
+        Inf _ _    -> pure Nothing
+        LevelUniv  -> pure Nothing
+        SizeUniv   -> pure Nothing
+        PiSort{}   -> pure Nothing
+        FunSort{}  -> pure Nothing
+        UnivSort{} -> pure Nothing
+        MetaS{}    -> pure Nothing
+        DefS{}     -> pure Nothing
+        DummyS{}   -> pure Nothing
 
     caseMaybe trFibrantDomain (return Nothing) $ \trA -> Just <$> do
-    [phi, u0] <- mapM (open . unArg) [ignoreBlocking (kanOpCofib cmd), kanOpBase cmd]
+    phi <- open . unArg $ ignoreBlocking $ kanOpCofib cmd
+    u0  <- open . unArg $ kanOpBase cmd
 
     glam (getArgInfo (fst $ famThing ab)) (absName $ snd $ famThing ab) $ \u1 -> do
       case (cmd, ab) of
@@ -329,7 +338,13 @@ doPathPKanOp (HCompOp phi u u0) (IsNot l) (IsNot (bA,x,y)) = do
   tHComp <- getTermLocal builtinHComp
 
   redReturn <=< runNamesT [] $ do
-    [l, u, u0, phi, bA, x, y] <- mapM (open . unArg) [l, u, u0, ignoreBlocking phi, bA, x, y]
+    l   <- open . unArg $ l
+    u   <- open . unArg $ u
+    u0  <- open . unArg $ u0
+    phi <- open . unArg $ ignoreBlocking phi
+    bA  <- open . unArg $ bA
+    x   <- open . unArg $ x
+    y   <- open . unArg $ y
 
     -- hcomp in path spaces is simply hcomp in the underlying space, but
     -- fixing the endpoints at (j ∨ ~ j) in the new direction to those
@@ -370,8 +385,13 @@ doPathPKanOp (TranspOp phi u0) (IsFam l) (IsFam (bA,x,y)) = do
     -- In reality to avoid a round-trip between primComp we use mkComp
     -- here.
     comp <- mkComp $ "transport for path types"
-    [l, u0, phi] <- traverse (open . unArg) [l, u0, ignoreBlocking phi]
-    [bA, x, y] <- mapM (\ a -> open . runNames [] $ lam "i" (const (pure $ unArg a))) [bA, x, y]
+    l   <- open . unArg $ l
+    u0  <- open . unArg $ u0
+    phi <- open . unArg $ ignoreBlocking phi
+    let lami = open . runNames [] . lam "i" . const . pure . unArg
+    bA  <- lami bA
+    x   <- lami x
+    y   <- lami y
     -- Γ ⊢ bA : (i : I) → Type (l i)
     -- Γ ⊢ x, y : (i : I) → bA i
 
@@ -428,7 +448,8 @@ primTransHComp cmd ts nelims = do
             -- regardless of how big the original term is, and
             -- isOneEmpty is *tiny*, so let's use that:
             IZero -> fmap (reduced . notBlocked . argN) . runNamesT [] $ do
-                [l,c] <- mapM (open . unArg) [famThing l, ignoreBlocking sc]
+                l <- open . unArg . famThing $ l
+                c <- open . unArg . ignoreBlocking $ sc
                 lam "i" $ \ i -> clP builtinIsOneEmpty <#> l <#> ilam "o" (\ _ -> c)
 
             -- Otherwise we have some interesting formula (though
@@ -463,7 +484,6 @@ primTransHComp cmd ts nelims = do
 
         mHComp <- getPrimitiveName' builtinHComp
         mGlue <- getPrimitiveName' builtinGlue
-        mId   <- getBuiltinName' builtinId
         pathV <- pathView'
 
         -- By cases on the family, determine what Kan operation we defer
@@ -508,10 +528,6 @@ primTransHComp cmd ts nelims = do
           d | PathType _ _ _ bA x y <- pathV (El __DUMMY_SORT__ d) -> do
             if nelims > 0 then doPathPKanOp operation l ((bA, x, y) <$ t) else fallback
 
-          -- Identity types:
-          Def q [Apply _ , Apply bA , Apply x , Apply y] | Just q == mId -> do
-            maybe fallback return =<< doIdKanOp operation l ((bA, x, y) <$ t)
-
           Def q es -> do
             info <- getConstInfo q
             let
@@ -556,7 +572,7 @@ primTransHComp cmd ts nelims = do
               -- compData knows what to do for the general cases.
               Datatype{dataPars = pars, dataIxs = ixs, dataPathCons = pcons, dataTransp = mtrD}
                 | and [null pcons && ixs == 0 | DoHComp  <- [cmd]], Just as <- allApplyElims es ->
-                  compData mtrD ((not $ null $ pcons) || ixs > 0) (pars + ixs) cmd l (as <$ t) sbA sphi u u0
+                  compData mtrD (not (null pcons) || ixs > 0) (pars + ixs) cmd l (as <$ t) sbA sphi u u0
 
               -- Is this an axiom with constrant transport? Then. Well. Transport is constant.
               Axiom constTransp | constTransp, [] <- es, DoTransp <- cmd -> redReturn $ unArg u0
@@ -568,7 +584,7 @@ primTransHComp cmd ts nelims = do
     allComponentsBack unview phi u p = do
             let
               boolToI b = if b then unview IOne else unview IZero
-              lamlam t = Lam defaultArgInfo (Abs "i" (Lam (setRelevance Irrelevant defaultArgInfo) (Abs "o" t)))
+              lamlam t = Lam defaultArgInfo (Abs "i" (Lam defaultIrrelevantArgInfo (Abs "o" t)))
             as <- decomposeInterval phi
             (flags,t_alphas) <- fmap unzip . forM as $ \ (bs,ts) -> do
                  let u' = listS bs' `applySubst` u
@@ -618,7 +634,8 @@ primTransHComp cmd ts nelims = do
             where
               su' = case view phi of
                      IZero -> notBlocked $ argN $ runNames [] $ do
-                                 [l,c] <- mapM (open . unArg) [l,ignoreBlocking sc]
+                                 l <- open . unArg $ l
+                                 c <- open . unArg . ignoreBlocking $ sc
                                  lam "i" $ \ i -> pure tEmpty <#> l
                                                               <#> ilam "o" (\ _ -> c)
                      _     -> su
@@ -640,7 +657,8 @@ primTransHComp cmd ts nelims = do
                             foldr (iMin . (\(i,b) -> applyUnless b iNeg $ var i)) iO (IntMap.toList m)
               runNamesT [] $ do
                 u <- open u
-                [l,c] <- mapM (open . unArg) [l,ignoreBlocking sc]
+                l <- open . unArg $ l
+                c <- open . unArg $ ignoreBlocking sc
                 phis <- mapM open phis
                 us   <- mapM (open . ignoreBlocking) us
                 lam "i" $ \ i -> do
@@ -695,7 +713,12 @@ primTransHComp cmd ts nelims = do
            io <- getTermLocal builtinIOne
            iz <- getTermLocal builtinIZero
            redReturn <=< runNamesT [] $ do
-             [l,bC,phi,psi,u,u0] <- mapM (open . unArg) [l,bC,ignoreBlocking sphi,psi,u,u0]
+             l   <- open . unArg $ l
+             bC  <- open . unArg $ bC
+             phi <- open . unArg $ ignoreBlocking sphi
+             psi <- open . unArg $ psi
+             u   <- open . unArg $ u
+             u0  <- open . unArg $ u0
              -- hcomp (sc 1) [psi |-> transp sc phi u] (transp sc phi u0)
              pure hcomp <#> (l <@> pure io) <#> (bC <@> pure io) <#> psi
                    <@> lam "j" (\ j -> ilam "o" $ \ o ->
@@ -723,7 +746,7 @@ primTransHComp cmd ts nelims = do
 -- The definition of it comes from 'mkComp'.
 primComp :: TCM PrimitiveImpl
 primComp = do
-  requireCubical CErased ""
+  requireCubical CErased
   t    <- runNamesT [] $
           hPi' "a" (primIntervalType --> els (pure LevelUniv) (cl primLevel)) $ \ a ->
           nPi' "A" (nPi' "i" primIntervalType $ \ i -> (sort . tmSort <$> (a <@> i))) $ \ bA ->
@@ -745,7 +768,11 @@ primComp = do
           _    -> do
             redReturnNoSimpl <=< runNamesT [] $ do
               comp <- mkComp (getBuiltinId PrimComp)
-              [l,c,phi,u,a0] <- mapM (open . unArg) [l,c,phi,u,a0]
+              l   <- open . unArg $ l
+              c   <- open . unArg $ c
+              phi <- open . unArg $ phi
+              u   <- open . unArg $ u
+              a0  <- open . unArg $ a0
               comp l c phi u a0
 
       _ -> __IMPOSSIBLE__
@@ -754,7 +781,7 @@ primComp = do
 -- TODO Andrea: keep reductions that happen under foralls?
 primFaceForall' :: TCM PrimitiveImpl
 primFaceForall' = do
-  requireCubical CErased ""
+  requireCubical CErased
   t <- (primIntervalType --> primIntervalType) --> primIntervalType
   return $ PrimImpl t $ primFun __IMPOSSIBLE__ 1 $ \case
     [phi] -> do
@@ -854,7 +881,7 @@ transpSysTel' flag delta us phi args = do
   ineg <- lift primINeg
   let
     noTranspError t = do
-      reportSDoc "cubical.prim.transpTel" 20 $ nest 2 $ (text "error type =" <+>) $
+      reportSDoc "cubical.prim.transpTel.error" 20 $ nest 2 $ (text "error type =" <+>) $
         addContext ("i" :: String, __DUMMY_DOM__) $ prettyTCM $ unAbs t
       lift . throwError =<< buildClosure t
     bapp :: forall m a. (Applicative m, Subst a) => m (Abs a) -> m (SubstArg a) -> m a
@@ -872,7 +899,7 @@ transpSysTel' flag delta us phi args = do
     combine l ty d [(psi,u)] = u
     combine l ty d ((psi,u):xs)
             = pure tPOr <#> l <@> psi <@> (foldr (\ x y -> pure imax <@> x <@> y) (pure iz) (map fst xs))
-                        <#> (ilam "o" $ \ _ -> ty) -- the type
+                        <#> ilam "o" (\ _ -> ty) -- the type
                         <@> u <@> (combine l ty d xs)
 
     gTransp :: Maybe (LM m Term) -> LM m (Abs Type) -> [(LM m Term,LM m (Abs Term))] -> LM m Term -> LM m Term -> LM m Term
@@ -925,19 +952,25 @@ transpSysTel' flag delta us phi args = do
               else return axi
 
         case s of
-          Type l -> do
-            l <- open $ lam_i (Level l)
+          Univ UType l -> do
+            l <- open $ lam_i $ Level l
             b' <- open b'
             axi <- open axi
             usxi <- mapM open usxi
             gTransp (Just l) b' (zip psis usxi) phi axi
-          Inf _ _  -> noTranspSort
-          SSet _  -> noTranspSort
-          SizeUniv -> noTranspSort
-          LockUniv -> noTranspSort
+          Univ _ _     -> noTranspSort
+          Inf _ _      -> noTranspSort
           IntervalUniv -> noTranspSort
-          Prop{}  -> noTranspSort
-          _ -> noTranspError b'
+          LevelUniv    -> noTranspSort
+          LockUniv     -> noTranspSort
+          SizeUniv     -> noTranspSort
+          PiSort{}     -> noTranspError b'
+          FunSort{}    -> noTranspError b'
+          UnivSort{}   -> noTranspError b'
+          MetaS{}      -> noTranspError b'
+          DefS{}       -> noTranspError b'
+          DummyS{}     -> noTranspError b'
+
     lam_i = Lam defaultArgInfo . Abs "i"
     go :: Telescope -> [[(Term,Term)]] -> Term -> Args -> ExceptT (Closure (Abs Type)) m Args
     go EmptyTel            [] _  []       = return []
@@ -948,18 +981,29 @@ transpSysTel' flag delta us phi args = do
       s <- reduce $ getSort t
       -- Γ ⊢ b : t[1]    Γ, i ⊢ bf : t[i]
       (b,bf) <- runNamesT [] $ do
+        let b' = Abs "i" $ unDom t
         l <- case s of
-               SSet _ -> return Nothing
-               IntervalUniv -> return Nothing
-               SizeUniv     -> return Nothing
-               LockUniv     -> return Nothing
-               Inf _ _ -> return Nothing
-               Type l -> Just <$> open (lam_i (Level l))
-               _ -> noTranspError (Abs "i" (unDom t))
-        t <- open $ Abs "i" (unDom t)
+          Univ UType l -> Just <$> open (lam_i (Level l))
+          Univ USSet _ -> pure Nothing
+          Univ UProp _ -> noTranspError b'  -- Issue #6060
+          Inf  UType _ -> pure Nothing
+          Inf  USSet _ -> pure Nothing
+          Inf  UProp _ -> noTranspError b'  -- Issue #7535
+          IntervalUniv -> pure Nothing
+          LevelUniv    -> pure Nothing
+          LockUniv     -> pure Nothing
+          SizeUniv     -> pure Nothing
+          PiSort{}     -> noTranspError b'
+          FunSort{}    -> noTranspError b'
+          UnivSort{}   -> noTranspError b'
+          MetaS{}      -> noTranspError b'
+          DefS{}       -> noTranspError b'
+          DummyS{}     -> noTranspError b'
+        t <- open b'
         u <- forM u $ \ (psi,upsi) -> do
               (,) <$> open psi <*> open (Abs "i" upsi)
-        [phi,a] <- mapM open [phi, unArg a]
+        phi <- open phi
+        a   <- open $ unArg a
         b <- gTransp l t u phi a
         bf <- bind "i" $ \ i -> do
                             gTransp ((<$> l) $ \ l -> lam "j" $ \ j -> l <@> (pure imin <@> i <@> j))
@@ -1011,26 +1055,23 @@ trFillTel' flag delta phi args r = do
 --           -> ExceptT (Closure (Abs Type)) TCM [Term]
 -- hFillTel' b delta sides base = undefined
 
-pathTelescope
-  :: forall m. (PureTCM m, MonadError TCErr m) =>
-  Telescope -- Δ
-  -> [Arg Term] -- lhs : Δ
-  -> [Arg Term] -- rhs : Δ
+pathTelescope :: forall m. (PureTCM m, MonadError TCErr m)
+  => Telescope  -- ^ Δ
+  -> [Arg Term] -- ^ lhs : Δ
+  -> [Arg Term] -- ^ rhs : Δ
   -> m Telescope
 pathTelescope tel lhs rhs = do
-  x <- runExceptT (pathTelescope' tel lhs rhs)
-  case x of
+  runExceptT (pathTelescope' tel lhs rhs) >>= \case
     Left t -> do
       enterClosure t $ \ t ->
                  typeError . GenericDocError =<<
                     (text "The sort of" <+> pretty t <+> text "should be of the form \"Set l\"")
     Right tel -> return tel
 
-pathTelescope'
-  :: forall m. (PureTCM m, MonadError (Closure Type) m) =>
-  Telescope -- Δ
-  -> [Arg Term] -- lhs : Δ
-  -> [Arg Term] -- rhs : Δ
+pathTelescope' :: forall m. (PureTCM m, MonadError (Closure Type) m)
+  => Telescope  -- ^ Δ
+  -> [Arg Term] -- ^ lhs : Δ
+  -> [Arg Term] -- ^ rhs : Δ
   -> m Telescope
 pathTelescope' tel lhs rhs = do
   pathp <- fromMaybe __IMPOSSIBLE__ <$> getTerm' builtinPathP
@@ -1049,8 +1090,10 @@ pathTelescope' tel lhs rhs = do
     (ExtendTel (a' <$ a) <$>) . runNamesT [] $ do
       let nm = (absName tel)
       tel <- open $ Abs "i" tel
-      [u,v] <- mapM (open . unArg) [u,v]
-      [lhs,rhs] <- mapM open [lhs,rhs]
+      u   <- open $ unArg u
+      v   <- open $ unArg v
+      lhs <- open lhs
+      rhs <- open rhs
       bind nm $ \ eq -> do
         lhs <- lhs
         rhs <- rhs
@@ -1231,20 +1274,31 @@ toCType ty = do
   sort <- reduce $ getSort ty
   case sort of
     Type l -> return $ Just $ LType (LEl l (unEl ty))
-    SSet{} -> do
+    SSet _ -> do
       t <- reduce (unEl ty)
       case t of
         Def q [] -> return $ Just $ ClosedType sort q
         _        -> return $ Nothing
-    _      -> return $ Nothing
+    Prop _       -> pure Nothing
+    Inf _ _      -> pure Nothing
+    IntervalUniv -> pure Nothing  -- ClosedType?  (See comment on CType.)
+    LevelUniv    -> pure Nothing  -- TODO: Or some ClosedType?
+    LockUniv     -> pure Nothing
+    SizeUniv     -> pure Nothing
+    PiSort{}     -> pure Nothing
+    FunSort{}    -> pure Nothing
+    UnivSort{}   -> pure Nothing
+    MetaS{}      -> pure Nothing
+    DefS{}       -> pure Nothing
+    DummyS{}     -> pure Nothing
+
 
 instance Subst CType where
   type SubstArg CType = Term
   applySubst rho (ClosedType s q) = ClosedType (applySubst rho s) q
   applySubst rho (LType t) = LType $ applySubst rho t
 
-hcomp
-  :: (HasBuiltins m, MonadError TCErr m, MonadReduce m, MonadPretty m)
+hcomp :: (HasBuiltins m, MonadError TCErr m, MonadReduce m, MonadPretty m)
   => NamesT m Type
   -> [(NamesT m Term, NamesT m Term)]
   -> NamesT m Term
@@ -1256,9 +1310,8 @@ hcomp ty sys u0 = do
   ty <- ty
   (l, ty) <- toLType ty >>= \case
     Just (LEl l ty) -> return (l, ty)
-    Nothing -> lift $ do -- TODO: support Setω properly
-      typeError . GenericDocError =<< sep
-        [ text "Cubical Agda: cannot generate hcomp clauses at type", prettyTCM ty ]
+    Nothing -> -- TODO: support Setω properly
+      typeError $ CannotGenerateHCompClause ty
   l <- open $ Level l
   ty <- open $ ty
   face <- (foldr max (pure iz) $ map fst $ sys)
