@@ -4,11 +4,10 @@ module Agda.TypeChecking.Records where
 
 import Prelude hiding (null)
 
-import Control.Monad
-import Control.Monad.Except
-import Control.Monad.Trans.Maybe
-import Control.Monad.Writer
-import Control.Applicative
+import Control.Applicative        ( (<|>) )
+import Control.Monad.Except       ( MonadError )
+import Control.Monad.Trans.Maybe  ( MaybeT(MaybeT), runMaybeT )
+import Control.Monad.Writer       ( Writer, runWriter, tell )
 
 import Data.Bifunctor
 import qualified Data.List as List
@@ -28,6 +27,7 @@ import Agda.Syntax.Scope.Base (isNameInScope)
 
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Monad
+import qualified Agda.TypeChecking.Monad.Base.Warning as W
 import Agda.TypeChecking.Pretty as TCM
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad () --instance only
@@ -35,23 +35,26 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Warnings
 import {-# SOURCE #-} Agda.TypeChecking.Primitive.Cubical.Base (isCubicalSubtype)
-
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike (eligibleForProjectionLike)
 
-import Agda.Utils.Either
+import Agda.Utils.Empty
 import Agda.Utils.Function (applyWhen)
 import Agda.Utils.Functor (for, ($>), (<&>))
 import Agda.Utils.Lens
 import Agda.Utils.List
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Pretty (prettyShow)
+import Agda.Syntax.Common.Pretty (prettyShow)
 import Agda.Utils.Singleton
 import Agda.Utils.Size
 
 import Agda.Utils.Impossible
 
+---------------------------------------------------------------------------
+-- * Tools to build record values
+
 mkCon :: ConHead -> ConInfo -> Args -> Term
 mkCon h info args = Con h info (map Apply args)
 
@@ -69,15 +72,15 @@ orderFields r fill axs fs = do
   --   , "  official fields: " <+> sep (map pretty xs)
   --   , "  provided fields: " <+> sep (map pretty ys)
   --   ]
-  unlessNull alien     $ warn $ TooManyFieldsWarning r missing
-  unlessNull duplicate $ warn $ DuplicateFieldsWarning
+  List1.unlessNull alien     $ warn $ W.TooManyFields r missing
+  List1.unlessNull duplicate $ warn $ W.DuplicateFields
   return $ for axs $ \ ax -> fromMaybe (fill ax) $ lookup (unArg ax) uniq
   where
     (uniq, duplicate) = nubAndDuplicatesOn fst fs   -- separating duplicate fields
     xs        = map unArg axs                       -- official fields (accord. record type)
     missing   = filter (not . hasElem (map fst fs)) xs  -- missing  fields
     alien     = filter (not . hasElem xs . fst) fs      -- spurious fields
-    warn w    = tell . singleton . w . map (second getRange)
+    warn w    = tell . singleton . w . fmap (second getRange)
 
 -- | Raise generated 'RecordFieldWarning's as warnings.
 warnOnRecordFieldWarnings :: Writer [RecordFieldWarning] a -> TCM a
@@ -179,10 +182,23 @@ insertMissingFieldsFail
 insertMissingFieldsFail r placeholder fs axs =
   failOnRecordFieldWarnings $ insertMissingFields r placeholder fs axs
 
+---------------------------------------------------------------------------
+-- * Query information about records from signature
+
+-- | Check if a name refers to a record.
+--   If yes, return record definition.
+{-# SPECIALIZE isRecord :: QName -> TCM (Maybe RecordData) #-}
+{-# SPECIALIZE isRecord :: QName -> ReduceM (Maybe RecordData) #-}
+isRecord :: HasConstInfo m => QName -> m (Maybe RecordData)
+isRecord r = do
+  getConstInfo r <&> theDef <&> \case
+    RecordDefn rd -> Just rd
+    _ -> Nothing
+
 -- | Get the definition for a record. Throws an exception if the name
 --   does not refer to a record or the record is abstract.
-getRecordDef :: (HasConstInfo m, ReadTCState m, MonadError TCErr m) => QName -> m Defn
-getRecordDef r = maybe err return =<< isRecord r
+getRecordDef :: (HasConstInfo m, ReadTCState m, MonadError TCErr m) => QName -> m RecordData
+getRecordDef r = fromMaybeM err $ isRecord r
   where err = typeError $ ShouldBeRecordType (El __DUMMY_SORT__ $ Def r [])
 
 -- | Get the record name belonging to a field name.
@@ -193,15 +209,12 @@ getRecordOfField d = caseMaybeM (isProjection d) (return Nothing) $
 
 -- | Get the field names of a record.
 getRecordFieldNames :: (HasConstInfo m, ReadTCState m, MonadError TCErr m)
-                    => QName -> m [Dom C.Name]
+  => QName
+  -> m [Dom C.Name]
 getRecordFieldNames r = recordFieldNames <$> getRecordDef r
 
-getRecordFieldNames_ :: (HasConstInfo m, ReadTCState m)
-                     => QName -> m (Maybe [Dom C.Name])
-getRecordFieldNames_ r = fmap recordFieldNames <$> isRecord r
-
-recordFieldNames :: Defn -> [Dom C.Name]
-recordFieldNames = map (fmap (nameConcrete . qnameName)) . recFields
+recordFieldNames :: RecordData -> [Dom C.Name]
+recordFieldNames = map (fmap (nameConcrete . qnameName)) . _recFields
 
 -- | Find all records with at least the given fields.
 findPossibleRecords :: [C.Name] -> TCM [QName]
@@ -221,10 +234,6 @@ findPossibleRecords fields = do
         _ -> False
     given = Set.fromList fields
 
--- | Get the field types of a record.
-getRecordFieldTypes :: QName -> TCM Telescope
-getRecordFieldTypes r = recTel <$> getRecordDef r
-
 -- | Get the field names belonging to a record type.
 getRecordTypeFields
   :: Type  -- ^ Record type.  Need not be reduced.
@@ -242,30 +251,19 @@ getRecordTypeFields t = do
 -- | Returns the given record type's constructor name (with an empty
 -- range).
 getRecordConstructor :: (HasConstInfo m, ReadTCState m, MonadError TCErr m) => QName -> m ConHead
-getRecordConstructor r = killRange . recConHead <$> getRecordDef r
-
--- | Check if a name refers to a record.
---   If yes, return record definition.
-{-# SPECIALIZE isRecord :: QName -> TCM (Maybe Defn) #-}
-{-# SPECIALIZE isRecord :: QName -> ReduceM (Maybe Defn) #-}
-isRecord :: HasConstInfo m => QName -> m (Maybe Defn)
-isRecord r = do
-  def <- theDef <$> getConstInfo r
-  return $ case def of
-    Record{} -> Just def
-    _        -> Nothing
+getRecordConstructor r = killRange . _recConHead <$> getRecordDef r
 
 -- | Reduce a type and check whether it is a record type.
 --   Succeeds only if type is not blocked by a meta var.
 --   If yes, return its name, parameters, and definition.
-isRecordType :: PureTCM m => Type -> m (Maybe (QName, Args, Defn))
+isRecordType :: PureTCM m => Type -> m (Maybe (QName, Args, RecordData))
 isRecordType t = either (const Nothing) Just <$> tryRecordType t
 
 -- | Reduce a type and check whether it is a record type.
 --   Succeeds only if type is not blocked by a meta var.
 --   If yes, return its name, parameters, and definition.
 --   If no, return the reduced type (unless it is blocked).
-tryRecordType :: PureTCM m => Type -> m (Either (Blocked Type) (QName, Args, Defn))
+tryRecordType :: PureTCM m => Type -> m (Either (Blocked Type) (QName, Args, RecordData))
 tryRecordType t = ifBlocked t (\ m a -> return $ Left $ Blocked m a) $ \ nb t -> do
   let no = return $ Left $ NotBlocked nb t
   case unEl t of
@@ -334,7 +332,7 @@ getDefType f t = do
               [ text $ "head d     = " ++ prettyShow d
               , "parameters =" <+> sep (map prettyTCM pars)
               ]
-            reportSLn "tc.deftype" 60 $ "parameters = " ++ show pars
+            reportSDoc "tc.deftype" 60 $ "parameters = " <+> pretty pars
             if length pars < npars then failure "does not supply enough parameters"
             else Just <$> a `piApplyM` pars
         _ -> failNotDef
@@ -348,7 +346,26 @@ getDefType f t = do
         , prettyTCM t
         , text $ "of its argument " ++ reason
         ]
-      return Nothing
+      reportSDoc "tc.deftype" 60 $ "raw type: " <+> pretty t
+      return $ case unEl t of
+        Dummy{} -> Just __DUMMY_TYPE__
+        _       -> Nothing
+
+-- | Apply a projection to an expression with a known type, returning
+--   the type of the projected value.
+--   The given type should either be a record type or a type eligible for
+--   the principal argument of a projection-like function.
+shouldBeProjectible :: (PureTCM m, MonadTCError m, MonadBlock m)
+                    => Term -> Type -> ProjOrigin -> QName -> m Type
+-- shouldBeProjectible t f = maybe failure return =<< projectionType t f
+shouldBeProjectible v t o f = do
+  t <- abortIfBlocked t
+  projectTyped v t o f >>= \case
+    Just (_ , _ , ft) -> return ft
+    Nothing -> case t of
+      El _ Dummy{} -> return __DUMMY_TYPE__
+      _ -> typeError $ ShouldBeRecordType t
+    -- TODO: more accurate error that makes sense also for proj.-like funs.
 
 -- | The analogue of 'piApply'.  If @v@ is a value of record type @t@
 --   with field @f@, then @projectTyped v t f@ returns the type of @f v@.
@@ -383,7 +400,7 @@ data ElimType
 instance PrettyTCM ElimType where
   prettyTCM (ArgT a)    = prettyTCM a
   prettyTCM (ProjT a b) =
-    "." TCM.<> parens (prettyTCM a <+> "->" <+> prettyTCM b)
+    "." <> parens (prettyTCM a <+> "->" <+> prettyTCM b)
 
 -- | Given a head and its type, compute the types of the eliminations.
 
@@ -402,6 +419,25 @@ typeElims a self (e : es) = do
       (ProjT dom a :) <$> typeElims a self es
     IApply{} -> __IMPOSSIBLE__
 
+-- | Given a term with a given type and a list of eliminations, returning the
+--   type of the term applied to the eliminations.
+eliminateType :: (PureTCM m) => m Empty -> Term -> Type -> Elims -> m Type
+eliminateType err = eliminateType' err . applyE
+
+eliminateType' :: (PureTCM m) => m Empty -> (Elims -> Term) -> Type -> Elims -> m Type
+eliminateType' err hd t [] = return t
+eliminateType' err hd t (e : es) = case e of
+  Apply v -> do
+    t' <- piApplyM' err t v
+    eliminateType' err (hd . (e:)) t' es
+  Proj o f -> reduce t >>= getDefType f >>= \case
+    Just a -> ifNotPiType a (\_ -> absurd <$> err) $ \_ c ->
+      eliminateType' err (hd . (e:)) (c `absApp` (hd [])) es
+    Nothing -> absurd <$> err
+  IApply _ _ r -> do
+    t' <- piApplyM' err t r
+    eliminateType' err (hd . (e:)) t' es
+
 -- | Check if a name refers to an eta expandable record.
 --
 -- The answer is no for a record type with an erased constructor
@@ -410,39 +446,36 @@ typeElims a self (e : es) = do
 {-# SPECIALIZE isEtaRecord :: QName -> ReduceM Bool #-}
 isEtaRecord :: HasConstInfo m => QName -> m Bool
 isEtaRecord r = do
-  isRec <- isRecord r
-  case isRec of
-    Nothing -> return False
-    Just r
-      | recEtaEquality r /= YesEta -> return False
-      | otherwise                  -> do
-        constructorQ <- getQuantity <$>
-                          getConstInfo (conName (recConHead r))
-        currentQ     <- viewTC eQuantity
-        return $ constructorQ `moreQuantity` currentQ
-
-isEtaCon :: HasConstInfo m => QName -> m Bool
-isEtaCon c = getConstInfo' c >>= \case
-  Left (SigUnknown err) -> __IMPOSSIBLE__
-  Left SigAbstract -> return False
-  Right def -> case theDef def of
-    Constructor {conData = r} -> isEtaRecord r
+  isRecord r >>= \case
+    Just r | isEtaRecordDef r -> do
+     constructorQ <- getQuantity <$> getConstInfo (conName (_recConHead r))
+     currentQ     <- viewTC eQuantity
+     return $ constructorQ `moreQuantity` currentQ
     _ -> return False
 
+isEtaRecordDef :: RecordData -> Bool
+isEtaRecordDef r = _recEtaEquality r == YesEta
+
+{-# SPECIALIZE isEtaCon :: QName -> TCM Bool #-}
+isEtaCon :: HasConstInfo m => QName -> m Bool
+isEtaCon c = isJust <$> isEtaRecordConstructor c
+
 -- | Going under one of these does not count as a decrease in size for the termination checker.
 isEtaOrCoinductiveRecordConstructor :: HasConstInfo m => QName -> m Bool
 isEtaOrCoinductiveRecordConstructor c =
-  caseMaybeM (isRecordConstructor c) (return False) $ \ (_, def) -> return $
-    recEtaEquality def == YesEta || recInduction def /= Just Inductive
+  caseMaybeM (isRecordConstructor c) (return False) $ \ (_, def) ->
+    pure (isEtaRecordDef def) `or2M`
+      return (_recInduction def /= Just Inductive)
       -- If in doubt about coinductivity, then yes.
 
 -- | Check if a name refers to a record which is not coinductive.  (Projections are then size-preserving)
-isInductiveRecord :: QName -> TCM Bool
-isInductiveRecord r = maybe False ((Just CoInductive /=) . recInduction) <$> isRecord r
+isInductiveRecord :: HasConstInfo m => QName -> m Bool
+isInductiveRecord r = maybe False ((Just CoInductive /=) . _recInduction) <$> isRecord r
 
 -- | Check if a type is an eta expandable record and return the record identifier and the parameters.
 isEtaRecordType :: (HasConstInfo m)
-                => Type -> m (Maybe (QName, Args))
+  => Type
+  -> m (Maybe (QName, Args))
 isEtaRecordType a = case unEl a of
   Def d es -> do
     let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
@@ -451,32 +484,25 @@ isEtaRecordType a = case unEl a of
 
 -- | Check if a name refers to a record constructor.
 --   If yes, return record definition.
-isRecordConstructor :: HasConstInfo m => QName -> m (Maybe (QName, Defn))
+--
+isRecordConstructor :: HasConstInfo m => QName -> m (Maybe (QName, RecordData))
 isRecordConstructor c = getConstInfo' c >>= \case
-  Left (SigUnknown err)        -> __IMPOSSIBLE__
-  Left SigAbstract             -> return Nothing
-  Right def -> case theDef $ def of
+  Left (SigUnknown err)     -> __IMPOSSIBLE__
+  Left SigCubicalNotErasure -> __IMPOSSIBLE__
+  Left SigAbstract          -> return Nothing
+  Right def                 -> case theDef $ def of
     Constructor{ conData = r } -> fmap (r,) <$> isRecord r
     _                          -> return Nothing
 
--- | Check if a constructor name is the internally generated record constructor.
---
---   Works also for abstract constructors.
-isGeneratedRecordConstructor :: (MonadTCEnv m, HasConstInfo m)
-                             => QName -> m Bool
-isGeneratedRecordConstructor c = ignoreAbstractMode $ do
-  caseMaybeM (isRecordConstructor c) (return False) $ \ (_, def) ->
-    case def of
-      Record{ recNamedCon = False } -> return True
-      _ -> return False
-
+isEtaRecordConstructor :: HasConstInfo m => QName -> m (Maybe (QName, RecordData))
+isEtaRecordConstructor c = isRecordConstructor c >>= \case
+  Nothing -> return Nothing
+  Just (d, def) -> if isEtaRecordDef def then return $ Just (d, def) else return Nothing
 
 -- | Turn off eta for unguarded recursive records.
 --   Projections do not preserve guardedness.
 unguardedRecord :: QName -> PatternOrCopattern -> TCM ()
-unguardedRecord q pat = modifySignature $ updateDefinition q $ updateTheDef $ \case
-  r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) $ NoEta pat }
-  _ -> __IMPOSSIBLE__
+unguardedRecord q pat = modifyRecEta q \ eta -> setEtaEquality eta $ NoEta pat
 
 -- | Turn on eta for non-recursive and inductive guarded recursive records,
 --   unless user declared otherwise.
@@ -491,9 +517,7 @@ updateEtaForRecord q = whenM etaEnabled $ do
       | otherwise -> False
     _ -> __IMPOSSIBLE__
 
-  when switchEta $ do
-    modifySignature $ updateDefinition q $ over (lensTheDef . lensRecord) $ \ d ->
-      d{ _recEtaEquality' = Inferred YesEta }
+  when switchEta $ modifyRecEta q $ const $ Inferred YesEta
 
 -- | Turn on eta for inductive guarded recursive records.
 --   Projections do not preserve guardedness.
@@ -508,7 +532,10 @@ nonRecursiveRecord = updateEtaForRecord
 --
 --   Precondition: record type identifier exists in signature.
 isRecursiveRecord :: QName -> TCM Bool
-isRecursiveRecord q = recRecursive . theDef . fromMaybe __IMPOSSIBLE__ . lookupDefinition q <$> getSignature
+isRecursiveRecord q =
+  getSignature <&> lookupDefinition q <&> fmap theDef <&> \case
+    Just (RecordDefn r) -> recRecursive_ r
+    _ -> __IMPOSSIBLE__
 
 {- | @etaExpandBoundVar i = (Δ, σ, τ)@
 
@@ -546,17 +573,17 @@ expandRecordVar i gamma0 = do
           " since its type " <+> prettyTCM a <+>
           " is not a record type"
         return Nothing
-  caseMaybeM (isRecordType a) failure $ \ (r, pars, def) -> case recEtaEquality def of
-    NoEta{} -> return Nothing
-    YesEta  -> Just <$> do
+  caseMaybeM (isRecordType a) failure $ \ (r, pars, def) -> case isEtaRecordDef def of
+    False -> return Nothing
+    True  -> Just <$> do
       -- Get the record fields @Γ₁ ⊢ tel@ (@tel = Γ'@).
       -- TODO: compose argInfo ai with tel.
-      let tel = recTel def `apply` pars
+      let tel = _recTel def `apply` pars
           m   = size tel
-          fs  = map argFromDom $ recFields def
+          fs  = map argFromDom $ _recFields def
       -- Construct the record pattern @Γ₁, Γ' ⊢ u := c ys@.
           ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-          u   = mkCon (recConHead def) ConOSystem ys
+          u   = mkCon (_recConHead def) ConOSystem ys
       -- @Γ₁, Γ' ⊢ τ₀ : Γ₁, x:_@
           tau0 = consS u $ raiseS m
       -- @Γ₁, Γ', Γ₂ ⊢ τ₀ : Γ₁, x:_, Γ₂@
@@ -616,20 +643,20 @@ curryAt t n = do
       -- This might trigger another call to @etaExpandProjectedVar@ later.
       -- A more efficient version does all the eta-expansions at once here.
       (r, pars, def) <- fromMaybe __IMPOSSIBLE__ <$> isRecordType a
-      if | NoEta _ <- recEtaEquality def -> __IMPOSSIBLE__
+      if | NoEta _ <- _recEtaEquality def -> __IMPOSSIBLE__
          | otherwise -> return ()
       -- TODO: compose argInfo ai with tel.
-      let tel = recTel def `apply` pars
+      let tel = _recTel def `apply` pars
           m   = size tel
-          fs  = map argFromDom $ recFields def
+          fs  = map argFromDom $ _recFields def
           ys  = zipWith (\ f i -> f $> var i) fs $ downFrom m
-          u   = mkCon (recConHead def) ConOSystem ys
+          u   = mkCon (killRange $ _recConHead def) ConOSystem ys
           b'  = raise m b `absApp` u
           t'  = gamma `telePi` (tel `telePi` b')
           gammai = map domInfo $ telToList gamma
           xs  = reverse $ zipWith (\ ai i -> Arg ai $ var i) gammai [m..]
           curry v = teleLam gamma $ teleLam tel $
-                      raise (n+m) v `apply` (xs ++ [Arg ai u])
+                      raise (n + m) v `apply` (xs ++ [Arg ai u])
           zs  = for fs $ fmap $ \ f -> Var 0 [Proj ProjSystem f]
           atel = sgTel $ (,) (absName b) <$> dom
           uncurry v = teleLam gamma $ teleLam atel $
@@ -652,35 +679,60 @@ curryAt t n = do
     where @tel@ is the record telescope instantiated at the parameters @pars@.
 -}
 etaExpandRecord :: (HasConstInfo m, MonadDebug m, ReadTCState m)
-                => QName -> Args -> Term -> m (Telescope, Args)
+  => QName       -- ^ Name of record type.
+  -> Args        -- ^ Parameters applied to record type.
+  -> Term        -- ^ Term to eta-expand.
+  -> m (Telescope, Args)
+     -- ^ Field types instantiated to parameters, field values.
 etaExpandRecord = etaExpandRecord' False
 
 -- | Eta expand a record regardless of whether it's an eta-record or not.
 forceEtaExpandRecord :: (HasConstInfo m, MonadDebug m, ReadTCState m, MonadError TCErr m)
-                     => QName -> Args -> Term -> m (Telescope, Args)
+  => QName       -- ^ Name of record type.
+  -> Args        -- ^ Parameters applied to record type.
+  -> Term        -- ^ Term to eta-expand.
+  -> m (Telescope, Args)
+     -- ^ Field types instantiated to parameters, field values.
 forceEtaExpandRecord = etaExpandRecord' True
 
+-- | Eta-expand a value at the given record type (must match).
 etaExpandRecord' :: (HasConstInfo m, MonadDebug m, ReadTCState m)
-                 => Bool -> QName -> Args -> Term -> m (Telescope, Args)
+  => Bool        -- ^ Force expansion, overriding '_recEtaEquality'?
+  -> QName       -- ^ Name of record type.
+  -> Args        -- ^ Parameters applied to record type.
+  -> Term        -- ^ Term to eta-expand.
+  -> m (Telescope, Args)
+     -- ^ Field types instantiated to parameters, field values.
 etaExpandRecord' forceEta r pars u = do
   def <- fromMaybe __IMPOSSIBLE__ <$> isRecord r
   (tel, _, _, args) <- etaExpandRecord'_ forceEta r pars def u
   return (tel, args)
 
+-- | Eta-expand a value at the given eta record type (must match).
 etaExpandRecord_ :: HasConstInfo m
-                 => QName -> Args -> Defn -> Term -> m (Telescope, ConHead, ConInfo, Args)
+  => QName       -- ^ Name of record type.
+  -> Args        -- ^ Parameters applied to record type.
+  -> RecordData  -- ^ Definition of record type.
+  -> Term        -- ^ Term to eta-expand.
+  -> m (Telescope, ConHead, ConInfo, Args)
+     -- ^ Field types instantiated to parameters, disassembled constructor term.
 etaExpandRecord_ = etaExpandRecord'_ False
 
+-- | Eta-expand a value at the given record type (must match).
 etaExpandRecord'_ :: HasConstInfo m
-                  => Bool -> QName -> Args -> Defn -> Term -> m (Telescope, ConHead, ConInfo, Args)
-etaExpandRecord'_ forceEta r pars def u = do
-  let Record{ recConHead     = con
-            , recFields      = xs
-            , recTel         = tel
-            } = def
-      tel' = apply tel pars
+  => Bool        -- ^ Force expansion, overriding '_recEtaEquality'?
+  -> QName       -- ^ Name of record type.
+  -> Args        -- ^ Parameters applied to record type.
+  -> RecordData  -- ^ Definition of record type.
+  -> Term        -- ^ Term to eta-expand.
+  -> m (Telescope, ConHead, ConInfo, Args)
+     -- ^ Field types instantiated to parameters, disassembled constructor term.
+etaExpandRecord'_ forceEta r pars
+    def@RecordData{ _recConHead = con, _recFields = xs, _recTel = tel }
+    u = do
+  let tel' = apply tel pars
   -- Make sure we do not expand non-eta records (unless forced to):
-  unless (recEtaEquality def == YesEta || forceEta) __IMPOSSIBLE__
+  unless (_recEtaEquality def == YesEta || forceEta) __IMPOSSIBLE__
   case u of
 
     -- Already expanded.
@@ -737,7 +789,7 @@ etaExpandAtRecordType t u = do
 {-# SPECIALIZE etaContractRecord :: QName -> ConHead -> ConInfo -> Args -> ReduceM Term #-}
 etaContractRecord :: HasConstInfo m => QName -> ConHead -> ConInfo -> Args -> m Term
 etaContractRecord r c ci args = if all (not . usableModality) args then fallBack else do
-  Just Record{ recFields = xs } <- isRecord r
+  RecordData{ _recFields = xs } <- fromMaybe __IMPOSSIBLE__ <$> isRecord r
   reportSDoc "tc.record.eta.contract" 20 $ vcat
     [ "eta contracting record"
     , nest 2 $ vcat
@@ -766,7 +818,7 @@ etaContractRecord r c ci args = if all (not . usableModality) args then fallBack
   -- @a@ is the constructor argument, @ax@ the corr. record field name
     -- skip irrelevant record fields by returning DontCare
     case (getRelevance a, hasElims $ unArg a) of
-      (Irrelevant, _)   -> Just Nothing
+      (Irrelevant{}, _)   -> Just Nothing
       -- if @a@ is the record field name applied to a single argument
       -- then it passes the check
       (_, Just (_, [])) -> Nothing  -- not a projection
@@ -775,16 +827,22 @@ etaContractRecord r c ci args = if all (not . usableModality) args then fallBack
         , unDom ax == f -> Just $ Just $ h es
       _                 -> Nothing
 
+{-# SPECIALIZE isSingletonRecord :: QName -> Args -> TCM Bool #-}
 -- | Is the type a hereditarily singleton record type? May return a
 -- blocking metavariable.
 --
 -- Precondition: The name should refer to a record type, and the
 -- arguments should be the parameters to the type.
-isSingletonRecord :: (PureTCM m, MonadBlock m) => QName -> Args -> m Bool
+isSingletonRecord :: (PureTCM m, MonadBlock m)
+  => QName   -- ^ Name of record type to check.
+  -> Args    -- ^ Parameters given to the record type.
+  -> m Bool
 isSingletonRecord r ps = isJust <$> isSingletonRecord' False r ps mempty
 
 isSingletonRecordModuloRelevance :: (PureTCM m, MonadBlock m)
-                                 => QName -> Args -> m Bool
+  => QName   -- ^ Name of record type to check.
+  -> Args    -- ^ Parameters given to the record type.
+  -> m Bool
 isSingletonRecordModuloRelevance r ps = isJust <$> isSingletonRecord' True r ps mempty
 
 -- | Return the unique (closed) inhabitant if exists.
@@ -811,18 +869,18 @@ isSingletonRecord' regardIrrelevance r ps rs = do
     caseMaybeM (isRecord r) no $ \ def -> do
       -- We might not know yet whether a record type is recursive because the positivity checker hasn't run yet.
       -- In this case, we pessimistically consider the record type to be recursive (@True@).
-      let recursive = maybe True (not . null) $ recMutual def
+      let recursive = maybe True (not . null) $ _recMutual def
       -- Andreas, 2022-03-23, issue #5823
       -- We may pass through terminating record types as often as we want.
       -- If the termination checker has not run yet, we pessimistically consider the record type
       -- to be non-terminating.
-      let nonTerminating = maybe True not $ recTerminates def
+      let nonTerminating = maybe True not $ _recTerminates def
       reportSDoc "tc.meta.eta" 30 $ vcat
         [ hsep [ prettyTCM r, "is recursive      :", prettyTCM recursive      ]
         , hsep [ prettyTCM r, "is non-terminating:", prettyTCM nonTerminating ]
         ]
-      fmap (mkCon (recConHead def) ConOSystem) <$> do
-        check (applyWhen (recursive && nonTerminating) (Set.insert r) rs) $ recTel def `apply` ps
+      fmap (mkCon (_recConHead def) ConOSystem) <$> do
+        check (applyWhen (recursive && nonTerminating) (Set.insert r) rs) $ _recTel def `apply` ps
   where
   -- Check that all entries of the constructor telescope are singletons.
   check :: Set QName -> Telescope -> m (Maybe [Arg Term])
@@ -867,7 +925,7 @@ isSingletonType' regardIrrelevance t rs = do
         record :: m (Maybe Term)
         record = runMaybeT $ do
           (r, ps, def) <- MaybeT $ isRecordType t
-          guard (YesEta == recEtaEquality def)
+          guard $ isEtaRecordDef def
           abstract tel <$> MaybeT (isSingletonRecord' regardIrrelevance r ps rs)
 
         -- Slightly harder case: η for Sub {level} tA phi elt.
@@ -893,6 +951,7 @@ isSingletonType' regardIrrelevance t rs = do
 
       (<|>) <$> record <*> subtype
 
+{-# SPECIALIZE isEtaVar :: Term -> Type -> TCM (Maybe Int) #-}
 -- | Checks whether the given term (of the given type) is beta-eta-equivalent
 --   to a variable. Returns just the de Bruijn-index of the variable if it is,
 --   or nothing otherwise.
@@ -927,13 +986,13 @@ isEtaVar u a = runMaybeT $ isEtaVarG u a Nothing []
           case (mi, is) of
             (Just i, _)     -> return i
             (Nothing, [])   -> mzero
-            (Nothing, i:is) -> guard (all (==i) is) >> return i
+            (Nothing, i:is) -> guard (all (== i) is) >> return i
         (_, Pi dom cod) -> addContext dom $ do
           let u'  = raise 1 u `apply` [argFromDom dom $> var 0]
               a'  = absBody cod
-              mi' = fmap (+1) mi
-              es' = (fmap . fmap) (+1) es ++ [Apply $ argFromDom dom $> 0]
-          (-1+) <$> isEtaVarG u' a' mi' es'
+              mi' = fmap (+ 1) mi
+              es' = (fmap . fmap) (+ 1) es ++ [Apply $ argFromDom dom $> 0]
+          (-1 +) <$> isEtaVarG u' a' mi' es'
         _ -> mzero
 
     -- `areEtaVarElims u a es es'` checks whether the given elims es (as applied
