@@ -5,7 +5,6 @@ module Agda.TypeChecking.Reduce
  ( Instantiate, instantiate', instantiate, instantiateWhen
  -- Recursive meta instantiation
  , InstantiateFull, instantiateFull', instantiateFull
- , instantiateFullExceptForDefinitions
  -- Check for meta (no reduction)
  , IsMeta, isMeta
  -- Reduction and blocking
@@ -17,7 +16,7 @@ module Agda.TypeChecking.Reduce
  , unfoldDefinitionE, unfoldDefinitionStep
  , unfoldInlined
  , appDef', appDefE'
- , abortIfBlocked, ifBlocked, isBlocked
+ , abortIfBlocked, ifBlocked, isBlocked, fromBlocked, blockOnError
  -- Simplification
  , Simplify, simplify, simplifyBlocked'
  -- Normalization
@@ -25,12 +24,12 @@ module Agda.TypeChecking.Reduce
  , slowNormaliseArgs
  ) where
 
-import Control.Monad ( (>=>), void )
+import Control.Monad.Except ( MonadError(..) )
 
+import Data.List ( intercalate )
 import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
-import qualified Data.IntMap as IntMap
 import Data.Foldable
 import Data.Traversable
 import Data.HashMap.Strict (HashMap)
@@ -40,13 +39,13 @@ import Agda.Interaction.Options
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
+import Agda.Syntax.Common.Pretty (prettyShow)
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.MetaVars
 import Agda.Syntax.Scope.Base (Scope)
 import Agda.Syntax.Literal
 
-import {-# SOURCE #-} Agda.TypeChecking.Irrelevance (workOnTypes, isPropM)
-import {-# SOURCE #-} Agda.TypeChecking.Level (reallyUnLevelView)
+import {-# SOURCE #-} Agda.TypeChecking.Irrelevance (isPropM)
 import Agda.TypeChecking.Monad hiding ( enterClosure, constructorForm )
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.CompiledClause
@@ -59,14 +58,14 @@ import {-# SOURCE #-} Agda.TypeChecking.Patterns.Match
 import {-# SOURCE #-} Agda.TypeChecking.Pretty
 import {-# SOURCE #-} Agda.TypeChecking.Rewriting
 import {-# SOURCE #-} Agda.TypeChecking.Reduce.Fast
+import {-# SOURCE #-} Agda.TypeChecking.Opacity
 
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
-import Agda.Utils.Maybe
+import Agda.Utils.List1 (List1)
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
-import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import qualified Agda.Utils.SmallSet as SmallSet
@@ -91,9 +90,11 @@ instantiateWhen p =
   localR (\env -> env { redPred = Just p }) .
   instantiateFull'
 
+{-# INLINE reduce #-}
 reduce :: (Reduce a, MonadReduce m) => a -> m a
 reduce = liftReduce . reduce'
 
+{-# INLINE reduceB #-}
 reduceB :: (Reduce a, MonadReduce m) => a -> m (Blocked a)
 reduceB = liftReduce . reduceB'
 
@@ -104,14 +105,7 @@ reduceWithBlocker a = ifBlocked a
   (\b a' -> return (b, a'))
   (\_ a' -> return (neverUnblock, a'))
 
--- Reduce a term and call the continuation. If the continuation is
--- blocked, the whole call is blocked either on what blocked the reduction
--- or on what blocked the continuation (using `blockedOnEither`).
-withReduced
-  :: (Reduce a, IsMeta a, MonadReduce m, MonadBlock m)
-  => a -> (a -> m b) -> m b
-withReduced a cont = ifBlocked a (\b a' -> addOrUnblocker b $ cont a') (\_ a' -> cont a')
-
+{-# INLINE normalise #-}
 normalise :: (Normalise a, MonadReduce m) => a -> m a
 normalise = liftReduce . normalise'
 
@@ -121,6 +115,7 @@ normalise = liftReduce . normalise'
 -- normaliseB :: (MonadReduce m, Reduce t, Normalise t) => t -> m (Blocked t)
 -- normaliseB = normalise >=> reduceB
 
+{-# INLINE simplify #-}
 simplify :: (Simplify a, MonadReduce m) => a -> m a
 simplify = liftReduce . simplify'
 
@@ -132,6 +127,7 @@ isFullyInstantiatedMeta m = do
     InstV inst -> noMetas <$> instantiateFull (instBody inst)
     _ -> return False
 
+{-# INLINABLE blockAll #-}
 -- | Blocking on all blockers.
 blockAll :: (Functor f, Foldable f) => f (Blocked a) -> Blocked (f a)
 blockAll bs = blockedOn block $ fmap ignoreBlocking bs
@@ -139,6 +135,7 @@ blockAll bs = blockedOn block $ fmap ignoreBlocking bs
         blocker NotBlocked{}  = alwaysUnblock
         blocker (Blocked b _) = b
 
+{-# INLINABLE blockAny #-}
 -- | Blocking on any blockers.
 blockAny :: (Functor f, Foldable f) => f (Blocked a) -> Blocked (f a)
 blockAny bs = blockedOn block $ fmap ignoreBlocking bs
@@ -148,6 +145,18 @@ blockAny bs = blockedOn block $ fmap ignoreBlocking bs
         blocker NotBlocked{}  = []
         blocker (Blocked b _) = [b]
 
+{-# SPECIALIZE blockOnError :: Blocker -> TCM a -> TCM a #-}
+-- | Run the given computation but turn any errors into blocked computations with the given blocker
+blockOnError :: MonadError TCErr m => Blocker -> m a -> m a
+blockOnError blocker f
+  | blocker == neverUnblock = f
+  | otherwise               = f `catchError` \case
+    TypeError{}         -> throwError $ PatternErr blocker
+    PatternErr blocker' -> throwError $ PatternErr $ unblockOnEither blocker blocker'
+    GenericException{}  -> __IMPOSSIBLE__
+    err@IOException{}   -> throwError err
+    ParserError{}       -> __IMPOSSIBLE__
+
 -- | Instantiate something.
 --   Results in an open meta variable or a non meta.
 --   Doesn't do any reduction, and preserves blocking tags (when blocking meta
@@ -159,6 +168,7 @@ class Instantiate t where
   instantiate' = traverse instantiate'
 
 instance Instantiate t => Instantiate [t]
+instance Instantiate t => Instantiate (List1 t)
 instance Instantiate t => Instantiate (Map k t)
 instance Instantiate t => Instantiate (Maybe t)
 instance Instantiate t => Instantiate (Strict.Maybe t)
@@ -169,6 +179,9 @@ instance Instantiate t => Instantiate (Elim' t)
 instance Instantiate t => Instantiate (Tele t)
 instance Instantiate t => Instantiate (IPBoundary' t)
 
+instance Instantiate () where
+    instantiate' () = pure ()
+
 instance (Instantiate a, Instantiate b) => Instantiate (a,b) where
     instantiate' (x,y) = (,) <$> instantiate' x <*> instantiate' y
 
@@ -201,9 +214,7 @@ instance Instantiate Term where
          _ | Just m' <- mvTwin mv, blocking ->
            instantiate' (MetaV m' es)
 
-         Open -> return t
-
-         OpenInstance -> return t
+         OpenMeta _ -> return t
 
          BlockedConst u
            | blocking  -> instantiate' . unBrave $
@@ -257,7 +268,8 @@ instance Instantiate Blocker where
   instantiate' (UnblockOnAny bs) = unblockOnAny . Set.fromList <$> mapM instantiate' (Set.toList bs)
   instantiate' b@(UnblockOnMeta x) =
     ifM (isInstantiatedMeta x) (return alwaysUnblock) (return b)
-  instantiate' b@UnblockOnProblem{} = return b
+  instantiate' (UnblockOnProblem pi) =
+    ifM (isProblemSolved pi) (return alwaysUnblock) (return $ UnblockOnProblem pi)
   instantiate' b@UnblockOnDef{} = return b
 
 instance Instantiate Sort where
@@ -289,7 +301,8 @@ instance Instantiate Constraint where
   instantiate' (LevelCmp cmp u v)   = uncurry (LevelCmp cmp) <$> instantiate' (u,v)
   instantiate' (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> instantiate' (a,b)
   instantiate' (UnBlock m)          = return $ UnBlock m
-  instantiate' (FindInstance m cs)  = FindInstance m <$> mapM instantiate' cs
+  instantiate' (FindInstance r m cs)   = FindInstance r m <$> mapM instantiate' cs
+  instantiate' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   instantiate' (IsEmpty r t)        = IsEmpty r <$> instantiate' t
   instantiate' (CheckSizeLtSat t)   = CheckSizeLtSat <$> instantiate' t
   instantiate' c@CheckFunDef{}      = return c
@@ -316,7 +329,7 @@ instance Instantiate EqualityView where
     <$> instantiate' t
   instantiate' (IdiomType t)            = IdiomType
     <$> instantiate' t
-  instantiate' (EqualityType s eq l t a b) = EqualityType
+  instantiate' (EqualityType r s eq l t a b) = EqualityType r
     <$> instantiate' s
     <*> return eq
     <*> mapM instantiate' l
@@ -388,6 +401,12 @@ isBlocked
   => t -> m (Maybe Blocker)
 isBlocked t = ifBlocked t (\m _ -> return $ Just m) (\_ _ -> return Nothing)
 
+-- | Throw a pattern violation if the argument is @Blocked@,
+--   otherwise return the value embedded in the @NotBlocked@.
+fromBlocked :: MonadBlock m => Blocked a -> m a
+fromBlocked (Blocked b _) = patternViolation b
+fromBlocked (NotBlocked _ x) = return x
+
 class Reduce t where
   reduce'  :: t -> ReduceM t
   reduceB' :: t -> ReduceM (Blocked t)
@@ -395,11 +414,14 @@ class Reduce t where
   reduce'  t = ignoreBlocking <$> reduceB' t
   reduceB' t = notBlocked <$> reduce' t
 
+  {-# MINIMAL reduce' | reduceB' #-}
+
 instance Reduce Type where
     reduce'  (El s t) = workOnTypes $ El s <$> reduce' t
     reduceB' (El s t) = workOnTypes $ fmap (El s) <$> reduceB' t
 
 instance Reduce Sort where
+    -- Does not return a 'NotBlocked' 'PiSort', 'FunSort', or 'UnivSort'.
     reduceB' s = do
       s <- instantiate' s
       let done | MetaS x _ <- s = return $ blocked x s
@@ -418,20 +440,24 @@ instance Reduce Sort where
               Right s -> reduceB' s
         FunSort s1 s2 -> reduceB' (s1 , s2) >>= \case
           Blocked b (s1',s2') -> return $ Blocked b $ FunSort s1' s2'
-          NotBlocked _ (s1',s2') -> case funSort' s1' s2' of
-            Left b -> return $ Blocked b $ FunSort s1' s2'
-            Right s -> reduceB' s
+          NotBlocked _ (s1',s2') -> do
+            case funSort' s1' s2' of
+              Left b -> return $ Blocked b $ FunSort s1' s2'
+              Right s -> reduceB' s
         UnivSort s1 -> reduceB' s1 >>= \case
           Blocked b s1' -> return $ Blocked b $ UnivSort s1'
           NotBlocked _ s1' -> case univSort' s1' of
             Left b -> return $ Blocked b $ UnivSort s1'
             Right s -> reduceB' s
-        Prop l     -> notBlocked . Prop <$> reduce l
-        Type l     -> notBlocked . Type <$> reduce l
-        Inf f n    -> done
-        SSet l     -> notBlocked . SSet <$> reduce l
+        Univ u l   -> notBlocked . Univ u <$> reduce l
+        Inf _ _    -> done
         SizeUniv   -> done
         LockUniv   -> done
+        LevelUniv  -> do
+          levelUniverseEnabled <- isLevelUniverseEnabled
+          if levelUniverseEnabled
+          then done
+          else return $ notBlocked (mkType 0)
         IntervalUniv -> done
         MetaS x es -> done
         DefS d es  -> done -- postulated sorts do not reduce
@@ -463,9 +489,9 @@ instance Reduce t => Reduce (Maybe t) where
 
 instance Reduce t => Reduce (Arg t) where
     reduce' a = case getRelevance a of
-      Irrelevant -> return a             -- Don't reduce' irr. args!?
-                                         -- Andreas, 2018-03-03, caused #2989.
-      _          -> traverse reduce' a
+      Irrelevant{} -> return a             -- Don't reduce' irr. args!?
+                                           -- Andreas, 2018-03-03, caused #2989.
+      _ -> traverse reduce' a
 
     reduceB' t = traverse id <$> traverse reduceB' t
 
@@ -516,11 +542,7 @@ instance Reduce Term where
   reduceB' = {-# SCC "reduce'<Term>" #-} maybeFastReduceTerm
 
 shouldTryFastReduce :: ReduceM Bool
-shouldTryFastReduce = (optFastReduce <$> pragmaOptions) `and2M` do
-  allowed <- asksTC envAllowedReductions
-  let optionalReductions = SmallSet.fromList [NonTerminatingReductions, UnconfirmedReductions]
-      requiredReductions = allReductions SmallSet.\\ optionalReductions
-  return $ (allowed SmallSet.\\ optionalReductions) == requiredReductions
+shouldTryFastReduce = optFastReduce <$> pragmaOptions
 
 maybeFastReduceTerm :: Term -> ReduceM (Blocked Term)
 maybeFastReduceTerm v = do
@@ -549,13 +571,13 @@ slowReduceTerm v = do
 --    and seems to save 2% sec on the standard library
 --      MetaV x args -> notBlocked . MetaV x <$> reduce' args
       MetaV x es -> iapp es
-      Def f es   -> flip reduceIApply es $ unfoldDefinitionE False reduceB' (Def f []) f es
+      Def f es   -> flip reduceIApply es $ unfoldDefinitionE reduceB' (Def f []) f es
       Con c ci es -> do
           -- Constructors can reduce' when they come from an
           -- instantiated module.
           -- also reduce when they are path constructors
           v <- flip reduceIApply es
-                 $ unfoldDefinitionE False reduceB' (Con c ci []) (conName c) es
+                 $ unfoldDefinitionE reduceB' (Con c ci []) (conName c) es
           traverse reduceNat v
       Sort s   -> done
       Level l  -> ifM (SmallSet.member LevelReductions <$> asksTC envAllowedReductions)
@@ -599,37 +621,37 @@ unfoldCorecursion :: Term -> ReduceM (Blocked Term)
 unfoldCorecursion v = do
   v <- instantiate' v
   case v of
-    Def f es -> unfoldDefinitionE True unfoldCorecursion (Def f []) f es
+    Def f es -> unfoldDefinitionE unfoldCorecursion (Def f []) f es
     _ -> slowReduceTerm v
 
 -- | If the first argument is 'True', then a single delayed clause may
 -- be unfolded.
 unfoldDefinition ::
-  Bool -> (Term -> ReduceM (Blocked Term)) ->
+  (Term -> ReduceM (Blocked Term)) ->
   Term -> QName -> Args -> ReduceM (Blocked Term)
-unfoldDefinition unfoldDelayed keepGoing v f args =
-  unfoldDefinitionE unfoldDelayed keepGoing v f (map Apply args)
+unfoldDefinition keepGoing v f args =
+  unfoldDefinitionE keepGoing v f (map Apply args)
 
 unfoldDefinitionE ::
-  Bool -> (Term -> ReduceM (Blocked Term)) ->
+  (Term -> ReduceM (Blocked Term)) ->
   Term -> QName -> Elims -> ReduceM (Blocked Term)
-unfoldDefinitionE unfoldDelayed keepGoing v f es = do
-  r <- unfoldDefinitionStep unfoldDelayed v f es
+unfoldDefinitionE keepGoing v f es = do
+  r <- unfoldDefinitionStep v f es
   case r of
     NoReduction v    -> return v
     YesReduction _ v -> keepGoing v
 
 unfoldDefinition' ::
-  Bool -> (Simplification -> Term -> ReduceM (Simplification, Blocked Term)) ->
+  (Simplification -> Term -> ReduceM (Simplification, Blocked Term)) ->
   Term -> QName -> Elims -> ReduceM (Simplification, Blocked Term)
-unfoldDefinition' unfoldDelayed keepGoing v0 f es = do
-  r <- unfoldDefinitionStep unfoldDelayed v0 f es
+unfoldDefinition' keepGoing v0 f es = do
+  r <- unfoldDefinitionStep v0 f es
   case r of
     NoReduction v       -> return (NoSimplification, v)
     YesReduction simp v -> keepGoing simp v
 
-unfoldDefinitionStep :: Bool -> Term -> QName -> Elims -> ReduceM (Reduced (Blocked Term) Term)
-unfoldDefinitionStep unfoldDelayed v0 f es =
+unfoldDefinitionStep :: Term -> QName -> Elims -> ReduceM (Reduced (Blocked Term) Term)
+unfoldDefinitionStep v0 f es =
   {-# SCC "reduceDef" #-} do
   traceSDoc "tc.reduce" 90 ("unfoldDefinitionStep v0" <+> pretty v0) $ do
   info <- getConstInfo f
@@ -646,7 +668,6 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
       dontUnfold = or
         [ defNonterminating info && SmallSet.notMember NonTerminatingReductions allowed
         , defTerminationUnconfirmed info && SmallSet.notMember UnconfirmedReductions allowed
-        , defDelayed info == Delayed && not unfoldDelayed
         , prp == Right True
         , isIrrelevant info
         , not defOk
@@ -716,8 +737,9 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
                           defaultResult -- non-terminating or delayed
         ([],[])        -> traceSLn "tc.reduce" 90 "reduceNormalE: no clauses or rewrite rules" $ do
           -- no definition for head
-          blk <- defBlocked <$> getConstInfo f
-          noReduction $ blk $> vfull
+          (defBlocked <$> getConstInfo f) >>= \case
+            Blocked{}    -> noReduction $ Blocked (UnblockOnDef f) vfull
+            NotBlocked{} -> defaultResult
         (cls,rewr)     -> do
           ev <- appDefE_ f v0 cls mcc rewr es
           debugReduce ev
@@ -770,7 +792,7 @@ reduceDefCopy f es = do
                 xs  = etaArgs ps es
                 n   = length xs
                 newes = raise n es ++ [ Apply $ var i <$ x | (i, x) <- zip (downFrom n) xs ]
-        if (defDelayed info == Delayed) || (defNonterminating info)
+        if defNonterminating info
          then return $ NoReduction ()
          else do
             ev <- liftReduce $ appDefE_ f v0 [cl] Nothing mempty $ map notReduced es'
@@ -793,13 +815,13 @@ reduceHead v = do -- ignoreAbstractMode $ do
     Def f es -> do
 
       abstractMode <- envAbstractMode <$> askTC
-      isAbstract <- treatAbstractly f
+      isAbstract <- not <$> hasAccessibleDef f
       traceSLn "tc.inj.reduce" 50 (
-        "reduceHead: we are in " ++ show abstractMode++ "; " ++ prettyShow f ++
+        "reduceHead: we are in " ++ show abstractMode ++ "; " ++ prettyShow f ++
         " is treated " ++ if isAbstract then "abstractly" else "concretely"
         ) $ do
       let v0  = Def f []
-          red = liftReduce $ unfoldDefinitionE False reduceHead v0 f es
+          red = liftReduce $ unfoldDefinitionE reduceHead v0 f es
       def <- theDef <$> getConstInfo f
       case def of
         -- Andreas, 2012-11-06 unfold aliases (single clause terminating functions)
@@ -807,7 +829,7 @@ reduceHead v = do -- ignoreAbstractMode $ do
         -- We restrict this to terminating functions to not make the
         -- type checker loop here on non-terminating functions.
         -- see test/fail/TerminationInfiniteRecord
-        Function{ funClauses = [ _ ], funDelayed = NotDelayed, funTerminates = Just True } -> do
+        Function{ funClauses = [ _ ], funTerminates = Just True } -> do
           traceSLn "tc.inj.reduce" 50 ("reduceHead: head " ++ prettyShow f ++ " is Function") $ do
           red
         Datatype{ dataClause = Just _ } -> red
@@ -815,7 +837,8 @@ reduceHead v = do -- ignoreAbstractMode $ do
         _                               -> return $ notBlocked v
     _ -> return $ notBlocked v
 
--- | Unfold a single inlined function.
+-- | Unfold as many copies as possible, and then potentially a single
+-- inline function.
 unfoldInlined :: PureTCM m => Term -> m Term
 unfoldInlined v = do
   inTypes <- viewTC eWorkingOnTypes
@@ -823,12 +846,31 @@ unfoldInlined v = do
     _ | inTypes -> return v -- Don't inline in types (to avoid unfolding of goals)
     Def f es -> do
       info <- getConstInfo f
-      let def = theDef info
-          irr = isIrrelevant $ defArgInfo info
-      case def of   -- Only for simple definitions with no pattern matching (TODO: maybe copatterns?)
-        Function{ funCompiled = Just Done{}, funDelayed = NotDelayed }
-          | def ^. funInline , not irr -> liftReduce $
-          ignoreBlocking <$> unfoldDefinitionE False (return . notBlocked) (Def f []) f es
+
+      let
+        def = theDef info
+        irr = isIrrelevant $ defArgInfo info
+        continue
+          | defCopy info = fmap notBlocked . unfoldInlined
+          | otherwise    = return . notBlocked
+
+      case def of
+        Function{} ->
+          reportSLn "tc.inline" 90 $
+            intercalate "\n"
+            [ "considering to inline " ++ prettyShow f
+            , "irr         = " ++ prettyShow irr
+            , "funInline   = " ++ prettyShow (def ^. funInline)
+            , "funCompiled = " ++ prettyShow (funCompiled def)
+            ]
+        _ -> pure ()
+
+      case def of -- Only for simple definitions with no pattern matching (TODO: maybe copatterns?)
+        Function{ funCompiled = Just Done{} }
+          | (defCopy info || def ^. funInline), not irr -> do
+            reportSLn "tc.inline" 70 $ "asking to inline " ++ prettyShow f
+            liftReduce $
+              ignoreBlocking <$> unfoldDefinitionE continue (Def f []) f es
         _ -> return v
     _ -> return v
 
@@ -890,13 +932,23 @@ appDefE'' v cls rewr es = traceSDoc "tc.reduce" 90 ("appDefE' v = " <+> pretty v
               nvars = size $ clauseTel cl
           -- if clause is underapplied, skip to next clause
           if length es < npats then goCls cls es else do
+            termCheck <- asksTC envTermCheckReducing
             let (es0, es1) = splitAt npats es
             (m, es0) <- matchCopatterns pats es0
             let es = es0 ++ es1
             case m of
-              No         -> goCls cls es
-              DontKnow b -> rewrite b (applyE v) rewr es
+              No               -> goCls cls es
+              -- Szumi, 2024-03-29, issue #7181:
+              -- If a lazy match is stuck and all non-lazy matches are conclusive,
+              -- then reduction should not be stuck on the current clause and it
+              -- should be fine to continue matching on the next clause.
+              -- This assumes it's impossible for a lazy match to be stuck if
+              -- all non-lazy matches succeed.
+              DontKnow OnlyLazy _ -> goCls cls es
+              DontKnow NonLazy  b -> rewrite b (applyE v) rewr es
               Yes simpl vs -- vs is the subst. for the variables bound in body
+                | termCheck && fromMaybe True (clauseRecursive cl) ->
+                    return $ NoReduction __IMPOSSIBLE__
                 | Just w <- body -> do -- clause has body?
                     -- TODO: let matchPatterns also return the reduced forms
                     -- of the original arguments!
@@ -909,6 +961,7 @@ instance Reduce a => Reduce (Closure a) where
     reduce' cl = do
         x <- enterClosure cl reduce'
         return $ cl { clValue = x }
+{-# SPECIALIZE reduce' :: Closure Constraint -> ReduceM (Closure Constraint) #-}
 
 instance Reduce Telescope where
   reduce' EmptyTel          = return EmptyTel
@@ -926,7 +979,8 @@ instance Reduce Constraint where
   reduce' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> reduce' (u,v)
   reduce' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> reduce' (a,b)
   reduce' (UnBlock m)           = return $ UnBlock m
-  reduce' (FindInstance m cs)   = FindInstance m <$> mapM reduce' cs
+  reduce' (FindInstance r m cs)   = FindInstance r m <$> mapM reduce' cs
+  reduce' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   reduce' (IsEmpty r t)         = IsEmpty r <$> reduce' t
   reduce' (CheckSizeLtSat t)    = CheckSizeLtSat <$> reduce' t
   reduce' c@CheckFunDef{}       = return c
@@ -956,7 +1010,7 @@ instance Reduce EqualityView where
     <$> reduce' t
   reduce' (IdiomType t)            = IdiomType
     <$> reduce' t
-  reduce' (EqualityType s eq l t a b) = EqualityType
+  reduce' (EqualityType r s eq l t a b) = EqualityType r
     <$> reduce' s
     <*> return eq
     <*> mapM reduce' l
@@ -1014,7 +1068,7 @@ instance Simplify Term where
     case v of
       Def f vs   -> iapp vs $ do
         let keepGoing simp v = return (simp, notBlocked v)
-        (simpl, v) <- unfoldDefinition' False keepGoing (Def f []) f vs
+        (simpl, v) <- unfoldDefinition' keepGoing (Def f []) f vs
         when (simpl == YesSimplification) $
           reportSDoc "tc.simplify'" 90 $
             pretty f <+> text ("simplify': unfolding definition returns " ++ show simpl) <+> pretty (ignoreBlocking v)
@@ -1045,12 +1099,11 @@ instance Simplify Sort where
         PiSort a s1 s2 -> piSort <$> simplify' a <*> simplify' s1 <*> simplify' s2
         FunSort s1 s2 -> funSort <$> simplify' s1 <*> simplify' s2
         UnivSort s -> univSort <$> simplify' s
-        Type s     -> Type <$> simplify' s
-        Prop s     -> Prop <$> simplify' s
+        Univ u s   -> Univ u <$> simplify' s
         Inf _ _    -> return s
-        SSet s     -> SSet <$> simplify' s
         SizeUniv   -> return s
         LockUniv   -> return s
+        LevelUniv  -> return s
         IntervalUniv -> return s
         MetaS x es -> MetaS x <$> simplify' es
         DefS d es  -> DefS d <$> simplify' es
@@ -1093,7 +1146,8 @@ instance Simplify Constraint where
   simplify' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> simplify' (u,v)
   simplify' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> simplify' (a,b)
   simplify' (UnBlock m)           = return $ UnBlock m
-  simplify' (FindInstance m cs)   = FindInstance m <$> mapM simplify' cs
+  simplify' (FindInstance r m cs)   = FindInstance r m <$> mapM simplify' cs
+  simplify' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   simplify' (IsEmpty r t)         = IsEmpty r <$> simplify' t
   simplify' (CheckSizeLtSat t)    = CheckSizeLtSat <$> simplify' t
   simplify' c@CheckFunDef{}       = return c
@@ -1136,7 +1190,7 @@ instance Simplify EqualityView where
     <$> simplify' t
   simplify' (IdiomType t)            = IdiomType
     <$> simplify' t
-  simplify' (EqualityType s eq l t a b) = EqualityType
+  simplify' (EqualityType r s eq l t a b) = EqualityType r
     <$> simplify' s
     <*> return eq
     <*> mapM simplify' l
@@ -1154,9 +1208,10 @@ class Normalise t where
   default normalise' :: (t ~ f a, Traversable f, Normalise a) => t -> ReduceM t
   normalise' = traverse normalise'
 
--- boring instances:
+-- Functor instances:
 
 instance Normalise t => Normalise [t]
+instance Normalise t => Normalise (List1 t)
 instance Normalise t => Normalise (Map k t)
 instance Normalise t => Normalise (Maybe t)
 instance Normalise t => Normalise (Strict.Maybe t)
@@ -1165,8 +1220,11 @@ instance Normalise t => Normalise (Strict.Maybe t)
 -- Elim' not included since it contains Arg
 instance Normalise t => Normalise (Named name t)
 instance Normalise t => Normalise (IPBoundary' t)
+instance Normalise t => Normalise (Ranged t)
 instance Normalise t => Normalise (WithHiding t)
 
+-- more boring instances:
+
 instance (Normalise a, Normalise b) => Normalise (a,b) where
     normalise' (x,y) = (,) <$> normalise' x <*> normalise' y
 
@@ -1196,12 +1254,11 @@ instance Normalise Sort where
         PiSort a s1 s2 -> piSort <$> normalise' a <*> normalise' s1 <*> normalise' s2
         FunSort s1 s2 -> funSort <$> normalise' s1 <*> normalise' s2
         UnivSort s -> univSort <$> normalise' s
-        Prop s     -> Prop <$> normalise' s
-        Type s     -> Type <$> normalise' s
+        Univ u s   -> Univ u <$> normalise' s
         Inf _ _    -> return s
-        SSet s     -> SSet <$> normalise' s
         SizeUniv   -> return SizeUniv
         LockUniv   -> return LockUniv
+        LevelUniv  -> return LevelUniv
         IntervalUniv -> return IntervalUniv
         MetaS x es -> return s
         DefS d es  -> return s
@@ -1275,7 +1332,8 @@ instance Normalise Constraint where
   normalise' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> normalise' (u,v)
   normalise' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> normalise' (a,b)
   normalise' (UnBlock m)           = return $ UnBlock m
-  normalise' (FindInstance m cs)   = FindInstance m <$> mapM normalise' cs
+  normalise' (FindInstance r m cs)   = FindInstance r m <$> mapM normalise' cs
+  normalise' (ResolveInstanceHead q) = return $ ResolveInstanceHead q
   normalise' (IsEmpty r t)         = IsEmpty r <$> normalise' t
   normalise' (CheckSizeLtSat t)    = CheckSizeLtSat <$> normalise' t
   normalise' c@CheckFunDef{}       = return c
@@ -1318,7 +1376,7 @@ instance Normalise EqualityView where
     <$> normalise' t
   normalise' (IdiomType t)            = IdiomType
     <$> normalise' t
-  normalise' (EqualityType s eq l t a b) = EqualityType
+  normalise' (EqualityType r s eq l t a b) = EqualityType r
     <$> normalise' s
     <*> return eq
     <*> mapM normalise' l
@@ -1341,6 +1399,7 @@ class InstantiateFull t where
 -- Traversables (doesn't include binders like Abs, Tele):
 
 instance InstantiateFull t => InstantiateFull [t]
+instance InstantiateFull t => InstantiateFull (List1 t)
 instance InstantiateFull t => InstantiateFull (HashMap k t)
 instance InstantiateFull t => InstantiateFull (Map k t)
 instance InstantiateFull t => InstantiateFull (Maybe t)
@@ -1396,21 +1455,23 @@ instance InstantiateFull ConHead where
 instance InstantiateFull DBPatVar where
     instantiateFull' = return
 
+instance InstantiateFull PrimitiveId where
+  instantiateFull' = return
+
 -- Rest:
 
 instance InstantiateFull Sort where
     instantiateFull' s = do
         s <- instantiate' s
         case s of
-            Type n     -> Type <$> instantiateFull' n
-            Prop n     -> Prop <$> instantiateFull' n
-            SSet n     -> SSet <$> instantiateFull' n
+            Univ u n   -> Univ u <$> instantiateFull' n
             PiSort a s1 s2 -> piSort <$> instantiateFull' a <*> instantiateFull' s1 <*> instantiateFull' s2
             FunSort s1 s2 -> funSort <$> instantiateFull' s1 <*> instantiateFull' s2
             UnivSort s -> univSort <$> instantiateFull' s
             Inf _ _    -> return s
             SizeUniv   -> return s
             LockUniv   -> return s
+            LevelUniv  -> return s
             IntervalUniv -> return s
             MetaS x es -> MetaS x <$> instantiateFull' es
             DefS d es  -> DefS d <$> instantiateFull' es
@@ -1474,6 +1535,12 @@ instance (Subst a, InstantiateFull a) => InstantiateFull (Abs a) where
 instance (InstantiateFull t, InstantiateFull e) => InstantiateFull (Dom' t e) where
     instantiateFull' (Dom i n b tac x) = Dom i n b <$> instantiateFull' tac <*> instantiateFull' x
 
+instance InstantiateFull ContextEntry where
+  instantiateFull' (CtxVar x a) = CtxVar x <$> instantiateFull' a
+
+instance InstantiateFull LetBinding where
+  instantiateFull' (LetBinding o v t) = LetBinding o <$> instantiateFull' v <*> instantiateFull' t
+
 -- Andreas, 2021-09-13, issue #5544, need to traverse @checkpoints@ map
 instance InstantiateFull t => InstantiateFull (Open t) where
   instantiateFull' (OpenThing checkpoint checkpoints modl t) =
@@ -1510,7 +1577,8 @@ instance InstantiateFull Constraint where
     LevelCmp cmp u v    -> uncurry (LevelCmp cmp) <$> instantiateFull' (u,v)
     SortCmp cmp a b     -> uncurry (SortCmp cmp) <$> instantiateFull' (a,b)
     UnBlock m           -> return $ UnBlock m
-    FindInstance m cs   -> FindInstance m <$> mapM instantiateFull' cs
+    FindInstance r m cs -> FindInstance r m <$> mapM instantiateFull' cs
+    ResolveInstanceHead q -> return $ ResolveInstanceHead q
     IsEmpty r t         -> IsEmpty r <$> instantiateFull' t
     CheckSizeLtSat t    -> CheckSizeLtSat <$> instantiateFull' t
     c@CheckFunDef{}     -> return c
@@ -1530,7 +1598,11 @@ instance InstantiateFull CompareAs where
   instantiateFull' AsTypes       = return AsTypes
 
 instance InstantiateFull Signature where
-  instantiateFull' (Sig a b c) = uncurry3 Sig <$> instantiateFull' (a, b, c)
+  instantiateFull' (Sig a b c d) = Sig
+    <$> instantiateFull' a
+    <*> instantiateFull' b
+    <*> instantiateFull' c
+    <*> pure d             -- The instance table only stores names
 
 instance InstantiateFull Section where
   instantiateFull' (Section tel) = Section <$> instantiateFull' tel
@@ -1559,16 +1631,15 @@ instance InstantiateFull NLPType where
     <*> instantiateFull' a
 
 instance InstantiateFull NLPSort where
-  instantiateFull' (PType x) = PType <$> instantiateFull' x
-  instantiateFull' (PProp x) = PProp <$> instantiateFull' x
-  instantiateFull' (PSSet x) = PSSet <$> instantiateFull' x
+  instantiateFull' (PUniv u x) = PUniv u <$> instantiateFull' x
   instantiateFull' (PInf f n) = return $ PInf f n
   instantiateFull' PSizeUniv = return PSizeUniv
   instantiateFull' PLockUniv = return PLockUniv
+  instantiateFull' PLevelUniv = return PLevelUniv
   instantiateFull' PIntervalUniv = return PIntervalUniv
 
 instance InstantiateFull RewriteRule where
-  instantiateFull' (RewriteRule q gamma f ps rhs t c) =
+  instantiateFull' (RewriteRule q gamma f ps rhs t c top) =
     RewriteRule q
       <$> instantiateFull' gamma
       <*> pure f
@@ -1576,13 +1647,14 @@ instance InstantiateFull RewriteRule where
       <*> instantiateFull' rhs
       <*> instantiateFull' t
       <*> pure c
+      <*> pure top
 
 instance InstantiateFull DisplayForm where
   instantiateFull' (Display n ps v) = uncurry (Display n) <$> instantiateFull' (ps, v)
 
 instance InstantiateFull DisplayTerm where
-  instantiateFull' (DTerm v)       = DTerm <$> instantiateFull' v
-  instantiateFull' (DDot  v)       = DDot  <$> instantiateFull' v
+  instantiateFull' (DTerm' v es)   = DTerm' <$> instantiateFull' v <*> instantiateFull' es
+  instantiateFull' (DDot' v es)    = DDot'  <$> instantiateFull' v <*> instantiateFull' es
   instantiateFull' (DCon c ci vs)  = DCon c ci <$> instantiateFull' vs
   instantiateFull' (DDef c es)     = DDef c <$> instantiateFull' es
   instantiateFull' (DWithApp v vs ws) = uncurry3 DWithApp <$> instantiateFull' (v, vs, ws)
@@ -1639,13 +1711,12 @@ instance InstantiateFull CompiledClauses where
   instantiateFull' (Case n bs) = Case n <$> instantiateFull' bs
 
 instance InstantiateFull Clause where
-    instantiateFull' (Clause rl rf tel ps b t catchall exact recursive unreachable ell wm) =
+    instantiateFull' (Clause rl rf tel ps b t catchall recursive unreachable ell wm) =
        Clause rl rf <$> instantiateFull' tel
        <*> instantiateFull' ps
        <*> instantiateFull' b
        <*> instantiateFull' t
        <*> return catchall
-       <*> return exact
        <*> return recursive
        <*> return unreachable
        <*> return ell
@@ -1668,45 +1739,27 @@ instance InstantiateFull RemoteMetaVariable where
     <*> instantiateFull' c
 
 instance InstantiateFull Interface where
-  instantiateFull' i = do
-    defs <- instantiateFull' (i ^. intSignature . sigDefinitions)
-    instantiateFullExceptForDefinitions'
-      (set (intSignature . sigDefinitions) defs i)
-
--- | Instantiates everything except for definitions in the signature.
-
-instantiateFullExceptForDefinitions' :: Interface -> ReduceM Interface
-instantiateFullExceptForDefinitions'
-  (Interface h s ft ms mod tlmod scope inside sig metas display userwarn
-     importwarn b foreignCode highlighting libPragmas filePragmas
-     usedOpts patsyns warnings partialdefs) =
-  Interface h s ft ms mod tlmod scope inside
-    <$> ((\s r -> Sig { _sigSections     = s
-                      , _sigDefinitions  = sig ^. sigDefinitions
-                      , _sigRewriteRules = r
-                      })
-         <$> instantiateFull' (sig ^. sigSections)
-         <*> instantiateFull' (sig ^. sigRewriteRules))
-    <*> instantiateFull' metas
-    <*> instantiateFull' display
-    <*> return userwarn
-    <*> return importwarn
-    <*> instantiateFull' b
-    <*> return foreignCode
-    <*> return highlighting
-    <*> return libPragmas
-    <*> return filePragmas
-    <*> return usedOpts
-    <*> return patsyns
-    <*> return warnings
-    <*> return partialdefs
-
--- | Instantiates everything except for definitions in the signature.
-
-instantiateFullExceptForDefinitions ::
-  MonadReduce m => Interface -> m Interface
-instantiateFullExceptForDefinitions =
-  liftReduce . instantiateFullExceptForDefinitions'
+  instantiateFull'
+    (Interface h s ft ms mod tlmod scope inside sig _ display userwarn
+         importwarn b foreignCode highlighting libPragmas filePragmas
+         usedOpts patsyns warnings partialdefs oblocks onames) = do
+    Interface h s ft ms mod tlmod scope inside
+      <$!> instantiateFull' sig
+      <*!> pure mempty               -- remote metas are dropped
+      <*!> instantiateFull' display
+      <*!> return userwarn
+      <*!> return importwarn
+      <*!> instantiateFull' b
+      <*!> return foreignCode
+      <*!> return highlighting
+      <*!> return libPragmas
+      <*!> return filePragmas
+      <*!> return usedOpts
+      <*!> return patsyns
+      <*!> return warnings
+      <*!> return partialdefs
+      <*!> return oblocks
+      <*!> return onames
 
 instance InstantiateFull a => InstantiateFull (Builtin a) where
     instantiateFull' (Builtin t) = Builtin <$> instantiateFull' t
@@ -1722,7 +1775,7 @@ instance InstantiateFull EqualityView where
     <$> instantiateFull' t
   instantiateFull' (IdiomType t)            = IdiomType
     <$> instantiateFull' t
-  instantiateFull' (EqualityType s eq l t a b) = EqualityType
+  instantiateFull' (EqualityType r s eq l t a b) = EqualityType r
     <$> instantiateFull' s
     <*> return eq
     <*> mapM instantiateFull' l
