@@ -119,9 +119,7 @@ compactDef bEnv def rewr = do
 
   -- WARNING: don't use isPropM here because it relies on reduction,
   -- which causes an infinite loop.
-  let isPrp = case getSort (defType def) of
-        Prop{} -> True
-        _      -> False
+  let isPrp = isProp $ getSort $ defType def
 
   shouldReduce <- shouldReduceDef (defName def)
   allowed <- asksTC envAllowedReductions
@@ -344,10 +342,10 @@ data FastCase c = FBranches
   , fsucBranch      :: Maybe c
   , flitBranches    :: Map Literal c
     -- ^ Map from literal to case subtree.
-  , fcatchAllBranch :: Maybe c
+  , fcatchallBranch :: Maybe c
     -- ^ (Possibly additional) catch-all clause.
   , ffallThrough    :: Bool
-    -- ^ (if True) In case of non-canonical argument use catchAllBranch.
+    -- ^ (if True) In case of non-canonical argument use catchallBranch.
   }
 
 --UNUSED Liang-Ting Chen 2019-07-16
@@ -356,7 +354,7 @@ data FastCase c = FBranches
 --                      , fconBranches    = Map.empty
 --                      , fsucBranch      = Nothing
 --                      , flitBranches    = Map.empty
---                      , fcatchAllBranch = Nothing
+--                      , fcatchallBranch = Nothing
 --                      , ffallThrough    = False }
 
 -- | Case tree with bodies.
@@ -383,7 +381,7 @@ fastCompiledClauses bEnv cc =
   case cc of
     Fail{}            -> FFail
     Done xs b         -> FDone xs b
-    Case (Arg _ n) Branches{ etaBranch = Just (c, cc), catchAllBranch = ca } ->
+    Case (Arg _ n) Branches{ etaBranch = Just (c, cc), catchallBranch = ca } ->
       FEta n (conFields c) (fastCompiledClauses bEnv $ content cc) (fastCompiledClauses bEnv <$> ca)
     Case (Arg _ n) bs -> FCase n (fastCase bEnv bs)
 
@@ -394,8 +392,8 @@ fastCase env (Branches proj con _ lit wild fT _) =
     , fconBranches    = Map.mapKeysMonotonic (nameId . qnameName) $ fmap (fastCompiledClauses env . content) (stripSuc con)
     , fsucBranch      = fmap (fastCompiledClauses env . content) $ flip Map.lookup con . conName =<< bSuc env
     , flitBranches    = fmap (fastCompiledClauses env) lit
-    , ffallThrough    = (Just True ==) fT
-    , fcatchAllBranch = fmap (fastCompiledClauses env) wild }
+    , ffallThrough    = Just True == fT
+    , fcatchallBranch = fmap (fastCompiledClauses env) wild }
   where
     stripSuc | Just c <- bSuc env = Map.delete (conName c)
              | otherwise          = id
@@ -467,9 +465,6 @@ fastReduce' norm v = do
     compactDef bEnv info rewr
   ReduceM $ \ redEnv -> reduceTm redEnv bEnv (memoQName constInfo) norm v
 
-unKleisli :: (a -> ReduceM b) -> ReduceM (a -> b)
-unKleisli f = ReduceM $ \ env x -> unReduceM (f x) env
-
 -- * Closures
 
 -- | The abstract machine represents terms as closures containing a 'Term', an environment, and a
@@ -603,19 +598,19 @@ data AM s = Eval (Closure s) !(ControlStack s)
             --   instance, long chains of 'suc' constructors.
           | Match QName FastCompiledClauses (Spine s) (MatchStack s) (ControlStack s)
             -- ^ @Match f cc spine stack ctrl@ Match the arguments @spine@ against the case tree
-            --   @cc@. The match stack contains a (possibly empty) list of 'CatchAll' frames and a
+            --   @cc@. The match stack contains a (possibly empty) list of 'Catchall' frames and a
             --   closure to return in case of a stuck match.
 
 -- | The control stack contains a list of continuations, i.e. what to do with
 --   the result of the current focus.
 type ControlStack s = [ControlFrame s]
 
--- | The control stack for matching. Contains a list of CatchAllFrame's and the closure to return in
+-- | The control stack for matching. Contains a list of CatchallFrame's and the closure to return in
 --   case of a stuck match.
-data MatchStack s = [CatchAllFrame s] :> Closure s
+data MatchStack s = [CatchallFrame s] :> Closure s
 infixr 2 :>, >:
 
-(>:) :: CatchAllFrame s -> MatchStack s -> MatchStack s
+(>:) :: CatchallFrame s -> MatchStack s -> MatchStack s
 (>:) c (cs :> cl) = c : cs :> cl
 -- Previously written as:
 --   c >: cs :> cl = c : cs :> cl
@@ -628,11 +623,11 @@ infixr 2 :>, >:
 --
 -- See https://ghc.haskell.org/trac/ghc/ticket/10018 which may be related.
 
-data CatchAllFrame s = CatchAll FastCompiledClauses (Spine s)
-                        -- ^ @CatchAll cc spine@. Case trees are not fully expanded, that is,
+data CatchallFrame s = Catchall FastCompiledClauses (Spine s)
+                        -- ^ @Catchall cc spine@. Case trees are not fully expanded, that is,
                         --   inner matches can be partial and covered by a catch-all at a higher
                         --   level. This catch-all is represented on the match stack as a
-                        --   @CatchAll@. @cc@ is the case tree in the catch-all case and @spine@ is
+                        --   @Catchall@. @cc@ is the case tree in the catch-all case and @spine@ is
                         --   the value of the pattern variables at the point of the catch-all.
 
 -- An Elim' with a hole.
@@ -989,8 +984,7 @@ reduceTm rEnv bEnv !constInfo normalisation =
               spine' <- elimsToSpine env es
               let (zs, env, !spine'') = buildEnv (instTel i) (spine' <> spine)
               runAM (evalClosure (lams zs (instBody i)) env spine'' ctrl)
-            Just Open{}                         -> __IMPOSSIBLE__
-            Just OpenInstance{}                 -> __IMPOSSIBLE__
+            Just OpenMeta{}                     -> __IMPOSSIBLE__
             Just BlockedConst{}                 -> __IMPOSSIBLE__
             Just PostponedTypeCheckingProblem{} -> __IMPOSSIBLE__
 
@@ -1136,8 +1130,8 @@ reduceTm rEnv bEnv !constInfo normalisation =
     -- Case: CaseK. Pattern matching against a value. If it's a stuck value the pattern match is
     -- stuck and we return the closure from the match stack (see stuckMatch). Otherwise we need to
     -- find a matching branch switch to the Match state. If there is no matching branch we look for
-    -- a CatchAll in the match stack, or fail if there isn't one (see failedMatch). If the current
-    -- branches contain a catch-all case we need to push a CatchAll on the match stack if picking
+    -- a Catchall in the match stack, or fail if there isn't one (see failedMatch). If the current
+    -- branches contain a catch-all case we need to push a Catchall on the match stack if picking
     -- one of the other branches.
     runAM' (Eval cl@(Closure (Value blk) t env spine) ctrl0@(CaseK f i bs spine0 spine1 stack : ctrl)) =
       {-# SCC "runAM.CaseK" #-}
@@ -1190,9 +1184,9 @@ reduceTm rEnv bEnv !constInfo normalisation =
 
         -- Push catch-all frame on the match stack if there is a catch-all (and we're not taking it
         -- right now).
-        catchallStack = case fcatchAllBranch bs of
+        catchallStack = case fcatchallBranch bs of
           Nothing -> stack
-          Just cc -> CatchAll cc catchallSpine >: stack
+          Just cc -> Catchall cc catchallSpine >: stack
 
         -- The matchX functions below all take an extra argument which is what to do if there is no
         -- appropriate branch in the case tree. ifJust is maybe with a different argument order
@@ -1205,8 +1199,8 @@ reduceTm rEnv bEnv !constInfo normalisation =
         matchCon' q ar = lookupCon q bs `ifJust` \ cc ->
           runAM (Match f cc (spine0 <> spine <> spine1) catchallStack ctrl)
 
-        -- Catch-all: Don't add a CatchAll to the match stack since this _is_ the catch-all.
-        matchCatchall = fcatchAllBranch bs `ifJust` \ cc ->
+        -- Catch-all: Don't add a Catchall to the match stack since this _is_ the catch-all.
+        matchCatchall = fcatchallBranch bs `ifJust` \ cc ->
           runAM (Match f cc catchallSpine stack ctrl)
 
         -- Matching literal: Switch to the Match state. There are no arguments to add to the spine.
@@ -1253,11 +1247,11 @@ reduceTm rEnv bEnv !constInfo normalisation =
             (_, [])                    -> done Underapplied -- matter for equality, but might for
             (spine0, Apply e : spine1) -> do                -- rewriting or 'with'.
               -- Replace e by its projections in the spine. And don't forget a
-              -- CatchAll frame if there's a catch-all.
+              -- Catchall frame if there's a catch-all.
               let projClosure (Arg ai f) = Closure Unevaled (Var 0 []) (extendEnv (unArg e) emptyEnv) [Proj ProjSystem f]
               projs <- mapM (createThunk . projClosure) fs
               let spine' = spine0 <> map (Apply . defaultArg) projs <> spine1
-                  stack' = caseMaybe ca stack $ \ cc -> CatchAll cc spine >: stack
+                  stack' = caseMaybe ca stack $ \ cc -> Catchall cc spine >: stack
               runAM (Match f cc spine' stack' ctrl)
             _ -> __IMPOSSIBLE__
 
@@ -1376,11 +1370,11 @@ reduceTm rEnv bEnv !constInfo normalisation =
     stuckMatch :: Blocked_ -> MatchStack s -> ControlStack s -> ST s (Blocked Term)
     stuckMatch blk (_ :> cl) ctrl = rewriteAM (Eval (mkValue blk cl) ctrl)
 
-    -- On a mismatch we find the next 'CatchAll' on the control stack and
+    -- On a mismatch we find the next 'Catchall' on the control stack and
     -- continue matching from there. If there isn't one we get an incomplete
     -- matching error (or get stuck if the function is marked partial).
     failedMatch :: QName -> MatchStack s -> ControlStack s -> ST s (Blocked Term)
-    failedMatch f (CatchAll cc spine : stack :> cl) ctrl = runAM (Match f cc spine (stack :> cl) ctrl)
+    failedMatch f (Catchall cc spine : stack :> cl) ctrl = runAM (Match f cc spine (stack :> cl) ctrl)
     failedMatch f ([] :> cl) ctrl
         -- Bad work-around for #3870: don't fail hard during instance search.
       | speculative          = rewriteAM (Eval (mkValue (NotBlocked (MissingClauses f) ()) cl) ctrl)
@@ -1454,8 +1448,8 @@ instance Pretty (AM s) where
                           , nest 2 $ pretty stack
                           , nest 2 $ prettyList ctrl ]
 
-instance Pretty (CatchAllFrame s) where
-  pretty CatchAll{} = "CatchAll"
+instance Pretty (CatchallFrame s) where
+  pretty Catchall{} = "Catchall"
 
 instance Pretty (MatchStack s) where
   pretty ([] :> _) = empty
