@@ -15,9 +15,8 @@ import Data.Function (on)
 import Data.Maybe
 
 import Control.Arrow (left)
-import Control.Monad
-import Control.Monad.Except
-import Control.Monad.Reader
+import Control.Monad.Except       ( MonadError(..), ExceptT(..), runExceptT )
+import Control.Monad.Reader       ( MonadReader(..), asks, runReaderT )
 import Control.Monad.Writer       ( MonadWriter(..), runWriterT )
 import Control.Monad.Trans.Maybe
 
@@ -41,7 +40,7 @@ import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views (asView, deepUnscope)
 import Agda.Syntax.Concrete (FieldAssignment'(..),LensInScope(..))
 import Agda.Syntax.Common as Common
-import Agda.Syntax.Info as A
+import qualified Agda.Syntax.Info as A
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
@@ -63,6 +62,7 @@ import Agda.TypeChecking.Patterns.Abstract
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records hiding (getRecordConstructor)
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Sort
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Telescope.Path
@@ -81,10 +81,13 @@ import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.List1 (List1, pattern (:|))
+import Agda.Utils.List2 (pattern List2)
 import qualified Agda.Utils.List1 as List1
+import qualified Agda.Utils.List2 as List2
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
+import qualified Agda.Syntax.Common.Pretty as P
 import Agda.Syntax.Common.Pretty (prettyShow)
 import Agda.Utils.Singleton
 import Agda.Utils.Size
@@ -93,13 +96,12 @@ import Agda.Utils.Tuple
 import Agda.Utils.Impossible
 import Agda.TypeChecking.Free (freeIn)
 
---UNUSED Liang-Ting Chen 2019-07-16
----- | Compute the set of flexible patterns in a list of patterns. The result is
-----   the deBruijn indices of the flexible patterns.
---flexiblePatterns :: [NamedArg A.Pattern] -> TCM FlexibleVars
---flexiblePatterns nps = do
---  forMaybeM (zip (downFrom $ length nps) nps) $ \ (i, Arg ai p) -> do
---    runMaybeT $ (\ f -> FlexibleVar (getHiding ai) (getOrigin ai) f (Just i) i) <$> maybeFlexiblePattern p
+-- | Extra read-only state for the LHS checker.
+--
+data LHSContext = LHSContext
+  { lhsRange       :: Range  -- ^ The range of the whole lhs of a clause.
+  , lhsContextSize :: Nat    -- ^ Original size of the context in which the lhs checker runs.
+  }
 
 -- | A pattern is flexible if it is dotted or implicit, or a record pattern
 --   with only flexible subpatterns.
@@ -129,7 +131,6 @@ instance IsFlexiblePattern A.Pattern where
         ifM (isNothing <$> isRecordConstructor c) (return OtherFlex) {-else-}
             (maybeFlexiblePattern qs)
       A.LitP{}  -> return OtherFlex
-      A.AnnP _ _ p -> maybeFlexiblePattern p
       _ -> mzero
 
 instance IsFlexiblePattern (I.Pattern' a) where
@@ -193,15 +194,11 @@ updateProblemEqs eqs = do
     updates = concat <.> traverse update
 
     update :: ProblemEq -> TCM [ProblemEq]
-    update eq@(ProblemEq A.WildP{} _ _) = return []
+    update eq@(ProblemEq A.WildP{} _ _) = return [eq]
     update eq@(ProblemEq p@A.ProjP{} _ _) = typeError $ IllformedProjectionPatternAbstract p
     update eq@(ProblemEq p@(A.AsP info x p') v a) =
       (ProblemEq (A.VarP x) v a :) <$> update (ProblemEq p' v a)
 
-    update eq@(ProblemEq p@(A.AnnP _ _ A.WildP{}) v a) = return [eq]
-    update eq@(ProblemEq p@(A.AnnP info ty p') v a) =
-      (ProblemEq (A.AnnP info ty (A.WildP patNoRange)) v a :) <$> update (ProblemEq p' v a)
-
     update eq@(ProblemEq p v a) = reduce v >>= constructorForm >>= \case
       Con c ci es -> do
         let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
@@ -219,7 +216,6 @@ updateProblemEqs eqs = do
         p <- expandLitPattern p
         case p of
           A.AsP{} -> __IMPOSSIBLE__
-          A.AnnP{} -> __IMPOSSIBLE__
           A.ConP cpi ambC ps -> do
             (c',_) <- disambiguateConstructor ambC d pars
 
@@ -247,7 +243,7 @@ updateProblemEqs eqs = do
 
             updates $ zipWith3 ProblemEq (map namedArg ps) (map unArg vs) bs
 
-          A.RecP pi fs -> do
+          A.RecP _ _ fs -> do
             axs <- map argFromDom . recFields . theDef <$> getConstInfo d
 
             -- Andreas, 2018-09-06, issue #3122.
@@ -259,7 +255,7 @@ updateProblemEqs eqs = do
 
             -- In fs omitted explicit fields are replaced by underscores,
             -- and the fields are put in the correct order.
-            ps <- insertMissingFieldsFail d (const $ A.WildP patNoRange) fs cxs
+            ps <- insertMissingFieldsFail d (const $ A.WildP empty) fs cxs
 
             -- We also need to insert missing implicit or instance fields.
             ps <- insertImplicitPatterns ExpandLast ps ctel
@@ -308,7 +304,6 @@ problemAllVariables problem =
     isSolved A.AbsurdP{}     = True
     -- recursive cases
     isSolved (A.AsP _ _ p)   = isSolved p
-    isSolved (A.AnnP _ _ p)  = isSolved p
     -- impossible:
     isSolved A.ProjP{}       = __IMPOSSIBLE__
     isSolved A.DefP{}        = __IMPOSSIBLE__
@@ -331,7 +326,6 @@ noShadowingOfConstructors problem@(ProblemEq p _ (Dom{domInfo = info, unDom = El
    A.DotP        {} -> return ()
    A.EqualP      {} -> return ()
    A.AsP _ _ p      -> noShadowingOfConstructors $ problem { problemInPat = p }
-   A.AnnP _ _ p     -> noShadowingOfConstructors $ problem { problemInPat = p }
    A.ConP        {} -> __IMPOSSIBLE__
    A.RecP        {} -> __IMPOSSIBLE__
    A.ProjP       {} -> __IMPOSSIBLE__
@@ -456,26 +450,26 @@ transferOrigins ps qs = do
     transfer :: A.Pattern -> DeBruijnPattern -> TCM DeBruijnPattern
     transfer p q = case (asView p , q) of
 
-      ((asB , anns , A.ConP pi _ ps) , ConP c (ConPatternInfo i r ft mb l) qs) -> do
+      ((asB , A.ConP pi _ ps) , ConP c (ConPatternInfo i r ft mb l) qs) -> do
         let cpi = ConPatternInfo (PatternInfo PatOCon asB) r ft mb l
         ConP c cpi <$> transfers ps qs
 
-      ((asB , anns , A.RecP pi fs) , ConP c (ConPatternInfo i r ft mb l) qs) -> do
+      ((asB , A.RecP _kwr pi fs) , ConP c (ConPatternInfo i r ft mb l) qs) -> do
         let Def d _  = unEl $ unArg $ fromMaybe __IMPOSSIBLE__ mb
             axs = map (nameConcrete . qnameName . unArg) (conFields c) `withArgsFrom` qs
             cpi = ConPatternInfo (PatternInfo PatORec asB) r ft mb l
-        ps <- insertMissingFieldsFail d (const $ A.WildP patNoRange) fs axs
+        ps <- insertMissingFieldsFail d (const $ A.WildP empty) fs axs
         ConP c cpi <$> transfers ps qs
 
-      ((asB , anns , p) , ConP c (ConPatternInfo i r ft mb l) qs) -> do
+      ((asB , p) , ConP c (ConPatternInfo i r ft mb l) qs) -> do
         let cpi = ConPatternInfo (PatternInfo (patOrig p) asB) r ft mb l
         return $ ConP c cpi qs
 
-      ((asB , anns , p) , VarP _ x) -> return $ VarP (PatternInfo (patOrig p) asB) x
+      ((asB , p) , VarP _ x) -> return $ VarP (PatternInfo (patOrig p) asB) x
 
-      ((asB , anns , p) , DotP _ u) -> return $ DotP (PatternInfo (patOrig p) asB) u
+      ((asB , p) , DotP _ u) -> return $ DotP (PatternInfo (patOrig p) asB) u
 
-      ((asB , anns , p) , LitP _ l) -> return $ LitP (PatternInfo (patOrig p) asB) l
+      ((asB , p) , LitP _ l) -> return $ LitP (PatternInfo (patOrig p) asB) l
 
       _ -> return q
 
@@ -493,7 +487,6 @@ transferOrigins ps qs = do
     patOrig A.DefP{}        = __IMPOSSIBLE__
     patOrig A.PatternSynP{} = __IMPOSSIBLE__
     patOrig A.WithP{}       = __IMPOSSIBLE__
-    patOrig A.AnnP{}        = __IMPOSSIBLE__
 
     matchingArgs :: NamedArg A.Pattern -> NamedArg DeBruijnPattern -> Bool
     matchingArgs p q
@@ -545,9 +538,6 @@ checkPatternLinearity eqs = do
               check (Map.insert x (u,unDom a) vars) eqs
         A.AsP _ x p ->
           check vars $ [ProblemEq (A.VarP x) u a, ProblemEq p u a] ++ eqs
-        A.AnnP _ _ A.WildP{} -> continue
-        A.AnnP r t p -> (ProblemEq (A.AnnP r t (A.WildP patNoRange)) u a:) <$>
-          check vars (ProblemEq p u a : eqs)
         A.WildP{}       -> continue
         A.DotP{}        -> continue
         A.AbsurdP{}     -> continue
@@ -575,7 +565,7 @@ computeLHSContext = go [] []
     go cxt _ []        EmptyTel = return cxt
     go cxt taken (x : xs) tel0@(ExtendTel a tel) = do
         name <- maybe (dummyName taken $ absName tel) return x
-        let e = (name,) <$> a
+        let e = CtxVar name a
         go (e : cxt) (name : taken) xs (absBody tel)
 
     dummyName taken s =
@@ -585,18 +575,20 @@ computeLHSContext = go [] []
 -- | Bind as patterns
 bindAsPatterns :: [AsBinding] -> TCM a -> TCM a
 bindAsPatterns []                ret = ret
-bindAsPatterns (AsB x v a m : asb) ret = do
+bindAsPatterns (AsB x v a : asb) ret = do
   reportSDoc "tc.lhs.as" 10 $ "as pattern" <+> prettyTCM x <+>
     sep [ ":" <+> prettyTCM a
         , "=" <+> prettyTCM v
         ]
-  addLetBinding (setModality m defaultArgInfo) Inserted x v a $ bindAsPatterns asb ret
+  addLetBinding' Inserted x v a $ bindAsPatterns asb ret
 
 -- | Since with-abstraction can change the type of a variable, we have to
 --   recheck the stripped with patterns when checking a with function.
 recheckStrippedWithPattern :: ProblemEq -> TCM ()
-recheckStrippedWithPattern (ProblemEq p v a) = checkInternal v CmpLeq (unDom a)
-  `catchError` \_ -> typeError $ IllTypedPatternAfterWithAbstraction p
+recheckStrippedWithPattern (ProblemEq p v a)
+  | A.WildP{} <- p = return ()
+  | otherwise      = checkInternal v CmpLeq (unDom a)
+      `catchError` \_ -> typeError $ IllTypedPatternAfterWithAbstraction p
 
 -- | Result of checking the LHS of a clause.
 data LHSResult = LHSResult
@@ -648,6 +640,8 @@ instance InstantiateFull LHSResult where
 checkLeftHandSide :: forall a.
      Call
      -- ^ Trace, e.g. 'CheckLHS' or 'CheckPattern'.
+  -> Range
+     -- ^ 'Range' of the entire left hand side, for error reporting.
   -> Maybe QName
      -- ^ The name of the definition we are checking.
   -> [NamedArg A.Pattern]
@@ -662,17 +656,17 @@ checkLeftHandSide :: forall a.
   -> (LHSResult -> TCM a)
      -- ^ Continuation.
   -> TCM a
-checkLeftHandSide call f ps a withSub' strippedPats =
+checkLeftHandSide call lhsRng f ps a withSub' strippedPats =
  Bench.billToCPS [Bench.Typing, Bench.CheckLHS] $
  traceCallCPS call $ \ ret -> do
 
   -- To allow module parameters to be refined by matching, we're adding the
   -- context arguments as wildcard patterns and extending the type with the
   -- context telescope.
-  cxt <- map (setOrigin Inserted) . reverse <$> getContext
-  let tel = telFromList' prettyShow cxt
-      cps = [ unnamed . A.VarP . A.mkBindName . fst <$> argFromDom d
-            | d <- cxt ]
+  cxt <- map (setOrigin Inserted) <$> getContext
+  let tel = contextToTel cxt
+      cps = [ argFromDom dom $> unnamed (A.VarP $ A.mkBindName $ unDom dom)
+            | (_,dom) <- contextVars cxt ]
       eqs0 = zipWith3 ProblemEq (map namedArg cps) (map var $ downFrom $ size tel) (flattenTel tel)
 
   let finalChecks :: LHSState a -> TCM a
@@ -680,8 +674,8 @@ checkLeftHandSide call f ps a withSub' strippedPats =
 
         reportSDoc "tc.lhs.top" 20 $ vcat
           [ "lhs: final checks with remaining equations"
-          , nest 2 $ if null eqs then "(none)" else addContext delta $ vcat $ map prettyTCM eqs
-          , "qs0 =" <+> addContext delta (prettyTCMPatternList qs0)
+          , nest 4 $ if null eqs then "(none)" else addContext delta $ vcat $ map prettyTCM eqs
+          , nest 2 $ "qs0 =" <+> addContext delta (prettyTCMPatternList qs0)
           ]
 
         unless (null rps) __IMPOSSIBLE__
@@ -706,6 +700,13 @@ checkLeftHandSide call f ps a withSub' strippedPats =
           -- transport. See #5448.
 
         arity_a <- arityPiPath a
+
+        reportSDoc "tc.lhs.top" 30 $ vcat
+          [ nest 2 $ "a        =" <+> prettyTCM a
+          , nest 2 $ "arity_a  =" <+> prettyTCM arity_a
+          , nest 2 $ "withSub' =" <+> prettyTCM withSub'
+          ]
+
         -- Compute substitution from the out patterns @qs0@
         let notProj ProjP{} = False
             notProj _       = True
@@ -765,7 +766,7 @@ checkLeftHandSide call f ps a withSub' strippedPats =
             asb          = asb0 ++ asb1
 
         -- Rename internal patterns with these names
-        let makeVar     = maybe deBruijnVar $ debruijnNamedVar . nameToArgName
+        let makeVar     = maybe deBruijnVar $ deBruijnNamedVar . nameToArgName
             ren         = parallelS $ zipWith makeVar (reverse vars) [0..]
 
         qs <- transferOrigins (cps ++ ps) $ applySubst ren qs0
@@ -828,7 +829,8 @@ checkLeftHandSide call f ps a withSub' strippedPats =
   let st = over (lhsProblem . problemEqs) (++ withEqs) st0
 
   -- doing the splits:
-  (result, block) <- unsafeInTopContext $ runWriterT $ (`runReaderT` (size cxt)) $ checkLHS f st
+  let initLHSContext = LHSContext { lhsRange = lhsRng, lhsContextSize = size cxt }
+  (result, block) <- unsafeInTopContext $ runWriterT $ (`runReaderT` initLHSContext) $ checkLHS f st
   return result
 
 -- | Check that this split will generate a modality-correct internal
@@ -837,28 +839,30 @@ checkLeftHandSide call f ps a withSub' strippedPats =
 -- necessarily an approximate check. We assume that any argument which
 -- (a) comes after and (b) mentions a dotted argument will be
 -- transported, which is probably an overestimate.
-conSplitModalityCheck
-  :: Modality
-  -- ^ Modality to check at
+conSplitModalityCheck ::
+     Range
+       -- ^ Range of the whole left hand side, for error reporting.
+  -> Modality
+       -- ^ Modality to check at.
   -> PatternSubstitution
-  -- ^ Substitution resulting from index unification. @Γ ⊢ ρ : Δ'@,
-  -- where @Δ'@ is the context we're in, and @Γ@ is the clause telescope
-  -- before unification.
-  -> Int       -- ^ Variable x at which we split
-  -> Telescope -- ^ The telescope @Γ@ itself
+      -- ^ Substitution resulting from index unification. @Γ ⊢ ρ : Δ'@,
+      -- where @Δ'@ is the context we're in, and @Γ@ is the clause telescope
+      -- before unification.
+  -> Int       -- ^ Variable @x@ at which we split.
+  -> Telescope -- ^ The telescope @Γ@ itself.
   -> Type      -- ^ Target type of the clause.
   -> TCM ()
-conSplitModalityCheck mod rho blocking gamma target = when (any ((/= defaultModality) . getModality) gamma) $ do
+conSplitModalityCheck lhsRng mod rho blocking gamma target = when (any ((/= defaultModality) . getModality) gamma) $ do
   reportSDoc "tc.lhs.top" 30 $ vcat
     [ "LHS modality check for modality: " <+> prettyTCM mod
     , "rho:    " <+> inTopContext (prettyTCM rho)
     , "gamma:  " <+> inTopContext (prettyTCM gamma)
-    , "target: " <+> prettyTCM target <+> parens (pretty target)
+    , "target: " <+> prettyTCM target
+    , "target (raw): " <+> pretty target
     , "Δ'target: " <+> prettyTCM (applyPatSubst rho target)
     , "blocking:" <+> prettyTCM blocking
     ]
-  case firstForced rho (length gamma) of
-    Just ix -> do
+  whenJust (firstForced rho (length gamma)) \ ix -> do
       -- We've found a forced argument. This means that the unifier has
       -- decided to kill a unification variable, and any of its
       -- occurrences in the generated term will be replaced by an
@@ -911,7 +915,6 @@ conSplitModalityCheck mod rho blocking gamma target = when (any ((/= defaultModa
         argn <- name arg
         when docheck $
           usableAtModality (IndexedClauseArg forced argn) mod ty'
-    Nothing -> pure ()
 
   -- ALways check the target clause type. Specifically, we check it both
   -- in Δ' and in Γ. The check in Δ' will sometimes let slip by a
@@ -919,8 +922,9 @@ conSplitModalityCheck mod rho blocking gamma target = when (any ((/= defaultModa
   -- the unifier likes to replace @0-variables for @ω-variables). A
   -- concrete case where this happens is #5468. Check in Δ' first since
   -- that will have the forced variable names.
-  usableAtModality IndexedClause mod (unEl (applyPatSubst rho target))
-  inTopContext $ addContext gamma $ usableAtModality IndexedClause mod (unEl target)
+  setCurrentRange lhsRng do
+    usableAtModality IndexedClause mod (unEl (applyPatSubst rho target))
+    inTopContext $ addContext gamma $ usableAtModality IndexedClause mod (unEl target)
   where
     -- Find the first dotted pattern in the substitution. "First" =
     -- "earliest bound", so counts down from the length of the
@@ -949,7 +953,6 @@ splitStrategy = filter shouldSplit
       A.AbsurdP{} -> False
 
       A.AsP _ _ p  -> shouldSplit $ problem { problemInPat = p }
-      A.AnnP _ _ p -> shouldSplit $ problem { problemInPat = p }
 
       A.ProjP{}       -> __IMPOSSIBLE__
       A.DefP{}        -> __IMPOSSIBLE__
@@ -958,8 +961,7 @@ splitStrategy = filter shouldSplit
 
 
 -- | The loop (tail-recursive): split at a variable in the problem until problem is solved
-checkLHS
-  :: forall tcm a. (MonadTCM tcm, PureTCM tcm, MonadWriter Blocked_ tcm, MonadError TCErr tcm, MonadTrace tcm, MonadReader Nat tcm)
+checkLHS :: forall tcm a. (MonadTCM tcm, PureTCM tcm, MonadWriter Blocked_ tcm, MonadError TCErr tcm, MonadTrace tcm, MonadReader LHSContext tcm)
   => Maybe QName      -- ^ The name of the definition we are checking.
   -> LHSState a       -- ^ The current state.
   -> tcm a
@@ -987,7 +989,7 @@ checkLHS mf = updateModality checkLHS_ where
 
     unlessM (optPatternMatching <$> getsTC getPragmaOptions) $
       unless (problemAllVariables problem) $
-        typeError $ GenericError $ "Pattern matching is disabled"
+        typeError NeedOptionPatternMatching
 
     let splitsToTry = splitStrategy $ problem ^. problemEqs
 
@@ -1030,14 +1032,13 @@ checkLHS mf = updateModality checkLHS_ where
       let pos = size tel - (i + 1)
           (delta1, tel'@(ExtendTel dom adelta2)) = splitTelescopeAt pos tel -- TODO:: tel' defined but not used
 
-      p <- liftTCM $ expandLitPattern p
+      p <- expandLitPattern p
       let splitOnPat = \case
             (A.LitP _ l)      -> splitLit delta1 dom adelta2 l
             p@A.RecP{}        -> splitCon delta1 dom adelta2 p Nothing
             p@(A.ConP _ c ps) -> splitCon delta1 dom adelta2 p $ Just c
             p@(A.EqualP _ ts) -> splitPartial delta1 dom adelta2 ts
             A.AsP _ _ p       -> splitOnPat p
-            A.AnnP _ _ p      -> splitOnPat p
 
             A.VarP{}        -> __IMPOSSIBLE__
             A.WildP{}       -> __IMPOSSIBLE__
@@ -1077,8 +1078,7 @@ checkLHS mf = updateModality checkLHS_ where
 
       -- Compute the new rest type by applying the projection type to 'self'.
       -- Note: we cannot be in a let binding.
-      f <- ifJust mf return $ hardTypeError $
-             GenericError "Cannot use copatterns in a let binding"
+      let f = fromMaybe __IMPOSSIBLE__ mf
       let self = Def f $ patternsToElims ip
       target' <- traverse (`piApplyM` self) projType
 
@@ -1141,11 +1141,17 @@ checkLHS mf = updateModality checkLHS_ where
     --   sigma = fails because several substitutions [[1/i],[1/j]] correspond to phi
     -- @
 
-    splitPartial :: Telescope     -- The types of arguments before the one we split on
-                 -> Dom Type      -- The type of the argument we split on
-                 -> Abs Telescope -- The types of arguments after the one we split on
-                 -> [(A.Expr, A.Expr)] -- [(φ₁ = b1),..,(φn = bn)]
-                 -> ExceptT TCErr tcm (LHSState a)
+    splitPartial ::
+         Telescope
+            -- The types of arguments before the one we split on.
+      -> Dom Type
+            -- The type of the argument we split on.
+      -> Abs Telescope
+            -- The types of arguments after the one we split on.
+      -> List1 (A.Expr, A.Expr)
+            -- [(φ₁ = b1),..,(φn = bn)]
+      -> ExceptT TCErr tcm (LHSState a)
+
     splitPartial delta1 dom adelta2 ts = do
 
       unless (domIsFinite dom) $ liftTCM $ addContext delta1 $
@@ -1175,7 +1181,7 @@ checkLHS mf = updateModality checkLHS_ where
       --             newContext = Γ Ξ
       --             cpSub = raiseS |Ξ|
       --
-      lhsCxtSize <- ask -- size of the context before checkLHS call.
+      lhsCxtSize <- asks lhsContextSize -- size of the context before checkLHS call.
       reportSDoc "tc.lhs.split.partial" 10 $ "lhsCxtSize =" <+> prettyTCM lhsCxtSize
 
       newContext <- liftTCM $ computeLHSContext names delta1
@@ -1184,39 +1190,28 @@ checkLHS mf = updateModality checkLHS_ where
       let cpSub = raiseS $ size newContext - lhsCxtSize
 
       (gamma,sigma) <- liftTCM $ updateContext cpSub (const newContext) $ do
-         ts <- forM ts $ \ (t,u) -> do
+         ts <- forM ts $ \ (lhs, rhs) -> do
                  reportSDoc "tc.lhs.split.partial" 10 $ "currentCxt =" <+> (prettyTCM =<< getContext)
-                 reportSDoc "tc.lhs.split.partial" 10 $ text "t, u (Expr) =" <+> prettyTCM (t,u)
-                 t <- checkExpr t tInterval
-                 u <- checkExpr u tInterval
+                 reportSDoc "tc.lhs.split.partial" 10 $ text "t, u (Expr) =" <+> prettyTCM (lhs, rhs)
+                 t <- checkExpr lhs tInterval
+                 u <- checkExpr rhs tInterval
                  reportSDoc "tc.lhs.split.partial" 10 $ text "t, u        =" <+> pretty (t, u)
-                 u <- intervalView =<< reduce u
-                 case u of
+                 reduce u >>= intervalView >>= \case
                    IZero -> primINeg <@> pure t
                    IOne  -> return t
-                   _     -> typeError $ GenericError $ "Only 0 or 1 allowed on the rhs of face"
+                   _     -> typeError $ ExpectedIntervalLiteral rhs
          -- Example: ts = (i=0) (j=1) will result in phi = ¬ i & j
-         phi <- case ts of
-                   [] -> do
-                     a <- reduce (unEl $ unDom dom)
-                     -- builtinIsOne is defined, since this is a precondition for having Partial
-                     isone <- fromMaybe __IMPOSSIBLE__ <$>  -- newline because of CPP
-                       getBuiltinName' builtinIsOne
-                     case a of
-                       Def q [Apply phi] | q == isone -> return (unArg phi)
-                       _           -> typeError $ BuiltinMustBeIsOne a
-
-                   _  -> foldl (\ x y -> primIMin <@> x <@> y) primIOne (map pure ts)
+         phi <- foldl (\ x y -> primIMin <@> x <@> y) primIOne (fmap pure ts)
          reportSDoc "tc.lhs.split.partial" 10 $ text "phi           =" <+> prettyTCM phi
          reportSDoc "tc.lhs.split.partial" 30 $ text "phi           =" <+> pretty phi
          phi <- reduce phi
          reportSDoc "tc.lhs.split.partial" 10 $ text "phi (reduced) =" <+> prettyTCM phi
-         refined <- forallFaceMaps phi (\ bs m t -> typeError $ GenericError $ "face blocked on meta")
+         refined <- forallFaceMaps phi (\ bs m t -> patternViolation m)
                             (\_ sigma -> (,sigma) <$> getContextTelescope)
          case refined of
            [(gamma,sigma)] -> return (gamma,sigma)
-           []              -> typeError $ GenericError $ "The face constraint is unsatisfiable."
-           _               -> typeError $ GenericError $ "Cannot have disjunctions in a face constraint."
+           []              -> typeError FaceConstraintUnsatisfiable
+           _               -> typeError FaceConstraintDisjunction
       itisone <- liftTCM primItIsOne
       -- substitute the literal in p1 and dpi
       reportSDoc "tc.lhs.faces" 60 $ text $ show sigma
@@ -1279,9 +1274,12 @@ checkLHS mf = updateModality checkLHS_ where
       unlessM (splittableCohesion info) $
         addContext delta1 $ softTypeError $ SplitOnUnusableCohesion dom
 
+      unless (splittablePolarity (getModalPolarity info)) $
+        addContext delta1 $ softTypeError $ SplitOnUnusablePolarity dom
+
       -- check that a is indeed the type of lit (otherwise fail softly)
       -- if not, fail softly since it could be instantiated by a later split.
-      suspendErrors $ equalType a =<< litType lit
+      suspendErrors $ addContext delta1 $ equalType a =<< litType lit
 
       -- Compute the new state
       let problem' = set problemEqs eqs' problem
@@ -1328,6 +1326,9 @@ checkLHS mf = updateModality checkLHS_ where
       unlessM (splittableCohesion info) $
         addContext delta1 $ softTypeError $ SplitOnUnusableCohesion dom
 
+      unless (splittablePolarity (getModalPolarity info)) $
+        addContext delta1 $ softTypeError $ SplitOnUnusablePolarity dom
+
       -- Should we attempt to compute a left inverse for this clause? When
       -- --cubical-compatible --flat-split is given, we don't generate a
       -- left inverse (at all). This means that, when the coverage checker
@@ -1338,13 +1339,16 @@ checkLHS mf = updateModality checkLHS_ where
       let genTrx = boolToMaybe ((getCohesion info == Flat)) SplitOnFlat
 
       -- We should be at a data/record type
-      (dr, d, pars, ixs) <- addContext delta1 $ isDataOrRecordType a
+      (dr, d, s, pars, ixs) <- addContext delta1 $ isDataOrRecordType a
       let isRec = case dr of
             IsData{}   -> False
             IsRecord{} -> True
 
       checkMatchingAllowed d dr  -- No splitting on coinductive constructors.
-      addContext delta1 $ checkSortOfSplitVar dr a delta2 (Just target)
+
+      -- Issue #7503: use principal sort for checking if split is ok
+      let a' = set lensSort s a
+      addContext delta1 $ checkSortOfSplitVar dr a' delta2 (Just target)
 
       -- Jesper, 2019-09-13: if the data type we split on is a strict
       -- set, we locally enable --with-K during unification.
@@ -1358,12 +1362,12 @@ checkLHS mf = updateModality checkLHS_ where
 
       -- Don't split on lazy (non-eta) constructor
       case focusPat of
-        A.ConP cpi _ _ | conPatLazy cpi == ConPatLazy ->
+        A.ConP cpi _ _ | A.conPatLazy cpi == A.ConPatLazy ->
           unlessM (isEtaRecord d) $ softTypeError $ ForcedConstructorNotInstantiated focusPat
         _ -> return ()
 
       -- The type of the constructor will end in an application of the datatype
-      (TelV gamma (El _ ctarget), boundary) <- liftTCM $ telViewPathBoundaryP b
+      (TelV gamma (El _ ctarget), boundary) <- liftTCM $ telViewPathBoundary b
       let Def d' es' = ctarget
           cixs = drop (size pars) $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es'
 
@@ -1377,9 +1381,10 @@ checkLHS mf = updateModality checkLHS_ where
         A.ConP _ _ ps -> do
           ps <- insertImplicitPatterns ExpandLast ps gamma
           return $ useNamesFromPattern ps gamma
-        A.RecP _ fs -> do
-          axs <- map argFromDom . recordFieldNames . theDef <$> getConstInfo d
-          ps <- insertMissingFieldsFail d (const $ A.WildP patNoRange) fs axs
+        A.RecP _ _ fs -> do
+          RecordDefn def <- theDef <$> getConstInfo d
+          let axs = map argFromDom $ recordFieldNames def
+          ps <- insertMissingFieldsFail d (const $ A.WildP empty) fs axs
           ps <- insertImplicitPatterns ExpandLast ps gamma
           return $ useNamesFromPattern ps gamma
         _ -> __IMPOSSIBLE__
@@ -1546,12 +1551,12 @@ checkLHS mf = updateModality checkLHS_ where
                      Quantity1{} -> __IMPOSSIBLE__
                      Quantityω{} -> q
 
-          liftTCM $ addContext delta' $ do
-            withoutK <- optWithoutK <$> pragmaOptions
-            cubical <- optCubicalCompatible <$> pragmaOptions
-            mod <- currentModality
-            when ((withoutK || cubical) && not (null ixs)) $
-              conSplitModalityCheck mod rho (length delta2) tel (unArg target)
+          unless (null ixs) $
+            whenM (withoutKOption `or2M` cubicalCompatibleOption) $ do
+              mod <- currentModality
+              lhsRng <- asks lhsRange
+              liftTCM $ addContext delta' $
+                conSplitModalityCheck lhsRng mod rho (length delta2) tel (unArg target)
 
           -- if rest type reduces,
           -- extend the split problem by previously not considered patterns
@@ -1576,8 +1581,7 @@ checkMatchingAllowed :: (MonadTCError m)
   -> m ()
 checkMatchingAllowed d = \case
   IsRecord InductionAndEta { recordInduction=ind, recordEtaEquality=eta }
-    | Just CoInductive <- ind -> typeError $
-        GenericError "Pattern matching on coinductive types is not allowed"
+    | Just CoInductive <- ind -> typeError SplitOnCoinductive
     | not $ patternMatchingAllowed eta -> typeError $ SplitOnNonEtaRecord d
     | otherwise -> return ()
   IsData -> return ()
@@ -1605,31 +1609,32 @@ hardTypeError = withCallerCallStack $ \loc -> liftTCM . typeError' loc
 type DataOrRecord = DataOrRecord' InductionAndEta
 
 -- | Check if the type is a data or record type and return its name,
---   definition, parameters, and indices. Fails softly if the type could become
+--   definition, sort, parameters, and indices. Fails softly if the type could become
 --   a data/record type by instantiating a variable/metavariable, or fail hard
 --   otherwise.
 isDataOrRecordType
   :: (MonadTCM m, PureTCM m)
   => Type
-  -> ExceptT TCErr m (DataOrRecord, QName, Args, Args)
+  -> ExceptT TCErr m (DataOrRecord, QName, Sort, Args, Args)
        -- ^ The 'Args' are parameters and indices.
 
 isDataOrRecordType a0 = ifBlocked a0 blocked $ \case
   ReallyNotBlocked -> \ a -> case unEl a of
 
     -- Subcase: split type is a Def.
-    Def d es -> liftTCM (theDef <$> getConstInfo d) >>= \case
+    Def d es -> liftTCM (getConstInfo d) >>= \def -> case theDef def of
 
-      Datatype{dataPars = np} -> do
+      Datatype{dataPars = np, dataSort = s} -> do
 
         whenM (isInterval a) $ hardTypeError =<< notData
 
         let (pars, ixs) = splitAt np $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-        return (IsData, d, pars, ixs)
+        return (IsData, d, s, pars, ixs)
 
       Record{ recInduction, recEtaEquality' } -> do
         let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-        return (IsRecord InductionAndEta {recordInduction=recInduction, recordEtaEquality=recEtaEquality' }, d, pars, [])
+        s <- shouldBeSort =<< defType def `piApplyM` pars
+        return (IsRecord InductionAndEta {recordInduction=recInduction, recordEtaEquality=recEtaEquality' }, d, s, pars, [])
 
       -- Issue #2253: the data type could be abstract.
       AbstractDefn{} -> hardTypeError $ SplitOnAbstract d
@@ -1713,14 +1718,14 @@ disambiguateProjection
 disambiguateProjection h ambD@(AmbQ ds) b = do
   -- If the target is not a record type, that's an error.
   -- It could be a meta, but since we cannot postpone lhs checking, we crash here.
-  caseMaybeM (liftTCM $ isRecordType $ unArg b) notRecord $ \(r, vs, def) -> case def of
-    Record{ recFields = fs, recInduction, recEtaEquality' = eta } -> do
+  caseMaybeM (liftTCM $ isRecordType $ unArg b) notRecord
+    \ (r, vs, RecordData{ _recFields = fs, _recInduction = ind, _recEtaEquality' = eta }) -> do
       reportSDoc "tc.lhs.split" 20 $ sep
-        [ text $ "we are of record type r  = " ++ prettyShow r
-        , text   "applied to parameters vs = " <+> prettyTCM vs
-        , text $ "and have fields       fs = " ++ prettyShow (map argFromDom fs)
+        [ "we are of record type r  = " <> pure (P.pretty r)
+        , "applied to parameters vs = " <> prettyTCM vs
+        , "and have fields       fs = " <> pure (P.pretty $ map argFromDom fs)
         ]
-      let comatching = recInduction == Just CoInductive
+      let comatching = ind == Just CoInductive
                     || copatternMatchingAllowed eta
       -- Try the projection candidates.
       -- First, we try to find a disambiguation that doesn't produce
@@ -1729,16 +1734,17 @@ disambiguateProjection h ambD@(AmbQ ds) b = do
           -- If this fails, we try again with constraints, but we require
           -- the solution to be unique.
           tryDisambiguate True fs r vs comatching $ \case
-            ([]   , []      ) -> __IMPOSSIBLE__
-            (err:_, []      ) -> throwError err
-            (_    , disambs@((d,a):_)) -> typeError $ AmbiguousProjection d (map fst disambs)
-    _ -> __IMPOSSIBLE__
-
+            (err:_, [] ) -> throwError err
+            ([]   , [] ) -> __IMPOSSIBLE__
+            (_    , [_]) -> __IMPOSSIBLE__
+            (_    , (d,_) : (d1,_) : disambs) ->
+              typeError $ AmbiguousProjection d $ d1 :| map fst disambs
   where
     tryDisambiguate constraintsOk fs r vs comatching failure = do
       -- Note that tryProj wraps TCM in an ExceptT, collecting errors
       -- instead of throwing them to the user immediately.
-      disambiguations <- mapM (runExceptT . tryProj constraintsOk fs r vs) ds
+      disambiguations :: List1 (Either TCErr (QName, (Arg Type, ArgInfo, Maybe TCState)))
+        <- mapM (runExceptT . tryProj constraintsOk fs r vs) ds
       case List1.partitionEithers disambiguations of
         (_ , (d, (a, ai, mst)) : disambs) | constraintsOk <= null disambs -> do
           mapM_ putTC mst -- Activate state changes
@@ -1845,8 +1851,8 @@ disambiguateConstructor ambC@(AmbQ cs) d pars = do
         -- meaning that only the parameters may differ,
         -- then throw more specific error.
         (_    , [_]) -> typeError $ CantResolveOverloadedConstructorsTargetingSameDatatype d cs
-        (_    , disambs@(((c,_,_) :| _) : _)) -> typeError $
-          AmbiguousConstructor c (map (conName . snd3) $ List1.concat disambs)
+        (_    , (d0@((c,_,_) :| _) : d1 : ds)) -> typeError $
+          AmbiguousConstructor c $ fmap (conName . snd3) $ List2.concat21 $ List2 d0 d1 ds
 
   where
     tryDisambiguate
@@ -2026,14 +2032,15 @@ checkSortOfSplitVar :: (MonadTCM m, PureTCM m, MonadError TCErr m,
                         LensSort a, PrettyTCM a, LensSort ty, PrettyTCM ty)
                     => DataOrRecord -> a -> Telescope -> Maybe ty -> m ()
 checkSortOfSplitVar dr a tel mtarget = do
-  liftTCM (reduce $ getSort a) >>= \case
-    Type{} -> whenM isTwoLevelEnabled checkFibrantSplit
-    Prop{} -> checkPropSplit
-    SSet{} -> return ()
-    Inf u _ -> when (univFibrancy u == IsFibrant) $ whenM isTwoLevelEnabled checkFibrantSplit
-    sa      -> softTypeError =<< do
+  let s = getSort a
+  sa <- liftTCM $ reduce s
+  case sortUniv sa of
+    Just UType{} -> whenM isTwoLevelEnabled checkFibrantSplit
+    Just UProp{} -> checkPropSplit
+    Just USSet{} -> return ()
+    Nothing      -> softTypeError =<< do
       liftTCM $ SortOfSplitVarError <$> isBlocked sa <*> sep
-        [ "Cannot split on datatype in sort" , prettyTCM (getSort a) ]
+        [ "Cannot split on datatype in sort" , prettyTCM s]
 
   where
     checkPropSplit
@@ -2064,13 +2071,13 @@ checkSortOfSplitVar dr a tel mtarget = do
 
     -- Cofibrant types are those that could be the domain of a fibrant
     -- pi type. (Notion by C. Sattler).
-    checkIsCoFibrant t = runBlocked (isCoFibrantSort t) >>= \case
+    checkIsCoFibrant t = isCoFibrantSort t >>= \case
       Left b      -> splitOnFibrantError' t $ Just b
       Right False -> unlessM (isInterval t) $
                        splitOnFibrantError' t $ Nothing
       Right True  -> return ()
 
-    checkIsFibrant t = runBlocked (isFibrant t) >>= \case
+    checkIsFibrant t = isFibrant' t >>= \case
       Left b      -> splitOnFibrantError $ Just b
       Right False -> splitOnFibrantError Nothing
       Right True  -> return ()
