@@ -1,10 +1,11 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE NondecreasingIndentation   #-}
 
 module Agda.TypeChecking.Rules.Record where
 
-import Prelude hiding (null)
+import Prelude hiding (null, not, (&&), (||))
 
-import Control.Monad
 import Data.Maybe
 import qualified Data.Set as Set
 
@@ -14,13 +15,11 @@ import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
-import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Info as Info
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Primitive
-import Agda.TypeChecking.Rewriting.Confluence
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Reduce
@@ -28,9 +27,9 @@ import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Polarity
 import Agda.TypeChecking.Warnings
-import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.CompiledClause (hasProjectionPatterns)
 import Agda.TypeChecking.CompiledClause.Compile
+import Agda.TypeChecking.InstanceArguments
 
 import Agda.TypeChecking.Rules.Data
   ( getGeneralizedParameters, bindGeneralizedParameters, bindParameters
@@ -40,16 +39,15 @@ import Agda.TypeChecking.Rules.Data
 import Agda.TypeChecking.Rules.Term ( isType_ )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl (checkDecl)
 
+import Agda.Utils.Boolean
+import Agda.Utils.Function ( applyWhen )
+import Agda.Utils.Lens
 import Agda.Utils.List (headWithDefault)
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Permutation
-import Agda.Utils.POMonoid
-import Agda.Utils.Pretty (render)
-import qualified Agda.Utils.Pretty as P
+import qualified Agda.Syntax.Common.Pretty as P
 import Agda.Utils.Size
-import Agda.Utils.WithDefault
 
 import Agda.Utils.Impossible
 
@@ -104,6 +102,10 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
             DataOrRecSig n -> n
             _              -> __IMPOSSIBLE__
 
+    -- If the record type is erased, then hard compile-time mode is
+    -- entered.
+    setHardCompileTimeModeIfErased' def $ do
+
     parNames <- getGeneralizedParameters gpars name
 
     bindGeneralizedParameters parNames t $ \ gtel t0 ->
@@ -151,15 +153,8 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
 
       -- Obtain name of constructor (if present).
       (hasNamedCon, conName) <- case con of
-        Just c  -> return (True, c)
-        Nothing -> do
-          m <- killRange <$> currentModule
-          -- Andreas, 2020-06-01, AIM XXXII
-          -- Using prettyTCM here jinxes the printer, see PR #4699.
-          -- r <- prettyTCM name
-          let r = P.pretty $ qnameName name
-          c <- qualify m <$> freshName_ (render r ++ ".constructor")
-          return (False, c)
+        A.NamedRecCon c -> return (True, c)
+        A.FreshRecCon c -> return (False, c)
 
       -- Add record type to signature.
       reportSDoc "tc.rec" 15 $ "adding record type to signature"
@@ -183,36 +178,31 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
           -- Andreas, 2016-09-20, issue #2197.
           -- Eta is inferred by the positivity checker.
           -- We should turn it off until it is proven to be safe.
-          haveEta      = maybe (Inferred $ NoEta patCopat) Specified eta
-          -- haveEta      = maybe (Inferred $ conInduction == Inductive && etaenabled) Specified eta
+          noEta    = Inferred $ NoEta patCopat
+          haveEta0 = maybe noEta Specified eta
           con = ConHead conName (IsRecord patCopat) conInduction $ map argFromDom fs
 
           -- A record is irrelevant if all of its fields are.
           -- In this case, the associated module parameter will be irrelevant.
-          -- See issue 392.
+          -- See issue #392.
           -- Unless it's been declared coinductive or no-eta-equality (#2607).
           recordRelevance
-            | Just NoEta{} <- eta         = Relevant
-            | CoInductive <- conInduction = Relevant
-            | otherwise                   = minimum $ Irrelevant : map getRelevance (telToList ftel)
+            | Just NoEta{} <- eta         = relevant
+            | CoInductive <- conInduction = relevant
+            | null (telToList ftel)       = relevant    -- #6270: eta unit types don't need to be irrelevant
+            | otherwise                   = minimum $ irrelevant : map getRelevance (telToList ftel)
 
       -- Andreas, 2017-01-26, issue #2436
       -- Disallow coinductive records with eta-equality
-      when (conInduction == CoInductive && theEtaEquality haveEta == YesEta) $ do
-        typeError . GenericDocError =<< do
-          sep [ "Agda doesn't like coinductive records with eta-equality."
-              , "If you must, use pragma"
-              , "{-# ETA" <+> prettyTCM name <+> "#-}"
-              ]
+      -- Andreas, 2024-06-14, PR #7300
+      -- Just make this a deadcode warning.
+      haveEta <-
+        if (conInduction == CoInductive && theEtaEquality haveEta0 == YesEta) then do
+          noEta <$ do
+            setCurrentRange eta0 $ warning $ CoinductiveEtaRecord name
+        else pure haveEta0
       reportSDoc "tc.rec" 30 $ "record constructor is " <+> prettyTCM con
 
-      -- Jesper, 2021-05-26: Warn when declaring coinductive record
-      -- but neither --guardedness nor --sized-types is enabled.
-      when (conInduction == CoInductive) $ do
-        guardedness <- collapseDefault . optGuardedness <$> pragmaOptions
-        sizedTypes  <- collapseDefault . optSizedTypes  <$> pragmaOptions
-        unless (guardedness || sizedTypes) $ warning $ NoGuardednessFlag name
-
       -- Add the record definition.
 
       -- Andreas, 2016-06-17, Issue #2018:
@@ -223,7 +213,7 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
       let npars = size tel
           telh  = fmap hideAndRelParams tel
       escapeContext impossible npars $ do
-        addConstant' name defaultArgInfo name t $
+        addConstant' name defaultArgInfo t $
             Record
               { recPars           = npars
               , recClause         = Nothing
@@ -244,20 +234,25 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
               , recComp           = emptyCompKit -- filled in later
               }
 
+        erasure <- optErasure <$> pragmaOptions
         -- Add record constructor to signature
-        addConstant' conName defaultArgInfo conName
-            (telh `abstract` contype) $
+        addConstant' conName defaultArgInfo
+             -- If --erasure is used, then the parameters are erased
+             -- in the constructor's type.
+            (applyWhen erasure (fmap $ applyQuantity zeroQuantity) telh
+             `abstract` contype) $
             Constructor
               { conPars   = npars
               , conArity  = size fs
               , conSrcCon = con
               , conData   = name
               , conAbstr  = Info.defAbstract i
-              , conInd    = conInduction
               , conComp   = emptyCompKit  -- filled in later
               , conProj   = Nothing       -- filled in later
               , conForced = []
               , conErased = Nothing
+              , conErasure = erasure
+              , conInline  = False
               }
 
       -- Declare the constructor as eligible for instance search
@@ -271,7 +266,7 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
         NotInstanceDef -> pure ()
 
       -- Check that the fields fit inside the sort
-      _ <- fitsIn uc [] contype s
+      _ <- fitsIn IsRecord_ conName uc [] contype s
 
       -- Check that the sort admits record declarations.
       checkDataSort name s
@@ -346,7 +341,10 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
       -- For checking the record declarations, hide the record parameters
       -- and the parameters of the parent modules.
       modifyContextInfo (hideOrKeepInstance . maybeErase) $ do
-        params <- getContext
+        -- If --erasure is used, then the parameters are erased in the
+        -- types of the projections.
+        erasure <- optErasure <$> pragmaOptions
+        params  <- applyWhen erasure (fmap $ applyQuantity zeroQuantity) <$> getContext
 
         -- Check the types of the fields and the other record declarations.
         addRecordVar $ withCurrentModule m $ do
@@ -358,14 +356,15 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
           -- See test/Succeed/ProjectionsTakeModuleTelAsParameters.agda.
           tel' <- do
             r <- headWithDefault __IMPOSSIBLE__ <$> getContext
-            return $ telFromList' nameToArgName $ reverse $ r : params
+            return $ contextToTel $ r : params
           setModuleCheckpoint m
           checkRecordProjections m name hasNamedCon con tel' ftel fields
 
 
       -- we define composition here so that the projections are already in the signature.
-      escapeContext impossible npars $ do
-        addCompositionForRecord name con tel (map argFromDom fs) ftel rect
+      whenM cubicalCompatibleOption do
+        escapeContext impossible npars do
+          addCompositionForRecord name haveEta con tel (map argFromDom fs) ftel rect
 
       -- The confluence checker needs to know what symbols match against
       -- the constructor.
@@ -378,28 +377,38 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
   -- then switch on pattern matching for no-eta-equality.
   -- Default is no pattern matching, but definition by copatterns instead.
   patCopat = maybe CopatternMatching (const PatternMatching) pat
-  eta      = (patCopat <$) <$> eta0
+  eta      = ((patCopat <$) . rangedThing) <$> eta0
 
 
 addCompositionForRecord
   :: QName       -- ^ Datatype name.
+  -> EtaEquality
   -> ConHead
   -> Telescope   -- ^ @Γ@ parameters.
   -> [Arg QName] -- ^ Projection names.
   -> Telescope   -- ^ @Γ ⊢ Φ@ field types.
   -> Type        -- ^ @Γ ⊢ T@ target type.
   -> TCM ()
-addCompositionForRecord name con tel fs ftel rect = do
+addCompositionForRecord name eta con tel fs ftel rect = do
   cxt <- getContextTelescope
   inTopContext $ do
 
     -- Record has no fields: attach composition data to record constructor
     if null fs then do
-      kit <- defineCompData name con (abstract cxt tel) [] ftel rect []
+      kit <- defineCompData name con (abstract cxt tel) [] ftel rect empty
       modifySignature $ updateDefinition (conName con) $ updateTheDef $ \case
         r@Constructor{} -> r { conComp = kit, conProj = Just [] }  -- no projections
         _ -> __IMPOSSIBLE__
 
+    -- No-eta record with pattern matching (i.e., withOUT copattern
+    -- matching): define composition as for a data type, attach it to
+    -- the record.
+    else if theEtaEquality eta == NoEta PatternMatching then do
+      kit <- defineCompData name con (abstract cxt tel) (unArg <$> fs) ftel rect empty
+      modifySignature $ updateDefinition name $ updateTheDef $ \case
+        r@Record{} -> r { recComp = kit }
+        _ -> __IMPOSSIBLE__
+
     -- Record has fields: attach composition data to record type
     else do
       -- If record has irrelevant fields but irrelevant projections are disabled,
@@ -421,14 +430,14 @@ defineCompKitR ::
   -> TCM CompKit
 defineCompKitR name params fsT fns rect = do
   required <- mapM getTerm'
-        [ builtinInterval
-        , builtinIZero
-        , builtinIOne
-        , builtinIMin
-        , builtinIMax
-        , builtinINeg
-        , builtinPOr
-        , builtinItIsOne
+        [ someBuiltin builtinInterval
+        , someBuiltin builtinIZero
+        , someBuiltin builtinIOne
+        , someBuiltin builtinIMin
+        , someBuiltin builtinIMax
+        , someBuiltin builtinINeg
+        , someBuiltin builtinPOr
+        , someBuiltin builtinItIsOne
         ]
   reportSDoc "tc.rec.cxt" 30 $ prettyTCM params
   reportSDoc "tc.rec.cxt" 30 $ prettyTCM fsT
@@ -477,7 +486,7 @@ defineKanOperationR cmd name params fsT fns rect = do
                   -- Γ = Δ, CompRArgs
                   -- pats = ... | phi = i1
                   -- body = u i1 itIsOne
-                  DoHComp  -> (2,Var 1 [] `apply` [argN io, setRelevance Irrelevant $ argN one])
+                  DoHComp  -> (2,Var 1 [] `apply` [argN io, setRelevance irrelevant $ argN one])
 
               p = ConP (ConHead io_name IsData Inductive [])
                        (noConPatternInfo { conPType = Just (Arg defaultArgInfo tInterval)
@@ -506,8 +515,7 @@ defineKanOperationR cmd name params fsT fns rect = do
                          , namedClausePats = pats
                          , clauseBody      = Just $ rhs
                          , clauseType      = Just $ argN t
-                         , clauseCatchall    = False
-                         , clauseExact       = Just True
+                         , clauseCatchall    = empty
                          , clauseRecursive   = Just False  -- definitely non-recursive!
                          , clauseUnreachable = Just False
                          , clauseEllipsis    = NoEllipsis
@@ -524,8 +532,7 @@ defineKanOperationR cmd name params fsT fns rect = do
                          , namedClausePats = pats
                          , clauseBody      = Just body
                          , clauseType      = Just $ argN (unDom clause_ty)
-                         , clauseCatchall    = False
-                         , clauseExact       = Just True
+                         , clauseCatchall    = empty
                          , clauseRecursive   = Nothing
                              -- Andreas 2020-02-06 TODO
                              -- Or: Just False;  is it known to be non-recursive?
@@ -555,7 +562,7 @@ defineKanOperationR cmd name params fsT fns rect = do
 
     [@con@  ]  name of the record constructor
 
-    [@tel@  ]  parameters and record variable r ("self")
+    [@tel@  ]  parameters (perhaps erased) and record variable r ("self")
 
     [@ftel@ ]  telescope of fields
 
@@ -608,17 +615,26 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
           , "coh   =" <+> (text . show) (getCohesion ai)
           ]
 
-      -- Cohesion check:
-      -- For a field `@c π : A` we would create a projection `π : .., (@(c^-1) r : R as) -> A`
-      -- So we want to check that `@.., (c^-1 . c) x : A |- x : A` is allowed by the modalities.
-      --
-      -- Alternatively we could create a projection `.. |- π r :c A`
-      -- but that would require support for a `t :c A` judgment.
-      if hasLeftAdjoint (UnderComposition (getCohesion ai))
-        then unless (getCohesion ai == Continuous)
-                    -- Andrea TODO: properly update the context/type of the projection when we add Sharp
-                    __IMPOSSIBLE__
-        else genericError $ "Cannot have record fields with modality " ++ show (getCohesion ai)
+      unless (getCohesion ai == Continuous) __IMPOSSIBLE__
+      -- Andreas, 2025-05-03, moved check to ConcreteToAbstract.checkFieldArgInfo.
+      -- -- Cohesion check:
+      -- -- For a field `@c π : A` we would create a projection `π : .., (@(c^-1) r : R as) -> A`
+      -- -- So we want to check that `@.., (c^-1 . c) x : A |- x : A` is allowed by the modalities.
+      -- --
+      -- -- Alternatively we could create a projection `.. |- π r :c A`
+      -- -- but that would require support for a `t :c A` judgment.
+      -- if hasLeftAdjoint (UnderComposition (getCohesion ai))
+      --   then unless (getCohesion ai == Continuous)
+      --               -- Atm, only Continuous has a left adjoint.
+      --               -- Andrea TODO: properly update the context/type of the projection when we add Sharp
+      --               __IMPOSSIBLE__
+      --   else typeError $ InvalidFieldModality (getCohesion ai)
+
+      -- For now, we forbid any polarity annotations on record fields (we would need to do as above,
+      -- and eta-equality or projection existence would be in danger).
+      unless (getModalPolarity ai `samePolarity` (withStandardLock MixedPolarity)) $
+        -- Andreas, 2025-05-03, already checked in ConcreteToAbstract.checkFieldArgInfo
+        __IMPOSSIBLE__
 
       -- The telescope tel includes the variable of record type as last one
       -- e.g. for cartesion product it is
@@ -677,10 +693,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
         -- 2012-04-02: DontCare instead of irrAxiom
 
         -- compute body modification for irrelevant projections
-        let bodyMod = case rel of
-              Relevant   -> id
-              NonStrict  -> id
-              Irrelevant -> dontCare
+        let bodyMod = applyWhen (isIrrelevant rel) dontCare
 
         let -- Andreas, 2010-09-09: comment for existing code
             -- split the telescope into parameters (ptel) and the type or the record
@@ -704,8 +717,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
                             , namedClausePats = [conp]
                             , clauseBody      = body
                             , clauseType      = cltype
-                            , clauseCatchall  = False
-                            , clauseExact       = Just True
+                            , clauseCatchall  = empty
                             , clauseRecursive   = Just False
                             , clauseUnreachable = Just False
                             , clauseEllipsis    = NoEllipsis
@@ -745,9 +757,19 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
 
         escapeContext impossible (size tel) $ do
           lang <- getLanguage
+          fun  <- emptyFunctionData
+          let -- It should be fine to mark a field with @ω in an
+              -- erased record type: the field will be non-erased, but
+              -- the projection will be erased. The following code
+              -- ensures that the use of addConstant does not trigger
+              -- a PlentyInHardCompileTimeMode warning.
+              ai' = flip mapQuantity ai $ \case
+                      Quantityω _ -> Quantityω QωInferred
+                      q           -> q
           addConstant projname $
-            (defaultDefn ai projname (killRange finalt) lang $ FunctionDefn
-              emptyFunctionData
+            (defaultDefn ai' projname (killRange finalt) lang $ FunctionDefn $
+             set funProj_ True $
+              fun
                 { _funClauses        = [clause]
                 , _funCompiled       = Just cc
                 , _funSplitTree      = mst
@@ -760,8 +782,11 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
               }
           computePolarity [projname]
 
-        case Info.defInstance info of
-          -- fields do not have an @instance@ keyword!?
+        addContext ftel1 case Info.defInstance info of
+          -- Instance projections have to be added with their types "qua
+          -- local variable" (i.e. the type you'd get were you to open
+          -- the record module), but this type has to be treated in the
+          -- context ftel1 otherwise it's nonsense
           InstanceDef _r -> addTypedInstance projname t
           NotInstanceDef -> pure ()
 
