@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 -- Andreas, Makoto, Francesco 2014-10-15 AIM XX:
 -- -O2 does not have any noticable effect on runtime
@@ -41,7 +42,7 @@ import Control.Monad.ST.Trans
 import Data.Array.IArray
 import Data.Array.IO
 import Data.Word
-import Data.Int (Int32)
+import Data.Word (Word32)
 import Data.ByteString.Lazy    ( ByteString )
 import Data.ByteString.Builder ( byteString, toLazyByteString )
 import qualified Data.ByteString.Lazy as L
@@ -51,16 +52,11 @@ import qualified Data.Binary.Get as B
 import qualified Data.Binary.Put as B
 import qualified Data.List as List
 import Data.Function (on)
-#if !(MIN_VERSION_base(4,11,0))
-import Data.Semigroup((<>))
-#endif
 
 import qualified Codec.Compression.GZip as G
 import qualified Codec.Compression.Zlib.Internal as Z
 
-#if __GLASGOW_HASKELL__ >= 804
 import GHC.Compact as C
-#endif
 
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
@@ -82,7 +78,7 @@ import Agda.Utils.Impossible
 -- 32-bit machines). Word64 does not have these problems.
 
 currentInterfaceVersion :: Word64
-currentInterfaceVersion = 20221119 * 10 + 0
+currentInterfaceVersion = 20250703 * 10 + 0
 
 -- | The result of 'encode' and 'encodeInterface'.
 
@@ -100,7 +96,7 @@ data Encoded = Encoded
 encode :: EmbPrj a => a -> TCM Encoded
 encode a = do
     collectStats <- hasProfileOption Profile.Serialize
-    newD@(Dict nD ltD stD bD iD dD _tD
+    newD@(Dict nD ltD stD bD iD dD
       _nameD
       _qnameD
       nC ltC stC bC iC dC tC
@@ -129,7 +125,7 @@ encode a = do
       statistics "A.Name"      nameC
     when collectStats $ do
       stats <- Map.fromListWith __IMPOSSIBLE__ . map (second toInteger) <$> do
-        liftIO $ H.toList stats
+        liftIO $ List.sort <$> H.toList stats
       modifyStatistics $ Map.unionWith (+) stats
     -- Encode hashmaps and root, and compress.
     bits1 <- Bench.billTo [ Bench.Serialization, Bench.BinaryEncode ] $
@@ -148,12 +144,12 @@ encode a = do
     statistics :: String -> IORef FreshAndReuse -> TCM ()
     statistics kind ioref = do
       FreshAndReuse fresh
-#ifdef DEBUG
+#ifdef DEBUG_SERIALISATION
                           reused
 #endif
                                  <- liftIO $ readIORef ioref
       tickN (kind ++ "  (fresh)") $ fromIntegral fresh
-#ifdef DEBUG
+#ifdef DEBUG_SERIALISATION
       tickN (kind ++ " (reused)") $ fromIntegral reused
 #endif
 
@@ -174,13 +170,17 @@ encode a = do
 --   where
 --   l h = List.map fst . List.sortBy (compare `on` snd) <$> H.toList h
 
-newtype ListLike a = ListLike { unListLike :: Array Int32 a }
+newtype ListLike a = ListLike { unListLike :: Array Word32 a }
 
 instance B.Binary a => B.Binary (ListLike a) where
   put = __IMPOSSIBLE__ -- Will never serialise this
   get = fmap ListLike $ runSTArray $ do
     n <- lift (B.get :: B.Get Int)
-    arr <- newArray_ (0, fromIntegral n - 1) :: STT s B.Get (STArray s Int32 a)
+    -- Andreas, 2024-10-15: If n is zero, create an empty array.
+    -- Since our indices are Word32, we need to represent it as [1..0] instead of the usual [0..-1].
+    if n <= 0 then (newArray_ (1,0) :: STT s B.Get (STArray s Word32 a)) else do
+
+    arr <- newArray_ (0, fromIntegral n - 1) :: STT s B.Get (STArray s Word32 a)
 
     -- We'd like to use 'for_ [0..n-1]' here, but unfortunately GHC doesn't unfold
     -- the list construction and so performs worse than the hand-written version.
@@ -207,46 +207,31 @@ decode s = do
   -- The decoder is (intended to be) strict enough to ensure that all
   -- such errors can be caught by the handler here.
 
-  (mf, r) <- liftIO $ E.handle (\(E.ErrorCall s) -> noResult s) $ do
-
+  res <- liftIO $ E.handle (\(E.ErrorCall s) -> pure $ Left s) $ do
     ((r, nL, ltL, stL, bL, iL, dL), s, _) <- return $ runGetState B.get s 0
-    if not (null s)
-     then noResult "Garbage at end."
-     else do
+    let ar = unListLike
+    when (not (null s)) $ E.throwIO $ E.ErrorCall "Garbage at end."
+    let nL' = ar nL
+    st <- St nL' (ar ltL) (ar stL) (ar bL) (ar iL) (ar dL)
+            <$> liftIO (newArray (bounds nL') MEEmpty)
+            <*> return mf <*> return incs
+    (r, st) <- runStateT (value r) st
+    let !mf = modFile st
+    return $ Right (mf, r)
 
-      let nL' = ar nL
-      st <- St nL' (ar ltL) (ar stL) (ar bL) (ar iL) (ar dL)
-              <$> liftIO (newArray (bounds nL') mempty)
-              <*> return mf <*> return incs
-      (r, st) <- runStateT (runExceptT (value r)) st
-      return (Just $ modFile st, r)
+  case res of
+    Left s -> do
+      reportSLn "import.iface" 5 $ "Error when decoding interface file: " ++ s
+      pure Nothing
 
-  forM_ mf (setTCLens stModuleToSource)
-
-  case r of
-    Right x -> do
-#if __GLASGOW_HASKELL__ >= 804
+    Right (mf, x) -> do
+      setTCLens stModuleToSource mf
       -- "Compact" the interfaces (without breaking sharing) to
       -- reduce the amount of memory that is traversed by the
       -- garbage collector.
       Bench.billTo [Bench.Deserialization, Bench.Compaction] $
         liftIO (Just . C.getCompact <$> C.compactWithSharing x)
-#else
-      return (Just x)
-#endif
-    Left err -> do
-      reportSLn "import.iface" 5 $ "Error when decoding interface file"
-      -- Andreas, 2014-06-11 deactivated debug printing
-      -- in order to get rid of dependency of Serialize on TCM.Pretty
-      -- reportSDoc "import.iface" 5 $
-      --   "Error when decoding interface file:"
-      --   $+$ nest 2 (prettyTCM err)
-      return Nothing
 
-  where
-  ar = unListLike
-
-  noResult s = return (Nothing, Left $ GenericError s)
 
 encodeInterface :: Interface -> TCM Encoded
 encodeInterface i = do
