@@ -1,13 +1,17 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.TypeChecking.Serialise.Instances.Abstract where
 
+import Control.Monad
+import Data.Void (Void)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Abstract.Pattern ( noDotOrEqPattern )
 import Agda.Syntax.Info
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.Fixity
@@ -17,6 +21,7 @@ import Agda.TypeChecking.Serialise.Instances.Common () --instance only
 
 import Agda.Utils.Functor
 import Agda.Utils.Lens
+import Agda.Utils.Null
 import Agda.Utils.Impossible
 
 -- Don't serialize the tactic.
@@ -30,35 +35,35 @@ instance EmbPrj Scope where
   value = valueN Scope
 
 instance EmbPrj DataOrRecordModule where
-  icod_ IsDataModule   = icodeN' IsDataModule
-  icod_ IsRecordModule = icodeN 0 IsRecordModule
+  icod_ IsDataModule   = pure 0
+  icod_ IsRecordModule = pure 1
 
-  value = vcase $ \case
-    []  -> valuN IsDataModule
-    [0] -> valuN IsRecordModule
-    _   -> malformed
+  value = \case
+    0 -> pure IsDataModule
+    1 -> pure IsRecordModule
+    _ -> malformed
 
 instance EmbPrj NameSpaceId where
-  icod_ PublicNS        = icodeN' PublicNS
-  icod_ PrivateNS       = icodeN 1 PrivateNS
-  icod_ ImportedNS      = icodeN 2 ImportedNS
+  icod_ PublicNS        = pure 0
+  icod_ PrivateNS       = pure 1
+  icod_ ImportedNS      = pure 2
 
-  value = vcase valu where
-    valu []  = valuN PublicNS
-    valu [1] = valuN PrivateNS
-    valu [2] = valuN ImportedNS
-    valu _   = malformed
+  value = \case
+    0 -> pure PublicNS
+    1 -> pure PrivateNS
+    2 -> pure ImportedNS
+    _ -> malformed
 
 instance EmbPrj Access where
-  icod_ (PrivateAccess UserWritten) = icodeN 0 ()
-  icod_ PrivateAccess{}             = icodeN 1 ()
-  icod_ PublicAccess                = icodeN' PublicAccess
+  icod_ (PrivateAccess _ UserWritten) = pure 0
+  icod_ PrivateAccess{}               = pure 1
+  icod_ PublicAccess                  = pure 2
 
-  value = vcase valu where
-    valu [0] = valuN $ PrivateAccess UserWritten
-    valu [1] = valuN $ PrivateAccess Inserted
-    valu []  = valuN PublicAccess
-    valu _   = malformed
+  value = \case
+    0 -> pure $ PrivateAccess empty UserWritten
+    1 -> pure $ privateAccessInserted
+    2 -> pure PublicAccess
+    _ -> malformed
 
 instance EmbPrj NameSpace where
   icod_ (NameSpace a b c) = icodeN' NameSpace a b c
@@ -143,17 +148,20 @@ instance EmbPrj KindOfName where
   --   valu _   = malformed
 
 instance EmbPrj BindingSource where
-  icod_ LambdaBound   = icodeN' LambdaBound
-  icod_ PatternBound  = icodeN 1 PatternBound
-  icod_ LetBound      = icodeN 2 LetBound
-  icod_ WithBound     = icodeN 3 WithBound
+  icod_ = \case
+    LambdaBound     -> pure 0
+    PatternBound _  -> pure 1
+    LetBound        -> pure 2
+    WithBound       -> pure 3
+    MacroBound      -> pure 4
 
-  value = vcase valu where
-    valu []  = valuN LambdaBound
-    valu [1] = valuN PatternBound
-    valu [2] = valuN LetBound
-    valu [3] = valuN WithBound
-    valu _   = malformed
+  value = \case
+    0 -> pure LambdaBound
+    1 -> pure $ PatternBound empty
+    2 -> pure LetBound
+    3 -> pure WithBound
+    4 -> pure MacroBound
+    _ -> malformed
 
 instance EmbPrj LocalVar where
   icod_ (LocalVar a b c)  = icodeN' LocalVar a b c
@@ -174,15 +182,14 @@ instance EmbPrj a => EmbPrj (A.Pattern' a) where
   icod_ (A.DefP p a b)        = icodeN 2 (A.DefP p) a b
   icod_ t@(A.WildP p)         = icodeN 3 t
   icod_ (A.AsP p a b)         = icodeN 4 (A.AsP p) a b
-  icod_ (A.DotP p a)          = icodeN 5 (A.DotP p) a
+  icod_ (A.DotP r a)          = icodeN 5 (A.DotP r) a
   icod_ t@(A.AbsurdP _)       = icodeN 6 t
   icod_ (A.LitP i a)          = icodeN 7 (A.LitP i) a
   icod_ (A.ProjP p a b)       = icodeN 8 (A.ProjP p) a b
   icod_ (A.PatternSynP p a b) = icodeN 9 (A.PatternSynP p) a b
-  icod_ (A.RecP p a)          = icodeN 10 (A.RecP p) a
+  icod_ (A.RecP r a b)        = icodeN 10 (A.RecP r) a b
   icod_ (A.EqualP _ a)        = __IMPOSSIBLE__
   icod_ (A.WithP i a)         = icodeN 11 (A.WithP i) a
-  icod_ (A.AnnP i a p)        = icodeN 12 (A.AnnP i) a p
 
   value = vcase valu where
     valu [0, a]       = valuN A.VarP a
@@ -195,13 +202,20 @@ instance EmbPrj a => EmbPrj (A.Pattern' a) where
     valu [7, a]       = valuN (A.LitP i) a
     valu [8, a, b]    = valuN (A.ProjP i) a b
     valu [9, a, b]    = valuN (A.PatternSynP i) a b
-    valu [10, a]      = valuN (A.RecP i) a
+    valu [10, a, b]   = valuN (A.RecP empty) a b
     valu [11, a]      = valuN (A.WithP i) a
-    valu [12, a, b]   = valuN (A.AnnP i) a b
     valu _            = malformed
 
     i = patNoRange
 
+-- | Hackish serialization for patterns that deletes dot patterns.
+--   So that we can serialize the 'WithClauseProjectionFixityMismatch'
+--   without having to define serialization of expressions.
+--
+instance {-# OVERLAPS #-} EmbPrj A.Pattern where
+  icod_ = icod_ <=< noDotOrEqPattern (return $ A.WildP empty)
+  value = fmap (__IMPOSSIBLE__ :: Void -> A.Expr) <.> value
+
 instance EmbPrj ParenPreference where
   icod_ PreferParen     = icodeN' PreferParen
   icod_ PreferParenless = icodeN 1 PreferParenless
@@ -236,8 +250,8 @@ instance EmbPrj Precedence where
     valu _         = malformed
 
 instance EmbPrj ScopeInfo where
-  icod_ (ScopeInfo a b c d e f g h i j) = icodeN' (\ a b c d e -> ScopeInfo a b c d e f g h i j) a b c d e
+  icod_ (ScopeInfo a b c d e f g h i j k) = icodeN' (\ a b c d e -> ScopeInfo a b c d e f g h i j k) a b c d e
 
-  value = valueN (\ a b c d e -> ScopeInfo a b c d e Map.empty Map.empty Set.empty Map.empty Map.empty)
+  value = valueN (\ a b c d e -> ScopeInfo a b c d e Map.empty Map.empty Set.empty Map.empty Map.empty Map.empty)
 
 instance EmbPrj NameOrModule
