@@ -5,24 +5,37 @@ module Agda.TypeChecking.Serialise.Instances.Errors where
 import Control.Monad
 
 import Agda.TypeChecking.Serialise.Base
+import Agda.TypeChecking.Serialise.Instances.Common   ( SerialisedRange(..) )
 import Agda.TypeChecking.Serialise.Instances.Internal () --instance only
 import Agda.TypeChecking.Serialise.Instances.Abstract () --instance only
 
-import Agda.Syntax.Concrete.Definitions (DeclarationWarning(..), DeclarationWarning'(..))
+import Agda.Syntax.Concrete.Definitions.Errors
+    ( DeclarationWarning(..), DeclarationWarning'(..), OpenOrImport(..) )
 import Agda.Syntax.Parser.Monad
 import Agda.TypeChecking.Monad.Base
+import qualified Agda.TypeChecking.Monad.Base.Warning as W
 import Agda.Interaction.Options
 import Agda.Interaction.Options.Warnings
 import Agda.Interaction.Library.Base
 import Agda.Termination.CutOff
-import Agda.Utils.Pretty
+import Agda.Syntax.Common.Pretty
 import Agda.Utils.ProfileOptions
 
 import Agda.Utils.Impossible
 
+instance EmbPrj IsAmbiguous where
+  icod_ = \case
+    YesAmbiguous a -> icodeN' YesAmbiguous a
+    NotAmbiguous   -> icodeN' NotAmbiguous
+
+  value = vcase \case
+    [a] -> valuN YesAmbiguous a
+    []  -> valuN NotAmbiguous
+    _   -> malformed
+
 instance EmbPrj TCWarning where
-  icod_ (TCWarning fp a b c d) = icodeN' TCWarning fp a b c d
-  value = valueN TCWarning
+  icod_ (TCWarning fp r a b c d) = icodeN' (\ fp -> TCWarning fp . underlyingRange) fp (SerialisedRange r) a b c d
+  value = valueN (\ fp -> TCWarning fp . underlyingRange)
 
 -- We don't need to serialise warnings that turn into errors
 instance EmbPrj Warning where
@@ -31,31 +44,25 @@ instance EmbPrj Warning where
     UnreachableClauses a b                -> icodeN 0 UnreachableClauses a b
     CoverageIssue a b                     -> __IMPOSSIBLE__
     NotStrictlyPositive a b               -> __IMPOSSIBLE__
+    ConstructorDoesNotFitInData{}         -> __IMPOSSIBLE__
     UnsolvedMetaVariables a               -> __IMPOSSIBLE__
     UnsolvedInteractionMetas a            -> __IMPOSSIBLE__
     UnsolvedConstraints a                 -> __IMPOSSIBLE__
+    InteractionMetaBoundaries a           -> __IMPOSSIBLE__
     OldBuiltin a b                        -> icodeN 1 OldBuiltin a b
     EmptyRewritePragma                    -> icodeN 2 EmptyRewritePragma
-    UselessPublic                         -> icodeN 3 UselessPublic
+    UselessPublic a                       -> icodeN 3 UselessPublic a
     UselessInline a                       -> icodeN 4 UselessInline a
-    GenericWarning a                      -> icodeN 5 GenericWarning a
-    GenericNonFatalError a                -> __IMPOSSIBLE__
+    InvalidCharacterLiteral a             -> __IMPOSSIBLE__
     SafeFlagPostulate a                   -> __IMPOSSIBLE__
     SafeFlagPragma a                      -> __IMPOSSIBLE__
-    SafeFlagNonTerminating                -> __IMPOSSIBLE__
-    SafeFlagTerminating                   -> __IMPOSSIBLE__
     SafeFlagWithoutKFlagPrimEraseEquality -> __IMPOSSIBLE__
-    SafeFlagNoPositivityCheck             -> __IMPOSSIBLE__
-    SafeFlagPolarity                      -> __IMPOSSIBLE__
-    SafeFlagNoUniverseCheck               -> __IMPOSSIBLE__
-    SafeFlagNoCoverageCheck               -> __IMPOSSIBLE__
-    SafeFlagInjective                     -> __IMPOSSIBLE__
-    SafeFlagEta                           -> __IMPOSSIBLE__
+    DuplicateRecordDirective a            -> icodeN 5 DuplicateRecordDirective a
     DeprecationWarning a b c              -> icodeN 6 DeprecationWarning a b c
     NicifierIssue a                       -> icodeN 7 NicifierIssue a
     InversionDepthReached a               -> icodeN 8 InversionDepthReached a
     UserWarning a                         -> icodeN 9 UserWarning a
-    AbsurdPatternRequiresNoRHS a          -> icodeN 10 AbsurdPatternRequiresNoRHS a
+    AbsurdPatternRequiresAbsentRHS        -> icodeN 10 AbsurdPatternRequiresAbsentRHS
     ModuleDoesntExport a b c d            -> icodeN 11 ModuleDoesntExport a b c d
     LibraryWarning a                      -> icodeN 12 LibraryWarning a
     CoverageNoExactSplit a b              -> icodeN 13 CoverageNoExactSplit a b
@@ -72,7 +79,7 @@ instance EmbPrj Warning where
     RewriteMaybeNonConfluent a b c        -> icodeN 24 RewriteMaybeNonConfluent a b c
     PragmaCompileErased a b               -> icodeN 25 PragmaCompileErased a b
     FixityInRenamingModule a              -> icodeN 26 FixityInRenamingModule a
-    NotInScopeW ns                        -> icodeN 27 NotInScopeW ns
+    NotInScopeW a                         -> icodeN 27 NotInScopeW a
     ClashesViaRenaming a b                -> icodeN 28 ClashesViaRenaming a b
     RecordFieldWarning a                  -> icodeN 29 RecordFieldWarning a
     UselessPatternDeclarationForRecord a  -> icodeN 30 UselessPatternDeclarationForRecord a
@@ -80,26 +87,72 @@ instance EmbPrj Warning where
     AsPatternShadowsConstructorOrPatternSynonym a -> icodeN 32 AsPatternShadowsConstructorOrPatternSynonym a
     DuplicateUsing a                      -> icodeN 33 DuplicateUsing a
     UselessHiding a                       -> icodeN 34 UselessHiding a
-    GenericUseless a b                    -> icodeN 35 GenericUseless a b
+    UselessPragma a b                     -> icodeN 35 UselessPragma a b
     RewriteAmbiguousRules a b c           -> icodeN 36 RewriteAmbiguousRules a b c
     RewriteMissingRule a b c              -> icodeN 37 RewriteMissingRule a b c
     ParseWarning a                        -> icodeN 38 ParseWarning a
-    NoGuardednessFlag a                   -> icodeN 39 NoGuardednessFlag a
+    UselessTactic                         -> icodeN 39 UselessTactic
     UnsupportedIndexedMatch f             -> icodeN 40 UnsupportedIndexedMatch f
     OptionWarning a                       -> icodeN 41 OptionWarning a
+    PlentyInHardCompileTimeMode a         -> icodeN 42 PlentyInHardCompileTimeMode a
+    NotAffectedByOpaque                   -> icodeN 43 NotAffectedByOpaque
+    UnfoldTransparentName nm              -> icodeN 44 UnfoldTransparentName nm
+    UselessOpaque                         -> icodeN 45 UselessOpaque
+    InlineNoExactSplit a b                -> icodeN 46 InlineNoExactSplit a b
+    FaceConstraintCannotBeHidden a        -> icodeN 47 FaceConstraintCannotBeHidden a
+    FaceConstraintCannotBeNamed a         -> icodeN 48 FaceConstraintCannotBeNamed a
+    PatternShadowsConstructor a b         -> icodeN 49 PatternShadowsConstructor a b
+    ConfluenceCheckingIncompleteBecauseOfMeta a -> icodeN 50 ConfluenceCheckingIncompleteBecauseOfMeta a
+    BuiltinDeclaresIdentifier a                 -> icodeN 51 BuiltinDeclaresIdentifier a
+    ConfluenceForCubicalNotSupported            -> icodeN 52 ConfluenceForCubicalNotSupported
+    -- We do not need to serialize compiler warnings:
+    PragmaCompileList                           -> __IMPOSSIBLE__
+    PragmaCompileMaybe                          -> __IMPOSSIBLE__
+    NoMain _                                    -> __IMPOSSIBLE__
+    IllegalRewriteRule a b                      -> icodeN 53 IllegalRewriteRule a b
+    MissingTypeSignatureForOpaque a b           -> icodeN 54 MissingTypeSignatureForOpaque a b
+    ConflictingPragmaOptions a b                -> icodeN 55 ConflictingPragmaOptions a b
+    CustomBackendWarning a b                    -> icodeN 56 CustomBackendWarning a b
+    CoinductiveEtaRecord a                      -> icodeN 57 CoinductiveEtaRecord a
+    WithClauseProjectionFixityMismatch a b c d  -> icodeN 58 WithClauseProjectionFixityMismatch a b c d
+    InvalidDisplayForm a b                      -> icodeN 59 InvalidDisplayForm a b
+    TooManyArgumentsToSort a b                  -> __IMPOSSIBLE__
+    NotARewriteRule a b                         -> icodeN 60 NotARewriteRule a b
+    PragmaCompileWrongName a b                  -> icodeN 61 PragmaCompileWrongName a b
+    PragmaCompileUnparsable a                   -> icodeN 62 PragmaCompileUnparsable a
+    PragmaCompileWrong a b                      -> icodeN 63 PragmaCompileWrong a b
+    PragmaExpectsUnambiguousConstructorOrFunction a b c ->
+      icodeN 64 PragmaExpectsUnambiguousConstructorOrFunction a b c
+    PragmaExpectsUnambiguousProjectionOrFunction a b c ->
+      icodeN 65 PragmaExpectsUnambiguousProjectionOrFunction a b c
+    PragmaExpectsDefinedSymbol a b              -> icodeN 66 PragmaExpectsDefinedSymbol a b
+    UnfoldingWrongName a                        -> icodeN 67 UnfoldingWrongName a
+    -- TODO: linearity
+    -- FixingQuantity a b c                        -> icodeN 68 FixingQuantity a b c
+    FixingRelevance a b c                       -> icodeN 69 FixingRelevance a b c
+    UnusedVariablesInDisplayForm a              -> icodeN 70 UnusedVariablesInDisplayForm a
+    HiddenNotInArgumentPosition a               -> __IMPOSSIBLE__
+    InstanceNotInArgumentPosition a             -> __IMPOSSIBLE__
+    MacroInLetBindings                          -> __IMPOSSIBLE__
+    AbstractInLetBindings                       -> __IMPOSSIBLE__
+    TopLevelPolarity a b                        -> __IMPOSSIBLE__
+    TooManyPolarities a b                       -> icodeN 71 TooManyPolarities a b
+    FixingCohesion a b c                        -> icodeN 72 FixingCohesion a b c
+    FixingPolarity a b c                        -> icodeN 73 FixingPolarity a b c
+    RewritesNothing                             -> icodeN 74 RewritesNothing
 
   value = vcase $ \ case
     [0, a, b]            -> valuN UnreachableClauses a b
     [1, a, b]            -> valuN OldBuiltin a b
     [2]                  -> valuN EmptyRewritePragma
-    [3]                  -> valuN UselessPublic
+    [3, a]               -> valuN UselessPublic a
     [4, a]               -> valuN UselessInline a
-    [5, a]               -> valuN GenericWarning a
+    [5, a]               -> valuN DuplicateRecordDirective a
     [6, a, b, c]         -> valuN DeprecationWarning a b c
     [7, a]               -> valuN NicifierIssue a
     [8, a]               -> valuN InversionDepthReached a
     [9, a]               -> valuN UserWarning a
-    [10, a]              -> valuN AbsurdPatternRequiresNoRHS a
+    [10]                 -> valuN AbsurdPatternRequiresAbsentRHS
     [11, a, b, c, d]     -> valuN ModuleDoesntExport a b c d
     [12, a]              -> valuN LibraryWarning a
     [13, a, b]           -> valuN CoverageNoExactSplit a b
@@ -116,7 +169,7 @@ instance EmbPrj Warning where
     [24, a, b, c]        -> valuN RewriteMaybeNonConfluent a b c
     [25, a, b]           -> valuN PragmaCompileErased a b
     [26, a]              -> valuN FixityInRenamingModule a
-    [27, ns]             -> valuN NotInScopeW ns
+    [27, a]              -> valuN NotInScopeW a
     [28, a, b]           -> valuN ClashesViaRenaming a b
     [29, a]              -> valuN RecordFieldWarning a
     [30, a]              -> valuN UselessPatternDeclarationForRecord a
@@ -124,21 +177,107 @@ instance EmbPrj Warning where
     [32, a]              -> valuN AsPatternShadowsConstructorOrPatternSynonym a
     [33, a]              -> valuN DuplicateUsing a
     [34, a]              -> valuN UselessHiding a
-    [35, a, b]           -> valuN GenericUseless a b
+    [35, a, b]           -> valuN UselessPragma a b
     [36, a, b, c]        -> valuN RewriteAmbiguousRules a b c
     [37, a, b, c]        -> valuN RewriteMissingRule a b c
     [38, a]              -> valuN ParseWarning a
-    [39, a]              -> valuN NoGuardednessFlag a
+    [39]                 -> valuN UselessTactic
     [40, a]              -> valuN UnsupportedIndexedMatch a
     [41, a]              -> valuN OptionWarning a
+    [42, a]              -> valuN PlentyInHardCompileTimeMode a
+    [43]                 -> valuN NotAffectedByOpaque
+    [44, a]              -> valuN UnfoldTransparentName a
+    [45]                 -> valuN UselessOpaque
+    [46, a, b]           -> valuN InlineNoExactSplit a b
+    [47, a]              -> valuN FaceConstraintCannotBeHidden a
+    [48, a]              -> valuN FaceConstraintCannotBeNamed a
+    [49, a, b]           -> valuN PatternShadowsConstructor a b
+    [50, a]              -> valuN ConfluenceCheckingIncompleteBecauseOfMeta a
+    [51, a]              -> valuN BuiltinDeclaresIdentifier a
+    [52]                 -> valuN ConfluenceForCubicalNotSupported
+    [53, a, b]           -> valuN IllegalRewriteRule a b
+    [54, a, b]           -> valuN MissingTypeSignatureForOpaque a b
+    [55, a, b]           -> valuN ConflictingPragmaOptions a b
+    [56, a, b]           -> valuN CustomBackendWarning a b
+    [57, a]              -> valuN CoinductiveEtaRecord a
+    [58, a, b, c, d]     -> valuN WithClauseProjectionFixityMismatch a b c d
+    [59, a, b]           -> valuN InvalidDisplayForm a b
+    [60, a, b]           -> valuN NotARewriteRule a b
+    [61, a, b]           -> valuN PragmaCompileWrongName a b
+    [62, a]              -> valuN PragmaCompileUnparsable a
+    [63, a, b]           -> valuN PragmaCompileWrong a b
+    [64, a, b, c]        -> valuN PragmaExpectsUnambiguousConstructorOrFunction a b c
+    [65, a, b, c]        -> valuN PragmaExpectsUnambiguousProjectionOrFunction a b c
+    [66, a, b]           -> valuN PragmaExpectsDefinedSymbol a b
+    [67, a]              -> valuN UnfoldingWrongName a
+    -- TODO: linearity
+    -- [68, a, b, c]        -> valuN FixingQuantity a b c
+    [69, a, b, c]        -> valuN FixingRelevance a b c
+    [70, a]              -> valuN UnusedVariablesInDisplayForm a
+    [71, a, b]           -> valuN TooManyPolarities a b
+    [72, a, b, c]        -> valuN FixingCohesion a b c
+    [73, a, b, c]        -> valuN FixingPolarity a b c
+    [74]                 -> valuN RewritesNothing
+    _ -> malformed
+
+instance EmbPrj UselessPublicReason
+
+instance EmbPrj IllegalRewriteRuleReason where
+  icod_ = \case
+    LHSNotDefinitionOrConstructor               -> icodeN 0 LHSNotDefinitionOrConstructor
+    VariablesNotBoundByLHS a                    -> icodeN 1 VariablesNotBoundByLHS a
+    VariablesBoundMoreThanOnce a                -> icodeN 2 VariablesBoundMoreThanOnce a
+    LHSReduces a b                              -> icodeN 3 LHSReduces a b
+    -- 4 was HeadSymbolIsProjection
+    HeadSymbolIsProjectionLikeFunction a        -> icodeN 5 HeadSymbolIsProjectionLikeFunction a
+    HeadSymbolIsTypeConstructor a               -> icodeN 6 HeadSymbolIsTypeConstructor a
+    HeadSymbolContainsMetas a                   -> icodeN 7 HeadSymbolContainsMetas a
+    ConstructorParametersNotGeneral a b         -> icodeN 8 ConstructorParametersNotGeneral a b
+    ContainsUnsolvedMetaVariables a             -> icodeN 9 ContainsUnsolvedMetaVariables a
+    BlockedOnProblems a                         -> icodeN 10 BlockedOnProblems a
+    RequiresDefinitions a                       -> icodeN 11 RequiresDefinitions a
+    DoesNotTargetRewriteRelation                -> icodeN 12 DoesNotTargetRewriteRelation
+    BeforeFunctionDefinition                    -> icodeN 13 BeforeFunctionDefinition
+    BeforeMutualFunctionDefinition a            -> icodeN 14 BeforeMutualFunctionDefinition a
+    DuplicateRewriteRule                        -> icodeN 15 DuplicateRewriteRule
+
+  value = vcase $ \case
+    [0]       -> valuN LHSNotDefinitionOrConstructor
+    [1, a]    -> valuN VariablesNotBoundByLHS a
+    [2, a]    -> valuN VariablesBoundMoreThanOnce a
+    [3, a, b] -> valuN LHSReduces a b
+    -- 4 was HeadSymbolIsProjection
+    [5, a]    -> valuN HeadSymbolIsProjectionLikeFunction a
+    [6, a]    -> valuN HeadSymbolIsTypeConstructor a
+    [7, a]    -> valuN HeadSymbolContainsMetas a
+    [8, a, b] -> valuN ConstructorParametersNotGeneral a b
+    [9, a]    -> valuN ContainsUnsolvedMetaVariables a
+    [10, a]   -> valuN BlockedOnProblems a
+    [11, a]   -> valuN RequiresDefinitions a
+    [12]      -> valuN DoesNotTargetRewriteRelation
+    [13]      -> valuN BeforeFunctionDefinition
+    [14, a]   -> valuN BeforeMutualFunctionDefinition a
+    [15]      -> valuN DuplicateRewriteRule
     _ -> malformed
 
 instance EmbPrj OptionWarning where
   icod_ = \case
-    OptionRenamed a b -> icodeN' OptionRenamed a b
+    OptionRenamed a b -> icodeN 0 OptionRenamed a b
+    WarningProblem a  -> icodeN 1 WarningProblem a
 
   value = vcase $ \case
-    [a, b] -> valuN OptionRenamed a b
+    [0, a, b] -> valuN OptionRenamed a b
+    [1, a]    -> valuN WarningProblem a
+    _ -> malformed
+
+instance EmbPrj WarningModeError where
+  icod_ = \case
+    Unknown a   -> icodeN 0 Unknown a
+    NoNoError a -> icodeN 1 NoNoError a
+
+  value = vcase $ \case
+    [0, a] -> valuN Unknown a
+    [1, a] -> valuN NoNoError a
     _ -> malformed
 
 instance EmbPrj ParseWarning where
@@ -146,21 +285,27 @@ instance EmbPrj ParseWarning where
     OverlappingTokensWarning a -> icodeN 0 OverlappingTokensWarning a
     UnsupportedAttribute a b   -> icodeN 1 UnsupportedAttribute a b
     MultipleAttributes a b     -> icodeN 2 MultipleAttributes a b
+    UnknownAttribute a b       -> icodeN 3 UnknownAttribute a b
+    UnknownPolarity a b        -> icodeN 4 UnknownPolarity a b
+    MisplacedAttributes a b    -> icodeN 5 MisplacedAttributes a b
 
   value = vcase $ \case
     [0, a]    -> valuN OverlappingTokensWarning a
     [1, a, b] -> valuN UnsupportedAttribute a b
     [2, a, b] -> valuN MultipleAttributes a b
+    [3, a, b] -> valuN UnknownAttribute a b
+    [4, a, b] -> valuN UnknownPolarity a b
+    [5, a, b] -> valuN MisplacedAttributes a b
     _ -> malformed
 
 instance EmbPrj RecordFieldWarning where
   icod_ = \case
-    DuplicateFieldsWarning a   -> icodeN 0 DuplicateFieldsWarning a
-    TooManyFieldsWarning a b c -> icodeN 1 TooManyFieldsWarning a b c
+    W.DuplicateFields a   -> icodeN 0 W.DuplicateFields a
+    W.TooManyFields a b c -> icodeN 1 W.TooManyFields a b c
 
   value = vcase $ \case
-    [0, a]       -> valuN DuplicateFieldsWarning a
-    [1, a, b, c] -> valuN TooManyFieldsWarning a b c
+    [0, a]       -> valuN W.DuplicateFields a
+    [1, a, b, c] -> valuN W.TooManyFields a b c
     _ -> malformed
 
 instance EmbPrj DeclarationWarning where
@@ -197,14 +342,24 @@ instance EmbPrj DeclarationWarning' where
     EmptyField r                      -> icodeN 23 EmptyField r
     ShadowingInTelescope nrs          -> icodeN 24 ShadowingInTelescope nrs
     InvalidCoverageCheckPragma r      -> icodeN 25 InvalidCoverageCheckPragma r
-    OpenPublicAbstract r              -> icodeN 26 OpenPublicAbstract r
-    OpenPublicPrivate r               -> icodeN 27 OpenPublicPrivate r
+    OpenImportAbstract r kwr a        -> icodeN 26 OpenImportAbstract r kwr a
+    OpenImportPrivate r kwr kwr' a    -> icodeN 27 OpenImportPrivate r kwr kwr' a
     EmptyConstructor a                -> icodeN 28 EmptyConstructor a
-    InvalidRecordDirective a          -> icodeN 29 InvalidRecordDirective a
-    InvalidConstructor a              -> icodeN 30 InvalidConstructor a
+    -- 29 removed
+    -- 30 removed
     InvalidConstructorBlock a         -> icodeN 31 InvalidConstructorBlock a
-    MissingDeclarations a             -> icodeN 32 MissingDeclarations a
+    MissingDataDeclaration a          -> icodeN 32 MissingDataDeclaration a
     HiddenGeneralize r                -> icodeN 33 HiddenGeneralize r
+    UselessMacro r                    -> icodeN 34 UselessMacro r
+    SafeFlagEta                    {} -> __IMPOSSIBLE__
+    SafeFlagInjective              {} -> __IMPOSSIBLE__
+    SafeFlagNoCoverageCheck        {} -> __IMPOSSIBLE__
+    SafeFlagNoPositivityCheck      {} -> __IMPOSSIBLE__
+    SafeFlagNoUniverseCheck        {} -> __IMPOSSIBLE__
+    SafeFlagNonTerminating         {} -> __IMPOSSIBLE__
+    SafeFlagPolarity               {} -> __IMPOSSIBLE__
+    SafeFlagTerminating            {} -> __IMPOSSIBLE__
+    EmptyPolarityPragma r             -> icodeN 35 EmptyPolarityPragma r
 
   value = vcase $ \case
     [0, a]   -> valuN UnknownNamesInFixityDecl a
@@ -233,16 +388,20 @@ instance EmbPrj DeclarationWarning' where
     [23,r]   -> valuN EmptyField r
     [24,nrs] -> valuN ShadowingInTelescope nrs
     [25,r]   -> valuN InvalidCoverageCheckPragma r
-    [26,r]   -> valuN OpenPublicAbstract r
-    [27,r]   -> valuN OpenPublicPrivate r
+    [26,r,kwr,a] -> valuN OpenImportAbstract r kwr a
+    [27,r,kwr,kwr',a] -> valuN OpenImportPrivate r kwr kwr' a
     [28,r]   -> valuN EmptyConstructor r
-    [29,r]   -> valuN InvalidRecordDirective r
-    [30,r]   -> valuN InvalidConstructor r
+    -- 29 removed
+    -- 30 removed
     [31,r]   -> valuN InvalidConstructorBlock r
-    [32,r]   -> valuN MissingDeclarations r
+    [32,r]   -> valuN MissingDataDeclaration r
     [33,r]   -> valuN HiddenGeneralize r
+    [34,r]   -> valuN UselessMacro r
+    [35,r]   -> valuN EmptyPolarityPragma r
     _ -> malformed
 
+instance EmbPrj OpenOrImport
+
 instance EmbPrj LibWarning where
   icod_ = \case
     LibWarning a b -> icodeN 0 LibWarning a b
@@ -290,25 +449,18 @@ instance EmbPrj InfectiveCoinfective where
     valu _   = malformed
 
 instance EmbPrj PragmaOptions where
-  icod_ = \case
-    PragmaOptions a b c d e f g h i j k l m n o p q r s t u v w x y z aa bb cc dd ee ff gg hh ii jj kk ll mm nn oo pp qq rr ss tt uu vv ww xx yy zz aaa bbb ccc ddd eee fff ggg hhh ->
-      icodeN' PragmaOptions a b c d e f g h i j k l m n o p q r s t u v w x y z aa bb cc dd ee ff gg hh ii jj kk ll mm nn oo pp qq rr ss tt uu vv ww xx yy zz aaa bbb ccc ddd eee fff ggg hhh
+  icod_    (PragmaOptions a b c d e f g h i j k l m n o p q r s t u v w x y z aa bb cc dd ee ff gg hh ii jj kk ll mm nn oo pp qq rr ss tt uu vv ww xx yy zz aaa bbb ccc ddd eee fff ggg hhh iii jjj kkk lll mmm nnn ooo ppp qqq rrr sss) =
+    icodeN' PragmaOptions a b c d e f g h i j k l m n o p q r s t u v w x y z aa bb cc dd ee ff gg hh ii jj kk ll mm nn oo pp qq rr ss tt uu vv ww xx yy zz aaa bbb ccc ddd eee fff ggg hhh iii jjj kkk lll mmm nnn ooo ppp qqq rrr sss
 
-  value = vcase $ \case
-    [a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z, aa, bb, cc, dd, ee, ff, gg, hh, ii, jj, kk, ll, mm, nn, oo, pp, qq, rr, ss, tt, uu, vv, ww, xx, yy, zz, aaa, bbb, ccc, ddd, eee, fff, ggg, hhh] ->
-      valuN PragmaOptions a b c d e f g h i j k l m n o p q r s t u v w x y z aa bb cc dd ee ff gg hh ii jj kk ll mm nn oo pp qq rr ss tt uu vv ww xx yy zz aaa bbb ccc ddd eee fff ggg hhh
-    _ -> malformed
+  value = valueN PragmaOptions
 
 instance EmbPrj ProfileOptions where
   icod_ opts = icode (profileOptionsToList opts)
   value = fmap profileOptionsFromList . value
 
-instance EmbPrj ProfileOption where
-  icod_ = icode . fromEnum
-  value = value >=> \ n -> if lo <= n && n <= hi then pure (toEnum n) else malformed
-    where
-      lo = fromEnum (minBound :: ProfileOption)
-      hi = fromEnum (maxBound :: ProfileOption)
+instance EmbPrj ProfileOption
+
+instance EmbPrj LHSOrPatSyn
 
 instance EmbPrj UnicodeOrAscii
 
@@ -322,211 +474,13 @@ instance EmbPrj ConfluenceCheck where
     valu _   = malformed
 
 instance EmbPrj WarningMode where
-  icod_ = \case
-    WarningMode a b -> icodeN' WarningMode a b
+  icod_ (WarningMode a b) = icodeN' WarningMode a b
 
-  value = vcase $ \case
-    [a, b]   -> valuN WarningMode a b
-    _ -> malformed
-
-instance EmbPrj WarningName where
-  icod_ = return . \case
-    OverlappingTokensWarning_                    -> 0
-    UnsupportedAttribute_                        -> 1
-    MultipleAttributes_                          -> 2
-    LibUnknownField_                             -> 3
-    EmptyAbstract_                               -> 4
-    EmptyConstructor_                            -> 5
-    EmptyField_                                  -> 6
-    EmptyGeneralize_                             -> 7
-    EmptyInstance_                               -> 8
-    EmptyMacro_                                  -> 9
-    EmptyMutual_                                 -> 10
-    EmptyPostulate_                              -> 11
-    EmptyPrimitive_                              -> 12
-    EmptyPrivate_                                -> 13
-    EmptyRewritePragma_                          -> 14
-    EmptyWhere_                                  -> 15
-    HiddenGeneralize_                            -> 16
-    InvalidCatchallPragma_                       -> 17
-    InvalidConstructor_                          -> 18
-    InvalidConstructorBlock_                     -> 19
-    InvalidCoverageCheckPragma_                  -> 20
-    InvalidNoPositivityCheckPragma_              -> 21
-    InvalidNoUniverseCheckPragma_                -> 22
-    InvalidRecordDirective_                      -> 23
-    InvalidTerminationCheckPragma_               -> 24
-    MissingDeclarations_                         -> 25
-    MissingDefinitions_                          -> 26
-    NotAllowedInMutual_                          -> 27
-    OpenPublicAbstract_                          -> 28
-    OpenPublicPrivate_                           -> 29
-    PolarityPragmasButNotPostulates_             -> 30
-    PragmaCompiled_                              -> 31
-    PragmaNoTerminationCheck_                    -> 32
-    ShadowingInTelescope_                        -> 33
-    UnknownFixityInMixfixDecl_                   -> 34
-    UnknownNamesInFixityDecl_                    -> 35
-    UnknownNamesInPolarityPragmas_               -> 36
-    UselessAbstract_                             -> 37
-    UselessInstance_                             -> 38
-    UselessPrivate_                              -> 39
-    AbsurdPatternRequiresNoRHS_                  -> 40
-    AsPatternShadowsConstructorOrPatternSynonym_ -> 41
-    CantGeneralizeOverSorts_                     -> 42
-    ClashesViaRenaming_                          -> 43
-    CoverageIssue_                               -> 44
-    CoverageNoExactSplit_                        -> 45
-    DeprecationWarning_                          -> 46
-    DuplicateUsing_                              -> 47
-    FixityInRenamingModule_                      -> 48
-    GenericNonFatalError_                        -> 49
-    GenericUseless_                              -> 50
-    GenericWarning_                              -> 51
-    IllformedAsClause_                           -> 52
-    InstanceArgWithExplicitArg_                  -> 53
-    InstanceWithExplicitArg_                     -> 54
-    InstanceNoOutputTypeName_                    -> 55
-    InversionDepthReached_                       -> 56
-    ModuleDoesntExport_                          -> 57
-    NoGuardednessFlag_                           -> 58
-    NotInScope_                                  -> 59
-    NotStrictlyPositive_                         -> 60
-    UnsupportedIndexedMatch_                        -> 61
-    OldBuiltin_                                  -> 62
-    PragmaCompileErased_                         -> 63
-    RewriteMaybeNonConfluent_                    -> 64
-    RewriteNonConfluent_                         -> 65
-    RewriteAmbiguousRules_                       -> 66
-    RewriteMissingRule_                          -> 67
-    SafeFlagEta_                                 -> 68
-    SafeFlagInjective_                           -> 69
-    SafeFlagNoCoverageCheck_                     -> 70
-    SafeFlagNonTerminating_                      -> 71
-    SafeFlagNoPositivityCheck_                   -> 72
-    SafeFlagNoUniverseCheck_                     -> 73
-    SafeFlagPolarity_                            -> 74
-    SafeFlagPostulate_                           -> 75
-    SafeFlagPragma_                              -> 76
-    SafeFlagTerminating_                         -> 77
-    SafeFlagWithoutKFlagPrimEraseEquality_       -> 78
-    TerminationIssue_                            -> 79
-    UnreachableClauses_                          -> 80
-    UnsolvedConstraints_                         -> 81
-    UnsolvedInteractionMetas_                    -> 82
-    UnsolvedMetaVariables_                       -> 83
-    UselessHiding_                               -> 84
-    UselessInline_                               -> 85
-    UselessPatternDeclarationForRecord_          -> 86
-    UselessPublic_                               -> 87
-    UserWarning_                                 -> 88
-    WithoutKFlagPrimEraseEquality_               -> 89
-    WrongInstanceDeclaration_                    -> 90
-    CoInfectiveImport_                           -> 91
-    InfectiveImport_                             -> 92
-    DuplicateFieldsWarning_                      -> 93
-    TooManyFieldsWarning_                        -> 94
-    OptionRenamed_                               -> 95
-
-  value = \case
-    0  -> return OverlappingTokensWarning_
-    1  -> return UnsupportedAttribute_
-    2  -> return MultipleAttributes_
-    3  -> return LibUnknownField_
-    4  -> return EmptyAbstract_
-    5  -> return EmptyConstructor_
-    6  -> return EmptyField_
-    7  -> return EmptyGeneralize_
-    8  -> return EmptyInstance_
-    9  -> return EmptyMacro_
-    10 -> return EmptyMutual_
-    11 -> return EmptyPostulate_
-    12 -> return EmptyPrimitive_
-    13 -> return EmptyPrivate_
-    14 -> return EmptyRewritePragma_
-    15 -> return EmptyWhere_
-    16 -> return HiddenGeneralize_
-    17 -> return InvalidCatchallPragma_
-    18 -> return InvalidConstructor_
-    19 -> return InvalidConstructorBlock_
-    20 -> return InvalidCoverageCheckPragma_
-    21 -> return InvalidNoPositivityCheckPragma_
-    22 -> return InvalidNoUniverseCheckPragma_
-    23 -> return InvalidRecordDirective_
-    24 -> return InvalidTerminationCheckPragma_
-    25 -> return MissingDeclarations_
-    26 -> return MissingDefinitions_
-    27 -> return NotAllowedInMutual_
-    28 -> return OpenPublicAbstract_
-    29 -> return OpenPublicPrivate_
-    30 -> return PolarityPragmasButNotPostulates_
-    31 -> return PragmaCompiled_
-    32 -> return PragmaNoTerminationCheck_
-    33 -> return ShadowingInTelescope_
-    34 -> return UnknownFixityInMixfixDecl_
-    35 -> return UnknownNamesInFixityDecl_
-    36 -> return UnknownNamesInPolarityPragmas_
-    37 -> return UselessAbstract_
-    38 -> return UselessInstance_
-    39 -> return UselessPrivate_
-    40 -> return AbsurdPatternRequiresNoRHS_
-    41 -> return AsPatternShadowsConstructorOrPatternSynonym_
-    42 -> return CantGeneralizeOverSorts_
-    43 -> return ClashesViaRenaming_
-    44 -> return CoverageIssue_
-    45 -> return CoverageNoExactSplit_
-    46 -> return DeprecationWarning_
-    47 -> return DuplicateUsing_
-    48 -> return FixityInRenamingModule_
-    49 -> return GenericNonFatalError_
-    50 -> return GenericUseless_
-    51 -> return GenericWarning_
-    52 -> return IllformedAsClause_
-    53 -> return InstanceArgWithExplicitArg_
-    54 -> return InstanceWithExplicitArg_
-    55 -> return InstanceNoOutputTypeName_
-    56 -> return InversionDepthReached_
-    57 -> return ModuleDoesntExport_
-    58 -> return NoGuardednessFlag_
-    59 -> return NotInScope_
-    60 -> return NotStrictlyPositive_
-    61 -> return UnsupportedIndexedMatch_
-    62 -> return OldBuiltin_
-    63 -> return PragmaCompileErased_
-    64 -> return RewriteMaybeNonConfluent_
-    65 -> return RewriteNonConfluent_
-    66 -> return RewriteAmbiguousRules_
-    67 -> return RewriteMissingRule_
-    68 -> return SafeFlagEta_
-    69 -> return SafeFlagInjective_
-    70 -> return SafeFlagNoCoverageCheck_
-    71 -> return SafeFlagNonTerminating_
-    72 -> return SafeFlagNoPositivityCheck_
-    73 -> return SafeFlagNoUniverseCheck_
-    74 -> return SafeFlagPolarity_
-    75 -> return SafeFlagPostulate_
-    76 -> return SafeFlagPragma_
-    77 -> return SafeFlagTerminating_
-    78 -> return SafeFlagWithoutKFlagPrimEraseEquality_
-    79 -> return TerminationIssue_
-    80 -> return UnreachableClauses_
-    81 -> return UnsolvedConstraints_
-    82 -> return UnsolvedInteractionMetas_
-    83 -> return UnsolvedMetaVariables_
-    84 -> return UselessHiding_
-    85 -> return UselessInline_
-    86 -> return UselessPatternDeclarationForRecord_
-    87 -> return UselessPublic_
-    88 -> return UserWarning_
-    89 -> return WithoutKFlagPrimEraseEquality_
-    90 -> return WrongInstanceDeclaration_
-    91 -> return CoInfectiveImport_
-    92 -> return InfectiveImport_
-    93 -> return DuplicateFieldsWarning_
-    94 -> return TooManyFieldsWarning_
-    95 -> return OptionRenamed_
-    _ -> malformed
+  value = valueN WarningMode
 
+-- Andreas, 2024-08-18
+-- Removed manual implementation of EmbPrj for this Enum type.
+instance EmbPrj WarningName
 
 instance EmbPrj CutOff where
   icod_ = \case
