@@ -3,6 +3,8 @@
 
 module Agda.TypeChecking.Serialise.Instances.Internal where
 
+import qualified Data.HashSet as HashSet
+import Control.Monad
 import Control.Monad.IO.Class
 
 import Agda.Syntax.Internal as I
@@ -15,7 +17,9 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Coverage.SplitTree
+import Agda.TypeChecking.DiscrimTree.Types
 
+import Agda.Utils.Functor
 import Agda.Utils.Permutation
 
 import Agda.Utils.Impossible
@@ -26,10 +30,15 @@ instance EmbPrj a => EmbPrj (Dom a) where
   value = valueN Dom
 
 instance EmbPrj Signature where
-  icod_ (Sig a b c) = icodeN' Sig a b c
+  icod_ (Sig a b c d) = icodeN' Sig a b c d
 
   value = valueN Sig
 
+instance EmbPrj InstanceTable where
+  icod_ (InstanceTable a b) = icodeN' InstanceTable a b
+
+  value = valueN InstanceTable
+
 instance EmbPrj Section where
   icod_ (Section a) = icodeN' Section a
 
@@ -141,35 +150,36 @@ instance EmbPrj IsFibrant where
   value 1 = return IsStrict
   value _ = malformed
 
+instance EmbPrj Univ where
+
 instance EmbPrj I.Sort where
-  icod_ (Type  a  ) = icodeN 0 Type a
-  icod_ (Prop  a  ) = icodeN 1 Prop a
-  icod_ SizeUniv    = icodeN 2 SizeUniv
-  icod_ (Inf f a)   = icodeN 3 Inf f a
-  icod_ (PiSort a b c) = icodeN 4 PiSort a b c
-  icod_ (FunSort a b) = icodeN 5 FunSort a b
-  icod_ (UnivSort a) = icodeN 6 UnivSort a
-  icod_ (DefS a b)   = icodeN 7 DefS a b
-  icod_ (SSet  a  ) = icodeN 8 SSet a
-  icod_ LockUniv    = icodeN 9 LockUniv
-  icod_ IntervalUniv = icodeN 10 IntervalUniv
-  icod_ (MetaS a b)  = icodeN 11 MetaS a b
-  icod_ (DummyS s)   = icodeN 12 DummyS s
+  icod_ = \case
+    Univ a b     -> icodeN 0  Univ a b
+    SizeUniv     -> icodeN 2  SizeUniv
+    Inf a b      -> icodeN 3  Inf a b
+    PiSort a b c -> icodeN 4  PiSort a b c
+    FunSort a b  -> icodeN 5  FunSort a b
+    UnivSort a   -> icodeN 6  UnivSort a
+    DefS a b     -> icodeN 7  DefS a b
+    LockUniv     -> icodeN 9  LockUniv
+    IntervalUniv -> icodeN 10 IntervalUniv
+    MetaS a b    -> icodeN 11 MetaS a b
+    DummyS s     -> icodeN 12 DummyS s
+    LevelUniv    -> icodeN 13 LevelUniv
 
   value = vcase valu where
-    valu [0, a]    = valuN Type  a
-    valu [1, a]    = valuN Prop  a
+    valu [0, a, b] = valuN Univ a b
     valu [2]       = valuN SizeUniv
-    valu [3, f, a] = valuN Inf f a
+    valu [3, a, b] = valuN Inf a b
     valu [4, a, b, c] = valuN PiSort a b c
     valu [5, a, b] = valuN FunSort a b
     valu [6, a]    = valuN UnivSort a
     valu [7, a, b] = valuN DefS a b
-    valu [8, a]    = valuN SSet a
     valu [9]       = valuN LockUniv
     valu [10]      = valuN IntervalUniv
     valu [11, a, b] = valuN MetaS a b
     valu [12, s]   = valuN DummyS s
+    valu [13]      = valuN LevelUniv
     valu _         = malformed
 
 instance EmbPrj DisplayForm where
@@ -184,34 +194,47 @@ instance EmbPrj a => EmbPrj (Open a) where
 
 instance EmbPrj CheckpointId where
   icod_ (CheckpointId a) = icode a
-  value n                = CheckpointId `fmap` value n
+  value n                = CheckpointId <$!> value n
 
 instance EmbPrj DisplayTerm where
-  icod_ (DTerm    a  )   = icodeN' DTerm a
-  icod_ (DDot     a  )   = icodeN 1 DDot a
+  icod_ (DTerm'   a b)   = icodeN' DTerm' a b
+  icod_ (DDot'    a b)   = icodeN 1 DDot' a b
   icod_ (DCon     a b c) = icodeN 2 DCon a b c
   icod_ (DDef     a b)   = icodeN 3 DDef a b
   icod_ (DWithApp a b c) = icodeN 4 DWithApp a b c
 
   value = vcase valu where
-    valu [a]          = valuN DTerm a
-    valu [1, a]       = valuN DDot a
+    valu [a, b]       = valuN DTerm' a b
+    valu [1, a, b]    = valuN DDot' a b
     valu [2, a, b, c] = valuN DCon a b c
     valu [3, a, b]    = valuN DDef a b
     valu [4, a, b, c] = valuN DWithApp a b c
     valu _            = malformed
 
 instance EmbPrj MutualId where
-  icod_ (MutId a) = icode a
-  value n         = MutId `fmap` value n
+  icod_ (MutualId a) = icode a
+  value n            = MutualId <$!> value n
 
 instance EmbPrj CompKit where
   icod_ (CompKit a b) = icodeN' CompKit a b
   value = valueN CompKit
 
-instance EmbPrj Definition where
-  icod_ (Defn a b c d e f g h i j k l m n o p q r s) = icodeN' Defn a b (P.killRange c) d e f g h i j k l m n o p q r s
+instance EmbPrj InstanceInfo where
+  icod_ (InstanceInfo a b) = icodeN' InstanceInfo a b
+  value = valueN InstanceInfo
 
+instance EmbPrj Definition where
+  icod_ (Defn a b c d e f g h i j k l m n o blocked r s) =
+    icodeN' Defn a b (P.killRange c) d e f g h i j k l m n o (ossify blocked) r s
+    where
+      -- Andreas, 2024-01-02, issue #7044:
+      -- After serialization, a definition can never be unblocked,
+      -- since all metas are ossified.
+      -- Thus, we turn any blocker into 'neverUnblock'.
+      ossify :: Blocked_ -> Blocked_
+      ossify = \case
+        b@NotBlocked{} -> b
+        Blocked b () -> Blocked neverUnblock ()
   value = valueN Defn
 
 instance EmbPrj NotBlocked where
@@ -229,11 +252,22 @@ instance EmbPrj NotBlocked where
     valu [3, a] = valuN MissingClauses a
     valu _      = malformed
 
-instance EmbPrj Blocked_ where
-  icod_ (NotBlocked a b) = icodeN' NotBlocked a b
-  icod_ Blocked{} = __IMPOSSIBLE__
+-- Andreas, 2024-01-02, issue #7044.
+-- We only serialize 'neverUnblock';
+-- other than that, there should not be any blockers left at serialization time.
+blockedToMaybe :: Blocked_ -> Maybe NotBlocked
+blockedToMaybe = \case
+  NotBlocked a ()       -> Just a
+  Blocked a ()
+    | a == neverUnblock -> Nothing
+    | otherwise         -> __IMPOSSIBLE__
 
-  value = valueN NotBlocked
+blockedFromMaybe :: Maybe NotBlocked -> Blocked_
+blockedFromMaybe = maybe (Blocked neverUnblock ()) (`NotBlocked` ())
+
+instance EmbPrj Blocked_ where
+  icod_ = icod_ . blockedToMaybe
+  value = blockedFromMaybe <.> value
 
 instance EmbPrj NLPat where
   icod_ (PVar a b)      = icodeN 0 PVar a b
@@ -267,6 +301,7 @@ instance EmbPrj NLPSort where
   icod_ PLockUniv   = icodeN 4 PLockUniv
   icod_ PIntervalUniv = icodeN 5 PIntervalUniv
   icod_ (PSSet a)   = icodeN 6 PSSet a
+  icod_ PLevelUniv = icodeN 7 PLevelUniv
 
   value = vcase valu where
     valu [0, a] = valuN PType a
@@ -276,10 +311,11 @@ instance EmbPrj NLPSort where
     valu [4]    = valuN PLockUniv
     valu [5]    = valuN PIntervalUniv
     valu [6, a] = valuN PSSet a
+    valu [7]    = valuN PLevelUniv
     valu _      = malformed
 
 instance EmbPrj RewriteRule where
-  icod_ (RewriteRule a b c d e f g) = icodeN' RewriteRule a b c d e f g
+  icod_ (RewriteRule a b c d e f g h) = icodeN' RewriteRule a b c d e f g h
 
   value = valueN RewriteRule
 
@@ -369,27 +405,45 @@ instance EmbPrj EtaEquality where
 
 instance EmbPrj ProjectionLikenessMissing
 
+instance EmbPrj BuiltinSort where
+  icod_ = \case
+    SortUniv  a      -> icodeN 0 SortUniv  a
+    SortOmega a      -> icodeN 1 SortOmega a
+    SortIntervalUniv -> icodeN 2 SortIntervalUniv
+    SortLevelUniv    -> icodeN 3 SortLevelUniv
+
+  value = vcase \case
+    [0, a] -> valuN SortUniv  a
+    [1, a] -> valuN SortOmega a
+    [2]    -> valuN SortIntervalUniv
+    [3]    -> valuN SortLevelUniv
+    _ -> malformed
+
 instance EmbPrj Defn where
   icod_ (Axiom       a)                                 = icodeN 0 Axiom a
-  icod_ (Function    a b s t u c d e f g h i j k l)     = icodeN 1 (\ a b s -> Function a b s t) a b s u c d e f g h i j k l
+  icod_ (Function    a b s t u c d e f g h i j k)       = icodeN 1 (\ a b s -> Function a b s t) a b s u c d e f g h i j k
   icod_ (Datatype    a b c d e f g h i j)               = icodeN 2 Datatype a b c d e f g h i j
   icod_ (Record      a b c d e f g h i j k l m)         = icodeN 3 Record a b c d e f g h i j k l m
-  icod_ (Constructor a b c d e f g h i j)               = icodeN 4 Constructor a b c d e f g h i j
-  icod_ (Primitive   a b c d e)                         = icodeN 5 Primitive a b c d e
+  icod_ (Constructor a b c d e f g h i j k)             = icodeN 4 Constructor a b c d e f g h i j k
+  icod_ (Primitive   a b c d e f)                       = icodeN 5 Primitive a b c d e f
   icod_ (PrimitiveSort a b)                             = icodeN 6 PrimitiveSort a b
   icod_ AbstractDefn{}                                  = __IMPOSSIBLE__
-  icod_ GeneralizableVar                                = icodeN 7 GeneralizableVar
-  icod_ DataOrRecSig{}                                  = __IMPOSSIBLE__
+  icod_ (GeneralizableVar a)                            = icodeN 7 GeneralizableVar a
+  icod_ (DataOrRecSig a)                                = icodeN 8 DataOrRecSig a
+    -- Andreas, 2024-10-27
+    -- DataOrRecSig is possible via unquoteDecl in meta-programming, see #7576
 
   value = vcase valu where
     valu [0, a]                                        = valuN Axiom a
-    valu [1, a, b, s, u, c, d, e, f, g, h, i, j, k, l] = valuN (\ a b s -> Function a b s Nothing) a b s u c d e f g h i j k l
+    valu [1, a, b, s, u, c, d, e, f, g, h, i, j, k]
+                                                       = valuN (\ a b s -> Function a b s Nothing) a b s u c d e f g h i j k
     valu [2, a, b, c, d, e, f, g, h, i, j]             = valuN Datatype a b c d e f g h i j
     valu [3, a, b, c, d, e, f, g, h, i, j, k, l, m]    = valuN Record   a b c d e f g h i j k l m
-    valu [4, a, b, c, d, e, f, g, h, i, j]             = valuN Constructor a b c d e f g h i j
-    valu [5, a, b, c, d, e]                            = valuN Primitive   a b c d e
+    valu [4, a, b, c, d, e, f, g, h, i, j, k]          = valuN Constructor a b c d e f g h i j k
+    valu [5, a, b, c, d, e, f]                         = valuN Primitive   a b c d e f
     valu [6, a, b]                                     = valuN PrimitiveSort a b
-    valu [7]                                           = valuN GeneralizableVar
+    valu [7, a]                                        = valuN GeneralizableVar a
+    valu [8, a]                                        = valuN DataOrRecSig a
     valu _                                             = malformed
 
 instance EmbPrj LazySplit where
@@ -421,16 +475,7 @@ instance EmbPrj a => EmbPrj (SplitTree' a) where
     valu [0, a, b, c] = valuN SplitAt a b c
     valu _            = malformed
 
-instance EmbPrj FunctionFlag where
-  icod_ FunStatic       = icodeN 0 FunStatic
-  icod_ FunInline       = icodeN 1 FunInline
-  icod_ FunMacro        = icodeN 2 FunMacro
-
-  value = vcase valu where
-    valu [0] = valuN FunStatic
-    valu [1] = valuN FunInline
-    valu [2] = valuN FunMacro
-    valu _   = malformed
+instance EmbPrj FunctionFlag
 
 instance EmbPrj a => EmbPrj (WithArity a) where
   icod_ (WithArity a b) = icodeN' WithArity a b
@@ -442,6 +487,20 @@ instance EmbPrj a => EmbPrj (Case a) where
 
   value = valueN Branches
 
+-- Opaque blocks are serialised in an abbreviated manner: We only need
+-- the enclosed definitions (3rd argument) and parent (4th argument) to
+-- compute the transitive closure during scope checking, never
+-- afterwards.
+instance EmbPrj OpaqueBlock where
+  icod_ (OpaqueBlock id uf _ _ r) =
+    icodeN' (\id uf ->
+      let !unfolding = HashSet.fromList uf
+      in OpaqueBlock id unfolding mempty Nothing)
+    id (HashSet.toList uf) r
+
+  value = valueN (\id uf -> let !unfolding = HashSet.fromList uf
+                            in OpaqueBlock id unfolding mempty Nothing)
+
 instance EmbPrj CompiledClauses where
   icod_ (Fail a)   = icodeN' Fail a
   icod_ (Done a b) = icodeN' Done a (P.killRange b)
@@ -478,7 +537,7 @@ instance EmbPrj TermHead where
     valu _      = malformed
 
 instance EmbPrj I.Clause where
-  icod_ (Clause a b c d e f g h i j k l) = icodeN' Clause a b c d e f g h i j k l
+  icod_ (Clause a b c d e f g h i j k) = icodeN' Clause a b c d e f g h i j k
 
   value = valueN Clause
 
@@ -507,6 +566,7 @@ instance EmbPrj I.PatOrigin where
   icod_ PatORec     = icodeN 6 PatORec
   icod_ PatOLit     = icodeN 7 PatOLit
   icod_ PatOAbsurd  = icodeN 8 PatOAbsurd
+  icod_ (PatOSplitArg a) = icodeN 9 PatOSplitArg a
 
   value = vcase valu where
     valu []     = valuN PatOSystem
@@ -518,6 +578,7 @@ instance EmbPrj I.PatOrigin where
     valu [6]    = valuN PatORec
     valu [7]    = valuN PatOLit
     valu [8]    = valuN PatOAbsurd
+    valu [9, a] = valuN PatOSplitArg a
     valu _      = malformed
 
 instance EmbPrj a => EmbPrj (I.Pattern' a) where
@@ -594,3 +655,31 @@ instance EmbPrj a => EmbPrj (Judgement a) where
 instance EmbPrj RemoteMetaVariable where
   icod_ (RemoteMetaVariable a b c) = icodeN' RemoteMetaVariable a b c
   value = valueN RemoteMetaVariable
+
+instance EmbPrj Key where
+  icod_ (RigidK x y) = icodeN 0 RigidK x y
+  icod_ (LocalK x y) = icodeN 1 LocalK x y
+  icod_ (PiK h)      = icodeN 2 PiK h
+  icod_ FlexK        = icodeN 3 FlexK
+  icod_ ConstK       = icodeN 4 ConstK
+  icod_ SortK        = icodeN 5 SortK
+
+  value = vcase valu where
+    valu [0, x, y] = valuN RigidK x y
+    valu [1, x, y] = valuN LocalK x y
+    valu [2, h]    = valuN PiK h
+    valu [3]       = valuN FlexK
+    valu [4]       = valuN ConstK
+    valu [5]       = valuN SortK
+    valu _         = malformed
+
+instance (EmbPrj a, Ord a) => EmbPrj (DiscrimTree a) where
+  icod_ EmptyDT        = icodeN' EmptyDT
+  icod_ (DoneDT s)     = icodeN' DoneDT s
+  icod_ (CaseDT i k s) = icodeN' CaseDT i k s
+
+  value = vcase valu where
+    valu []        = valuN EmptyDT
+    valu [a]       = valuN DoneDT a
+    valu [i, k, s] = valuN CaseDT i k s
+    valu _         = malformed
