@@ -1,14 +1,12 @@
+{-# LANGUAGE CPP #-}
 
 module Agda.Utils.Monad
     ( module Agda.Utils.Monad
-    , when, unless, MonadPlus(..)
-    , (<$>), (<*>)
-    , (<$)
+    , module X
+    , (<$>), (<*>) , (<$)
     )
     where
 
-import Control.Applicative    ( liftA2 )
-import Control.Monad          ( MonadPlus(..), guard, unless, when )
 import Control.Monad.Except   ( MonadError(catchError, throwError) )
 import Control.Monad.Identity ( runIdentity )
 import Control.Monad.State    ( MonadState(get, put) )
@@ -22,20 +20,80 @@ import Data.Maybe
 import Data.Monoid
 
 import Agda.Utils.Applicative
+import Agda.Utils.Boolean
 import Agda.Utils.Either
 import Agda.Utils.Null (empty, ifNotNullM)
 import Agda.Utils.Singleton
 
 import Agda.Utils.Impossible
 
+-- Reexport Control.Monad
+import Control.Applicative as X
+  ( liftA2
+  )
+import Control.Monad as X
+  ( MonadPlus(..), (<$!>), (>=>), (<=<)
+  , filterM, foldM, forM, forM_
+  , join
+  , liftM2, liftM3, liftM4
+  , msum
+  , void
+  , zipWithM, zipWithM_
+  )
+import Control.Monad.Trans as X
+  ( MonadTrans, lift
+  )
+
+
+---------------------------------------------------------------------------
+-- Vendor some new functions from mtl-2.3.1
+
+#if MIN_VERSION_mtl(2,3,1)
+import Control.Monad.Except as X ( tryError, withError )
+#endif
+
+#if !MIN_VERSION_mtl(2,3,1)
+-- | 'MonadError' analogue to the 'Control.Exception.try' function.
+tryError :: MonadError e m => m a -> m (Either e a)
+tryError action = (Right <$> action) `catchError` (pure . Left)
+
+-- | 'MonadError' analogue to the 'withExceptT' function.
+-- Modify the value (but not the type) of an error.  The type is
+-- fixed because of the functional dependency @m -> e@.  If you need
+-- to change the type of @e@ use 'mapError' or 'modifyError'.
+withError :: MonadError e m => (e -> e) -> m a -> m a
+withError f action = tryError action >>= either (throwError . f) pure
+#endif
+
 ---------------------------------------------------------------------------
 
 -- | Binary bind.
 (==<<) :: Monad m => (a -> b -> m c) -> (m a, m b) -> m c
 k ==<< (ma, mb) = ma >>= \ a -> k a =<< mb
 
+-- | Strict `ap`
+(<*!>) :: Monad m => m (a -> b) -> m a -> m b
+(<*!>) mf ma = do
+  f <- mf
+  a <- ma
+  pure $! f a
+{-# INLINE (<*!>) #-}
+infixl 4 <*!>
+
 -- Conditionals and monads ------------------------------------------------
 
+{-# SPECIALIZE when :: Monad m => Bool -> m () -> m () #-}
+when :: (IsBool b, Monad m) => b -> m () -> m ()
+when b m = ifThenElse b m $ pure ()
+
+{-# SPECIALIZE unless :: Monad m => Bool -> m () -> m () #-}
+unless :: (IsBool b, Monad m) => b -> m () -> m()
+unless b m = ifThenElse b (pure ()) m
+
+{-# SPECIALIZE guard :: MonadPlus m => Bool -> m () #-}
+guard :: (IsBool b, MonadPlus m) => b -> m ()
+guard b = ifThenElse b (pure ()) mzero
+
 whenM :: Monad m => m Bool -> m () -> m ()
 whenM c m = c >>= (`when` m)
 
@@ -61,8 +119,11 @@ and2M ma mb = ifM ma mb (return False)
 andM :: (Foldable f, Monad m) => f (m Bool) -> m Bool
 andM = Fold.foldl' and2M (return True)
 
-allM :: (Foldable f, Monad m) => f a -> (a -> m Bool) -> m Bool
-allM xs f = Fold.foldl' (\b -> and2M b . f) (return True) xs
+allM :: (Foldable f, Monad m) => (a -> m Bool) -> f a -> m Bool
+allM f = Fold.foldl' (\ b -> and2M b . f) (return True)
+
+forallM :: (Foldable f, Monad m) => f a -> (a -> m Bool) -> m Bool
+forallM xs f = Fold.foldl' (\b -> and2M b . f) (return True) xs
 
 -- | Lazy monadic disjunction.
 or2M :: Monad m => m Bool -> m Bool -> m Bool
@@ -71,8 +132,11 @@ or2M ma = ifM ma (return True)
 orM :: (Foldable f, Monad m) => f (m Bool) -> m Bool
 orM = Fold.foldl' or2M (return False)
 
-anyM :: (Foldable f, Monad m) => f a -> (a -> m Bool) -> m Bool
-anyM xs f = Fold.foldl' (\b -> or2M b . f) (return False) xs
+anyM :: (Foldable f, Monad m) => (a -> m Bool) -> f a -> m Bool
+anyM f = Fold.foldl' (\ b -> or2M b . f) (return False)
+
+existsM :: (Foldable f, Monad m) => f a -> (a -> m Bool) -> m Bool
+existsM xs f = anyM f xs
 
 -- | Lazy monadic disjunction with @Either@  truth values.
 --   Returns the last error message if all fail.
@@ -139,13 +203,22 @@ forMM_ = flip mapMM_
 
 -- Lists and monads -------------------------------------------------------
 
+concatMapM :: Monad m => (a -> m [b]) -> [a] -> m [b]
+concatMapM f xs = concat <$> Trav.mapM f xs
+
 -- | A monadic version of @'mapMaybe' :: (a -> Maybe b) -> [a] -> [b]@.
 mapMaybeM :: Monad m => (a -> m (Maybe b)) -> [a] -> m [b]
-mapMaybeM f xs = catMaybes <$> Trav.mapM f xs
+mapMaybeM f = go where
+  go []     = return []
+  go (a:as) = f a >>= \case
+    Nothing -> go as
+    Just b  -> do {!bs <- go as; pure (b : bs)}
+{-# INLINE mapMaybeM #-}
 
 -- | A version of @'mapMaybeM'@ with a computation for the input list.
 mapMaybeMM :: Monad m => (a -> m (Maybe b)) -> m [a] -> m [b]
 mapMaybeMM f m = mapMaybeM f =<< m
+{-# INLINE mapMaybeMM #-}
 
 -- | The @for@ version of 'mapMaybeM'.
 forMaybeM :: Monad m => [a] -> (a -> m (Maybe b)) -> m [b]
@@ -170,7 +243,7 @@ dropWhileEndM p (x : xs) = ifNotNullM (dropWhileEndM p xs) (return . (x:)) $ {-e
 -- | A ``monadic'' version of @'partition' :: (a -> Bool) -> [a] -> ([a],[a])
 partitionM :: (Functor m, Applicative m) => (a -> m Bool) -> [a] -> m ([a], [a])
 partitionM f =
-  foldr (\ x mlr -> bool (first (x:)) (second (x:)) <$> f x <*> mlr)
+  foldr (\ x -> liftA2 (bool (second (x:)) (first (x:))) $ f x)
         (pure empty)
 
 -- MonadPlus -----------------------------------------------------------------
