@@ -2,14 +2,23 @@
 
 module Agda.Utils.String where
 
+import Control.Monad.Except
 import Control.Monad.Reader
 import Control.Monad.State
+import Control.Monad.Trans.Maybe
+import Control.Monad.Writer
 
 import Data.Char
 import qualified Data.List as List
 import Data.String
 
+import Agda.Utils.Function (applyUnless)
 import Agda.Utils.List
+import Agda.Utils.List1 (String1, fromList)
+import Agda.Utils.Size  (Sized, natSize)
+
+instance IsString String1 where
+  fromString = fromList
 
 -- | 'quote' adds double quotes around the string, replaces newline
 -- characters with @\n@, and escapes double quotes and backslashes
@@ -82,6 +91,11 @@ indent i = unlines . map (List.genericReplicate i ' ' ++) . lines
 unwords1 :: [String] -> String
 unwords1 = unwords . filter (not . null)
 
+-- | Append an @"s"@ to the second argument if the first has cardinality @/= 1@.
+
+pluralS :: Sized a => a -> String -> String
+pluralS xs = applyUnless (natSize xs == 1) (++ "s")
+
 -- | Show a number using comma to separate powers of 1,000.
 
 showThousandSep :: Show a => a -> String
@@ -99,8 +113,17 @@ rtrim = List.dropWhileEnd isSpace
 trim :: String -> String
 trim = rtrim . ltrim
 
+instance (IsString (m a), Monad m) => IsString (ExceptT e m a) where
+  fromString = lift . fromString
+
+instance (IsString (m a), Monad m) => IsString (MaybeT m a) where
+  fromString = lift . fromString
+
 instance (IsString (m a), Monad m) => IsString (ReaderT r m a) where
   fromString = lift . fromString
 
 instance (IsString (m a), Monad m) => IsString (StateT s m a) where
   fromString = lift . fromString
+
+instance (IsString (m a), Monad m, Monoid w) => IsString (WriterT w m a) where
+  fromString = lift . fromString
