@@ -1,10 +1,20 @@
 {-# OPTIONS_GHC -Wunused-imports #-}
 
 -- | Strict tries (based on "Data.Map.Strict" and "Agda.Utils.Maybe.Strict").
+--
+-- Note that if 'delete' or 'adjust' are used, one may end up with non-canonical
+-- subtries, i.e., those that do not contain any value (@Data.List.null . toList@)
+-- but may be @Eq@-different to other empty tries (@/= empty@).
 
 module Agda.Utils.Trie
   ( Trie(..)
   , empty, singleton, everyPrefix, insert, insertWith, union, unionWith
+  -- Andreas 2024-11-16:
+  -- Deletion does not clean up the trie, thus violates equational laws
+  -- with the given Eq that does not quotient out the different representations
+  -- of empty subtries.
+  -- E.g. @delete k . insert k v . delete k /= delete k@.
+  -- @adjust@ has a similar problem.
   , adjust, delete
   , toList, toAscList, toListOrderedBy
   , lookup, member, lookupPath, lookupTrie
@@ -31,11 +41,15 @@ import Agda.Utils.Lens
 --
 --   With the strict 'Maybe' type, 'Trie' is also strict in 'v'.
 data Trie k v = Trie !(Strict.Maybe v) !(Map k (Trie k v))
-  deriving ( Show
-           , Eq
-           , Functor
-           , Foldable
-           )
+  deriving
+    ( Show
+    , Eq
+        -- Andreas, 2024-11-16: The derived Eq isn't extensional.
+        -- It disguishes different representation of the empty Trie.
+        -- (E.g. the empty map and the map that only contains Nothing values.)
+    , Functor
+    , Foldable
+    )
 
 instance (NFData k, NFData v) => NFData (Trie k v) where
   rnf (Trie a b) = rnf a `seq` rnf b
@@ -43,7 +57,10 @@ instance (NFData k, NFData v) => NFData (Trie k v) where
 -- | Empty trie.
 instance Null (Trie k v) where
   empty = Trie Strict.Nothing Map.empty
-  null (Trie v t) = null v && null t
+  null (Trie v t) = null v && all null t
+    -- Andreas, 2024-11-16: since we allow non-canoncial tries,
+    -- we have to check that every subtrie is empty,
+    -- not just that there are no subtries.
 
 -- | Helper function used to implement 'singleton' and 'everyPrefix'.
 singletonOrEveryPrefix :: Bool -> [k] -> v -> Trie k v
@@ -83,11 +100,25 @@ insert k v t = (singleton k v) `union` t
 insertWith :: (Ord k) => (v -> v -> v) -> [k] -> v -> Trie k v -> Trie k v
 insertWith f k v t = unionWith f (singleton k v) t
 
+-- Andreas, 2024-11-16: delete does not clean up empty subtrees,
+-- so it is not correct wrt. @Eq@.
+
 -- | Delete value at key, but leave subtree intact.
+--
+-- Disclaimer: may return a non-canoncial trie
+-- because it does not clean up subtries that become empty.
+--
 delete :: Ord k => [k] -> Trie k v -> Trie k v
 delete path = adjust path (const Strict.Nothing)
 
+-- Andreas, 2024-11-16: adjust does not clean up empty subtrees,
+-- so it is not correct wrt. @Eq@.
+
 -- | Adjust value at key, leave subtree intact.
+--
+-- Disclaimer: may return a non-canoncial trie
+-- because it does not clean up subtries that become empty.
+--
 adjust ::
   Ord k =>
   [k] -> (Strict.Maybe v -> Strict.Maybe v) -> Trie k v -> Trie k v
