@@ -1,9 +1,7 @@
 {-# OPTIONS_GHC -Wunused-imports #-}
 
-{-# LANGUAGE GADTs                #-}
 {-# LANGUAGE DataKinds            #-}
 {-# LANGUAGE PolyKinds            #-}
-{-# LANGUAGE TypeOperators        #-}
 
 -- We need undecidable instances for the definition of @Foldr@,
 -- and @Domains@ and @CoDomain@ using @If@ for instance.
@@ -59,7 +57,7 @@ data ConsMap1 :: (Function k l -> Type) -> k -> Function [l] [l] -> Type
 type instance Apply (ConsMap0 f)    a = ConsMap1 f a
 type instance Apply (ConsMap1 f a) tl = Apply f a ': tl
 
-type family Constant (b :: l) (as :: [k]) :: [l] where
+type family Constant (b :: Type) (as :: [k]) :: [Type] where
   Constant b as = Map (Constant1 b) as
 
 ------------------------------------------------------------------
